/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor;

import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;
import org.xquark.extractor.common.Constants;
import org.xquark.extractor.common.MessageLibrary;
import org.xquark.extractor.common.Trace;

/**
 * SAX2 content handler implementation that performs configuration file loading.
 */
public class ExtractorConfigurationFileHandler extends DefaultHandler
implements Constants {
	
	/* private "status" constants */
	private final byte IN_CATALOG = 1;
	private final byte IN_SCHEMA = 2;
	private final byte IN_TABLE = 1;
	private final byte IN_COLUMN = 1;

	private static final String RCSRevision = "$Revision: 1.3 $";
	private static final String RCSName = "$Name:  $";

	/* copied for defaultAccessorhandler */
	protected Properties xmlproperties;
	protected Extractor accessor;

	private Selection current_site;
	private Selection current_catalog;
	private Selection current_schema;
	private Selection current_table;
	private Selection current_column;
	private String _currentView = null;
	private StringBuffer _currentChars = new StringBuffer();
	private Selection _selection = null;
	private String _dataSourceName = null;
	private Properties _substitutions = null;
	private Integer _nameCase = new Integer(CASE_MIXED);
	private Map _viewDefinitions = null;
	private byte includeType = Selection.NONE;


	public ExtractorConfigurationFileHandler() {
		xmlproperties = new Properties();
	}

	//	public ExtractorHandler (SQLWrapper accessor) /* @modified : yucheng 21/01/2001*/
	public ExtractorConfigurationFileHandler(Extractor accessor) {
		this.accessor = accessor;
		xmlproperties = new Properties();
	}

	private Map getViewDefinitionsInstance() {
		Map retVal = null;
		if (null == _viewDefinitions) {
			_viewDefinitions = new HashMap();

		}
		retVal = _viewDefinitions;
		return retVal;
	}

	public void startElement(String namespaceURI, String localName, String qualifiedName, Attributes atts) throws SAXException {
		Trace.enter(this, "startElement(...)", qualifiedName);
		// DefaultWrapperHandler_startElement(namespaceURI, localName, qualifiedName, atts) ;
		_currentChars.setLength(0);
		if (localName.equalsIgnoreCase(LABEL_DATASOURCE)) {
			String name = atts.getValue(LABEL_SPEC_NAME);
			if (null != accessor) {
				accessor.setName(name);
			}
			_dataSourceName = name;
			setProperty(LABEL_SPEC_NAME, name);

			current_site = new Selection(name, null, Selection.NONE, null);
			xmlproperties.put("selections", current_site);
			includeType = Selection.INCLUDES;
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_SUBST)) {
			_substitutions = new Properties();
			xmlproperties.put("substitutions", _substitutions);
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_SUBST_CHAR)) {
			_substitutions.setProperty(atts.getValue(LABEL_SPEC_SUBST_VALUE), atts.getValue(LABEL_SPEC_SUBST_SUBST));
		}
		//		else if (localName.equalsIgnoreCase(LABEL_SPEC_SITE)) {
		//			String name = atts.getValue(LABEL_SPEC_NAME) ;
		//			current_site = new Selection(name,null,Selection.NONE,null) ;
		//		}
		else if (localName.equalsIgnoreCase(LABEL_SPEC_CATALOG)) {
			String name = atts.getValue(LABEL_SPEC_NAME);
			current_catalog = new Selection(name, null, includeType, null);
			current_site.addToMap(name, current_catalog);
			includeType = Selection.INCLUDES;
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_SCHEMA)) {
			String name = atts.getValue(LABEL_SPEC_NAME);
			String targetNamespace = atts.getValue(LABEL_SPEC_TARGETNAMESPACE);
			current_schema = new Selection(name, targetNamespace, includeType, null);

			String efd = atts.getValue(LABEL_SPEC_ELEMENTFORMDEFAULT);
			int elementFormDefault;
			if ("qualified".equalsIgnoreCase(efd)) {
				elementFormDefault = Selection.QUALIFIED;
			} else {
				elementFormDefault = Selection.UNQUALIFIED;
			}
			current_schema.setElementFormDefault(elementFormDefault);

			current_catalog.addToMap(name, current_schema);
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_TABLE)) {
			String name = atts.getValue(LABEL_SPEC_NAME);
			if (null != name) {
				current_table = new Selection(name, null, includeType, null);
				String alias = atts.getValue(LABEL_SPEC_ALIAS);
				if (null != alias && !Selection.verifyName(alias)) {
					throw new SAXException(MessageLibrary.getMessage("C_INV_NCNAME", alias));
				}
				current_table.setAlias(alias);
				current_schema.addToMap(name, current_table);
			} else {
				name = atts.getValue(LABEL_SPEC_REGEX);
				if (null != name) {
					current_table = new Selection(name, true, null, includeType, null);
					current_schema.addToMap(name, current_table);
				} else {
					//                    throw new Exception("invalid table element in configuration file.");
				}
			}
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_COLUMN)) {
			String name = atts.getValue(LABEL_SPEC_NAME);

			if (null != name) {
				current_column = new Selection(name, null, includeType, null);
				String alias = atts.getValue(LABEL_SPEC_ALIAS);
				if (null != alias && !Selection.verifyName(alias)) {
					throw new SAXException(MessageLibrary.getMessage("C_INV_NCNAME", alias));
				}
				current_column.setAlias(alias);
				current_table.addToMap(name, current_column);
			} else {
				name = atts.getValue(LABEL_SPEC_REGEX);
				if (null != name) {
					current_column = new Selection(name, true, null, includeType, null);
					current_table.addToMap(name, current_table);
				} else {
					//                    throw new Exception("invalid table element in configuration file.");
				}
			}

			String isPK = atts.getValue(LABEL_SPEC_PK_COLUMN);

			if (null != isPK && isPK.equalsIgnoreCase("true")) {
				current_column.setPKColumn(true);
			}

		} else if (localName.equalsIgnoreCase(LABEL_SPEC_INCLUDES)) {
			includeType = Selection.INCLUDES;
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_EXCLUDES)) {
			includeType = Selection.EXCLUDES;
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_VIEW)) {

			_currentView = atts.getValue(LABEL_SPEC_NAME);
		}

		//        Trace.exit(this,"startElement()", qualifiedName);
	}

	public void endElement(String namespaceURI, String localName, String qualifiedName) throws SAXException {
		///*DEBUG*/ System.out.println ("SQLWrapperHandler:endElement("+namespaceURI+","+localName+","+qualifiedName+")") ;
		//		Trace.enter(this,"endElement(...)", qualifiedName);
		if (localName.equalsIgnoreCase(LABEL_SPEC_DESCRIPTION)) {
			setProperty(LABEL_SPEC_DESCRIPTION, _currentChars.toString());
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_CASE)) {
			setProperty(LABEL_SPEC_CASE, _currentChars.toString());
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_SUBST_NAMECASE)) {
			String cur_car = _currentChars.toString();
			if (cur_car.equalsIgnoreCase(CASE_LOWER_TOKEN)) {
				_nameCase = new Integer(CASE_LOWER);
			} else if (cur_car.equalsIgnoreCase(CASE_UPPER_TOKEN)) {
				_nameCase = new Integer(CASE_UPPER);
			} else if (cur_car.equalsIgnoreCase(CASE_MIXED_TOKEN)) {
				_nameCase = new Integer(CASE_MIXED);
			}
			xmlproperties.put(LABEL_SPEC_SUBST_NAMECASE, _nameCase);
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_URL)) {
			setProperty(LABEL_SPEC_URL, _currentChars.toString());
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_USER)) {
			setProperty(LABEL_SPEC_USER, _currentChars.toString());
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_PASSWORD)) {
			setProperty(LABEL_SPEC_PASSWORD, _currentChars.toString());
		} else if (localName.equalsIgnoreCase(LABEL_SPEC_VIEW)) {
			getViewDefinitionsInstance().put(_currentView, _currentChars.toString());
		}
		//        else {
		//            Trace.warn(this,"endElement(...)", MessageLibrary.getMessage("C_ELM_UNK", qualifiedName));
		//        }
		Trace.exit(this, "endElement(...)", qualifiedName);
	}

	public void endDocument() {

		//	System.out.println ("======================================") ;
		//System.out.println ("CONTAINERS") ;
		//Iterator iterator = containers.keySet ().iterator () ;
		//while (iterator.hasNext ())
		//{
		//	String n = (String) iterator.next () ;
		//	System.out.println ("Container name : " + n) ;
		//	HashMap hm = (HashMap) containers.get (n) ;
		//
		//	System.out.println (" TABLES") ;
		//	Iterator iterator2 = hm.keySet ().iterator () ;
		//	while (iterator2.hasNext ())
		//	{
		//		String n2 = (String) iterator2.next () ;
		//		System.out.println ("  Table name : " + n2) ;
		//		ArrayList hm2 = (ArrayList) hm.get (n2) ;
		//
		//		for (int i = 0 ; i < hm2.size () ; i ++)
		//		{
		//			System.out.println ("    column : " + hm2.get (i)) ;
		//		}
		//
		//	}
		//	System.out.println ("======================================") ;
		//}
		if (null != accessor) {
			// ((Extractor) accessor).setSelection(current_site) ;
		}
		_selection = current_site;
	}

	public Properties getProperties() {
		return xmlproperties;
	}

	public Properties getSubstitutions() {
		return _substitutions;
	}

	/**
	 * Set a property of the accessor.
	 *
	 * @param a property of the accessor.
	 */
	protected void setProperty(String key, String value) {
		if (key == null) {
			System.err.println("Key " + key + " is null.");
			return;
		}
		if (value == null) {
			System.err.println("Value " + value + " of " + key + " is null.");
			return;
		}
		xmlproperties.setProperty(key, value);
	}

	/**
	 * Return  a property of the accessor.
	 *
	 * @return a property of the accessor.
	 */
	protected String getProperty(String key) {
		return (String) xmlproperties.getProperty(key);
	}

	public void characters(char ch[], int start, int length) throws SAXException {
		String cur_car = new String(ch, start, length);
		cur_car.trim();
		_currentChars.append(cur_car);
	}

	public String getDataSourceName() {
		return _dataSourceName;
	}

	public Selection getSelection() {
		return _selection;
	}

	public Map getViewDefinitions() {
		return _viewDefinitions;
	}

	public void error(SAXParseException spe) throws SAXException {
		throw spe;
	}
}
