/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor;

import java.sql.*;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import org.xquark.extractor.common.Log;
import org.xquark.extractor.common.MessageLibrary;
import org.xquark.extractor.common.Trace;
import org.xquark.extractor.metadata.MetaDataManager;
import org.xquark.util.misc.Debug;
import org.xquark.xml.xdbc.*;

/**
 * Extractor implementation of {@link org.xquark.xml.xdbc.XMLConnection}.
 */
public class ExtractorReadOnlyConnection
    extends DefaultReadOnlyXMLConnection
    implements XMLConnection {

    public static final String URL_PREFIX = "xdbc:xquark:extractor:";

    private static final String RCSRevision = "$Revision: 1.5 $";
    private static final String RCSName = "$Name:  $";

    /* a instantance of java.sql.Connection */
    Connection _jdbcConnection;
    Configuration _config;

    private String _user = null;

    /* the Extractor instance who created this conncetion*/
    private Extractor _extractor = null;

    /* */
    private boolean _isClosed = false;

    /* when this XMLConnection is constructed with a JDBC Connection ,
     * the JDBC Connection is called borrowed connection */
    private boolean _borrowedConnection = false;

    /* a list all active XMLStatement that this XMLConnection has created */
    private List _activeStatementList = new ArrayList();

    public static void main(String[] args) throws Exception {
        ExtractorReadOnlyConnection e =
            new ExtractorReadOnlyConnection(null, null);
    }

    public ExtractorReadOnlyConnection(
        Connection jdbcConnection,
        Configuration config)
        throws XMLDBCException {
        this(jdbcConnection, config, false);
    }

    public ExtractorReadOnlyConnection(
        Connection jdbcConnection,
        Configuration config,
        boolean borrowed)
        throws XMLDBCException {
        Trace.enter(
            this,
            "ExtractorReadOnlyConnection(Connection jdbcConnection, Configuration config, boolean borrowed)");

        _jdbcConnection = jdbcConnection;
        _borrowedConnection = borrowed;

        try {
            DatabaseMetaData dbm = _jdbcConnection.getMetaData();
            _user = dbm.getUserName();
        } catch (SQLException e) {
            throw new XMLDBCException(e.getMessage(), e);
        }

        _config = config;
        Trace.exit(
            this,
            "ExtractorReadOnlyConnection(Connection jdbcConnection, Configuration config, boolean borrowed)");
    }

    /**
     * Constructor used by xmlizer command line. Configuration is built with
     * the default and Extractor is called statically to build MetadataManager.
     * @param jdbcConnection
     * @throws XMLDBCException
     */
    public ExtractorReadOnlyConnection(Connection jdbcConnection)
        throws XMLDBCException {
        Trace.enter(
            this,
            "ExtractorReadOnlyConnection(Connection jdbcConnection)");

        try {
            _jdbcConnection = jdbcConnection;
            _borrowedConnection = true;
            DatabaseMetaData dbm = _jdbcConnection.getMetaData();
            String jdbcUri = dbm.getURL();
            _user = dbm.getUserName();
            String key = jdbcUri + "|" + _user;
            
			Properties properties =
				(Properties) Extractor._defaultProperties.clone();
			Extractor.checkJdbcConnection(
				jdbcConnection.getMetaData(),
				properties);

            _config = (Configuration) Extractor._configMap.get(key);

            if (null == _config) {
                _config = new Configuration();
                _config.properties = properties;

                _config.uri = new XMLDBCUri(URL_PREFIX + jdbcUri);

                /* load only one schema */
                getMetaData(true);
                
                //Debug.print(this.getMetaData().getMetaData().getAsString());
                
                Extractor._configMap.put(key, _config);
            }
        } catch (SQLException ex) {
            throw new XMLDBCException(
                MessageLibrary.getMessage("D_ACC_FAIL", ex.getMessage()),
                ex);
        }

        Trace.exit(
            this,
            "ExtractorReadOnlyConnection(Connection jdbcConnection)");
    }

    public Connection getJdbcConnection() {
        return _jdbcConnection;
    }

    MetaDataManager getMetaDataManagerForAlgebra() {
        return _config.metaDataManager;
    }

    Statement createJdbcStatement() throws SQLException {
        // TODO (YS)  ResultSet.TYPE_SCROLL_INSENSITIVE,ResultSet.CONCUR_READ_ONLY is added to facilitate debug
        Statement retVal = null;
        if (org.xquark.util.misc.Debug.ON || Extractor.USE_SCROLLABLE_RESULTSET) {
            retVal =
                _jdbcConnection.createStatement(
                    ResultSet.TYPE_SCROLL_INSENSITIVE,
                    ResultSet.CONCUR_READ_ONLY);
        } else {
            retVal =
                _jdbcConnection.createStatement(
                    ResultSet.TYPE_FORWARD_ONLY,
                    ResultSet.CONCUR_READ_ONLY);
        }
        return retVal;
    }

    public void close() throws XMLDBCException {
        Trace.enter(this, "close()");
        try {
            _isClosed = true;
            if (!_borrowedConnection) {
                _jdbcConnection.close();
            }
            XMLStatement statement;
            for (int i = _activeStatementList.size() - 1; i > -1; i--) {
                statement = (XMLStatement) _activeStatementList.get(i);
                statement.close();
            }
            Log.message(MessageLibrary.getMessage("S_CLS_CONN"));
            // TODO (YS): close related statement and resultset
        } catch (Exception ex) {}
        Trace.exit(this, "close()");
    }

    public void commit() throws XMLDBCException, XMLDBCNotSupportedException {
        Trace.enter(this, "commit()");
        if (!_isClosed) {
            try {
                _jdbcConnection.commit();
            } catch (Exception ex) {
                throw new XMLDBCException(ex.getMessage(), ex);
            }
        } else {
            throw new XMLDBCException("XMLConnection is already closed.");
        }
        Trace.exit(this, "commit()");
    }

    public boolean isClosed() throws XMLDBCException {
        return _isClosed;
    }

    public String getUserName() throws XMLDBCException {
        return _user;
    }

    public String getURL() throws XMLDBCException {
        if (_config == null || _config.uri == null)
            return null;
        return _config.uri.getUriString();
    }

    void statementClosed(XMLStatement closedStatement) throws XMLDBCException {
        Trace.enter(this, "statementClosed(XMLStatement closedStatement)");
        _activeStatementList.remove(closedStatement);
        Trace.exit(this, "statementClosed(XMLStatement closedStatement)");
    }

	public XMLStatement createStatement() throws XMLDBCException {
		Trace.enter(this, "createStatement()");

		XMLStatement retVal = createStatement(XQUERY_STRING_TYPE);

		Trace.exit(this, "createStatement()");
		return retVal;
	}

    public XMLStatement createStatement(short queryType)
        throws XMLDBCException {
        Trace.enter(this, "createStatement(short queryType)");
        if (isClosed()) {
            throw new XMLDBCException(MessageLibrary.getMessage("CON_CLOSED"));
        }

        ExtractorStatement retVal = null;
        String dbName =
            _config.properties.getProperty(
                ExtractorConfigurationFileHandler.LABEL_SPEC_DATABASE_NAME);
        String dbVersion =
            _config.properties.getProperty(
                ExtractorConfigurationFileHandler.LABEL_SPEC_DATABASE_VERSION);

		/* DBMS version release switch (Statement implementations */
        if (-1 < dbName.indexOf("Oracle")) {
            if (-1 < dbVersion.indexOf("Release 8.1")) {
                retVal =
                    new org.xquark.extractor.oracle.ExtractorStatement(this);
            } else if (-1 < dbVersion.indexOf("Release 7.3.4")) {
                retVal =
                    new org.xquark.extractor.oracle.ExtractorOracle7Statement(
                        this);
            } else {	// default Oracle is Oracle 8i
                retVal =
                    new org.xquark.extractor.oracle.ExtractorStatement(this);
            }
        } else if (-1 < dbName.indexOf("Microsoft")) {
            retVal =
                new org.xquark.extractor.microsoft.ExtractorStatement(this);
        } else if (-1 < dbName.indexOf("Adaptive Server Enterprise")) {
            retVal = new org.xquark.extractor.sybase.ExtractorStatement(this);
        } else if (-1 < dbName.indexOf("PROGRESS")) {
            retVal = new org.xquark.extractor.progress.ExtractorStatement(this);
		} else if (-1 < dbName.indexOf("MySQL")) {
			retVal = new org.xquark.extractor.mysql.ExtractorStatement(this);
        } else {
            //		retVal = new org.xquark.extractor.microsoft.ExtractorStatement(this);
            throw new XMLDBCNotSupportedException(
                MessageLibrary.getMessage("D_N_SUP_DB_VENDOR", dbName));
        }

        retVal.setConfig(_config);
        _activeStatementList.add(retVal);

        Trace.exit(this, "createStatement(short queryType)");
        return retVal;
    }

    public PreparedXMLStatement prepareStatement(String query)
        throws XMLDBCException, XMLDBCNotSupportedException {
        throw new XMLDBCNotSupportedException(
            "PreparedXMLStatement prepareStatement(String query) is not supported.",
            null);
    }

    public PreparedXMLStatement prepareStatement(String query, short queryType)
        throws XMLDBCException, XMLDBCNotSupportedException {
        throw new XMLDBCNotSupportedException("PreparedXMLStatement prepareStatement(String query) is not supported.");
    }

    Configuration getConfiguration() {
        return _config;
    }

    /**
     * Gets the metadata regarding this connection's data source. A
     * connection's data source is able to provide information
     * describing itself.
     * @return an XMLDataSourceMetaData object for this connection.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public XMLDataSourceMetaData getMetaData() throws XMLDBCException {
        return _config.metaDataForParser;
    }

    public XMLDataSourceMetaData getMetaData(boolean refresh)
        throws XMLDBCException {
        Trace.enter(
            this,
            "getMetaData(boolean refresh)",
            refresh ? "Refresh" : "NoRefresh");

        XMLDataSourceMetaData retVal = null;

        if (refresh) {
            try {
                _config.metaDataManager =
                    Extractor.createMetaDataManager(_jdbcConnection, _config);
                _config.metaDataForParser =
                    _config.metaDataManager.buildMetaDataForParser(
                        _config.properties.getProperty(
                            ExtractorConfigurationFileHandler.LABEL_SPEC_NAME));
                _config.metaDataForParser.setConnection(this);
                retVal = _config.metaDataForParser;
            } catch (Exception ex) {
                throw new XMLDBCException(ex.getMessage(), ex);
            }
        } else {
            retVal = _config.metaDataForParser;
        }

        Trace.exit(
            this,
            "getMetaData(boolean refresh)",
            refresh ? "Refresh" : "NoRefresh");
        return retVal;
    }

    protected void finalize() throws Exception {
        close();
    }
}
