/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor;

import java.io.StringReader;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.*;

import org.xquark.extractor.algebra.*;
import org.xquark.extractor.common.*;
import org.xquark.extractor.metadata.MetaDataManager;
import org.xquark.extractor.sql.AbstractGenSqlVisitor;
import org.xquark.extractor.sql.SqlExpression;
import org.xquark.xml.xdbc.*;
import org.xquark.xquery.metadata.MetaDataImpl;
import org.xquark.xquery.metadata.StaticContext;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.parser.util.DependNode;
import org.xquark.xquery.parser.util.TransformXQueryExpression;
import org.xquark.xquery.xdbc.*;

abstract public class ExtractorStatement implements XMLStatementInternal {
    final static String KEY_NAME = "key_name";

    private static final String RCSRevision = "$Revision: 1.3 $";
    private static final String RCSName = "$Name:  $";

    protected ExtractorReadOnlyConnection _connection;
    protected MetaDataImpl _metadata = null;
    protected MetaDataManager _metadataManager = null;
    protected Properties _properties = null;
    protected Map _viewDefinitions = null;
    protected TypeVisitor _typeVisitor = null;
    protected xqueryParser _parser = null;
    protected GenAlgebraVisitor _genAlgebraVisitor = null;
    protected AlgebraTypeVisitor _algebraTypeVisitor = null;
    protected RemoveIntermediateProjectVisitor _removeProjectVisitor = null;
    protected BuildReferredAttributes _brav = null;
    protected AnalyzePredicateVisitor _apv = null;
    protected AbstractGenSqlVisitor _genSqlVisitor = null;
    protected HashMap selections = null;
    protected Connection _jdbcConnection = null;
    protected Statement _jdbcStatement = null;
    protected XMLResultSet _currentResultSet = null;
    protected boolean _hasResultSet = false;
    protected boolean _isClosed = false;

    public ExtractorStatement(ExtractorReadOnlyConnection connection)
        throws XMLDBCException {
        try {
            setConnection(connection);
            setConfig(connection.getConfiguration());

            createParser();
            createVisitors();

            _jdbcConnection = connection._jdbcConnection;
            _jdbcStatement = _connection.createJdbcStatement();
        } catch (Exception ex) {
            throw new XMLDBCException("Can not create statement", ex);
        }
    }

    protected void createParser() throws ParseException, TypeException {
        _parser = new xqueryParser(new StringReader("dummy"));
        _parser.setFactory(
            "org.xquark.xquery.parser.ParserFactoryMediator");
        _parser.setViewDefinitions(_viewDefinitions);
        // change   04/07/2003
        //_typeVisitor = new TypeVisitor(_metadata.getSchemaManager(), null, _metadata);
		_typeVisitor = new TypeVisitor();

    }

    abstract protected void createVisitors();

    protected Statement getJdbcStatement() throws SQLException {
        return _connection.createJdbcStatement();
    }

    public void setConfig(Configuration configuration) {
        _metadata = configuration.metaDataForParser;
        _metadataManager = configuration.metaDataManager;
        _properties = configuration.properties;
        _viewDefinitions = configuration.viewDefinitions;
    }

    /**
     * Sets the value of the _connection property.
     * @param aConnection the new value of the _connection property
     */
    public void setConnection(ExtractorReadOnlyConnection aConnection) {
        _connection = aConnection;
    }
    
    /**
     * Access method for the _connection property.
     * @return   the current value of the _connection property
     */
    public XMLConnection getConnection() {
        return _connection;
    }

    public boolean execute(String query) throws XMLDBCException {
        _currentResultSet = null;
        _currentResultSet = executeXQuery(query);
        return true;
    }

    public boolean execute(String query, int queryType)
        throws XMLDBCException {
        _currentResultSet = null;
        _currentResultSet = executeXQuery(query);
        return true;
    }

    public XMLResultSet executeQuery(String query) throws XMLDBCException {
        _currentResultSet = null;
        _currentResultSet = executeXQuery(query);
        return _currentResultSet;
    }

    public XMLResultSet executeQuery(String query, int queryType)
        throws XMLDBCException {
        if (queryType == XMLConnection.MEDIATOR_STRING_TYPE) // QUESTION : a subset of XQuery ?
            _parser.setRenameVars(false); // TO DO : was to avoid that extractor rename vars used by mediator (
        _currentResultSet = null;
        _currentResultSet = executeXQuery(query);
        return _currentResultSet;
    }

    public XMLDocumentSet executeDocumentQuery(String query)
        throws XMLDBCException {
        throw new XMLDBCNotSupportedException("executeDocumentQuery not yet supported");
    }

    public XMLDocumentSet executeDocumentQuery(String query, int queryType)
        throws XMLDBCException {
        throw new XMLDBCNotSupportedException("executeDocumentQuery not yet supported");
    }

    public XMLResultSet getResultSet() throws XMLDBCException {
        return _currentResultSet;
    }

    public XMLDocumentSet getDocumentSet() throws XMLDBCException {
        throw new XMLDBCNotSupportedException("executeDocumentQuery not yet supported");
    }

    public void close() throws XMLDBCException {
        Trace.enter(this, "close()");
        if (!_isClosed) {
            _metadata = null;
            _metadataManager = null;
            _properties = null;
            _typeVisitor = null;
            _parser = null;
            _genAlgebraVisitor = null;
            _algebraTypeVisitor = null;
            _removeProjectVisitor = null;
            _brav = null;
            _apv = null;
            _genSqlVisitor = null;
            selections = null;
            _jdbcConnection = null;
            try {
                _jdbcStatement.close();
            } catch (Exception ex) {
                Log.exception(ex);
            }
            _jdbcStatement = null;
            if (null != _currentResultSet) {
                _currentResultSet.close();
                _currentResultSet = null;
            }
            _connection.statementClosed(this);
            _isClosed = true;
        }
        Trace.exit(this, "close()");
    }

    public boolean isClosed() throws XMLDBCException {
        return _isClosed;
    }

    /**
     * Execute the Xquery request and return an XML document as result.
     *
     * @param xquery the query to execute.
     * @return the XML result document.
     */
    protected XMLResultSet executeXQuery(String xquery)
        throws XMLDBCException {
        Trace.enter(this, "executeXQueryFile(String xquery)");
        Trace.log(this, "executeXQueryFile(String xquery)", xquery);

        if (isClosed()) {
            throw new XMLDBCException(MessageLibrary.getMessage("ST_CLOSED"));
        }

        XMLResultSet retVal = null;

        long time = System.currentTimeMillis();
        _parser.ReInit(new StringReader(xquery));
        time = System.currentTimeMillis() - time;
        //Debug.print("-------------------- >>>>>>>>  Parser initialisation : "+time+" ms");

        time = System.currentTimeMillis();

		
        XQueryFile result = null;
        try {
			// initializing typevisitor
			StaticContext staticContext = new StaticContext(_metadata, _metadata.getSchemaManager());
			_typeVisitor.setContext(staticContext);
			// parsing request
            result = _parser.Start(_typeVisitor);
        } catch (Exception ex) {
            throw new XMLDBCException(ex.getMessage(), ex);
        }
        time = System.currentTimeMillis() - time;
        //Debug.print("-------------------- >>>>>>>>  Parser execution : "+time+" ms");

        XQueryUnit unit = result.getUnit(0);	// Parser does not consider more than 1 unit

        List exprList = unit.getExpressions();

        if (exprList.size() == 1) {
            retVal = executeXQuery((XQueryExpression) exprList.get(0));
        } else if (1 < exprList.size()) {
            retVal = new XResultSetUnion(exprList, this);
        }
        return retVal;
    }

    public XMLResultSet executeXQuery(XQueryExpression xExpr)
        throws XMLDBCException {
        XMLResultSet retVal = null;

        try {
        	// add test  
        	//System.out.println("_metadata.getCollectionNames() = " + _metadata.getCollectionNames());
			//_metadata.getCollectionRootNodes();
        	//_metadata.getCollectionRootNodes();
        	//_metadata.getDocumentRootNode();        	
			//System.out.println("_metadata.getInputRootNodes() = " + _metadata.getInputRootNodes());
        	
            /* transforms XQueryExpression */
            
            TransformXQueryExpression transformExpr = xExpr.transformExtractor(_typeVisitor);
                
            // NOTES:
            // 1. normalization (no typing considered) AW,
            // 2. typing (for the case it is lost...),
            // 3. restructuration (=normalization considering types) LS,
            // 4. canonization (separates Qmem & Qdb) LS,
            // 5. typing Qdb (will be removed by LS ?),
            // 6. get Bound vars on QMem,
            // 7. patches on Qdb for specific needs of end modules (no agreement could be found).
            
			// Outcoming Qdbs: Flat, without tags but with let (correlation).
			// Multiple unrelated queries are not really processed (1 expression for 1 unit for 1 file).
			// Anyway, has changed in the new version of specs.
            
            ArrayList idVarList = transformExpr.getIdVarList();
            List Qdb = transformExpr.getQDB();
            XQueryExpression Qmem = transformExpr.getQMEM();

            Trace.log("=================< QDBs >=================");
            for (int m = 0; m < Qdb.size(); m++) {
                Trace.log(((Variable) Qdb.get(m)).getExpression());
            }
            Trace.log("=================< QMEM >=================");
            Trace.log(Qmem.toSimpleString());

            XDBCResultSetInterface xdbcResultSet = null;
            DependNode root = transformExpr.getDependMap();

            if (Qdb.size() > 1) {
                if (root != null) {	// Multiple expressions imbricated in return with dependant variables.
                    xdbcResultSet = executeSOU(root, idVarList);
                } else { // For instance, 2 FLWORs side-by-side in an element constructor.

                    int[] idPosition = new int[idVarList.size()];
                    for (int m = 0; m < idPosition.length; m++) {
                        idPosition[m] = m;
                    }

                    List resultSetList = new ArrayList();
                    FLWRExpression xQuery = null;
                    SynchronizedResultSet subResultSet = null;
                    for (int m = 0; m < Qdb.size(); m++) {
                        xQuery =
                            (FLWRExpression) ((Variable) Qdb.get(m))
                                .getExpression();
                        subResultSet = executeSingleQDB(xQuery, true);
                        subResultSet.setIdPosition(idPosition);
                        resultSetList.add(subResultSet);
                    }
                    xdbcResultSet = new UnionResultSet(resultSetList);
                }
            } else if (Qdb.size() == 1) { // Base single query : expression with potential let (correlation).
                xdbcResultSet =
                    executeSingleQDB(
                        ((Variable) Qdb.get(0)).getExpression(),
                        false);
            }

            retVal =
                new XResultSetImpl(
                    xdbcResultSet,
					Qmem,
					idVarList,
                    this);
        } catch (NotSupport ex) {
            throw new XMLDBCNotSupportedException(ex.getMessage(), ex);
        } catch (SqlWrapperException ex) {
            throw new XMLDBCException(ex.getMessage(), ex);
        } catch (AssertionException ex) {
            throw new XMLDBCException(
                MessageLibrary.getMessage("G_INT_LOG_ERR", ex.getMessage()),
                ex);
        } catch (SQLException ex) {
            throw new XMLDBCException(
                MessageLibrary.getMessage("E_ERR_SQL", ex.getMessage()),
                ex);
        } catch (ParseException ex) {
            throw new XMLDBCException(ex.getMessage(), ex);
        } catch (Exception ex) {
            throw new XMLDBCException(ex.getMessage(), ex);
        }

        return retVal;
    }

	/** Performs recursively Qdb (adding the suitable sorts for SOU) following the
	 * dependency tree and build an homothetic result set tree with columns making
	 * row ids positionned in order for reconstruction to synchronize the SOU
	 * execution.
	 * @param node
	 * @param idVarList
	 * @return
	 * @throws Exception
	 */
    protected SynchronizedResultSet executeSOU(DependNode node, List idVarList)
        throws Exception {
        Trace.enter(this, "executeSOU(DependNode node, List idVarList)");

        SynchronizedResultSet retVal = null;
        FLWRExpression xQuery = node.getExpression();

        /* rebuild orderby list */
        ArrayList newSortExprList = new ArrayList();
        List varLevels = node.varLevels();
        List varsOfCurrentLevel = null;
        List sortExprLevels = node.sortExprLevels();
        List sortExprsOfCurrentLevel = null;
        int levelNumber = varLevels.size();
        for (int i = 0; i < varLevels.size(); i++) {
            sortExprsOfCurrentLevel = (List) sortExprLevels.get(i);
            if (null != sortExprsOfCurrentLevel) {
                newSortExprList.addAll(sortExprsOfCurrentLevel);
            }
            varsOfCurrentLevel = (List) varLevels.get(i);
            PlaceHolder placeholder = new PlaceHolder(varsOfCurrentLevel);
            newSortExprList.add(placeholder);
        }
        xQuery.setOrderBy(newSortExprList);

        retVal = executeSingleQDB(xQuery, true);

        /* calculate the position put the ids */
        int[] idPosition =
            calculateIdPosition(node.getExpression().getVariables(), idVarList);
        retVal.setIdPosition(idPosition);

        List childQueryList = node.getSons();
        if (null != childQueryList && !childQueryList.isEmpty()) {
            DependNode childQuery = null;
            List childResultSetList = new ArrayList();
            SynchronizedResultSet childResultSet = null;
            for (int i = 0; i < childQueryList.size(); i++) {
                childQuery = (DependNode) childQueryList.get(i);
                childResultSet = executeSOU(childQuery, idVarList);
                childResultSetList.add(childResultSet);
            }
            retVal.setSlaveList(childResultSetList);
        }

        Trace.exit(this, "executeSOU(DependNode node, List idVarList)");
        return retVal;
    }
	
	/** Retrieves id position in result set by var name.
	 * 
	 * @param varList
	 * @param idVarList
	 * @return
	 */
    private int[] calculateIdPosition(List varList, List idVarList) {
        Trace.enter(this, "calculateIdPosition(List varList, List idVarList)");
        int[] retVal = new int[idVarList.size()];
        int idNum = -1;
        int idPosition = 0;
        int idVarIndex = -1;

        Object var = null;
        for (int i = 0; i < varList.size(); i++) {
            var = varList.get(i);
            idPosition = (int) idVarList.indexOf(var);
            if (-1 != idPosition) {
                /* there is a identifier on this */
                idNum++;
                retVal[idNum] = idPosition;
            }
        }

        Trace.exit(
            this,
            "calculateIdPosition(List varList, List idVarList)",
            retVal.toString());
        return retVal;
    }

    protected SynchronizedResultSet executeSingleQDB(
        XQueryExpression xQuery,
        boolean keyNeeded)
        throws Exception {
        Trace.enter(
            this,
            "executeSingleQDB(XQueryExpression xQuery, boolean keyNeeded)");

        SynchronizedResultSet retVal = null;
        SqlExpression sql = null;
        Expression algebraTree = null;
        Mapper mapper = null;

		/* Algebra tree building */
        _genAlgebraVisitor.reinit();

        if (keyNeeded) {	// Used when Sorted Outer Union is needed to identify rows
            List souArgList = new ArrayList();
            souArgList.add(xQuery);
            _genAlgebraVisitor.noticSortedOuterUnion(souArgList);
        }
        xQuery.accept(_genAlgebraVisitor);	// QUESTION : do we know what subtype can be root (I guess from the XQuery spec)
        algebraTree = _genAlgebraVisitor.getAlgebraTree();

		// QUESTION : what's this visitor ?
		// INFO : YS types of visitors : GenAlgebraVisitor (a ParserVisitor impl),
		// AlgebraVisitor, GenSQLVisitor
        _removeProjectVisitor.reinit();
        algebraTree.accept(_removeProjectVisitor);

		// QUESTION : what's this visitor ?
        algebraTree.accept(_brav);

        /* resolve invalid predicates */
        _apv.reinit();
        algebraTree.accept(_apv);

		// QUESTION : what's this visitor ?
        _algebraTypeVisitor.reinit();
        algebraTree.accept(_algebraTypeVisitor);

        FormatDateTime.formatDateTimes(algebraTree);

        mapper = algebraTree.getMapper();

        _genSqlVisitor.reinit();
        sql = algebraTree.accept(_genSqlVisitor);

        XMLDBCException exception = null;

        List requestList = sql.generateRequestList();
        Trace.log("=================< SQL >=================");
        for (int i = 0; i < requestList.size(); i++) {
            Trace.log(requestList.get(i));
        }
        Trace.log("================< Mapper >================");
        Trace.log(mapper.pprint());

        try {
            retVal =
                new SynchronizedResultSet(
                    requestList,
                    mapper,
                    getJdbcStatement());
        } catch (XMLDBCException ex) {
            exception = ex;
        } catch (SQLException ex) {
            exception = new XMLDBCException(ex.getMessage(), ex);
        }

        try {
            Statement jdbcStatement = getJdbcStatement();
            sql.cleanAll(jdbcStatement);
            jdbcStatement.close();
        } catch (Exception ex) {}

        if (null != exception) {
            throw exception;
        }

        Trace.exit(
            this,
            "executeSingleQDB(XQueryExpression xQuery, boolean keyNeeded)");
        return retVal;
    }


    protected void finalize() throws Exception {
        close();
    }
    
    // does nothing
	public void setBaseURI(String baseURI) {
	}
	public String getBaseURI() {
		return null;
	}
}
