/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor;

import java.sql.*;
import java.util.ArrayList;
import java.util.List;

import org.xml.sax.ContentHandler;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.ext.LexicalHandler;
import org.xml.sax.helpers.AttributesImpl;
import org.xquark.extractor.algebra.MapItem;
import org.xquark.extractor.algebra.Mapper;
import org.xquark.extractor.common.*;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xquery.xdbc.XDBCResultSetInterface;



public final class SynchronizedResultSet implements XDBCResultSetInterface {

	private static final String RCSRevision = "$Revision: 1.3 $";
	private static final String RCSName     = "$Name:  $";

	static final Object NULLID = new Object();
	static final AttributesImpl  NULL_ATTRIBUTE = new AttributesImpl();

	/* a enveloped JDBC ResultSet providing lookAhead capacity */
	private LookAheadJdbcResultSet _lookAheadRes;

	/** a mapper mapping XQuery Expressions(String) to attribute names in the JDBC ResultSet */
	private Mapper _mapper;

	private boolean _hasIdentifiers = false;

	/** identifier values in current tuple. (a identifier may be composed of multiple attributes.
	 * In this case, these attribute values  will be placed in a List )
	 */
	Object[] _aoThisId;

	/* identifier values in next tuple. */
	private Object[] _aoThisNextId;

	/** when getIdentifiers(Object[] ids) or getNextIdentifiers(Object[] ids) is called,
	 * the identifiers in this result set should be put into the arry "ids" at proper position.
	 * _aiIPosition indicate the position where a identier should be placed.
	 */
	private int[] _aiIdPosition;


	/** a list of SychronizedResultSet that depand on this SychronizedResultSet */
	private List _slaveList;

	/** the result set on which this one depends. When this one is root, _father = null. */
	private SynchronizedResultSet _father;


	/** indicate which slave result set is active(it will response getIdentifers(..), generate(..),... ).
	 * Rangs from -1 to _maxSlaveIndex.  -1 means that this result set is active.
	 */
	private int _curSlaveIndex = 0;

	/* indicate the index of the last slave available in <code>_slaveList<code> */
	private int _maxSlaveIndex;

	/* = true this result set is synchronized with its father. */
	private boolean _isSynchronizedWithFather   = false;

	/* = true when this resultset is closed */
	private boolean _isClosed = false;

	private ContentHandler  _contenthandler     = null;
	private LexicalHandler  _lexicalhandler     = null;
	private ErrorHandler    _errorhandler       = null;

	/* a JDBC Statement used to execute SQL <code>requestList<code> */
	private Statement _jdbcStatement = null;

	/**
	 * Constract a new SynchronizedResultSet.
	 * @param requestList a list of SQL Statement.
	 * @param mapper a mapping from XQuery paths to attribute names.
	 * @param jdbcStatement a JDBC Statement used to execute SQL <code>requestList<code>.
	 * this <code>jdbcStatement<code> will be close() either when the all the result is
	 * scaned or when this SynchronizedResultSet is closed.
	 * @throws XMLDBCException  if failed in executing the <code>requestList<code>.
	 */
	public  SynchronizedResultSet(List requestList, Mapper mapper, Statement jdbcStatement) throws XMLDBCException  {
		Trace.enter(this, "SynchronizedResultSet(List requestList, Mapper mapper, Statement jdbcStatement)");
		try {
			ResultSet jdbcResultSet = executeRequests(requestList, jdbcStatement);

			_lookAheadRes = new LookAheadJdbcResultSet(jdbcResultSet);

			SQLWarning warning = null;
			warning = jdbcStatement.getWarnings();
			if (null != warning) {
				throw warning;
			}
			warning = jdbcResultSet.getWarnings();
			if (null != warning) {
				throw warning;
			}
		}
		catch (SQLException ex) {
			throw new XMLDBCException(ex.getMessage(),ex);
		}

		mapper.optimize();
		setMapper(mapper);

		_jdbcStatement = jdbcStatement;
		Trace.exit(this, "SynchronizedResultSet(List requestList, Mapper mapper, Statement jdbcStatement)");
	}

	/**
	 * setter methode of _aiIdPosition.
	 * @param idPosition  when getIdentifiers(Object[] ids) or getNextIdentifiers(Object[] ids) is called,
	 * the identifiers in this result set should be put into the arry "ids" at proper position.
	 * idPosition  indicates the position where a identier should be placed. eg. [0, 2 , 3, 0 , 0] means
	 * the first id in this result set should be the first element in <code>ids<code>, the second id should
	 * be the third element, the third id the fourth element. trailling "0" are ignored.
	 */
	public void setIdPosition(int[] idPosition) {
		_aiIdPosition = idPosition;
	}

	/**
	 * setter method of _mapper
	 * @param mapper
	 */
	public void setMapper(Mapper mapper) {
		_mapper = mapper;

		List idMapItemList = mapper.getIdentifierList();
		if (null != idMapItemList && !idMapItemList.isEmpty()) {
			int idNumber = mapper.getIdentifierList().size();
			_aoThisId = new Object[idNumber];
			_aoThisNextId = new Object[idNumber];

			_hasIdentifiers = true;
		}
		else {
			_hasIdentifiers = false;
		}
	}

	/**
	 * setter method of _father.
	 * @param father
	 */
	void setFather(SynchronizedResultSet father) {
		_father = father;
	}

	/**
	 * setter methode of _slaveList.
	 * @param slaveList
	 */
	public void setSlaveList(List slaveList) {
		_slaveList = slaveList;

		if (null != _slaveList) {
			_maxSlaveIndex = _slaveList.size() - 1;
			SynchronizedResultSet slave = null;
			for (int i = 0; i < _maxSlaveIndex + 1; i++) {
				slave = (SynchronizedResultSet)_slaveList.get(i);
				slave.setFather(this);
			}
		}
	}

	/**
	 * execute a list of SQL request.
	 * @param requestList
	 * @param jdbcStatement
	 * @return a ResultSet contains data return by the last request in <code>requestList<code>.
	 * @throws SQLException
	 */
	protected ResultSet executeRequests(List requestList, Statement jdbcStatement) throws SQLException {
		Trace.enter(this, "executeRequests(List requestList, Statement jdbcStatement)");
		ResultSet retVal = null;

		String request = null;
		int number = requestList.size();
		for (int i = 0; i < number-1; i++) {
			request = (String)requestList.get(i);
			jdbcStatement.executeUpdate(request);
		}
		request = (String)requestList.get(number-1);
		retVal = jdbcStatement.executeQuery(request);

		Trace.exit(this, "executeRequests(List requestList, Statement jdbcStatement)");
		return retVal;
	}

	/**
	 * Releases this SynchronizedResultSet object's database and JDBC resources immediately
	 * instead of waiting for this to happen when it is automatically closed.
	 * @throws XMLDBCException if a database access error occurs
	 */
	public void close() throws XMLDBCException {
		Trace.enter(this, "close()");
		if (!_isClosed) {
			/* close slave result set */
			if (!isLeaf()) {
				SynchronizedResultSet slave = null;
				for (int i = 0; i < _slaveList.size(); i++) {
					slave = (SynchronizedResultSet)_slaveList.get(i);
					slave.close();
				}
			}

			/* close this result set */
			try {
				if (_jdbcStatement!=null) {
					_jdbcStatement.close();
					_jdbcStatement = null;
				}
				if (_lookAheadRes != null) {
					_lookAheadRes.close();
					_lookAheadRes = null;
				}
			}
			catch (SQLException ex) {
				throw new XMLDBCException(ex.getMessage(), ex);
			}
			_isClosed = true;
		}
		Trace.exit(this, "close()");
	}

	public boolean next() throws XMLDBCException {
		Trace.enter(this, "next()");
		boolean retVal = false;

		if (isLeaf()) {
			retVal =  nextInThis();
		}
		else if (isRoot()) {
			retVal = next_internalNode();
		}
		else /* internal node node */ {
			retVal = next_internalNode();
		}

		Trace.exit(this, "next()", retVal ?"true" : "false");
		return retVal;
	}

	/**
	 * Looks for (in this SynchronizedResultSet and this slaves)and move to next result
	 * ( a tuple whoes identifiers correspond to father identifiers when this result set
	 * is not root. if not found, stay in current tuple.
	 * @return true when found.
	 * @throws XMLDBCException
	 */
	protected boolean next_internalNode() throws XMLDBCException {
		Trace.enter(this, "nextAsIntermediate()");
		boolean retVal = false;

		if (!_isSynchronizedWithFather || _curSlaveIndex == -1) {
			retVal = nextInThis();
			if (retVal) {
				_curSlaveIndex = 0;
				retVal = nextInSlaves();
				if (!retVal) {
					_curSlaveIndex= -1;
					retVal = true;
				}
			}
		}
		else {
			retVal = nextInSlaves();
			if (!retVal) {
				retVal = nextInThis();
				if (retVal) {
					_curSlaveIndex = 0;
					retVal = nextInSlaves();
					if (!retVal) {
						_curSlaveIndex= -1;
						retVal = true;
					}
				}
			}
		}
		Trace.exit(this, "nextAsIntermediate()");
		return retVal;
	}


	/**
	 * Looks for (in this SynchronizedResultSet only) and move to next result
	 * ( a tuple whoes identifiers correspond to father identifiers when this
	 * result set is not root. if not found, stay in current tuple.
	 * @return true when found.
	 * @throws XMLDBCException
	 */
	protected boolean nextInThis() throws XMLDBCException {
		Trace.enter(this, "nextInThis()");
		boolean retVal = false;

		if (! isRoot()) {
			/*  this resultset has a father */
			if (_lookAheadRes.hasNext()) {
				_lookAheadRes.lookAhead();
				feedIdentifiers(_aoThisNextId);
				_lookAheadRes.goBack();
				if (isSynchronized(_father._aoThisId, _aoThisNextId)) {
					try {
						_lookAheadRes.next();
					}
					catch (SQLException ex) {
						throw new XMLDBCException(MessageLibrary.getMessage("D_ACC_FAIL",ex.getMessage()), ex);
						/* throw XMLDBCException() */
					}
					for (int i = 0; i < _aoThisNextId.length; i++) {
						_aoThisId[i] = _aoThisNextId[i];
					}
					retVal = true;
				}
			}
		}
		else {
			/* this resultset is the root. */
			if (_lookAheadRes.hasNext()) {
				try {
					_lookAheadRes.next();
				}
				catch (SQLException ex) {
					throw new XMLDBCException(MessageLibrary.getMessage("D_ACC_FAIL",ex.getMessage()), ex);
				}
				feedIdentifiers(_aoThisId);
				retVal = true;
			}
		}

		if (retVal) {
			notifySlaves();
			_isSynchronizedWithFather = true;
		}

		Trace.exit(this, "nextInThis()");
		return retVal;
	}


	/**
	 * Looks for (in this SynchronizedResultSet and this slaves)and move to next result
	 * ( a tuple whoes identifiers correspond to father identifiers when this result set
	 * is not root. if not found, stay in current tuple. if a result is  found in a slave,
	 * point _curSlaveIndex to it.
	 * @return true when found.
	 * @throws XMLDBCException
	 */
	protected boolean nextInSlaves() {
		Trace.enter(this, "nextInSlaves()");
		boolean retVal = false;

		SynchronizedResultSet curSlave = null;
		for (int i=_curSlaveIndex; i<=_maxSlaveIndex; i++) {
			curSlave = (SynchronizedResultSet)_slaveList.get(i);
			try {
				retVal = curSlave.next();
			}
			catch (XMLDBCException ex) {
				continue;
			}
			if (retVal) {
				_curSlaveIndex = i;
				retVal = true;
				break;
			}
		}

		Trace.exit(this, "nextInSlaves()");
		return retVal;
	}


	/** Call <code>fatherIdsChanged()<code> method on all slaves.
	 */
	protected void notifySlaves() {
		Trace.enter(this, "notifySlaves()");
		if (null != _slaveList)	{
			SynchronizedResultSet slave = null;
			for (int i = 0; i < _slaveList.size(); i++)	{
				slave = (SynchronizedResultSet)_slaveList.get(i);
				slave.fatherIdsChanged();
			}
		}
		Trace.exit(this, "notifySlaves()");

	}

	/**
	 * Set _isSynchronizedWithFather to <code>false<code>;
	 * This method should be called by the father whenenver his id values change.
	 */
	protected void fatherIdsChanged() {
		_isSynchronizedWithFather = false;
	}

	public boolean hasNext() throws XMLDBCException{
		Trace.enter(this, "hasNext()");
		if (_isClosed ) throw new XMLDBCException(MessageLibrary.getMessage("RS_CLOSED"));
		boolean retVal = false;

		retVal = _lookAheadRes.hasNext();
		/* we do not care about whether the next tuple is the one to be returned later */

		if (!retVal) {
			SynchronizedResultSet curSlave = null;
			if (_curSlaveIndex != -1 && !isLeaf()) {
				for (int i=_curSlaveIndex; i<=_maxSlaveIndex; i++){
					curSlave = (SynchronizedResultSet)_slaveList.get(i);
					retVal = curSlave.hasNext();
					/* again, we do not care about whether the next tuple is the one to be returned later */
					if (retVal){
						break;
					}
				}
			}
		}

		Trace.exit(this, "hasNext()", retVal ? "true" : "false");
		return retVal;
	}


	public void getIdentifiers(Object[] ids) throws XMLDBCException {
		Trace.enter(this, "getIdentifiers(Object[] ids)");
		if (_isClosed ) throw new XMLDBCException(MessageLibrary.getMessage("RS_CLOSED"));
		if (!_hasIdentifiers) throw new XMLDBCException(MessageLibrary.getMessage("RS_NO_ID"), null);

		if (!isLeaf() && _curSlaveIndex >=0  ) {
			/* return ids in current Slave */
			SynchronizedResultSet curSlave =(SynchronizedResultSet)_slaveList.get(_curSlaveIndex);
			curSlave.getIdentifiers(ids);
		}
		else {
			fillIdsWithCache(ids, _aoThisId);
		}

		StringBuffer idsBuf = new StringBuffer();
		printArray(idsBuf,ids);
		Trace.message(this, "getIdentifiers(Object[] ids) ids:", idsBuf.toString());
		Trace.exit(this, "getIdentifiers(Object[] ids)");
	}

	private void printArray(StringBuffer buffer, Object[] array) {
		buffer.append('[');
		for (int i = 0; i < array.length; i++) {
			Object item = array[i];
			if (item instanceof Object[]) {
				printArray(buffer, array);
			}
			if (null != item) {
				buffer.append(item.toString());
			}

			buffer.append(",\t");
		}
		buffer.append(']');
	}

	/**
	 * Get identifiers values from <code>_lookAheadRes<code>, and put them in to <code>ids<code>.
	 * @param ids
	 * @throws XMLDBCException if no identifiers are provided in this <code>SynchronizedResultSet<code>.
	 */
	protected void feedIdentifiers(Object[] ids) throws XMLDBCException {
		Trace.enter(this, "feedIdentifiers(Object[] ids)");

		List lmiIdItemList  = _mapper.getIdentifierList();
		Object id = null;
		if (null != lmiIdItemList) {
			MapItem idItem = null;
			for (int i = 0; i < lmiIdItemList.size(); i++ )			{
				idItem = (MapItem)lmiIdItemList.get(i);
				id = getIdentifiers(idItem);
				ids[i] = id;
			}
		}

		Trace.exit(this, "feedIdentifiers(Object[] ids)");
	}

	/**
	 * get identifier value at cursor position(<code>BUFFER<code> or <code>RESULTSET<code>) in <code>_lookAheadRes<code>.
	 * @param idItem a MapItem representing a identifier(might).
	 * @return a Object contains identifier value.
	 * @throws XMLDBCException database access failed.
	 */
	protected Object getIdentifiers(MapItem idItem) throws XMLDBCException {
		Trace.enter(this,"getIdentifiers(MapItem idItem)");
		Object retVal = null;

		try {
			if (idItem.isLeafPath()) {
				retVal = _lookAheadRes.getString(idItem.getItemName());
				if (null == retVal) {
					retVal = NULLID;
				}
			}
			else {
				List idAttrs = new ArrayList();
				List children = idItem.getChildItemList();
				for (int j = 0; j < children.size(); j++) {
					MapItem child = (MapItem)children.get(j);
					String attrValue = _lookAheadRes.getString(child.getItemName());
					idAttrs.add(attrValue);
				}
				retVal	= idAttrs;
			}
		}
		catch (SQLException ex) {
			throw new XMLDBCException(MessageLibrary.getMessage("D_ACC_FAIL",
			ex.getMessage()), ex);
		}
		Trace.exit(this,"getIdentifiers(MapItem idItem)");
		return retVal;
	}


	/**
	 * fill the array ids with next (global, logicaly)identifiers in this result set
	 * or slave result set at proper positions.
	 * @param ids
	 * @throws XMLDBCException
	 * this function should be called only when this result set is root.
	 */
	public void getNextIdentifiers(Object[] ids) throws XMLDBCException {
		Trace.enter(this, "getNextIdentifiers(Object[] ids)");
		if (_isClosed ) throw new XMLDBCException(MessageLibrary.getMessage("RS_CLOSED"));
		if (!_hasIdentifiers) throw new XMLDBCException(MessageLibrary.getMessage("RS_NO_ID"), null);

		boolean retVal = false;
		if (!_isSynchronizedWithFather) {
			/* the "cursor" is at before-first*/
			if (_lookAheadRes.lookAhead()) {
				feedIdentifiers(_aoThisNextId);
				_lookAheadRes.goBack();
				if (!isLeaf()) {
					retVal = getNextIdentifiersInSlaves(_aoThisNextId, ids, 0);
					if (!retVal) {
						fillIdsWithCache(ids, _aoThisNextId);
						retVal = true;
					}
				}
				else {
					fillIdsWithCache(ids, _aoThisNextId);
				}
			}
			else {
				/* this result set is empty*/
				throw new XMLDBCException("getNextIdentifiers(Object[] ids) while there isn't next result ",null);
			}
		}
		else {
			if (!isLeaf()) {
				/* retVal = false; */
				retVal = getNextIdentifiersInSlaves(_aoThisId, ids, _curSlaveIndex);
				if (!retVal) {
					if (_lookAheadRes.lookAhead()) {
						feedIdentifiers(_aoThisNextId);
						_lookAheadRes.goBack();
						retVal = getNextIdentifiersInSlaves(_aoThisNextId, ids, 0);
						if (!retVal) {
							fillIdsWithCache(ids, _aoThisNextId);
							retVal = true;
						}
					}
					else {
						/* this result set has no more result empty*/
						throw new XMLDBCException("getNextIdentifiers(Object[] ids) called, while there isn't next result ",null);
					}
				}
			}
			else {
				if (_lookAheadRes.lookAhead()) {
					feedIdentifiers(_aoThisNextId);
					fillIdsWithCache(ids, _aoThisNextId);
					_lookAheadRes.goBack();
				}
				else {
					/* this result set has no more result empty*/
					throw new XMLDBCException("getNextIdentifiers(Object[] ids) called,  while there isn't next result ",null);
				}
			}
		}
		StringBuffer idsBuf = new StringBuffer();
		printArray(idsBuf,ids);
		Trace.message(this, "getIdentifiers(Object[] ids) ids:", idsBuf.toString());
		Trace.exit(this, "getNextIdentifiers(Object[] ids)", ids);
	}


	/**
	 * Looks for next identifiers in this SynchronizedResultSet and its slaves(when there are).
	 * @param fatherIds identifiers whith whom the return values should be synchronized.
	 * @param ids a array where the identifiers found should be put.
	 * @param startIndex begin index of slave resultset from which.
	 * @return true if found.
	 * @throws XMLDBCException
	 */
	protected boolean getNextIdentifiersOf(Object[] fatherIds, Object[] ids) throws XMLDBCException {
		Trace.enter(this, "protected boolean getNextIdentifiersOf(Object[] fatherIds, Object[] ids)");
		boolean retVal = false;

		if (isLeaf()) {
			retVal = getNextIdentifiersOf_leaf(fatherIds, ids);
		}
		else {
			retVal = getNextIdentifiersOf_internalNode(fatherIds, ids);
		}

		Trace.exit(this, "protected boolean getNextIdentifiersOf(Object[] fatherIds, Object[] ids)");
		return retVal;
	}

	/**
	 * Looks for next identifiers in this SynchronizedResultSet and its slaves. This function
	 * should be called only when this result set is not root node.
	 * @param fatherIds identifiers whith whom the return values should be synchronized.
	 * @param ids a array where the identifiers found should be put.
	 * @param startIndex begin index of slave resultset from which.
	 * @return true if found.
	 * @throws XMLDBCException
	 */
	protected boolean getNextIdentifiersOf_internalNode(Object[] fatherIds, Object[] ids) throws XMLDBCException {
		Trace.enter(this, "getNextIdentifiersOf_internalNode(Object[] fatherIds, Object[] ids)");

		boolean retVal = false;
		if (!_isSynchronizedWithFather) {
			if (_lookAheadRes.lookAhead()) {
				feedIdentifiers(_aoThisNextId);
				_lookAheadRes.goBack();
				if (isSynchronized(fatherIds, _aoThisNextId)) {
					retVal = getNextIdentifiersInSlaves(_aoThisNextId, ids, 0);
					if (!retVal) {
						fillIdsWithCache(ids, _aoThisNextId);
						retVal = true;
					}
				}

			}
		}
		else {
			/* this result set is already synchronized with his father */
			retVal = getNextIdentifiersInSlaves(_aoThisId, ids, _curSlaveIndex);
			if (!retVal) {
				if (_lookAheadRes.lookAhead()) {
					feedIdentifiers(_aoThisNextId); /*/** @todo  */
					_lookAheadRes.goBack();
					if (isSynchronized(fatherIds, _aoThisNextId)) {
						retVal = getNextIdentifiersInSlaves(_aoThisNextId, ids, 0);
						if (!retVal) {
							fillIdsWithCache(ids, _aoThisNextId);
							retVal = true;
						}
					}
				}
			}
		}

		Trace.exit(this, "getNextIdentifiersOf_internalNode(Object[] fatherIds, Object[] ids)");
		return retVal;
	}

	protected void fillIdsWithCache(Object[] ids, Object[] cache) {
		/* fill ids with _thisNextId */
		for (int i = 0; i < ids.length; i++) {
			ids[i] = null;
		}

		for (int i = 0; i < cache.length; i++) {
			ids[_aiIdPosition[i]] = cache[i];
		}

	}

	/**
	 * Looks for next identifiers in slave result set.
	 * @param fatherIds identifiers whith whom the return values should be synchronized.
	 * @param ids a array where the identifiers found should be put.
	 * @param startIndex begin index of slave resultset from which.
	 * @return true if found.
	 * @throws XMLDBCException
	 */
	protected boolean getNextIdentifiersInSlaves(Object[] fatherIds, Object[] ids, int startIndex) throws XMLDBCException{
		Trace.enter(this, "getNextIdentifiersInSlaves(Object[] ids)");
		boolean retVal = false;

		SynchronizedResultSet curSlave = null;
		if (startIndex != -1) {
			for (int i = startIndex; i<= _maxSlaveIndex; i++) {
				curSlave = (SynchronizedResultSet)_slaveList.get(i);
				retVal = curSlave.getNextIdentifiersOf(fatherIds, ids);
				if (retVal)	{
					break;
				}
			}
		}

		Trace.exit(this, "getNextIdentifiersInSlaves(Object[] ids)");
		return retVal;
	}


	/**
	 * Looks for next identifiers in this SynchronizedResultSet only.
	 * @param fatherIds identifiers whith whom the return values should be synchronized.
	 * @param ids a array where the identifiers found should be put.
	 * @param startIndex begin index of slave resultset from which.
	 * @return true if found.
	 * @throws XMLDBCException
	 */
	protected boolean getNextIdentifiersOf_leaf(Object[] fatherIds, Object[] ids) throws XMLDBCException {
		Trace.enter(this, "getNextIdentifiersOf_leaf(Object[] fatherIds, Object[] ids)");

		boolean retVal = false;
		if (_lookAheadRes.lookAhead()) {
			feedIdentifiers(_aoThisNextId);
			_lookAheadRes.goBack();
			if (isSynchronized(fatherIds, _aoThisNextId)) {
				fillIdsWithCache(ids, _aoThisNextId);
				retVal = true;
			}
		}

		Trace.exit(this, "getNextIdentifiersOf_leaf(Object[] fatherIds, Object[] ids)");
		return retVal;
	}

	/**
	 * Implements <code>generate(String path,int nodeAccessor,String loopID)<code>.
	 * @param path
	 * @param nodeAccessor
	 * @param loopID
	 * @throws XMLDBCException
	 */
	public void generate(String path,int nodeAccessor,String loopID) throws XMLDBCException {
		generate(path,nodeAccessor);
	}
	public void generate(String path,int nodeAccessor) throws XMLDBCException {
		/* argument <code> String loopID <code> is always ignored */
		Trace.enter(this, "generate(String path,int nodeAccessor,String loopID)", path);
		if (_isClosed ) throw new XMLDBCException(MessageLibrary.getMessage("RS_CLOSED"));

		if (_contenthandler != null) {
			if (_curSlaveIndex != -1 && !isLeaf()) {
				/* a slave is active */
				SynchronizedResultSet curSlave = (SynchronizedResultSet)_slaveList.get(_curSlaveIndex);
				curSlave.generate(path, nodeAccessor);
			}
			else {
				/* this result set is active */
				MapItem mapItem  = _mapper.map(path);
				Debug.assertTrue(null!=mapItem, "null!=mapItem");
				switch (nodeAccessor) {
					case 0:
						generate(mapItem);
						break;
					case XDBCResultSetInterface.NODE_ACCESSOR_TEXT:
						/* mapItem should represent a element of simple type */
						String value = fetch(mapItem);
						if (null != value) {
							try {
								_contenthandler.characters(value.toCharArray(), 0, value.length());
							}
							catch (SAXException ex) {
								throw new XMLDBCException(ex.getMessage(), ex);
							}
						}
						break;
					case XDBCResultSetInterface.NODE_ACCESSOR_NODE:
						if (mapItem.isLeafPath()) {
							generate(mapItem);
						}
						else {
							List childItemList = mapItem.getChildItemList();
							for (int i = 0; i < childItemList.size(); i++) {
								mapItem = (MapItem)childItemList.get(i);
								generate(mapItem);
							}
						}
						break;
					default:
						throw new XMLDBCException("Internal error : error argument in function call");
				}
			}
		}
		Trace.exit(this, "generate(String path,int nodeAccessor,String loopID)", path);
	}

	/**
	 * Generates XML data for given <code>MapItem<code>.
	 * @param mapItem
	 * @throws XMLDBCException
	 */
	protected void generate(MapItem mapItem) throws XMLDBCException {
		Trace.enter(this, "generate(MapItem mapItem)");
		try {
			String tag = mapItem.getElementName();
			String namespace = mapItem.getNamespace();
			if (mapItem.isLeafPath()) {
				/* mapItem represent an element of simple type */
				String value =  _lookAheadRes.getString(mapItem.getItemName());
				if (value != null) {
					if (null != tag) {
						_contenthandler.startElement((namespace==null)?"":namespace, tag, "", NULL_ATTRIBUTE);
						_contenthandler.characters(value.toCharArray(), 0, value.length());
						_contenthandler.endElement(namespace, tag, "");
					}
					else {
						_contenthandler.characters(value.toCharArray(), 0, value.length());
					}
				}
			}
			else {
				/* mapItem represent an element of complex type */
				List childItemList = mapItem.getChildItemList();
				_contenthandler.startElement((namespace==null)?"":namespace, tag ,"", NULL_ATTRIBUTE);
				for (int i = 0; i < childItemList.size(); i++) {
					generate((MapItem) childItemList.get(i));
				}
				_contenthandler.endElement(namespace, tag ,"");
			}
		}
		catch (Exception ex) {
			throw new XMLDBCException(ex.getMessage(),ex);
		}
		Trace.exit(this, "localGenerate(String path)");
	}

	/**
	 * Implements <code>fetch(String path, int nodeAccessor, String loopID)<code> method.
	 * @param path
	 * @param nodeAccessor
	 * @param loopID
	 * @return
	 * @throws XMLDBCException
	 */
	public String fetch(String path, int nodeAccessor, String loopID) throws XMLDBCException {
		return fetch(path, nodeAccessor);
	}
	public String fetch(String path, int nodeAccessor) throws XMLDBCException {
		Trace.enter(this, "fetch(String path, int nodeAccessor, String loopID)", path);
		if (_isClosed ) throw new XMLDBCException(MessageLibrary.getMessage("RS_CLOSED"));
		String retVal = null;


		if (_curSlaveIndex != -1 && !isLeaf()) {
			/* a slave is active */

			SynchronizedResultSet curSlave = (SynchronizedResultSet)_slaveList.get(_curSlaveIndex);
			retVal = curSlave.fetch(path, nodeAccessor);
		}
		else {
			/* this result set is active */
			MapItem mapItem = _mapper.map(path);
			if (mapItem != null) {
				retVal = fetch(mapItem);
			}
			else {
				throw new XMLDBCException("(fetch) This path \"" + path + "\" can not be found in the result set.", null);
			}
		}
		Trace.exit(this, "fetch(String path, int nodeAccessor, String loopID)", path);
		return retVal;
	}

	/**
	 * Get the value(no tag, no attribute) of a path represented by <code>mapItem<code>.
	 * @param mapItem
	 * @return
	 * @throws XMLDBCException
	 */
	protected String fetch(MapItem mapItem) throws XMLDBCException {
		Trace.enter(this, "fetch(MapItem mapItem)");
		String retVal = null;
		try {
			retVal = _lookAheadRes.getString(mapItem.getItemName());
		}
		catch (SQLException ex) {
			throw new XMLDBCException(MessageLibrary.getMessage("D_ACC_FAIL"), ex);
		}

		Trace.exit(this, "fetch(MapItem mapItem)");
		return retVal;
	}

	/**
	 * Return turn if <code>thisId<code> is synchronized with <code>fatherId<code>.
	 * @param fatherId
	 * @param thisId
	 * @return turn if <code>thisId<code> is synchronized with <code>fatherId<code>.
	 */
	protected boolean isSynchronized(Object[] fatherId, Object[] thisId) {
		Trace.enter(this, "isSynchronized(Object[] fatherId, Object[] thisId)");
		boolean retVal = true;

		if (null != fatherId && null != thisId) {
			for (int i = 0; i < fatherId.length; i++) {
				retVal = fatherId[i].equals(thisId[i]);
				if (!retVal) {
					break;
				}
			}
		}
		else {
			throw new SqlWrapperException(
			MessageLibrary.getMessage("IE_INV_ARG","isSynchronized(Object[] fatherId, Object[] thisId)"),
			null);
		}

		Trace.exit(this, "isSynchronized(Object[] fatherId, Object[] thisId)");
		return retVal;
	}

	public boolean isRoot()	{
		return null == _father;
	}

	public boolean isLeaf()	{
		return null == _slaveList;
	}

	public void setContentHandler(ContentHandler handler) {
		if (handler == null) throw new NullPointerException("ContentHandler can not be null.");
		this._contenthandler = handler;

		if (!isLeaf()) {
			SynchronizedResultSet slave = null;
			for (int i = 0; i < _slaveList.size(); i++) {
				slave = (SynchronizedResultSet)_slaveList.get(i);
				slave.setContentHandler(_contenthandler);
			}
		}
	}


	public void setLexicalHandler(LexicalHandler handler) {
		if (handler == null) throw new NullPointerException("LexicalHandler can not be null.");
		this._lexicalhandler = handler;

		if (!isLeaf()) {
			SynchronizedResultSet slave = null;
			for (int i = 0; i < _slaveList.size(); i++) {
				slave = (SynchronizedResultSet)_slaveList.get(i);
				slave.setLexicalHandler(_lexicalhandler);
			}
		}
	}


	public void setErrorHandler(ErrorHandler handler) {
		if (handler == null) throw new NullPointerException("ErrorHandler can not be null.");
		this._errorhandler = handler;

		if (!isLeaf()) {
			SynchronizedResultSet slave = null;
			for (int i = 0; i < _slaveList.size(); i++) {
				slave = (SynchronizedResultSet)_slaveList.get(i);
				slave.setErrorHandler(_errorhandler);
			}
		}
	}


}
