/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor.algebra;

import java.sql.Types;
import java.util.ArrayList;
import java.util.List;

import org.xquark.extractor.common.Debug;
import org.xquark.extractor.common.SqlWrapperException;
import org.xquark.extractor.common.Trace;
import org.xquark.extractor.metadata.DataType;
import org.xquark.extractor.metadata.DbPrimitiveType;
import org.xquark.extractor.xfunctions.*;


/**
 * Browse an expression tree to set the sql type field.
 * 1. SqlTypeAtom are built for attributes basing on database metadata and
 * on XML type for literals.
 * 2. SqlTypeStructure are built for simple relations.
 * 3. and SqlTypeCombinedStructure for joins.
 * 4. Predefined types are used for other algebra nodes, e.g., logical expressions.
 */
public final class AlgebraTypeVisitor extends DefaultAlgebraVisitor implements Constants {
	private static final String RCSRevision = "$Revision: 1.3 $";
	private static final String RCSName     = "$Name:  $";

	private DbPrimitiveType ptArray = new DbPrimitiveType(Types.ARRAY);
	private DbPrimitiveType ptBigInt = new DbPrimitiveType(Types.BIGINT);
	private DbPrimitiveType ptBinary = new DbPrimitiveType(Types.BINARY);
	private DbPrimitiveType ptBlob = new DbPrimitiveType(Types.BLOB);
	private DbPrimitiveType ptBit = new DbPrimitiveType(Types.BIT);
	private DbPrimitiveType ptBoolean = new DbPrimitiveType(DbPrimitiveType.BOOLEAN);
	private DbPrimitiveType ptChar = new DbPrimitiveType(Types.CHAR);
	private DbPrimitiveType ptClob = new DbPrimitiveType(Types.CLOB);
	private DbPrimitiveType ptDate = new DbPrimitiveType(Types.DATE);
	private DbPrimitiveType ptDecimal = new DbPrimitiveType(Types.DECIMAL);
	private DbPrimitiveType ptDistinct = new DbPrimitiveType(Types.DISTINCT);
	private DbPrimitiveType ptDouble = new DbPrimitiveType(Types.DOUBLE);
	private DbPrimitiveType ptFloat = new DbPrimitiveType(Types.FLOAT);
	private DbPrimitiveType ptInteger = new DbPrimitiveType(Types.INTEGER);
	private DbPrimitiveType ptJavaObject = new DbPrimitiveType(Types.JAVA_OBJECT);
	private DbPrimitiveType ptLongVarBanary = new DbPrimitiveType(Types.LONGVARBINARY);
	private DbPrimitiveType ptLongVarChar = new DbPrimitiveType(Types.LONGVARCHAR);
	private DbPrimitiveType ptNull = new DbPrimitiveType(Types.NULL);
	private DbPrimitiveType ptNumeric = new DbPrimitiveType(Types.NUMERIC);
	private DbPrimitiveType ptOther = new DbPrimitiveType(Types.OTHER);
	private DbPrimitiveType ptReal = new DbPrimitiveType(Types.REAL);
	private DbPrimitiveType ptRef = new DbPrimitiveType(Types.REF);
	private DbPrimitiveType ptSmallInt = new DbPrimitiveType(Types.SMALLINT);
	private DbPrimitiveType ptStruct = new DbPrimitiveType(Types.STRUCT);
	private DbPrimitiveType ptTime = new DbPrimitiveType(Types.TIME);
	private DbPrimitiveType ptTimeStamp = new DbPrimitiveType(Types.TIMESTAMP);
	private DbPrimitiveType ptTinyInt = new DbPrimitiveType(Types.TINYINT);
	private DbPrimitiveType ptVarBinary = new DbPrimitiveType(Types.VARBINARY);
	private DbPrimitiveType ptVarChar = new DbPrimitiveType(Types.VARCHAR);

	public AlgebraTypeVisitor() {
	}

	public void reinit () {
	}

	public void visit(Expression arg) throws SqlWrapperException {
		Trace.enter(this,"visit(Expression arg)", arg);
		Debug.assertTrue (false , "NYI!!" + arg.toString());
		Trace.exit(this,"visit(Expression arg)", arg);
	}

	public void visit(Attribute arg) throws SqlWrapperException {
		Trace.enter(this,"visit(Attribute arg)", arg);

		DataType dType = arg.getAttribute().getType();
		String  attrName = arg.getAttribute().getName();
		SqlType type = new SqlTypeAtom(dType,attrName,SqlType.ONE);
		arg.setType(type);

		Trace.exit(this,"visit(Attribute arg)", arg);
	}

	public void visit(Rowid arg) throws SqlWrapperException {
		Trace.enter(this,"visit(OracleRowid arg)", arg);

		SqlTypeAtom type = new SqlTypeAtom(new DbPrimitiveType(DbPrimitiveType.ORACLE_ROWID), arg.getName());
		arg.setType(type);

		Trace.exit(this,"visit(OracleRowid arg)", arg);
	}

	public void visit(AttributeExpression arg) throws SqlWrapperException {
		Trace.enter(this,"visit(AttributeExpression arg)", arg);
		SqlTypeAtom type = null;

		Expression underlyinExpr = arg.getUnderlyinExpr();
		if (underlyinExpr != null && !arg.getName().equals("*")) {
			// AttributeExpression may be created during rewritting without XQuery expression...
			underlyinExpr.accept(this);
			type = (SqlTypeAtom)underlyinExpr.getType().clone();
			type.setName(arg.getAttribute());
		}

		arg.setType(type);

		Trace.exit(this,"visit(AttributeExpression arg)", arg);
	}


//    public void visit(BinaryOperator arg) throws SqlWrapperException {
//        Trace.enter(this,"visit(BinaryOperator arg)", arg);
//        Debug.assertTrue (false , "NYI!!");
//        Trace.exit(this,"visit(BinaryOperator arg)", arg);
//    }

//    public void visit(BinaryAtomicOp arg) throws SqlWrapperException {
//        Trace.enter(this,"visit(BinaryAtomicOp arg)", arg);
//        Debug.assertTrue (false , "NYI!!");
//        Trace.exit(this,"visit(BinaryAtomicOp arg)", arg);
//    }

//    public void visit(BinaryAlgebra arg) throws SqlWrapperException {
//        Trace.enter(this,"visit(BinaryAlgebra arg)", arg);
//        Debug.assertTrue (false , "NYI!!");
//        Trace.exit(this,"visit(BinaryAlgebra arg)", arg);
//    }

	public void visit(BinOpArithmetic arg) throws SqlWrapperException {
		Trace.enter(this,"visit(BinOpArithmetic arg)", arg);

		Expression lOprnd = arg.getLeftOperand();
		Expression rOprnd = arg.getRightOperand();
		lOprnd.accept(this);
		rOprnd.accept(this);

		SqlType type = new SqlTypeAtom( ptDouble,null,SqlType.ONE);
		arg.setType(type);

		Trace.exit(this,"visit(BinOpArithmetic arg)", arg);
	}

	public void visit(BinOpBoolean arg) throws SqlWrapperException {
		Trace.enter(this,"visit(BinOpBoolean arg)", arg);

		Expression lOprnd = arg.getLeftOperand();
		Expression rOprnd = arg.getRightOperand();
		lOprnd.accept(this);
		rOprnd.accept(this);

		DbPrimitiveType pType = new DbPrimitiveType(Types.BIT);
		SqlType type = new SqlTypeAtom( ptBit,null,SqlType.ONE);
		arg.setType(type);

		Trace.exit(this,"visit(BinOpBoolean arg)", arg);
	}

	public void visit(BinOpCompare arg) throws SqlWrapperException {
		Trace.enter(this,"visit(BinOpCompare arg)", arg);

		Expression lOprnd = arg.getLeftOperand();
		Expression rOprnd = arg.getRightOperand();
		lOprnd.accept(this);
		rOprnd.accept(this);

		SqlType type = new SqlTypeAtom( ptBit,null,SqlType.ONE);
		arg.setType(type);

		Trace.exit(this,"visit(BinOpCompare arg)", arg);
	}

	public void visit(BinOpCompareAny arg) throws SqlWrapperException {
		Trace.enter(this,"visit(BinOpCompareAny arg)", arg);

		Expression lOprnd = arg.getLeftOperand();
		Expression rOprnd = arg.getRightOperand();
		lOprnd.accept(this);
		rOprnd.accept(this);

		SqlType type = new SqlTypeAtom( ptBit,null,SqlType.ONE);
		arg.setType(type);

		Trace.exit(this,"visit(BinOpCompareAny arg)", arg);
	}

	public void visit(BinOpDifference arg) throws SqlWrapperException {
		Trace.enter(this,"visit(BinOpDifference arg)", arg);

		Expression lOprnd = arg.getLeftOperand();
		Expression rOprnd = arg.getRightOperand();
		lOprnd.accept(this);
		rOprnd.accept(this);

		SqlType type = (SqlType)lOprnd.getType().clone();
		arg.setType(type);

		Trace.exit(this,"visit(BinOpDifference arg)", arg);
	}

	public void visit(BinOpIntersect arg) throws SqlWrapperException {
		Trace.enter(this,"visit(BinOpIntersect arg)", arg);

		Expression lOprnd = arg.getLeftOperand();
		Expression rOprnd = arg.getRightOperand();
		lOprnd.accept(this);
		rOprnd.accept(this);

		SqlType type = (SqlType)lOprnd.getType().clone();
		arg.setType(type);

		Trace.exit(this,"visit(BinOpIntersect arg)", arg);
	}

	public void visit(BinOpMerge arg) throws SqlWrapperException {
		Trace.enter(this,"visit(BinOpMerge arg)", arg);

		Expression lOprnd = arg.getLeftOperand();
		Expression rOprnd = arg.getRightOperand();
		lOprnd.accept(this);
		rOprnd.accept(this);

		SqlType type = (SqlType)lOprnd.getType().clone();
		arg.setType(type);

		Trace.exit(this,"visit(BinOpMerge arg)", arg);
	}

	public void visit(BinOpOuterJoin arg) throws SqlWrapperException {
		Trace.enter(this,"visit(BinOpOuterJoin arg)", arg);

		Trace.message(this,"visit(BinOpOuterJoin arg)","Typing operandes...");
		Expression lOprnd = arg.getLeftOperand();
		Expression rOprnd = arg.getRightOperand();
		lOprnd.accept(this);
		rOprnd.accept(this);

		SqlType leftType = (SqlType)lOprnd.getType().clone();
		SqlType rightType = (SqlType)rOprnd.getType().clone();
		SqlTypeCombinedStructure type = new SqlTypeCombinedStructure();
		type.addStructure(leftType);
		type.addStructure(rightType);
		arg.setType(type);

		Trace.message(this,"visit(BinOpOuterJoin arg)","Typing predicates...");
		List predicateList = arg.getPredicateList();
		if (null != predicateList) {
			Expression predicate = null;
			for (int i = 0; i < predicateList.size(); i++) {
				predicate = (Expression)predicateList.get(i);
				predicate.accept(this);
			}
		}

		Trace.exit(this,"visit(BinOpOuterJoin arg)", arg);
	}

	public void visit(BinOpUnion arg) throws SqlWrapperException {
		Trace.enter(this,"visit(BinOpUnion arg)", arg);

		Expression lOprnd = arg.getLeftOperand();
		Expression rOprnd = arg.getRightOperand();
		lOprnd.accept(this);
		rOprnd.accept(this);

		SqlType type = (SqlType)lOprnd.getType().clone();
		arg.setType(type);

		Trace.exit(this,"visit(BinOpUnion arg)", arg);
	}

	public void visit(DummyTable arg) throws SqlWrapperException {
		Trace.enter(this,"visit(DummyTable arg)", arg);

		arg.setType(new SqlTypeAtom(null,null,SqlType.MANY));

		Trace.exit(this,"visit(DummyTable arg)", arg);
	}


	public void visit(FunAggregate arg) throws SqlWrapperException {
		Trace.enter(this,"visit(FunAggregate arg)", arg);

		Expression oprnd = arg.getOperand();
		oprnd.accept(this);

		SqlType type = null;
		int oprtr = arg.getOperator();
		switch (oprtr) {
			case AVG:
				type = new SqlTypeAtom( ptDouble,null,SqlType.ONE);
				break;
			case COUNT :
				type = new SqlTypeAtom( ptInteger,null,SqlType.ONE);
				break;
			case MIN:
				type = (SqlType)oprnd.getType().clone();
				break;
			case MAX:
				type = (SqlType)oprnd.getType().clone();
				break;
			case SUM:
				type = (SqlType)oprnd.getType().clone();
				break;
			default:
				Debug.assertTrue(false,"internal error");
				break;
		}
		arg.setType(type);
		Trace.exit(this,"visit(FunAggregate arg)", arg);
	}

	public void visit(IfThenElse arg) throws SqlWrapperException {
		Trace.enter(this,"visit(IfThenElse arg)", arg);
		Expression iif = arg.getIf();
		iif.accept(this);

		Expression tthen = arg.getThen();
		tthen.accept(this);

		Expression eelse = arg.getElse();
		eelse.accept(this);
		/** @todo check type : iff : boolean, tthen = eelse */

		int tthenCode = ((SqlTypeAtom)tthen.getType()).getType().getPrimitiveTypeCode();
		int eelseCode = ((SqlTypeAtom)eelse.getType()).getType().getPrimitiveTypeCode();
		Debug.assertTrue(tthenCode == eelseCode, "tthenCode == eelseCode");

		arg.setType( (SqlTypeAtom)tthen.getType().clone());

		Trace.exit(this,"visit(IfThenElse arg)", arg);
	}

	public void visit(Join arg) throws SqlWrapperException {
		Trace.enter(this,"visit(Join arg)", arg);

		Trace.message(this,"visit(Join arg)","Typing operands...");
		List operandList = arg.getOperandList();
		SqlType type = null;
		Expression operand = null;
		if (1 < operandList.size()) {
			type = new SqlTypeCombinedStructure();
			for (int i = 0; i < operandList.size(); i++) {
				operand = (Expression)operandList.get(i);
				operand.accept(this);
				SqlType operandType = (SqlType)operand.getType().clone();
				((SqlTypeCombinedStructure)type).addStructure(operandType);
			}
		}
		else {
			operand = (Expression)operandList.get(0);
			operand.accept(this);
			type = (SqlType)operand.getType().clone();
		}
		arg.setType(type);

		Trace.message(this,"visit(Join arg)","Typing predicates...");
		List predicateList = arg.getPredicateList();
		if (null != predicateList) {
			Expression predicate = null;
			for (int i = 0; i < predicateList.size(); i++) {
				predicate = (Expression)predicateList.get(i);
				predicate.accept(this);
			}
		}

		Trace.exit(this,"visit(Join arg)", arg);
	}

//    public void visit(Literal arg) throws SqlWrapperException {
//        Trace.enter(this,"visit(Literal arg)", arg);
//        Trace.exit(this,"visit(Literal arg)", arg);
//    }

	public void visit(LitBoolean arg) throws SqlWrapperException {
		Trace.enter(this,"visit(LitBoolean arg)", arg);
		SqlType type = new SqlTypeAtom( ptBit,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this,"visit(LitBoolean arg)", arg);
	}

	public void visit(LitDate arg) throws SqlWrapperException {
		Trace.enter(this,"visit(LitDate arg)", arg);
		SqlType type = new SqlTypeAtom( ptDate,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this,"visit(LitDate arg)", arg);
	}

	public void visit(LitDecimal arg) throws SqlWrapperException{
		Trace.enter(this,"visit(LitDecimal arg)", arg);
		SqlType type = new SqlTypeAtom( ptDecimal,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this,"visit(LitDecimal arg)", arg);
	}

	public void visit(LitDouble arg) throws SqlWrapperException {
		Trace.enter(this,"visit(LitDouble arg)", arg);
		SqlType type = new SqlTypeAtom( ptDouble,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this,"visit(LitDouble arg)", arg);
	}

	public void visit(LitFloat arg) throws SqlWrapperException {
		Trace.enter(this,"visit(LitFloat arg)", arg);
		SqlType type = new SqlTypeAtom( ptFloat,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this,"visit(LitFloat arg)", arg);
	}

	public void visit(LitInteger arg) throws SqlWrapperException {
		Trace.enter(this,"visit(LitInteger arg)", arg);
		SqlType type = new SqlTypeAtom( ptInteger,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this,"visit(LitInteger arg)", arg);
	}


	public void visit(LitList arg) throws SqlWrapperException {
		Trace.enter(this,"visit(LitList arg)", arg);
		SqlTypeAtom type = null;
		List list = arg.getLitList();
		if (null != list && !list.isEmpty()) {
			Literal element = (Literal)list.get(0);
			element.accept(this);
			type = (SqlTypeAtom)element.getType();
			for (int i = 1; i < list.size(); i++) {
				element = (Literal)list.get(i);
				element.setType(type);
			}
			type = new SqlTypeAtom(type.getType().getPrimitiveType(), null, SqlType.MANY);
		}
		else {
			type = new SqlTypeAtom(ptNull, null, SqlType.MANY);
		}

		arg.setType(type);
		Trace.exit(this,"visit(LitList arg)", arg);
	}

	public void visit(LitString arg) throws SqlWrapperException {
		Trace.enter(this,"visit(LitString arg)", arg);
		SqlType type = new SqlTypeAtom( ptChar,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this,"visit(LitString arg)", arg);
	}

	public void visit(LitNull arg) throws SqlWrapperException {
		Trace.enter(this,"visit(LitNull arg)", arg);
		SqlType type = new SqlTypeAtom( ptNull,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this,"visit(LitNull arg)", arg);
	}

	public void visit(RenameItem arg) throws SqlWrapperException {
		Trace.enter(this,"visit(RenameItem arg)", arg);

		Expression operand = arg.getOperand();
		operand.accept(this);
		SqlTypeAtom type = (SqlTypeAtom)operand.getType();
		type.setName(arg.getName());
		arg.setType(type);

		Trace.exit(this,"visit(RenameItem arg)", arg);
	}

	public void visit(RenameRelation arg) throws SqlWrapperException {
		Trace.enter(this,"visit(RenameRelation arg)", arg);

		Expression operand = arg.getOperand();
		operand.accept(this);
		SqlType oprndType = operand.getType();
		SqlType type = null;
		if (oprndType instanceof SqlTypeAtom) {
			type = (SqlTypeAtom)oprndType.clone();
			((SqlTypeAtom)type).setRelationName(arg.getName());
		}
		else if (oprndType instanceof SqlTypeStructure) {
			type = (SqlTypeStructure)oprndType.clone();
			((SqlTypeStructure)type).setName(arg.getName());
		}
		else if (oprndType instanceof SqlTypeCombinedStructure ) {
			List attrList = ((SqlTypeCombinedStructure)oprndType).getAttributes();
			type = new SqlTypeStructure(arg.getName(),attrList,SqlType.MANY );
		}

		arg.setType(type);

		Trace.exit(this,"visit(RenameRelation arg)", arg);
	}

	public void visit(SortSpecification arg) throws SqlWrapperException {
		Trace.enter(this,"visit(SortSpecification arg)", arg);

		Expression expr = arg.getSortExpression();
		expr.accept(this);
		arg.setType(expr.getType());

		Trace.exit(this,"visit(SortSpecification arg)", arg);
	}

	public void visit(Table arg) throws SqlWrapperException {
		Trace.enter(this,"visit(Table arg)", arg);

		List attrList = arg.getAttributes();
		List attrTypeList = new ArrayList();
		SqlTypeStructure type = new SqlTypeStructure();
		type.setMultiplicity(SqlType.MANY);

		for (int i = 0; i < attrList.size(); i++) {
			Expression attr = (Expression)attrList.get(i);
			attr.accept(this);
			SqlTypeAtom attrType = (SqlTypeAtom)attr.getType();
			type.addAttribute(attrType);
		}
		arg.setType(type);

		Trace.exit(this,"visit(Table arg)", arg);
	}

	public void visit(TempValue arg) throws SqlWrapperException {
		Trace.enter(this,"visit(TempValue arg)", arg);

		Expression operand = arg.getOperand();
		operand.accept(this);

		SqlType type = operand.getType();
		arg.setType(type);

		Trace.exit(this,"visit(TempValue arg)", arg);
	}

//    public void visit(UnaryOperator arg) throws SqlWrapperException {
//        Trace.enter(this,"visit(UnaryOperator arg)", arg);
//        Trace.exit(this,"visit(UnaryOperator arg)", arg);
//    }

//    public void visit(UnaryAlgebra arg) throws SqlWrapperException {
//        Trace.enter(this,"visit(UnaryAlgebra arg)", arg);
//        Trace.exit(this,"visit(UnaryAlgebra arg)", arg);
//    }

//    public void visit(UnaryAtomicOp arg) throws SqlWrapperException {
//        Trace.enter(this,"visit(UnaryAtomicOp arg)", arg);
//        Trace.exit(this,"visit(UnaryAtomicOp arg)", arg);
//    }

//    public void visit(UnOpAggregate arg) throws SqlWrapperException {
//        Trace.enter(this,"visit(UnaryAtomicOp arg)", arg);
//        Trace.exit(this,"visit(UnaryAtomicOp arg)", arg);
//    }


	public void visit(UnOpAggregate arg) throws SqlWrapperException {
		Trace.enter(this,"visit(UnOpAggregate arg)", arg);
		visit((UnOpProject)arg);
		SqlType  type = arg.getType();
		type.setMultiplicity(SqlType.ONE);
		/* arg.setType(type); */

		Trace.exit(this,"visit(UnOpAggregate arg)", arg);
	}

	public void visit(UnOpExists arg) throws SqlWrapperException {
		Trace.enter(this,"visit(UnOpExists arg)", arg);

		Expression operand = arg.getOperand();
		operand.accept(this);

		SqlType type = new SqlTypeAtom( ptBit,null,SqlType.ONE);
		arg.setType(type);

		Trace.exit(this,"visit(UnOpExists arg)", arg);
	}


	public void visit(UnOpIsNull arg) throws SqlWrapperException {
		Trace.enter(this,"visit(UnOpIsNull arg)", arg);

		Expression operand = arg.getOperand();
		operand.accept(this);

		SqlType type = new SqlTypeAtom( ptBit,null,SqlType.ONE);
		arg.setType(type);

		Trace.exit(this,"visit(UnOpIsNull arg)", arg);
	}

	public void visit(UnOpGroup arg) throws SqlWrapperException {
		Trace.enter(this,"visit(UnOpGroup arg)", arg);

		Expression operand = arg.getOperand();
		operand.accept(this);

		SqlType type = (SqlType)operand.getType().clone();
		arg.setType(type);

		Trace.exit(this,"visit(UnOpGroup arg)", arg);
	}

	public void visit(UnOpMinus arg) throws SqlWrapperException {
		Trace.enter(this,"visit(UnOpMinus arg)", arg);

		Expression operand = arg.getOperand();
		operand.accept(this);

		SqlType type = operand.getType();
		arg.setType(type);

		Trace.exit(this,"visit(UnOpMinus arg)", arg);
	}

	public void visit(UnOpNot arg) throws SqlWrapperException {
		Trace.enter(this,"visit(UnOpNot arg)", arg);

		Expression operand = arg.getOperand();
		operand.accept(this);

		SqlType type = new SqlTypeAtom( ptBit,null,SqlType.ONE);
		arg.setType(type);

		Trace.exit(this,"visit(UnOpNot arg)", arg);
	}

	public void visit(UnOpProject arg) throws SqlWrapperException {
		Trace.enter(this,"visit(UnOpProject arg)", arg);

		Trace.message(this,"visit(UnOpProject arg)","typing operand...");
		Expression operand = arg.getOperand();
		operand.accept(this);

		Trace.message(this,"visit(UnOpProject arg)","typing select list...");
		List selectList = arg.getItemList();
		List itemTypeList = new ArrayList();

		for (int i = 0; i < selectList.size(); i++) {
			Expression item = (Expression)selectList.get(i);
			item.accept(this);
			SqlTypeAtom itemType = (SqlTypeAtom)item.getType().clone();
			itemTypeList.add(itemType);
		}

		SqlType type = null;
		if (1 < itemTypeList.size()) {
			type = new SqlTypeStructure(itemTypeList,SqlType.MANY);
		}
		else {
			type = (SqlTypeAtom)itemTypeList.get(0);
			type.setMultiplicity(SqlType.MANY);
		}

		arg.setType(type);
		Trace.exit(this,"visit(UnOpProject arg)", arg);
	}

	public void visit(UnOpRestrict arg) throws SqlWrapperException {
		Trace.enter(this,"visit(UnOpRestrict arg)", arg);

		Trace.message(this,"visit(UnOpRestrict arg)","Typing operand...");
		Expression operand = arg.getOperand();
		operand.accept(this);
		SqlType type = (SqlType)operand.getType().clone();
		arg.setType(type);

		Trace.message(this,"visit(UnOpRestrict arg)","Typing predicates...");
		List predicateList = arg.getPredicateList();
		if (null != predicateList) {
			Expression predicate = null;
			for (int i = 0; i < predicateList.size(); i++) {
				predicate = (Expression)predicateList.get(i);
				predicate.accept(this);
			}
		}

		Trace.exit(this,"visit(UnOpRestrict arg)", arg);
	}

	public void visit(UnOpSort arg) throws SqlWrapperException {
		Trace.enter(this,"visit(UnOpSort arg)", arg);

		Trace.message(this,"visit(UnOpSort arg)","Typing operands...");
		Expression operand = arg.getOperand();
		operand.accept(this);
		SqlType type = (SqlType)operand.getType().clone();
		arg.setType(type);

		Trace.message(this,"visit(UnOpSort arg)","Typing sort specifications...");
		List list  = arg.getSortSpecificationList();
		SortSpecification   sortSpec = null;
		Expression          sortExpr = null;
		Expression predicate = null;
		for (int i = 0; i < list.size(); i++) {
			sortSpec = (SortSpecification)list.get(i);
			sortExpr = sortSpec.getSortExpression();
			sortExpr.accept(this);
		}

		Trace.exit(this,"visit(UnOpSort arg)", arg);
	}

	public void visit(AfTrim arg) throws SqlWrapperException {
		Trace.enter(this, "visit(AfTrim arg)", arg);
		Expression para = (Expression)arg.getArgument(0);
		para.accept(this);

		SqlType type = new SqlTypeAtom( ptChar, null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this, "visit(AfTrim arg)", arg);
	}

	public void visit(AfCast arg) throws SqlWrapperException {
		Trace.enter(this, "visit(AfCast arg)", arg);
		Expression para = (Expression)arg.getExpression();
		para.accept(this);

		SqlTypeAtom type = new SqlTypeAtom(arg.getTargetType(), arg.getName());
		type.setName(arg.getName());
		arg.setType(type);
		Trace.exit(this, "visit(AfCast arg)", arg);
	}

	public void visit(XfCast arg) throws SqlWrapperException {
		Trace.enter(this, "visit(XfCast arg)", arg);
		Expression para = (Expression)arg.getExpression();
		para.accept(this);

		SqlType type = new SqlTypeAtom(arg.getTargetType(),null);
		arg.setType(type);
		Trace.exit(this, "visit(XfCast arg)", arg);
	}

	public void visit(XfAbs arg) throws SqlWrapperException {
		Trace.enter(this, "visit(XfAbs arg)", arg);
		Expression para = (Expression)arg.getArgument(0);
		para.accept(this);

		SqlType type = (SqlTypeAtom)((SqlTypeAtom)para.getType()).clone();
		arg.setType(type);
		Trace.exit(this, "visit(XfAbs arg)", arg);
	}

	public void visit(XfCeiling arg) throws SqlWrapperException {
		Trace.enter(this, "visit(XfCeiling arg)", arg);
		Expression para = (Expression)arg.getArgument(0);
		para.accept(this);

		SqlType type = new SqlTypeAtom( ptInteger,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this, "visit(XfCeiling arg)", arg);
	}

	public void visit(XfConcat arg) throws SqlWrapperException {
		Trace.enter(this, "visit(XfConcat arg)", arg);

		Expression parameter;
		for (int i = 0; i < arg.getArgumentNumber(); i++) {
			parameter = (Expression)arg.getArgument(i);
			parameter.accept(this);
		}

		SqlType type = new SqlTypeAtom( ptChar,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this, "visit(XfConcat arg)", arg);
	}

	public void visit(XfContains arg) throws SqlWrapperException {
		Trace.enter(this, "visit(XfContains arg)", arg);
		Expression para = (Expression)arg.getArgument(0);
		para.accept(this);
		para = (Expression)arg.getArgument(1);
		para.accept(this);

		SqlType type = new SqlTypeAtom( ptBoolean,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this, "visit(XfContains arg)", arg);
	}

	public void visit(XfCurrentDateTime arg) throws SqlWrapperException {
		Trace.enter(this, "visit(XfCurrentDateTime arg)", arg);
		SqlType type = new SqlTypeAtom( ptTimeStamp,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this, "visit(XfCurrentDateTime arg)", arg);
	}

	public void visit(XfDate arg) throws SqlWrapperException{
		Trace.enter(this, "visit(XfDate arg)", arg);
		Expression para = (Expression)arg.getArgument(0);
		para.accept(this);

		SqlType type = new SqlTypeAtom( ptDate,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this, "visit(XfDate arg)", arg);
	}

	public void visit(XfDateTime arg) throws SqlWrapperException{
		Trace.enter(this, "visit(XfDateTime arg)", arg);
		Expression para = (Expression)arg.getArgument(0);
		para.accept(this);

		SqlType type = new SqlTypeAtom( ptTimeStamp,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this, "visit(XfDateTime arg)", arg);
	}

	public void visit(XfEndsWith arg) throws SqlWrapperException {
		Trace.enter(this, "visit(XfEndsWith arg)", arg);
		Expression para = (Expression)arg.getArgument(0);
		para.accept(this);
		para = (Expression)arg.getArgument(1);
		para.accept(this);

		SqlType type = new SqlTypeAtom( ptBoolean,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this, "visit(XfEndsWith arg)", arg);
	}

	public void visit(XfFloor arg) throws SqlWrapperException {
		Trace.enter(this, "visit(XfFloor arg)", arg);
		Expression para = (Expression)arg.getArgument(0);
		para.accept(this);

		SqlType type = new SqlTypeAtom( ptInteger,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this, "visit(XfFloor arg)", arg);
	}

	public void visit(XfLowerCase arg) throws SqlWrapperException {
		Trace.enter(this, "visit(XfLowerCase arg)", arg);
		Expression para = (Expression)arg.getArgument(0);
		para.accept(this);

		SqlType type = new SqlTypeAtom( ptChar, null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this, "visit(XfLowerCase arg)", arg);
	}

	public void visit(XfRound arg) throws SqlWrapperException {
		Trace.enter(this, "visit(XfRound arg)", arg);
		Expression para = (Expression)arg.getArgument(0);
		para.accept(this);

		SqlType type = new SqlTypeAtom( ptInteger,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this, "visit(XfRound arg)", arg);
	}

	public void visit(XfStartsWith arg) throws SqlWrapperException {
		Trace.enter(this, "visit(XfStartsWith arg)", arg);
		Expression para = (Expression)arg.getArgument(0);
		para.accept(this);
		para = (Expression)arg.getArgument(1);
		para.accept(this);

		SqlType type = new SqlTypeAtom( ptBoolean,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this, "visit(XfStartsWith arg)", arg);
	}

	public void visit(XfStringLength arg) throws SqlWrapperException {
		Trace.enter(this, "visit(XfStringLength arg)", arg);
		Expression para = (Expression)arg.getArgument(0);
		para.accept(this);

		SqlType type = new SqlTypeAtom( ptInteger,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this, "visit(XfStringLength arg)", arg);
	}

	public void visit(XfSubString arg) throws SqlWrapperException {
		Trace.enter(this, "visit(XfContains arg)", arg);

		Expression para = (Expression)arg.getArgument(0);
		para.accept(this);
		para = (Expression)arg.getArgument(1);
		para.accept(this);
		para = (Expression)arg.getArgument(2);
		if (null != para) {
			para.accept(this);
		}

		SqlType type = new SqlTypeAtom( ptChar,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this, "visit(XfContains arg)", arg);
	}

	public void visit(XfTime arg) throws SqlWrapperException{
		Trace.enter(this, "visit(XfCurrentDateTime arg)", arg);
		Expression para = (Expression)arg.getArgument(0);
		para.accept(this);
		SqlType type = new SqlTypeAtom( ptTime,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this, "visit(XfCurrentDateTime arg)", arg);
	}

	public void visit(XfUpperCase arg) throws SqlWrapperException {
		Trace.enter(this, "visit(XfUpperCase arg)", arg);
		Expression para = (Expression)arg.getArgument(0);
		para.accept(this);

		SqlType type = new SqlTypeAtom( ptChar,null,SqlType.ONE);
		arg.setType(type);
		Trace.exit(this, "visit(XfUpperCase arg)", arg);
	}

}
