/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor.algebra;

import java.sql.Types;
import java.util.*;

import org.xquark.extractor.common.Debug;
import org.xquark.extractor.common.Trace;
import org.xquark.extractor.metadata.DataType;
import org.xquark.extractor.metadata.DbPrimitiveType;
import org.xquark.schema.*;
import org.xquark.xquery.parser.QName;
import org.xquark.xquery.parser.XQueryExpression;
import org.xquark.xquery.typing.*;

/**
 * This class is used by the GenAlgebraVisitor to perform a "quick" translation
 * of XML typing coming from the query into an SQL (JDBC) data type. The mapping
 * is defined statically and used by {@link org.xquark.extractor.algebra.GenAlgebraVisitor}
 * to check the query is consistent from a SQL type point of view and mostly to
 * guide algebra construction during operator generation, e.g. according to the
 * fact that one operator is an atom or not (a constant or a column).
 */
// TODO: check it is not redundant with front-end processing. Especially when multiplicity
// if set by the new QType flag.
public final class TypeInterpreter {
	private static final String RCSRevision = "$Revision: 1.2 $";
	private static final String RCSName = "$Name:  $";

	/** The static type mapping table between XML and SQL atom types */
	private static final int[] CODE_TABLE = { -1, Types.VARCHAR, Types.BIT, Types.DECIMAL, Types.DOUBLE, Types.FLOAT, -1, Types.TIMESTAMP, Types.TIME, Types.DATE, -1, -1, -1, -1, -1 - 1, -1, -1, -1, -1, -1 };
	/**
	public static final int STRING = 1;
	public static final int BOOLEAN = 2;
	public static final int DECIMAL = 3;
	public static final int DOUBLE = 4;
	public static final int FLOAT = 5;
	public static final int DURATION = 6;
	public static final int DATE_TIME = 7;
	public static final int TIME = 8;
	public static final int DATE = 9;
	public static final int GYEAR_MONTH = 10;
	public static final int GYEAR = 11;
	public static final int GMONTH_DAY = 12;
	public static final int GDAY = 13;
	public static final int GMONTH = 14;
	public static final int HEX_BINARY = 15;
	public static final int BASE64_BINARY = 16;
	public static final int ANY_URI = 17;
	public static final int QNAME = 18;
	public static final int NOTATION = 19;
	 */
	private int sqlTypeCode(int schemaTypeCode) {
		int retVal = -1;
		if (schemaTypeCode < 1 || schemaTypeCode > 20) {
			retVal = -1;
		} else {
			retVal = CODE_TABLE[schemaTypeCode];
		}

		return retVal;
	}

	public TypeInterpreter() {}

	public SqlType interprete(QType qtype) {
		Trace.enter(this, "interprete (QType qtype)");
		Debug.assertTrue(null != qtype, "null != qtype");

		SqlType retVal = null;
		if (qtype instanceof QTypeAtom)
			retVal = interprete((QTypeAtom) qtype);
		if (qtype instanceof QTypeText)
			retVal = interprete((QTypeText) qtype);
		else if (qtype instanceof QTypeDocument)
			retVal = interprete((QTypeDocument) qtype);
		else if (qtype instanceof QTypeElement)
			retVal = interprete((QTypeElement) qtype);
		else if (qtype instanceof QTypeSequence)
			retVal = interprete((QTypeSequence) qtype);

		// start add   17/07/2003
		if (retVal != null) {
			if (qtype.isMultiple())
				retVal.setMultiplicity(SqlType.MANY);
			else
				retVal.setMultiplicity(SqlType.ONE);
		}
		// end add   17/07/2003
		Trace.exit(this, "interprete (QType qtype)");
		return retVal;
	}

	public SqlType interprete(QTypeAtom qtype) {
		Trace.enter(this, "interprete (QTypeAtom qtype)");

		SqlType retVal = null;

		// verify QTypeAtom is a atomic Value
		//Debug.assertTrue(qtype.isAtom(), "qtype.isAtom()");
		retVal = interprete(qtype.getSimpleType());

		Trace.exit(this, "interprete (QTypeAtom qtype)");
		return retVal;
	}

	// new   02/09/2003
	public SqlType interprete(QTypeText qtype) {
		Trace.enter(this, "interprete (QTypeText qtype)");

		SqlType retVal = null;

		retVal = interprete(qtype.getSimpleType());

		Trace.exit(this, "interprete (QTypeText qtype)");
		return retVal;
	}

	public SqlType interprete(QTypeElement qte) {
		Trace.enter(this, "interprete(QTypeElement qte)");
		SqlType retVal = null;
		Type type = qte.getType();
		if (type instanceof SimpleType) { // qtype contains a SimpleType
			SqlTypeAtom sqlAtom;
			sqlAtom = (SqlTypeAtom) interprete((SimpleType) type);
			if (null != sqlAtom) {
				// set name
				XQueryExpression name = qte.getName();
				Debug.assertTrue((name instanceof QName), "(name instanceof QName)");

				String nameStr = ((QName) name).getName();
				sqlAtom.setName(nameStr);
				retVal = sqlAtom;
			}
		} else { // QTypeElement contains a ComplexType
			retVal = interprete((ComplexType) type);
		}

		Trace.exit(this, "interprete (QTypeElement qtype)");
		return retVal;
	}

	public SqlType interprete(QTypeSequence qtype) {
		Trace.enter(this, "interprete(QTypeSequence qtype)");
		SqlType retVal = null;

		SqlType sqlType = interprete((QType) qtype.getList().get(0));
		if (sqlType.isRelation()) {
			retVal = VerifyRelation(qtype);
		} else if (sqlType.isAtom()) {
			retVal = VerifyColumn(qtype);
		} else {
			Debug.assertTrue(false, "NYI");
		}

		Trace.enter(this, "interprete(QTypeSequence qtype)");
		return retVal;
	}

	protected SqlType VerifyColumn(QTypeSequence qtype) {
		Trace.enter(this, "VerifyColumn(QTypeSequence qtype)");

		SqlTypeStructure retVal = new SqlTypeStructure(null);

		SqlType sqlType = null;
		Iterator iter = qtype.getList().iterator();
		while (iter.hasNext()) {
			//Debug.assertTrue(.........);
			sqlType = interprete((QType) iter.next());
			if (null == sqlType) {
				retVal = null;
				break;
			}
			if (sqlType.isAtom()) {
				retVal.addAttribute((SqlTypeAtom) sqlType);
			} else if (sqlType.isTuple()) {
				retVal.addAttributes(((SqlTypeStructure) sqlType).getAttributes());
			} else {
				retVal = null;
				break;
			}
		}

		Trace.exit(this, "VerifyColumn(QTypeSequence qtype)");
		return retVal;
	}

	protected SqlType VerifyRelation(QTypeSequence qtype) {
		Trace.enter(this, "VerifyRelation(QTypeSequence qtype)");
		SqlType retVal = null;

		SqlType sqlType = null;
		Iterator iter = qtype.getList().iterator();
		while (iter.hasNext()) {
			//Debug.assertTrue(.........);
			sqlType = interprete((QType) iter.next());
			if (null == sqlType && !sqlType.isRelation()) {
				retVal = null;
				break;
			} else if (null == retVal) {
				retVal = sqlType;
			}

		}

		Trace.exit(this, "VerifyRelation(QTypeSequence qtype)");
		return retVal;
	}

	public SqlType interprete(QTypeDocument qtype) {
		Trace.enter(this, "interprete(QTypeDocument qtype)");

		SqlType retVal = interprete(qtype.getQType());

		// start change   17/07/2003
		//		if (retVal != null) { 
		//			if (SqlType.MANY == retVal.getMultiplicity()) {
		//				retVal = null;
		//			} else {
		//				retVal.setMultiplicity(SqlType.MANY);
		//			}
		//		}
		// end change   17/07/2003

		Trace.exit(this, "interprete(QTypeDocument qtype)");
		return retVal;
	}

	public SqlType interprete(SimpleType stype) {
		Trace.enter(this, "iinterprete(SimpleType stype)");
		SqlTypeAtom retVal;
		Debug.assertTrue(SimpleType.VARIETY_ATOMIC == stype.getVariety(), "SimpleType.VARIETY_ATOMIC==stype.getVariety()");

		DataType dt = new DbPrimitiveType(sqlTypeCode(stype.getPrimitive().getType()));
		retVal = new SqlTypeAtom(dt, null, SqlType.ONE);

		Trace.exit(this, "interprete", "SimpleType");
		return retVal;
	}

	public SqlType interprete(ComplexType stype) {
		Trace.enter(this, "interprete(SimpleType stype)");

		SqlTypeStructure retVal = null;
		retVal = new SqlTypeStructure(null, SqlType.ONE);

		Collection elements = getElementDeclarations(stype);
		ElementDeclaration element;
		SqlType elementType;
		Iterator iter = elements.iterator();
		Debug.assertTrue(iter.hasNext(), "iter.hasNext()");
		while (iter.hasNext()) {
			element = (ElementDeclaration) iter.next();
			elementType = interprete(element);
			if (elementType.isAtom()) {
				retVal.addAttribute((SqlTypeAtom) elementType);
			} else {
				retVal = null;
				break;
			}
		}
		Trace.exit(this, "interprete(SimpleType stype)");
		return retVal;
	}

	public SqlType interprete(ElementDeclaration stype) {
		Trace.enter(this, "interprete(ElementDeclaration stype)");

		/** @todo  NYI!*/
		SqlType retVal = null;

		Type schemaType = stype.getType();
		if (schemaType.isSimpleType()) {
			retVal = interprete((SimpleType) schemaType);
			((SqlTypeAtom) retVal).setName(stype.getName());
		} else {
			retVal = interprete((ComplexType) schemaType);
		}

		//        retVal.setName();

		Trace.exit(this, "interprete(ElementDeclaration stype)");
		return retVal;
	}

	private List getElementDeclarations(ComplexType complextype) {
		Trace.enter(this, "getElementDeclarations(ComplexType complextype)");

		List retVal = new ArrayList();
		ContentModel contentmodel = complextype.getContentModel();
		Particle particle = (Particle) contentmodel.getModel();
		Object term = particle.getTerm();
		if (term instanceof ElementDeclaration)
			retVal.add(term);
		else if (term instanceof ModelGroup) {
			ModelGroup modelgroup = (ModelGroup) term;
			int compositor = modelgroup.getCompositor();
			Debug.assertTrue(compositor == ModelGroup.SEQUENCE, "compositor == ModelGroup.SEQUENCE");

			for (int i = 0; i < modelgroup.size(); i++) {
				Particle part = (Particle) modelgroup.get(i);
				ElementDeclaration elemDecl = (ElementDeclaration) part.getTerm();
				retVal.add(elemDecl);
			}
		}

		Trace.exit(this, "getElementDeclarations(ComplexType complextype)");
		return retVal;
	}

	//    public List getEelementNames (Qtype qType)
	//    {
	//
	//        if (qtype instanceof QTypeAtom)
	//            return interprete ((QTypeAtom)qtype);
	//        else if (qtype instanceof QTypeCollection)
	//            return interprete ((QTypeCollection)qtype);
	//        else if (qtype instanceof QTypeElement)
	//             return interprete ((QTypeElement)qtype);
	//        if (qtype instanceof QTypeSequence)     return interprete ((QTypeSequence)qtype);
	//
	//        return null ;
	//    }
	//	  public static void main(String []args) {
	//	  }
}
