/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor.metadata;

import java.sql.*;
import java.util.*;

import org.xquark.extractor.Selection;
import org.xquark.extractor.common.*;

/**
 * Base class for metadata path tree building from JDBC metadata applying
 * filters defined in the configuration file. This class will be derived for
 * every RDBMS supported.
 */
public abstract class Loader {

	private static final String RCSRevision = "$Revision: 1.5 $";
	private static final String RCSName = "$Name:  $";

	/* Maintains a map from XML names to DBMS name for table and columns */
	public Properties _tableMap = new Properties();

	protected DatabaseMetaData _databaseMetaData;
	protected Connection _connection;
	protected HashSet _nameSpaces = new HashSet();
	protected QNameEncoder _encoder = null;

	public Loader() {
	}

	public Loader(DatabaseMetaData databaseMetaData) {
		_databaseMetaData = databaseMetaData;
	}

	public void setConnection(Connection connection) {
		_connection = connection;
	}

	public void setEncoder(QNameEncoder encoder) {
		_encoder = encoder;
	}

	public DatabaseMetaData getDatabaseMetaData() {
		return _databaseMetaData;
	}

	public void setDatabaseMetaData(DatabaseMetaData aDatabaseMetaData) {
		_databaseMetaData = aDatabaseMetaData;
	}

	/**
	 * Main entry point performing path tree construction. Browse JDBC catalogs
	 * calling sub-procedure for schemas if matching the selection.
	 * @param selection A user-defined filter used for catalogs.
	 * @return The root metadata node created for the source (Site).
	 * @throws MetaDataException
	 * @see org.xquark.extractor.metadata.NamedNode
	 */
	public Site loadSite(Selection selection, boolean doesPrepared) throws MetaDataException {
		Trace.enter(this, "loadSite(Selection selection)");

		Site retVal = new Site(null);
		List children = new ArrayList();
		List subSelections = new ArrayList();
		NamedNode catalog = null;
		ResultSet resultSet = null;
		String name = null;
		_nameSpaces.clear();

		try {
			/* JDBC catalog browsing */
			resultSet = _databaseMetaData.getCatalogs();

			if (null == selection || selection.noSubElementDef()) {
				// not indications of which catalogs to be loaded, so all catalog will be loaded
				while (resultSet.next()) {
					name = resultSet.getString("TABLE_CAT");
					children.add(name);
					subSelections.add(null);

				}
				resultSet.close();
			} else { // filter catalogs
				Selection subSelection = null;

				while (resultSet.next()) {
					name = resultSet.getString("TABLE_CAT");
					subSelection = selection.includesSubElement(name);
					if (subSelection != null) {
						// TODO : No indications left by YS on this...
						//                        if (CASE_TOLOWER == _nameCase ) name = name.toLowerCase();
						//                        else if (CASE_TOUPPER== _nameCase) name = name.toUpperCase();
						children.add(name);
						subSelections.add(subSelection);
					}
				}
				resultSet.close();

				/* For databases in which there is not concept of catalog, add a
				 * "null" in children.
				 */
				if (children.isEmpty()) {
					children.add(null);
					subSelection = selection.includesSubElement(null);
					subSelections.add(subSelection);
				}
			}

			/* Browse filtered catalogs for building */
			for (int i = 0; i < children.size(); i++) {
				String child = (String) children.get(i);
				Selection subSelection = (Selection) subSelections.get(i);
				retVal.addChild(loadCatalog(child, subSelection, doesPrepared));
			}

		} catch (SQLException ex) {
			throw new MetaDataException(ex.getMessage(), ex);
		}
		Trace.exit(this, "loadSite(Selection selection)");
		return retVal;
	}

	/**
	 * Create the Catalog metadata node for a specific catalog. Browse JDBC
	 * schemas calling sub-procedure for schemas if matching the selection.
	 * @param catalogName The catalog browsed.
	 * @param selection A user-defined filter used for schemas.
	 * @return the {@link org.xquark.extractor.metadata.Catalog Catalog} node
	 * built.
	 * @throws MetaDataException
	 */
	protected Catalog loadCatalog(String catalogName, Selection selection, boolean doesPrepared) throws MetaDataException {
		Trace.enter(this, "loadCatalog(String name, Selection selection)", catalogName);

		List children = new ArrayList();
		List subSelections = new ArrayList();
		ResultSet resultSet = null;
		String name = null;

		/* Catalog node creation */
		Catalog retVal = null;
		if (catalogName == null || catalogName.length() == 0)
			retVal = new Catalog(null);
		else
			retVal = new Catalog(catalogName);

		/* JDBC schema browsing */
		try {
			resultSet = _databaseMetaData.getSchemas();

			if (selection == null || selection.noSubElementDef()) {
				// not indications of which catalogs to be loaded, so all catalog will be loaded
				while (resultSet.next()) {
					name = resultSet.getString("TABLE_SCHEM");

					children.add(name);
					subSelections.add(null);

				}
				resultSet.close();
			} else {
				Selection subSelection = null;
				while (resultSet.next()) {
					name = resultSet.getString("TABLE_SCHEM");
					subSelection = selection.includesSubElement(name);
					if (subSelection != null) {
						String targetNameSpace = subSelection.getTargetNamespace();
						//                        if (targetNameSpace == null) targetNameSpace = "";
						if (_nameSpaces.contains(targetNameSpace))
							throw new MetaDataException("Same target namespace : [" + targetNameSpace + "] used twice");
						_nameSpaces.add(targetNameSpace);

						children.add(name);
						subSelections.add(subSelection);
					}
				}
				resultSet.close();

				/* For databases in which there is not concept of schema, add a
				 * "null" in children.
				 */
				if (children.isEmpty()) {
					children.add(null);
					subSelection = selection.includesSubElement(null);
					subSelections.add(subSelection);
				}
			}

			/* Browse filtered schema for building */
			for (int i = 0; i < children.size(); i++) {
				String child = (String) children.get(i);
				Selection subSelection = (Selection) subSelections.get(i);
				retVal.addChild(loadSchema(catalogName, child, subSelection, doesPrepared));
			}
		} catch (Exception ex) {
			throw new MetaDataException(ex.getMessage(), ex);
		}

		Trace.exit(this, "loadCatalog(String name, Selection selection)", catalogName);
		return retVal;

	}

	/**
	 * Create the Schema metadata node for a specific schema. Browse JDBC
	 * tables encoding names if demanded by user and calling sub-procedures if 
	 * table is matching the selection for both tables and keys.
	 * @param catalogName The catalog that contains the schema.
	 * @param schemaName The schema browsed.
	 * @param selection A user-defined filter used for tables.
	 * @return the {@link org.xquark.extractor.metadata.Schema Schema} node
	 * built.
	 * @throws MetaDataException
	 */
	protected Schema loadSchema(String catalogName, String schemaName, Selection selection, boolean doesPrepared) throws MetaDataException {
		Trace.enter(this, "loadSchema ( String catalogName, String schemaName, Selection selection )", schemaName);

		List children = new ArrayList();
		List subSelections = new ArrayList();
		NamedNode catalog = null;
		ResultSet resultSet = null;
		String name = null;
		_nameSpaces.clear();
		String alias = null;

		/* Schema node creation */
		Schema retVal = new Schema(schemaName);
		if (selection != null) {
			retVal.setTargetNamespace(selection.getTargetNamespace());
			retVal.setElementFormDefault(selection.getElementFormDefault());
		}

		/* JDBC tables browsing */
		try {
			String[] tableTypes = { "TABLE", "VIEW", "SYNONYM" };
			String tableType = null;
			resultSet = _databaseMetaData.getTables(catalogName, schemaName, null, tableTypes);

			if (selection == null || selection.noSubElementDef()) {
				/* not indications of which tables to be loaded, so all tables will be loaded */
				while (resultSet.next()) {
					tableType = resultSet.getString("TABLE_TYPE");
					name = resultSet.getString("TABLE_NAME");
					alias = _encoder.encode(name);
					if (!Selection.verifyName(alias)) {
						Log.warn(MessageLibrary.getMessage("C_INV_NCNAME", alias));
						Log.warn(MessageLibrary.getMessage("M_OBJ_N_LOADED", tableType + " " + schemaName + "." + name));
						continue; // out of selection
					}

					if (tableType.equalsIgnoreCase("TABLE")) {
						children.add(new Relation(name, alias, Table.TABLE));
						subSelections.add(null);
					} else if (tableType.equalsIgnoreCase("VIEW")) {
						children.add(new Relation(name, alias, Table.VIEW));
						subSelections.add(null);
					} else if (tableType.equalsIgnoreCase("SYNONYM")) {
						children.add(new Relation(name, alias, Table.SYNONYM));
						subSelections.add(null);
					} else {
						Log.warn(MessageLibrary.getMessage("M_OBJ_N_LOADED", tableType + " " + schemaName + "." + name));
					}
				}
				resultSet.close();
			} else {
				/* */
				Selection subSelection = null;
				while (resultSet.next()) {
					name = resultSet.getString("TABLE_NAME");
					subSelection = selection.includesSubElement(name);
					if (subSelection != null) {
						tableType = resultSet.getString("TABLE_TYPE");
						alias = subSelection.getAlias();
						if (alias == null) {
							alias = _encoder.encode(name);
							if (!Selection.verifyName(alias)) {
								Log.warn(MessageLibrary.getMessage("C_INV_NCNAME", alias));
								Log.warn(MessageLibrary.getMessage("M_OBJ_N_LOADED", tableType + " " + schemaName + "." + name));
								continue; // out of selection
							}
						}

						if (tableType.equalsIgnoreCase("TABLE")) {
							children.add(new Relation(name, alias, Table.TABLE));
							subSelections.add(subSelection);
						} else if (tableType.equalsIgnoreCase("VIEW")) {
							children.add(new Relation(name, alias, Table.VIEW));
							subSelections.add(subSelection);
						} else if (tableType.equalsIgnoreCase("SYNONYM")) {
							children.add(new Relation(name, alias, Table.SYNONYM));
							subSelections.add(subSelection);
						} else {
							Log.warn(MessageLibrary.getMessage("M_OBJ_N_LOADED", tableType + " " + schemaName + "." + name));
						}
					}
				}
				resultSet.close();
			}

			// TODO : temporary code (written by YS)
			String path = null; /* the prefix of key in _tableMap Properties */
			if (null != catalogName)
				path = catalogName;
			if (schemaName != null)
				if (path == null)
					path = schemaName;
				else
					path += '.' + schemaName;

			String encodedName = null;

			/* browse "relation" objects for table and key loading */
			for (int i = 0; i < children.size(); i++) {
				Relation child = ((Relation) children.get(i));
				Selection subSelection = (Selection) subSelections.get(i);

				putTable(path, child.alias, child.name); // register table

				NamedNode childNode = null;
				if (Table.TABLE == child.type) {
					childNode = loadTable(catalogName, schemaName, child.name, child.alias, subSelection, doesPrepared);
					if (childNode != null) {
						loadKeys(catalogName, schemaName, (Table) childNode, subSelection);
					}
				} else {
					childNode = loadTable(catalogName, schemaName, child.name, child.alias, subSelection, doesPrepared);
				}
				if (childNode != null) {
					retVal.addChild(childNode);
				}
			}
		} catch (Exception ex) {
			throw new MetaDataException(ex.getMessage(), ex);
		}

		Trace.exit(this, "loadSchema ( String catalogName, String schemaName, Selection selection )", schemaName);
		return retVal;
	}

	/**
	 * Create the table and columns metadata nodes for a specific table. Browse
	 * JDBC {@link java.sql.ResultSetMetaData ResultSetMetaData} for column
	 * information encoding names if demanded by user.
	 * @param catalogName
	 * @param schemaName
	 * @param origTableName
	 * @param tableName
	 * @param selection
	 * @return
	 * @throws MetaDataException
	 */
	protected Table loadTable(String catalogName, String schemaName, String origTableName, String tableName, Selection selection, boolean doesPrepared) throws MetaDataException {
		Trace.enter(this, "loadTable(String catalogName, String schemaName, String origTableName, String tableName, Selection selection)", origTableName);

		/* Table node creation */
		Table retTable = new Table(origTableName, tableName);
		retTable.setType(Table.TABLE);

		if (!doesPrepared) {
			Statement stmt = null;
			ResultSetMetaData rsetdata = null;
			/* Get dummy SQL query string to explore resultset metadata (columns)*/
			try {
				stmt = _connection.createStatement();
				stmt.execute(createQuery(catalogName, schemaName, origTableName));
				rsetdata = stmt.getResultSet().getMetaData();
			} catch (SQLException ex) {
				Log.error(ex);
				Log.warn(MessageLibrary.getMessage("M_OBJ_N_LOADED", "Table : " + schemaName + "." + origTableName));
				if (null != stmt) {
					/* close the working cursor */
					try {
						stmt.close();
					} catch (Exception excp) {
						Log.warn("Failed to close PreparedStatement.");
					}
				}
				Trace.exit(this, "loadTable(String catalogName, String schemaName, String origTableName, String tableName, Selection selection)", origTableName);
				return null;
			}

			/* Browse result set metadata */
			try {
				Attribute attribute = null;
				String column_name;
				Selection subSelection = null;
				String alias = null;
				List pkAttributes = new ArrayList();
				// list of column nodes forming PK
				boolean currentColumnIsInPK = false;

				Trace.message(this, "loadTable(...)", "Loading columns...");

				// TODO : temporary code (written by YS)
				String path = null; /* the prefix of key in _tableMap Properties */
				if (null != catalogName) {
					path = catalogName + '.' + schemaName + '.' + tableName;
				} else {
					path = schemaName + '.' + tableName;
				}

				for (int i = 1; i <= rsetdata.getColumnCount(); i++) {
					column_name = ((String) rsetdata.getColumnName(i));
					currentColumnIsInPK = false;
					alias = null;

					if (selection == null || selection.noSubElementDef()) {
						alias = _encoder.encode(column_name);
						if (!Selection.verifyName(alias)) {
							Log.warn(MessageLibrary.getMessage("C_INV_NCNAME", alias));
							Log.warn(MessageLibrary.getMessage("M_OBJ_N_LOADED", schemaName + "." + origTableName + "." + column_name));
							continue; // out of selection
						}
					} else {
						subSelection = selection.includesSubElement(column_name);
						if (subSelection == null) {
							continue; // out of selection
						} else {
							alias = subSelection.getAlias();
							if (alias == null) {
								alias = _encoder.encode(column_name);
								if (!Selection.verifyName(alias)) {
									Log.warn(MessageLibrary.getMessage("C_INV_NCNAME", alias));
									Log.warn(MessageLibrary.getMessage("M_OBJ_N_LOADED", schemaName + "." + origTableName + "." + column_name));
									continue;
								}
							}
							if (subSelection.getPKColumn()) {
								currentColumnIsInPK = true;
							}
						}
					}
					int data_type = (int) rsetdata.getColumnType(i);
					String genericType = MetaDataManager.getGenericType(data_type);
					if (genericType == null) {
						// TODO this was done for special purposes (BNF)
						String columnClassName = rsetdata.getColumnTypeName(i);
						if ("TEXT".equalsIgnoreCase(columnClassName)) {
							data_type = java.sql.Types.VARCHAR;
						} else {
						Log.warn(MessageLibrary.getMessage("M_OBJ_N_LOADED", schemaName + "." + origTableName + "." + column_name));
						continue;
						}
					}

					putAttribute(path, alias, column_name); // register column

					/* Attribute (column) node creation */
					attribute = new Attribute(column_name, alias);
					// QUESTION : why attributes ? columns are mapped to elements !
					attribute.setType(new DbPrimitiveType(data_type));

					if (rsetdata.isNullable(i) == ResultSetMetaData.columnNoNulls)
						attribute.setNullable(false);
					else
						attribute.setNullable(true);

					/* Add column to metadata tree */
					retTable.addChild(attribute);

					/* Add column metadata node to Table node PK list as defined 
					 * by user in configuration file. 
					 */
					// QUESTION : the user PK is the first so used in priority
					if (currentColumnIsInPK) {
						pkAttributes.add(attribute);
					}
				}

				/* Add PK column metadata node list to Table node */
				if (!pkAttributes.isEmpty()) {
					retTable.addKey(pkAttributes);
				}

				if (stmt != null) {
					stmt.close();
					stmt = null;
				}
			} catch (SQLException ex) {
				Log.exception(ex);
				throw new MetaDataException(ex.getMessage());
			}
		} else {

			PreparedStatement pstmt = null;
			ResultSetMetaData rsetdata = null;
			/* Get dummy SQL query string to explore resultset metadata (columns)*/
			try {
				pstmt = createPreparedStatement(_connection, createQuery(catalogName, schemaName, origTableName));
				pstmt.execute();
				rsetdata = pstmt.getMetaData();
			} catch (SQLException ex) {
				Log.error(ex);
				Log.warn(MessageLibrary.getMessage("M_OBJ_N_LOADED", "Table : " + schemaName + "." + origTableName));
				if (null != pstmt) {
					/* close the working cursor */
					try {
						pstmt.close();
					} catch (Exception excp) {
						Log.warn("Failed to close PreparedStatement.");
					}
				}
				Trace.exit(this, "loadTable(String catalogName, String schemaName, String origTableName, String tableName, Selection selection)", origTableName);
				return null;
			}

			/* Browse result set metadata */
			try {
				Attribute attribute = null;
				String column_name;
				Selection subSelection = null;
				String alias = null;
				List pkAttributes = new ArrayList();
				// list of column nodes forming PK
				boolean currentColumnIsInPK = false;

				Trace.message(this, "loadTable(...)", "Loading columns...");

				// TODO : temporary code (written by YS)
				String path = null; /* the prefix of key in _tableMap Properties */
				if (null != catalogName) {
					path = catalogName + '.' + schemaName + '.' + tableName;
				} else {
					path = schemaName + '.' + tableName;
				}

				for (int i = 1; i <= rsetdata.getColumnCount(); i++) {
					column_name = ((String) rsetdata.getColumnName(i));
					currentColumnIsInPK = false;
					alias = null;

					if (selection == null || selection.noSubElementDef()) {
						alias = _encoder.encode(column_name);
						if (!Selection.verifyName(alias)) {
							Log.warn(MessageLibrary.getMessage("C_INV_NCNAME", alias));
							Log.warn(MessageLibrary.getMessage("M_OBJ_N_LOADED", schemaName + "." + origTableName + "." + column_name));
							continue; // out of selection
						}
					} else {
						subSelection = selection.includesSubElement(column_name);
						if (subSelection == null) {
							continue; // out of selection
						} else {
							alias = subSelection.getAlias();
							if (alias == null) {
								alias = _encoder.encode(column_name);
								if (!Selection.verifyName(alias)) {
									Log.warn(MessageLibrary.getMessage("C_INV_NCNAME", alias));
									Log.warn(MessageLibrary.getMessage("M_OBJ_N_LOADED", schemaName + "." + origTableName + "." + column_name));
									continue;
								}
							}
							if (subSelection.getPKColumn()) {
								currentColumnIsInPK = true;
							}
						}
					}

					int data_type = (int) rsetdata.getColumnType(i);
					String genericType = MetaDataManager.getGenericType(data_type);
					if (genericType == null) {
						// TODO this was done for special purposes (BNF)
						String columnClassName = rsetdata.getColumnTypeName(i);
						if ("TEXT".equalsIgnoreCase(columnClassName)) {
							data_type = java.sql.Types.VARCHAR;
						} else {
						Log.warn(MessageLibrary.getMessage("M_OBJ_N_LOADED", schemaName + "." + origTableName + "." + column_name));
						continue;
						}
					}

					putAttribute(path, alias, column_name); // register column

					/* Attribute (column) node creation */
					attribute = new Attribute(column_name, alias);
					// QUESTION : why attributes ? columns are mapped to elements ! Column type seems unused.
					attribute.setType(new DbPrimitiveType(data_type));

					if (rsetdata.isNullable(i) == ResultSetMetaData.columnNoNulls)
						attribute.setNullable(false);
					else
						attribute.setNullable(true);

					/* Add column to metadata tree */
					retTable.addChild(attribute);

					/* Add column metadata node to Table node PK list as defined 
					 * by user in configuration file. 
					 */
					// QUESTION : the user PK is the first so used in priority
					if (currentColumnIsInPK) {
						pkAttributes.add(attribute);
					}
				}

				/* Add PK column metadata node list to Table node */
				if (!pkAttributes.isEmpty()) {
					retTable.addKey(pkAttributes);
				}

				if (pstmt != null) {
					pstmt.close();
					pstmt = null;
				}
			} catch (SQLException ex) {
				Log.exception(ex);
				throw new MetaDataException(ex.getMessage());
			}

		}
		Trace.exit(this, "loadTable(String catalogName, String schemaName, String origTableName, String tableName, Selection selection)", origTableName);
		return retTable;
	}

	/**
	 * Add the real Primary Key to Table metadata node basing on JDBC metadata.
	 * @param catalogName
	 * @param schemaName
	 * @param table
	 * @param selection
	 * @throws MetaDataException
	 */
	protected void loadKeys(String catalogName, String schemaName, Table table, Selection selection) throws MetaDataException {
		Trace.enter(this, "loadKeys(String catalogName, String schemaName, Table table, Selection selection)", table.getName());

		try {
			//		    /* load Primary Keys for this table */
			//            Trace.message( this,"loadTable(...)", "creating pseudoAttribute column...");
			String column_name = null;
			String alias = null;
			Selection subSelection = null;
			Attribute attribute = null;
			List key = new ArrayList();

			ResultSet resultSet = _databaseMetaData.getPrimaryKeys(catalogName, schemaName, table.getName());

			String path = null; /* the prefix of key in _tableMap Properties */
			if (null != catalogName) {
				path = catalogName + '.' + schemaName + '.' + table.getName();
			} else {
				path = schemaName + '.' + table.getName();
			}

			/* Loop on columns of the PK for the specified table */
			while (resultSet.next()) {
				column_name = resultSet.getString("COLUMN_NAME");

				/* generate and verify alias*/
				alias = null;
				if (selection == null || selection.noSubElementDef()) {
					alias = _encoder.encode(column_name);
				} else {
					subSelection = selection.includesSubElement(column_name);
					if (subSelection == null) {
						alias = _encoder.encode(column_name);
					} else {
						alias = subSelection.getAlias();
						if (null == alias) {
							alias = _encoder.encode(column_name);
						}
					}
				}

				Trace.message(this, "loadKeys(String catalogName, String schemaName, Table table, Selection selection)", column_name + "--" + alias);

				attribute = (Attribute) table.findChild(column_name);
				if (attribute == null) {
					putAttribute(path, alias, column_name);
					attribute = new Attribute(column_name, alias);
				}
				key.add(attribute);
			}
			resultSet.close();

			/* Add PK to Table metadata node */
			if (!key.isEmpty()) {
				table.addKey(key);
			}
		} catch (SQLException ex) {
			throw new MetaDataException(ex.getMessage());
		}
		Trace.exit(this, "loadKeys(String catalogName, String schemaName, Table table, Selection selection)", table.getName());
	}

	/**
	 * Create a dummy costless query in order to explore result set JDBC
	 * metadata that seems more reliable on most drivers than exploring
	 * database metadata.
	 * @param catalogName
	 * @param schemaName
	 * @param origTableName
	 * @return
	 */
	protected String createQuery(String catalogName, String schemaName, String origTableName) {
		StringBuffer retVal = new StringBuffer();

		retVal.append("SELECT * FROM ");
		if (null != catalogName && 0 < catalogName.trim().length()) {
			retVal.append(catalogName.trim());
			retVal.append(".");
		}
		if (null != schemaName && 0 < schemaName.trim().length()) {
			retVal.append(schemaName.trim());
			retVal.append(".");
		}
		if (null != origTableName && 0 < origTableName.trim().length()) {
			retVal.append(origTableName.trim());
		}
		retVal.append(" WHERE 0=1");
		return retVal.toString();
	}

	/**
	 * Isolated to be overridden.
	 * @param jdbcConnection
	 * @param command
	 * @return
	 * @throws SQLException
	 */
	protected PreparedStatement createPreparedStatement(Connection jdbcConnection, String command) throws SQLException {
		PreparedStatement retVal = jdbcConnection.prepareStatement(command);
		return retVal;
	}

	/**
	 * Add the table to map between XML names and native DBMS names.
	 * @param path
	 * @param encodedTableName
	 * @param originalName
	 * @throws MetaDataException
	 */
	protected void putTable(String path, String encodedTableName, String originalName) throws MetaDataException {
		String key = null;
		key = path + '.' + encodedTableName;

		Object value = null;
		value = _tableMap.getProperty(key);
		if (null == value) {
			_tableMap.setProperty(key, originalName);
		} else {
			throw new MetaDataException(MessageLibrary.getMessage("M_NAME_CONF", key, originalName));
		}
	}

	/**
	 * Add the column to map between XML names and native DBMS names.
	 * @param path
	 * @param encodedAttributeName
	 * @param originalName
	 * @throws MetaDataException
	 */
	protected void putAttribute(String path, String encodedAttributeName, String originalName) throws MetaDataException {
		String key = null;
		key = path + '.' + encodedAttributeName;

		Object value = null;
		value = _tableMap.getProperty(key);
		if (null == value) {
			_tableMap.setProperty(key, originalName);
		} else {
			throw new MetaDataException(MessageLibrary.getMessage("M_NAME_CONF", key, originalName));
		}
	}

	/**
	 * Struct-like data class.
	 */
	class Relation {

		public String name;
		public String alias;
		public int type;

		public Relation(String name, String alias, int type) {
			this.name = name;
			this.alias = alias;
			this.type = type;
		}

		public Relation(String name, int type) {
			this.name = name;
			this.type = type;
			this.alias = name;
		}
	}
}
