/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor.mysql;

import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;

import org.xquark.extractor.ExtractorReadOnlyConnection;
import org.xquark.extractor.SynchronizedResultSet;
import org.xquark.extractor.algebra.*;
import org.xquark.extractor.common.Trace;
import org.xquark.extractor.sql.SqlExpression;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xquery.parser.XQueryExpression;

public class ExtractorStatement extends org.xquark.extractor.ExtractorStatement{
	private static final String RCSRevision = "$Revision: 1.3 $";
	private static final String RCSName     = "$Name:  $";

	private PreEvaluationVisitor _pev;
	private MysqlRemoveSubQueries _rsv;

    public ExtractorStatement(ExtractorReadOnlyConnection connection) throws XMLDBCException {
        super(connection);
		Trace.enter(this, "ExtractorStatement(ExtractorReadOnlyConnection connection)");
		Trace.exit(this, "ExtractorStatement(ExtractorReadOnlyConnection connection)");
    }

    protected void createVisitors()
    {
        TypeInterpreter ti = new TypeInterpreter();
        _genAlgebraVisitor = new GenAlgebraVisitor( ti,_metadataManager);

        _algebraTypeVisitor     = new AlgebraTypeVisitor();
        _removeProjectVisitor   = new RemoveIntermediateProjectVisitor();
        _brav                   = new BuildReferredAttributes();
        _apv                    = new AnalyzePredicateVisitor();

        _genSqlVisitor = new org.xquark.extractor.mysql.MySqlGenSqlVisitor();
        _genSqlVisitor.setMetadataManager(_metadataManager);

		_pev = new PreEvaluationVisitor();
		_rsv = new MysqlRemoveSubQueries();
    }

	protected SynchronizedResultSet executeSingleQDB(
		XQueryExpression xQuery,
		boolean keyNeeded)
		throws Exception {
		Trace.enter(
			this,
			"executeSingleQDB(XQueryExpression xQuery, boolean keyNeeded)");

		SynchronizedResultSet retVal = null;
		SqlExpression sql = null;
		Expression algebraTree = null;
		Mapper mapper = null;

		/* Algebra tree building */
		_genAlgebraVisitor.reinit();

		if (keyNeeded) {	// Used when Sorted Outer Union is needed to identify rows
			List souArgList = new ArrayList();
			souArgList.add(xQuery);
			_genAlgebraVisitor.noticSortedOuterUnion(souArgList);
		}
		xQuery.accept(_genAlgebraVisitor);
		algebraTree = _genAlgebraVisitor.getAlgebraTree();

		// Clean useless cascaded projection that are systematically generated 
		// because of Qdb intermediate let clauses
		_removeProjectVisitor.reinit();
		algebraTree.accept(_removeProjectVisitor);

		// Browse algebra tree to set the _refferedAttributes variable on nodes
		algebraTree.accept(_brav);

		/* resolve invalid predicates */
		_apv.reinit();
		algebraTree.accept(_apv);

		/* Transforms a SELECT or WHERE with aggregation in the select clause into an outer join */		
		_rsv.reinit(algebraTree);
		algebraTree.accept(_rsv);
		algebraTree = _rsv.getRewrittenQuery();	// root can be changed
		
		// TODO: check if batchs could not be used when multiple statements are issued

		/* Sets the type field for every tree node */
		_algebraTypeVisitor.reinit();
		algebraTree.accept(_algebraTypeVisitor);

		// QUESTION : what's this visitor ?
		_pev.setGenSqlVisitor(_genSqlVisitor);
		_pev.setJdbcStatement(_jdbcStatement);
		algebraTree.accept(_pev);

		/* Formats top-level projection to retrieve a schema valid output for dates */
		FormatDateTime.formatDateTimes(algebraTree);

		mapper = algebraTree.getMapper();

		/* Generates sql from semantically supported algebra */
		_genSqlVisitor.reinit();
		sql = algebraTree.accept(_genSqlVisitor);

		XMLDBCException exception = null;

		List requestList = sql.generateRequestList();
		Trace.log("=================< SQL >=================");
		for (int i = 0; i < requestList.size(); i++) {
			Trace.log(requestList.get(i));
		}
		Trace.log("================< Mapper >================");
		Trace.log(mapper.pprint());

		try {
			retVal =
				new SynchronizedResultSet(
					requestList,
					mapper,
					getJdbcStatement());
		} catch (XMLDBCException ex) {
			exception = ex;
		} catch (SQLException ex) {
			exception = new XMLDBCException(ex.getMessage(), ex);
		}

		try {
			Statement jdbcStatement = getJdbcStatement();
			sql.cleanAll(jdbcStatement);
			jdbcStatement.close();
		} catch (Exception ex) {}

		if (null != exception) {
			throw exception;
		}

		Trace.exit(
			this,
			"executeSingleQDB(XQueryExpression xQuery, boolean keyNeeded)");
		return retVal;
	}
}
