/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor.mysql;

import java.util.*;

import org.xquark.extractor.algebra.*;
import org.xquark.extractor.common.Debug;
import org.xquark.extractor.common.SqlWrapperException;
import org.xquark.extractor.common.Trace;

/**
 * This visitor rewrites SQL queries so that nested subqueries in SELECT or
 * WHERE are transformed into piped LEFT OUTER JOINS ...GROUP BY... HAVING.
 * 
 * This visitor
 * Rules pipe :
 * 
 * 1st step :
 *  
 * Scan predicates (in restrictions and joins) for aggregates and set operators
 * (IN, ANY, EXIST, NOT EXIST) and pull up them as "SELECT in SELECT".
 * 
 * 2nd step :
 * 
 * "SELECT in SELECT" are rewritten as "LEFT OUTER JOIN...GROUP BY... HAVING"
 */
public class MysqlRemoveSubQueries extends RemoveNestedAggregation {

    private static final String RCSRevision = "$Revision: 1.3 $";
    private static final String RCSName = "$Name:  $";

	protected Expression _rewrittenQuery;
	protected Mapper _topMapper; // keep the original because making a new one may be dangerous (original expressions do not always exist...)
    protected FindQuerySIgnatureVisitor _keyNodesFinder = new FindQuerySIgnatureVisitor();
	protected SplitPredicateVisitor _splitVisitor = null;  // initialized on demand
	protected RemoveAggregateFunctionVisitor _removeAggrVisitor = null;  // initialized on demand
	protected RenameAttributeExpressionVisitor _renameVisitor = null;
    protected Stack _subQueryStack = new Stack();
	protected Expression _current;
    public MysqlRemoveSubQueries() {}

    public void reinit(Expression query) {
        _rewrittenQuery = query;
        _topMapper = query.getMapper();
    }

    public Expression getRewrittenQuery() {
        return _rewrittenQuery;
    }

	public void setRewrittenQuery(SubQueryNode query) {
		if (query._type == SubQueryNode.TYPE_MAIN) {
			((Expression)query._query).setMapper(_topMapper);
			if (query._referringNode == null)
				_rewrittenQuery = (Expression)query._query;
			else
				_rewrittenQuery = query._referringNode;
		}
		// STOPPED HERE: either keep original rename item and use the original mapper or recreate a mapper but make it insensible to key without ref XQuery (simpler)...
		// UC11: What to do with independant aggregation ? A cartesian product on a temp table ?
	}

    public void visit(UnOpProject arg) {
        Trace.enter(this, "visit(UnOpProject arg)", arg);
        browseProjection(arg, SubQueryNode.CONTEXT_PROJECT);
        Trace.exit(this, "visit(UnOpProject arg)", arg);
    }

    public void visit(UnOpAggregate arg) {
        Trace.enter(this, "visit(UnOpProject arg)", arg);
        browseProjection(arg, SubQueryNode.CONTEXT_AGGR);
        Trace.exit(this, "visit(UnOpProject arg)", arg);
    }

    private void browseProjection(UnOpProject arg, byte context) {
        SubQueryNode query = null;
        
        if (_subQueryStack.isEmpty())
            query = new SubQueryNode(arg);
        else {
			SubQueryNode parentQuery = (SubQueryNode) _subQueryStack.peek();
			/* a subquery must have been already created */
			if (parentQuery.getType() != SubQueryNode.TYPE_PREDICATE_EXIST)
				query = new SubQueryNode(arg, context);
        }

		if (query != null)
        	query.setBrowsingStep(SubQueryNode.STEP_OPERAND);

        arg.getOperand().accept(this);

		if (query != null)
	        query.setBrowsingStep(SubQueryNode.STEP_ITEMS);
	        
        List list = arg.getParameterList();
        if (null != list) {
            for (int i = 0; i < list.size(); i++) {
                _current = (Expression) list.get(i);
                _current.accept(this);
            }
        }

		if (query != null)
	        setRewrittenQuery(query.rewrite());
    }
    
    public void visit(UnOpRestrict arg) {
        Trace.enter(this, "visit(UnOpRestrict arg)", arg);
        SubQueryNode query = (SubQueryNode) _subQueryStack.peek();

        query.setBrowsingStep(SubQueryNode.STEP_PREDICATES);
        List list = arg.getPredicateList();
        if (null != list) {
            for (int i = 0; i < list.size(); i++) {
                _current = (Expression) list.get(i);
                _current.accept(this);
            }
        }

        query.setBrowsingStep(SubQueryNode.STEP_OPERAND);
        arg.getOperand().accept(this);

        Trace.exit(this, "visit(UnOpRestrict arg)", arg);
    }

    public void visit(Join arg) {
        Trace.enter(this, "visit(Join arg)", arg);
        SubQueryNode query = (SubQueryNode) _subQueryStack.peek();

        query.setBrowsingStep(SubQueryNode.STEP_PREDICATES);
        List list = arg.getPredicateList();
        if (null != list) {
            for (int i = 0; i < list.size(); i++) {
                _current = (Expression) list.get(i);
                _current.accept(this);
            }
        }

        query.setBrowsingStep(SubQueryNode.STEP_OPERAND);
        list = arg.getOperandList();
        if (null != list) {
            for (int i = 0; i < list.size(); i++) {
                _current = (Expression) list.get(i);
                _current.accept(this);
            }
        }

        Trace.exit(this, "visit(Join arg)", arg);
    }

    public void visit(BinOpCompareAny arg) {
        SubQueryNode query =
            new SubQueryNode(
                (Relation) arg.getRightOperand(),
                SubQueryNode.CONTEXT_ANY);
        super.visit((BinaryAtomicOp) arg);
        query.rewrite();
    }

    public void visit(UnOpExists arg) {
        SubQueryNode query = new SubQueryNode((Relation) arg.getOperand(), SubQueryNode.CONTEXT_EXIST);
        super.visit((UnaryAtomicOp) arg);
        query.rewrite();
    }

    /**
     * 2nd step of rewriting:
     * 1. Detects aggregations in projection.
     * 2. Aggregations are the browsed to detect those based on the same relation
     * (table or restriction on the same table) and grouped following this criteria.
     * 3. Redundant aggregation from the same relation items are removed preserving
     * original mapping.
     * 4. Remaining relations in projection (nested aggreagations) are then
     * transformed into a outer join and a groupby.
     * 5. Outer join predicates (out of join predicates) are splitted in two: predicates
     * containing references to aggregates aliases are moved into a new restriction nodes
     * set above group by node, becoming a having close.
     * @param arg
     */
    /**
     * 1st step of rewriting:
     * 1. Detects aggregations and assimilated in restrictions. During this process
     * '= ANY' (i.e. IN which do not exist in algebra) predicates are processed like
     * EXIST. Predicates involving the aggregates and COUNT(*) > 0 (or = 0 if a NOT
     * is involved) are gathered for the HAVING clause.
     * 2. Aggregations and assimilated are renamed and pulled up to the surrounding
     * projection keeping. Join predicates not involving outer relation columns
     * should be kept in the predicate of outer relation (multi-level nesting).
     * 3. 2nd step is then piped because of depth-first trasversal of algebra
     * tree.
     * @param expr the restriction node browsed
     * @param predicateList the list of expressions that will be scanned for
     * aggregations
     */

    protected SubQueryNode transformSubquery(SubQueryNode query) {
        //
        return query;
    }
    /**
     * Creates "flatened" outer join tree for a correlated aggregation.
     * @param mainRelation the outer relation to join
     * @param subRelation the subquery relation without the projection
     * @param projectiontAttributes attributes that will be retrieved with the 
     * aggregation. The aggregation will be added to this list in order to retrieve
     * data in the next surrounding outer join.
     * @param groupByList attribute that will be used for grouping
     * @return A relation with no aliasing performed.
     * @throws CloneNotSupportedException
     */
    protected Expression createOuterJoin(
        Expression mainRelation,
        InnerRelationInfo subQuery,
        List projectiontAttributes,
        List groupByList,
        boolean distinct,
        boolean mainIsAggregate) {

        Trace.enter(this, "createOuterJoin(...)");

		/* Collect main relation predicates */
		List newRelationPredicates = pickOutPredicates(mainRelation);
		if (mainRelation instanceof UnOpRestrict)
			mainRelation = ((UnOpRestrict)mainRelation).getOperand(); // pickOutPredicates could not remove it on top

		/* Build new relation */
		Expression retRelation = null;
		if (groupByList == null) {	// main relation is a DummyTable
			retRelation = subQuery._relation;
			newRelationPredicates.addAll(subQuery._predicates);
		}
		else if (subQuery._predicates.isEmpty())	{// create a standard join instead // TODO: uncorrelated aggregates should be removed before
			if (mainRelation instanceof Join) {	// In the case main relation is already a join
				((Join)mainRelation).addOperand(subQuery._relation);
				retRelation = mainRelation;
			}
			else {
				List operands = new ArrayList(2);
				operands.add(mainRelation);
				operands.add(subQuery._relation);
				retRelation =
					new Join(operands, subQuery._predicates);
			}
		}
		else {
			retRelation =
				new BinOpOuterJoin(mainRelation,subQuery._relation,
					Constants.LEFT_OUTER_JOIN, subQuery._predicates);
		}

        // NOTE: As MySQL support predicate (out of join predicates) in ON clause (and its better
        // semantically because predicate in WHERE could remove OUTER JOIN result "null tuples") all
        // predicates (join an not) are put in ON

        /* Pull up both predicates from main relation (first Outer join only since it's piped after) */
        retRelation = new UnOpRestrict(retRelation, newRelationPredicates);

        /* create a groupby node using PK + result columns(not necessary if a PK ? Pas d'erreur en tout cas avec MySQL) */
		if (groupByList != null)
        	retRelation = new UnOpGroup(retRelation, groupByList);

        /* Create a having clause for aggregates in where */
        if (subQuery._havingClauses != null
            && !subQuery._havingClauses.isEmpty())
            retRelation =
                new UnOpRestrict(retRelation, subQuery._havingClauses, true);

        /* create projection node(s) */
        if (mainIsAggregate) {
            // in this case, create 2 projections (temp table) to avoid aggregates to interfere
            // first projection must contain all second's item but without aggregate
            // remove aggregation items and collect aggregate function parameters
            if (_removeAggrVisitor == null) {
				_removeAggrVisitor = new RemoveAggregateFunctionVisitor();
            }
            else {
            	_removeAggrVisitor.reinit();
            }

			List aggrItems = new ArrayList();
            ListIterator lit = projectiontAttributes.listIterator();
            Expression item = null;
            while (lit.hasNext()) {
            	item = (Expression) lit.next(); 
            	_removeAggrVisitor.newItem();
                item.accept(_removeAggrVisitor);
                if (_removeAggrVisitor.aggregateFound()) {
					lit.remove();
					aggrItems.add(item);
                }
            }

			// Clone the remaining projection attribute for the second projection
			List proj2Items = null;
			try {
				proj2Items =
					AlgebraTools.clone(projectiontAttributes, new HashMap()); // TODO: a simple deep clone
			
			/* create the outer join projection (first) */
			UnOpProject proj1 = null;
			if (subQuery._aggrItems != null)
				proj1 =	new UnOpAggregate(retRelation, projectiontAttributes, distinct);
			else
				proj1 =	new UnOpProject(retRelation, projectiontAttributes, distinct);
					
			RenameRelation rel1 = new RenameRelation(proj1);
					
			/* browse the collected items to insert them in the projections and
			 * substitue the alias in the initial item */
			Iterator it = _removeAggrVisitor.getCollectedItems().iterator();
			FunAggregate funAggr = null;
			AttributeExpression att = null;
			
			while (it.hasNext()) {
				item = (Expression) it.next();
				funAggr = (FunAggregate) item.getFather();	// save the parent before it is changed
				item = proj1.addItem(item);	// can be substituted with an alias if conflict
				att = new AttributeExpression(rel1, item.getName());
				att.setUnderlyingExpr(item);
				funAggr.setOperand(att);	// set the potential alias for aggregate function	
			}
			
			proj2Items.addAll(aggrItems);
			
			/* add outer join aggregates */
			if (subQuery._aggrItems != null) {
				RenameItem alias = null;
				it = subQuery._aggrItems.iterator();
				while (it.hasNext()) {
					item = (Expression) it.next();
					alias = (RenameItem) proj1.addItem(item);
					att = new AttributeExpression(rel1, alias.getName());
					att.setUnderlyingExpr(item);
					att.setOrginalXExpr(item.getOrginalXExpr()); // QUESTION: not useful now original mapper is preserved
					proj2Items.add(att);	
				}
			}
			
			/* A second projection is piped to avoid aggregation "collision" */
			// 2nd projection creation
			retRelation = new UnOpAggregate(rel1, proj2Items, false);
			}
			catch (Exception ex) {
				throw new SqlWrapperException(ex.getMessage(), ex);
			}
        }
        else {
			/* create the outer join projection */
			if (subQuery._aggrItems != null) {
				projectiontAttributes.addAll(subQuery._aggrItems);
				retRelation =
					new UnOpAggregate(retRelation, projectiontAttributes, distinct);
			}
			else if (!projectiontAttributes.isEmpty())
				retRelation =
					new UnOpProject(retRelation, projectiontAttributes, distinct);
        }
        
                
        Trace.exit(this, "createOuterJoin(...)");
        return retRelation;
    }

    /**
     * Pick out all remaining predicates (join predicates have been removed) on
     * an outer join.
     * @param tree outer join to browse
     * @return a list of predicates.
     */
    protected List pickOutPredicates(Expression tree) {
        Trace.enter(this, "pickOutPredicates(Expression tree)", tree);
        List retVal = new ArrayList();

        /* find out UnOpRestrict nodes */
        List joinPredicateList = new ArrayList();
        _restrictionVisitor.reinit();
        tree.accept(_restrictionVisitor);
        List restrictionList = _restrictionVisitor.getRestrictions();

        /* browse found nodes */
        UnOpRestrict restrict = null;
        Set vti = null;

        Expression expr;
        for (int i = 0; i < restrictionList.size(); i++) {
            expr = (Expression) restrictionList.get(i);

            if (expr instanceof UnOpRestrict) {
                restrict = (UnOpRestrict) expr;
                retVal.addAll(restrict.getPredicateList());
                /* Remove the restriction node if possible and compact potential piled relation alias*/
                Expression father = restrict.getFather();
                if (father != null)
					father.replaceChild(restrict, restrict.getOperand());
            }
            else
                Debug.nyi("pickOutPredicates(Expression tree)");
        }

        Trace.exit(this, "pickOutPredicates(Expression tree)", tree);
        return retVal;
    }

    /**
     * Find out in a list of items or predicates any Subquery and build a corresponding
     * temporary structure to store them in a list waiting for the next step of
     * rewriting.
     * @param exprList the list of items or predicates to be scanned. Matching items
     * are removed from the list.
     * @param subQueries the initial list of subquery expressions.
     * @param isPredicate flag allowing to know if a having clause must be generated
     * @return the subquery list augmented with newly detected.
     */

    protected class SubQueryNode {
        /* context of discovery */
        public static final byte CONTEXT_PROJECT = 0;
        public static final byte CONTEXT_AGGR = 1;
        public static final byte CONTEXT_ANY = 2;
        public static final byte CONTEXT_EXIST = 3;

        /* subquery types */
        public static final byte TYPE_MAIN = 0;
        public static final byte TYPE_SELECT_AGGR = 1;
        public static final byte TYPE_PREDICATE_AGGR = 2;
        public static final byte TYPE_PREDICATE_EXIST = 3;
        public static final byte TYPE_PREDICATE_ANY = 4;
        public static final byte TYPE_FROM = 5; // not processed here

        /* browsing step */
        public static final byte STEP_NONE = 0;
        public static final byte STEP_ITEMS = 1;
        public static final byte STEP_PREDICATES = 2;
        public static final byte STEP_OPERAND = 3;
        public static final byte STEP_DONE = 4;

        public byte _type = TYPE_MAIN;

        public byte _browsingStep = STEP_NONE;

        /**
         * Base expression. Predicate if predicate subquery.
         */
        private Expression _anchor = null;

        /**
         * the Rename surrounding subquery.
         */
        private UnaryOperator _renamedQuery = null;

        /**
         * The parent node of subquery expression (exception made of Rename).
         */
        private Expression _referringNode = null;

        /**
         * The subquery.
         */
        private Relation _query = null;

        /**
         * SubQueryNode.
         */
        private List _subQueryList = new ArrayList();

        public SubQueryNode(Expression mainQuery) {
            _type = TYPE_MAIN;
            _query = (UnaryAlgebra) mainQuery;
            _referringNode = mainQuery.getFather();

			// add subquery to stack
			_subQueryStack.push(this);
        }

        /* parentQuery not null */
        public SubQueryNode(Relation root, byte context) {
            SubQueryNode parentQuery = (SubQueryNode) _subQueryStack.peek();

            _anchor = _current;
            Expression subQuery = (Expression) root;
            _referringNode = subQuery.getFather();

            /* set members according to context */
            switch (context) {
                case CONTEXT_PROJECT :
					/* set type */
                    // by default we consider that a simple projection is either root, either in from
                    _type = TYPE_FROM; // nothing performed
                    // rename may be parent
                    _query = root;
                    if (_referringNode instanceof RenameRelation) {
                        _renamedQuery = (UnaryOperator) subQuery.getFather();
                        _referringNode = _renamedQuery.getFather();
                    }

                    break;

                case CONTEXT_AGGR :
                    /* set type */
                    switch (parentQuery.getBrowsingStep()) {
                        case STEP_ITEMS :
                            _type = TYPE_SELECT_AGGR;
							_renamedQuery = (UnaryOperator) _referringNode;
							_referringNode = _renamedQuery.getFather();
								// assume there is always a RenameItem
                            break;
                        case STEP_PREDICATES :
                            _type = TYPE_PREDICATE_AGGR; // nothing performed
                            break;
                        default :
                            _type = TYPE_FROM; // nothing performed
                    }
                    // rename may be parent
                    _query = root;

                    break;

                case CONTEXT_ANY :
                    /* set type */
                    _type = TYPE_PREDICATE_ANY;
                    // rename may be itself
                    if (root instanceof RenameRelation) {
                        _renamedQuery = (UnaryOperator) subQuery;
                        _query =
                            (UnaryAlgebra) ((RenameRelation) root).getOperand();
                    } else
                        _query = root;

                    break;
                case CONTEXT_EXIST :
                    /* set type */
                    _type = TYPE_PREDICATE_EXIST;
                    // rename may be itself
                    if (root instanceof RenameRelation) {
                        _renamedQuery = (UnaryOperator) subQuery;
                        _query =
                            (UnaryAlgebra) ((RenameRelation) root).getOperand();
                    } else
                        _query = root;
                    break;

                default :
                    break;
            }

            // create an alias if missing
            if (_referringNode == null) {
                switch (_type) {
                    case TYPE_PREDICATE_ANY :
                    case TYPE_PREDICATE_EXIST :
                    case TYPE_FROM :
                        _referringNode =
                            new RenameRelation((Expression) _query);
                        break;

                    default :
                        Debug.nyi(
                            "Aggregates should have surrounding RenameItem.");
                }
            }

            // canonize subquery
            canonize();

            // add subquery to parent if to be rewritten
            if (_type != TYPE_FROM)
                parentQuery.addSubQuery(this);

            // add subquery to stack
            _subQueryStack.push(this);
        }

        /**
         * 
         */
        private void canonize() {

            /**
             * ANY (i.e. IN) predicate subquery processing. Transform it in a EXIST
             * by adding a join predicate to the subquery then pipe the EXIST process.
             */
            switch (_type) {
                case TYPE_PREDICATE_ANY :
                    /* transform this BinOpCompareAny in an 'homothetic' BinOpCompare with
                     * on the one hand, the simple attribute expression and the other the
                     * attribute expression of the subquery (remove the projection) as
                     * operands and add this as a new subquery predicate
                     */
                    // by convention left operand is the scalar and right the subquery
                    UnOpProject subQueryProjection = (UnOpProject) _query;
                    BinOpCompareAny original = (BinOpCompareAny) _referringNode;
                    AttributeExpression attExpr1 =
                        (AttributeExpression) original.getLeftOperand(),
                        attExpr2 =
                            (AttributeExpression) subQueryProjection
                                .getItemList()
                                .get(
                                0);
                    BinOpCompare newPredicate =
                        new BinOpCompare(
                            original.getOperator(),
                            attExpr1,
                            attExpr2);
                    // QUESTION: is it really correct to use straight the original operator

                    /* the referred attribute field (used by exist processing) */
                    Set refAtts = new HashSet();
                    refAtts.add(attExpr1);
                    refAtts.add(attExpr2);
                    newPredicate.setReferredAttributes(refAtts);

                    /* Find the restriction, join or create it if missing and add the predicate */
                    _keyNodesFinder.reinit();
                    subQueryProjection.accept(_keyNodesFinder);
                    PredicateHolder restrict =
                        _keyNodesFinder.getRestriction();
                    if (restrict == null) { // insert a restriction node
                        Expression relation = _keyNodesFinder.getRelation();
                        relation.getFather().replaceChild(
                            relation,
                            new UnOpRestrict(relation, newPredicate));
                    } else {
                        restrict.addPredicate(newPredicate);
                    }

                    /* Modify this node */
                    _type = TYPE_PREDICATE_EXIST;
                    _query = (Relation) subQueryProjection.getOperand();
                    _referringNode = new UnOpExists((Expression)_query);	// used later to fetch "not"
                    break;

                case TYPE_PREDICATE_EXIST :
                    /* remove Potential UnOpNot */
                    // TODO: check. in fact we consider that normalization prevent NOT(NOT EXISTS())
                    if (_anchor instanceof UnOpNot) {
                        UnOpExists exists = (UnOpExists) _referringNode;
                        // Note : could check UnOpNot's operand is exist...
                        //_anchor.getFather().replaceChild(exists); // _anchor is kept for predicate removal
                        exists.setNot(exists.getNot() ? false : true);
                    }
                    
                    /* remove useless projection (should not happen ?) */
                    if (_query instanceof UnOpProject) {
                    	UnOpProject project = (UnOpProject)_query; 
                    	_query = (Relation) project.getOperand();
                    	if (_referringNode != null)
                    		_referringNode.replaceChild(project, (Expression) _query);
                    	else if (_renamedQuery != null)
                    		_renamedQuery.setOperand((Expression) _query);
                    }
					break;

                default :
                    }
        }

        public SubQueryNode rewrite() {

            // process dummy queries without subqueries but with a predicate (which MySQL does not handle)
            // introduces an UnaryAlgebra chain on a RenameRelation which carries out at generation time
            if (_subQueryList.isEmpty()) {
            	_keyNodesFinder.reinit();
            	((Expression)_query).accept(_keyNodesFinder);
				
            	if (_keyNodesFinder.getRestriction() instanceof UnOpRestrict) {
            		UnOpRestrict restrict = (UnOpRestrict) _keyNodesFinder.getRestriction();
            		if (_keyNodesFinder.getRelation() instanceof DummyTable) {
            			/* clone the project to put it in the inner relation */
						UnOpProject innerProject = null;
            			try {
            				innerProject = (UnOpProject) _keyNodesFinder.getProjection().clone();
						}
						catch (Exception ex) {
							throw new SqlWrapperException(ex.getMessage(), ex);
						}
						
						/* create a rename for the inner relation */
						RenameRelation rename = new RenameRelation(innerProject);
						
						/* insert the new sequence of nodes above the dummy table */
						DummyTable dummy = (DummyTable) _keyNodesFinder.getRelation();
						dummy.getFather().replaceChild(dummy, rename);
						innerProject.setOperand(dummy);
            		}
            	}
            }
            else { // process rewritting of this relation with one level of subquery
                Iterator it = _subQueryList.iterator();
                SubQueryNode subQuery = null;
                InnerRelationInfo relation = null;
                List subRelations = new ArrayList();
                
				/* remove subqueries before complex EXIST rewritting that browses predicates */
				// NOTE : do it now not to interfere with tree algebra iteration !!!
                while (it.hasNext()) {
                    subQuery = (SubQueryNode) it.next();

                    switch (subQuery.getType()) {
                        case TYPE_SELECT_AGGR :
                            ((UnOpProject) subQuery._anchor.getFather()).removeItem(
							subQuery._anchor);
                            break;

                        case TYPE_PREDICATE_AGGR :
                        case TYPE_PREDICATE_ANY : // normally canonization should have remove it
                        case TYPE_PREDICATE_EXIST :
                            /* remove predicate from outer relation */
                            ((PredicateHolder) subQuery._anchor
                                    .getFather())
                                    .removePredicate(
							subQuery._anchor);
                            break;

                        default :
                    }
                }

				it = _subQueryList.iterator();
                while (it.hasNext()) {
                    subQuery = (SubQueryNode) it.next();

                    switch (subQuery.getType()) {
                        case TYPE_PREDICATE_AGGR :
                            relation = processPredicateAggr(subQuery);
                            break;
                        case TYPE_PREDICATE_EXIST :
                            relation = processExist(subQuery);
                            break;
                        case TYPE_SELECT_AGGR :
                            relation = processSelectAggr(subQuery);
                            break;

                        default :
                            Debug.nyi(
                                "The subquery type "
                                    + subQuery.getType()
                                    + " should not occur at such a stage of rewritting.");
                    }
                    subRelations.add(relation);
                }
                
                /* Optimization: limit the number of potential OUTER joins by grouping identical relations */
                subRelations = partitionSubQueriesByRelation(subRelations);
                
				/* find out PK and other parameters for outer join creation*/
				boolean distinct = false, mainIsAggregate = _query instanceof UnOpAggregate;
				List items = new ArrayList(); // this list will be augmented at every piped outer join
				Expression mainRelation = (Expression) _query;
				if (_query instanceof UnOpProject) {
//					if (_type != TYPE_PREDICATE_EXIST) {
					UnOpProject main = (UnOpProject)_query;
					distinct = main.getDistinct();
					items.addAll(main.getItemList());
					mainRelation = ((UnaryOperator)mainRelation).getOperand();
				}
				
				// Note: performed on relation and not root projection because
				// the visitor add keys to items
				_addKeysVisitor.reinit();
				mainRelation.accept(_addKeysVisitor);
				List keys = ((Relation)mainRelation).getKeys();

				/* Create a subtree(containing UnOpAggregate and UnOpgroup) for each aggregation */
				Expression outerJoin = null;
				
				for (int i = 0; i < subRelations.size(); i++) {
					relation = (InnerRelationInfo)subRelations.get(i);
					
					/* Builds 1st outer join with all projection columns */
					if (i == 0) {
						outerJoin = createOuterJoin (mainRelation, relation, items, keys, distinct, mainIsAggregate);
					}
					else {
						// rename the previous relation
						outerJoin = new RenameRelation(outerJoin);
						outerJoin = createOuterJoin (outerJoin, relation, items, keys, distinct, mainIsAggregate);
					}
				}
				_query = (Relation)outerJoin;
            }


            // remove from stack
            _subQueryStack.pop();

			//connect the original variable for reconstruction if topmost
            if (_subQueryStack.isEmpty()) {
                if (_referringNode != null) {
                    ((UnaryOperator) _referringNode).setOperand(
                        (Expression) _query);
                }
                else if (_renamedQuery != null) {
                    _renamedQuery.setOperand((Expression) _query);
                }
            }
			
            return this;
        }

        /**
         * @param subQuery
         * @return
         */
        private InnerRelationInfo processSelectAggr(SubQueryNode subQuery) {
			/* connect to the original XQuery expr */ // Assuming there is only one
			List itemList = ((UnOpAggregate)subQuery._query).getItemList();
			RenameItem item = (RenameItem) itemList.get(0);
			item.setOrginalXExpr(subQuery._renamedQuery.getOrginalXExpr());	// TODO: useless now original mapper is preserved
			// in order to keep the original alias for mapper, set the supressed subquery on into the item
			item.setName(subQuery._renamedQuery.getName());
			subQuery._renamedQuery.setOperand(item.getOperand());
        	
            return new InnerRelationInfo(itemList, ((UnOpAggregate)subQuery._query).getOperand());
        }

        /**
         * @param node
         * @param subQuery
         * @return
         */
        private InnerRelationInfo processExist(SubQueryNode subQuery) {
			
			Set outerTables = _query.visibleTableInstances();
			List projectAttributes = new ArrayList();
			
			/* EXTERNAL/OUTER JOIN PREDICATES
			 * Browse main query predicates to find external join predicates with
			 * the outer table in order to add outer table attributes to
			 * projection (do it first not to interfere with the possible
			 * introduction of a join that may add  predicates too)
			 */
			 PredicateHolder holder = (PredicateHolder) subQuery._anchor.getFather();
			 Iterator it = holder.getPredicateList().iterator(), it2 = null;
			 Expression predicate = null;
			 AttributeExpression att = null, outerAtt = null;
			 int nbOatt = 0, i = 0;
             
			 while (it.hasNext()) {
				predicate = (Expression) it.next();
				it2 = predicate.getReferredAttributes().iterator();
				outerAtt = null;
				i = nbOatt = 0;
				while (it2.hasNext()) {
					att = (AttributeExpression) it2.next();
					i++;
					if (outerTables.contains(att.getTableInstance())) {
						nbOatt++;
						outerAtt = att;
					}
				}
             	
				try {
					if (i == 2 && nbOatt == 1)
						projectAttributes.add(outerAtt.clone());
				}
				catch (Exception ex) {
					throw new SqlWrapperException(ex.getMessage(), ex);
				}
			 }

			/* EXTERNAL/INNER JOIN PREDICATES
			/* gather external conditions without removing them */
			ExternalJoinInfo sortInfo =
				sortSubQueryPredicates(
					(Expression)subQuery._query,
					outerTables
					);
                    
			/* Piped Exists: add external table in the current exist (if foreign
			 * , i.e. external, tables where found in subquery predicates)
			 * in order to apply piping exist rewriting basic rule */
            if (!sortInfo._externalTables.isEmpty()) {
            	if (_type != TYPE_PREDICATE_EXIST)
            		Debug.nyi("External join predicate should not occur out of exist subquery"); // TODO: find another exception raising method
            	
            	// find external attribute expression and clone the referred table, and create a new alias
				List renamedExternalTables = null;
				
                try {
                    renamedExternalTables =
                        AlgebraTools.clone(sortInfo._externalTables, new HashMap());
                    // Note: external tables contains renames
                    // TODO: add a classical clone to algebra nodes (deep), without map...
                }
                catch (Exception ex) {
                    throw new SqlWrapperException(ex.getMessage(), ex);
                }

            	// Change table aliases
            	RenameRelation rTable = null, oTable = null;
            	List keys = null;
            	AttributeExpression rAtt = null, oAtt;

            	int len = renamedExternalTables.size();
            	RenameAttributeExpressionVisitor renameVisitor = new RenameAttributeExpressionVisitor();
            	
            	/* insert a restriction node for external join */
				if (!(_query instanceof UnOpRestrict))
					_query = new UnOpRestrict((Expression)_query, new ArrayList());

            	for (i = 0; i < len; i++) {
					oTable = (RenameRelation) sortInfo._externalTables.get(i);
					rTable = (RenameRelation) renamedExternalTables.get(i);
            		rTable.setName(GenAlgebraVisitor.getUniqueRelationName());
            		
            		// get PK or equivalent
            		keys = oTable.getKeys();
            		if (keys == null) {
	            		_addKeysVisitor.reinit();
	            		oTable.accept(_addKeysVisitor);
						keys = oTable.getKeys(); // returns a list of attribute expressions without table part
            		}
            		
            		/* Browse keys to buid projection and join predicates */
            		it = keys.iterator();
            		while (it.hasNext()) {
            			oAtt = (AttributeExpression) it.next();
            			oAtt.setTableInstance(oTable);
            			rAtt = new AttributeExpression(rTable, oAtt.getName());
            			rAtt.setUnderlyingExpr(oAtt.getUnderlyinExpr());
            			projectAttributes.add(rAtt);
						// 2 different attribute expressions for projection and predicates
						// must be created in case of late renaming (temporary table)
						rAtt = new AttributeExpression(rTable, oAtt.getName());	// TODO: make a clone when it is cleaned
						rAtt.setUnderlyingExpr(oAtt.getUnderlyinExpr());
						((UnOpRestrict)_query).addPredicate(new BinOpCompare(Constants.EQ_COMPOP,
						rAtt, oAtt));
            		}
            		
					/* Replace external table by the added one into external join predicates */
					renameVisitor.reinit(Collections.singletonList(oTable), rTable);
					AlgebraTools.applyVisitorOnCollection(renameVisitor, sortInfo._externalJoinPredicates);
            	}
            	
            	/* add an artificial natural join (using keys) for these tables */
            	Expression outerTable = (Expression) outerTables.iterator().next();
            	Join join = null;
            	if (outerTable.getFather() instanceof Join) {
            		join = (Join)outerTable.getFather();
            	}
            	else {
					renamedExternalTables.add(outerTable); // Must be one if no join
            		join = new Join();
            		outerTable.getFather().replaceChild(outerTable, join);
            	}
            	join.addOperandList(renamedExternalTables);
            }
            
            	
			// add a projection with external joins attributes involved (in order to keep only needed columns...)
			if (!projectAttributes.isEmpty()) {
				UnOpProject project = null;
				if (project instanceof UnOpProject)
					project = (UnOpProject) _query;
				else
					project = new UnOpProject((Expression)_query);
				_referringNode.replaceChild((Expression)_query, project);
				_query = project;
				project.addItemList(projectAttributes);
			}
            
            /* build the having predicate */
			AttributeExpression innerTupleRepresentative = null;
            try {
                innerTupleRepresentative = (AttributeExpression)
                	findNonNullAttribute(subQuery._query.providedTableInstances()).clone();
            }
            catch (Exception ex) {
                throw new SqlWrapperException(ex.getMessage(), ex);
            }
            
            BinOpCompare having =
                new BinOpCompare(
                    ((UnOpExists) subQuery._referringNode).getNot()
                        ? Constants.EQ_COMPOP
                        : Constants.GT_COMPOP,
                    new FunAggregate(
                        FunAggregate.COUNT,
						innerTupleRepresentative,
                		// TODO: in fact should be a list but FunAggregate is an UnaryOperator
                		false
                		),
                	new LitInteger(0));
                
            /* remove predicate from outer relation */
			((PredicateHolder)subQuery._anchor.getFather()).removePredicate(subQuery._anchor);
			
			return new InnerRelationInfo(
					(Expression)subQuery._query,
					Collections.singletonList(having));
        }

        /**
         * @param subQuery
         * @return
         */
        private InnerRelationInfo processPredicateAggr(SubQueryNode subQuery) {
        	/* Replace subquery by an alias in the predicate */
			Expression aggr = ((RenameItem)((UnOpAggregate)subQuery._query).getItemList().get(0)).getOperand();
			// must be only one in a predicate
			subQuery._referringNode.replaceChild((Expression) subQuery._query, aggr);

            return new InnerRelationInfo(
            		((UnOpAggregate)subQuery._query).getOperand(),
            		Collections.singletonList(subQuery._anchor));
        }

        /**
         * Create a partition between aggregations based on incoming relation and not
         * aggregation itself and merge subqueries. The alias kept is the surrounding
         * one if the aggreagtion has one item (in case it is referenced).
         * @param aggregationList
         * @return a list of {@link AggregatingExpression}. 
         */
        private List partitionSubQueriesByRelation(List subQueryList) {
            ArrayList ret = new ArrayList();

            int size;
            Iterator it = subQueryList.iterator();
            InnerRelationInfo sub1, sub2;
            int i;

            /* partitions and merge in a single pass */
            while (it.hasNext()) {
                size = ret.size(); // sentinel for expression storage
                sub1 = (InnerRelationInfo) it.next();

                /* browse existing partitions to find a match for unsorted */
                for (i = 0; i < size; i++) {
                    sub2 = (InnerRelationInfo) ret.get(i);
                    if (sub2.merge(sub1))
                        break;
                }
                /* no equivalent relation already encountered */
                if (i == size) {
                    ret.add(sub1);
                }
            }
            return ret;
        }

        /**
         * @return
         */
        public byte getBrowsingStep() {
            return _browsingStep;
        }

        /**
         * @return
         */
        public byte getType() {
            return _type;
        }
        
        /**
         * @param b
         */
        public void setBrowsingStep(byte b) {
            _browsingStep = b;
        }

        public void addSubQuery(SubQueryNode subQuery) {
            _subQueryList.add(subQuery);
        }
    }

    protected class InnerRelationInfo {
        public List _aggrItems;
        public RenameRelation _relation;
		public List _havingClauses;
		public List _predicates;	// used for building the outer join predicate (all are put in the ON clause)

        public InnerRelationInfo(List items, Expression relation) {
        	this(relation, items, null);
        }

		public InnerRelationInfo(Expression relation, List havingClauses) {
			this(relation, null, havingClauses);
		}
		
		private InnerRelationInfo(Expression relation, List items, List havingClauses) {
			_aggrItems = items;
			_havingClauses = havingClauses;	
			_relation = prepare(relation);	// after setting other fields
		}
			
		/**
		 * Predicates are picked up and relation may be renamed if needed.
		 * Subqueries are unconditionnaly renamed because of potential splitting
		 * in multiple Sql statements with temporary tables or views. Renaming,
		 * considering the SQL generating algorithm, would be uneasy. 
		 * @param relation
		 * @return
		 */
		private RenameRelation prepare(Expression relation) {
			/* Pick up predicate and remove restriction nodes carrying them */
			_predicates = pickOutPredicates(relation);
			// visitor could not remove the top node and multiple node may be piled
			while (relation instanceof UnOpRestrict
				&& !(((UnOpRestrict)relation).getOperand() instanceof UnOpGroup)) // the top node may be a having not to be discarded	
				relation = ((UnOpRestrict)relation).getOperand();			
			
			while (relation instanceof RenameRelation)	// no need to rename relation
				return (RenameRelation) relation;
				
			RenameRelation rename = new RenameRelation(relation);
			
			/* Prepare rename visitor */
			if (_renameVisitor == null)
				_renameVisitor = new RenameAttributeExpressionVisitor();
			_renameVisitor.reinit(((Relation)relation).visibleTableInstances(), rename);
			
			/* Rename items */
			if  (_aggrItems != null)
				AlgebraTools.applyVisitorOnCollection(_renameVisitor, _aggrItems);

			/* Rename predicates */
			if  (_predicates != null)
				AlgebraTools.applyVisitorOnCollection(_renameVisitor, _predicates);

			/* Rename havings */
			if  (_havingClauses != null)
				AlgebraTools.applyVisitorOnCollection(_renameVisitor, _havingClauses);

			return rename;
		}

        /* If relation clauses are identical, items and havings are merged to produce one relation */
        public boolean merge(InnerRelationInfo merged) {
            if (_relation.deepEquals(merged._relation)) { // relation are the same
            	
				/* Prepare rename visitor */	// Note: we assume there is one...
				RenameRelation rename = (RenameRelation)_relation.providedTableInstances().iterator().next(); 
				if (_renameVisitor == null)
					_renameVisitor = new RenameAttributeExpressionVisitor();
				_renameVisitor.reinit(merged._relation.providedTableInstances(),	rename);
			
            	/* merge aggregates */
                if (merged._aggrItems != null) {
					/* Rename items */
					AlgebraTools.applyVisitorOnCollection(_renameVisitor, merged._aggrItems);
                    if (_aggrItems == null)
                        _aggrItems = merged._aggrItems;
                    else
						_aggrItems.addAll(merged._aggrItems);
                }
				/* Compare and merge havings */
                if (merged._havingClauses != null) {
                    if (_havingClauses == null)
                        _havingClauses = merged._havingClauses;
                    else if (
                        !AlgebraTools.areExprListEquivalent(
                            _havingClauses,
                            merged._havingClauses)) {
                        // create new one cause they are final singletons...
                        _havingClauses = new ArrayList(_havingClauses);
                        _havingClauses.addAll(merged._havingClauses);
                    }
                }
				return true;
            }
            else
            	return false;
        }
    }
    
    /**
     * this class scans(by means of simple scan) a algebre tree and find out key
     * nodes of the query (does not browse subqueries).
     */

    protected class FindQuerySIgnatureVisitor extends DefaultSimpleVisitor {
        private static final String RCSRevision = "$Revision: 1.3 $";
        private static final String RCSName = "$Name:  $";

		private UnOpProject project = null;
		private PredicateHolder restrict = null;
        private Expression relation = null; // relation if no restriction found

        public FindQuerySIgnatureVisitor() {}

        public void reinit() {
			project = null;
            restrict = null;
			relation = null;
        }

		public UnOpProject getProjection() {
			return project;
		}

		public PredicateHolder getRestriction() {
			return restrict;
		}

        public Expression getRelation() {
            return relation;
        }

		public void visit(UnOpProject arg) {
			super.visit((UnaryOperator)arg);
			project = arg;
		}

		public void visit(UnOpRestrict arg) {
			super.visit((UnaryOperator)arg);
			restrict = arg;
		}

        public void visit(Join arg) {
            restrict = arg;
            relation = arg;
        }

		public void visit(Table arg) {
			relation = arg;
		}

		public void visit(DummyTable arg) {
			relation = arg;
		}

        public void visit(BinaryAlgebra arg) {
            relation = arg;
        }

        public void visit(BinOpOuterJoin arg) {
            relation = arg;
        }
    }

	protected class RemoveAggregateFunctionVisitor extends DefaultSimpleVisitor {
		private static final String RCSRevision = "$Revision: 1.3 $";
		private static final String RCSName = "$Name:  $";

		private List _collectedItems = new ArrayList();
		private boolean _aggrFound = false;
		
		public void reinit() {
			_collectedItems.clear();
			_aggrFound = false;
		}
		
		public List getCollectedItems() {
			return _collectedItems;
		}
		
		public boolean aggregateFound() {
			return _aggrFound;
		}
		
		public void newItem() {
			_aggrFound = false;
		} 
		
		public void visit(FunAggregate arg) {
			super.visit(arg);
			_aggrFound = true;
			_collectedItems.add(arg.getOperand());
		}
	}
	
    protected ExternalJoinInfo sortSubQueryPredicates(
        Expression subQuery,
        Set outerTables) {
        Trace.enter(this, "sortSubQueryPredicates()", subQuery);

		ExternalJoinInfo ret = new ExternalJoinInfo();
		
        /* find out UnOpRestrict nodes */
        _restrictionVisitor.reinit();
        subQuery.accept(_restrictionVisitor);
        List restrictionList = _restrictionVisitor.getRestrictions();

        /* browse found nodes */
        List predicateList = null;

		Set innerTables = ((Relation) subQuery).visibleTableInstances();
        // TODO: if does not exist : generate a view with automatic numbering...

        for (int i = 0; i < restrictionList.size(); i++) {
            Expression expr = (Expression) restrictionList.get(i);
            predicateList = ((PredicateHolder) expr).getPredicateList();

            /* Split: join predicates may be intermingled NO THEY AREN'T ?*/ // TODO: to it at algebra generation time 
            predicateList = splitPredicateList(predicateList);

            Expression predicate = null;
            Set allTables = null;
            Iterator it = predicateList.iterator();
            
            while (it.hasNext()) {
                predicate = (Expression) it.next();
                // get all referred tables
                allTables = predicate.getReferredTableInstances();
                // a copy is performed by the method

                // remove all inner tables: if empty simple predicates (ignored)
                allTables.removeAll(innerTables);
                allTables.removeAll(outerTables);

                // external join predicate
                if (!allTables.isEmpty()) {
					ret._externalTables.addAll(allTables);
					ret._externalJoinPredicates.add(predicate);
                }
            }
        }

        Trace.exit(this, "sortSubQueryPredicates()", subQuery);
        return ret;
    }

	protected class ExternalJoinInfo {
		public List _externalTables = new ArrayList();
		public List _externalJoinPredicates= new ArrayList();
	}

    /**
     * @param innerTables
     * @return
     */
    private AttributeExpression findNonNullAttribute(Set relations) {

        Iterator it = relations.iterator(), itColumns;
        AttributeExpression attExp = null;

        while (it.hasNext() && attExp == null) {
            attExp = ((Relation) it.next()).findNonNullAttribute();
        }

        Debug.assertTrue(attExp != null,
            "Should find a solution for rewritting if no non-null column is provided");
        // TODO
        return attExp;
    }

    /**
     * @param predicateList
     * @return
     */
    private List splitPredicateList(List predicateList) {
        int len = predicateList.size();

        if (_splitVisitor == null)
            _splitVisitor = new SplitPredicateVisitor();
        else
            _splitVisitor.reinit();

		AlgebraTools.applyVisitorOnCollection(_splitVisitor, predicateList);

        return _splitVisitor.getPredicateList();
    }

    protected class SplitPredicateVisitor extends DefaultAlgebraVisitor {
        private static final String RCSRevision = "$Revision: 1.3 $";
        private static final String RCSName = "$Name:  $";

        private List _predicates = null;

        public SplitPredicateVisitor() {
            reinit();
        }

        public void reinit() {
            _predicates = new ArrayList();
        }

        public List getPredicateList() {
            return _predicates;
        }

        public void visit(BinOpBoolean arg) {
            // perform recursion only on cascaded BinOpBoolean "AND"
            if (arg.getOperator() == Constants.AND) {
                arg.getLeftOperand().accept(this);
                arg.getRightOperand().accept(this);
            } else
                _predicates.add(arg);
        }

        public void visit(Expression arg) {
            // not a Binary boolean AND operator: gather the predicate as is
            _predicates.add(arg);
        }
    }
}
