/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor.oracle;

import org.xquark.extractor.algebra.*;
import org.xquark.extractor.common.MessageLibrary;
import org.xquark.extractor.common.SqlWrapperException;
import org.xquark.extractor.common.Trace;
import org.xquark.extractor.sql.AbstractGenSqlVisitor;
import org.xquark.extractor.sql.SqlExpression;
import org.xquark.extractor.sql.SqlFactory;

public final class OracleGenSqlVisitor extends AbstractGenSqlVisitor {

	private static final String RCSRevision = "$Revision: 1.3 $";
    private static final String RCSName     = "$Name:  $";


    public OracleGenSqlVisitor() {
        super((SqlFactory) OracleFactory.getInstance());
    }


    public SqlExpression visit(BinOpArithmetic arg)
    {
        Trace.enter(this,"visit(BinOpArithmetic arg)");

        SqlExpression  retVal = null;

        Expression      aLOprnd = arg.getLeftOperand();
        SqlExpression   sLOprnd = aLOprnd.accept(this);

        Expression      aROprnd = arg.getRightOperand();
        SqlExpression   sROprnd = aROprnd.accept(this);

        int oprtr = arg.getOperator();

		retVal = _factory.createBinOpArithmetic(oprtr, sLOprnd, sROprnd);

		if (false /** @todo   INT_DIVIDE_ARITHMETICS == oprtr &&
			aLOprnd.getType().isInteger() &&
			aROprnd.getType().isInteger()*/ ) {
			/** @todo pay attention that the round funcion on negative values */
			retVal = _factory.createSfRound(retVal);
		}

        Trace.exit(this,"visit(BinOpArithmetic arg)");
        return retVal;
    }

    protected SqlExpression visitCompareString(BinOpCompare arg)
    {
        Trace.enter(this,"visitCompareString(BinOpCompare arg)");
        SqlExpression  retVal = null;

        Expression      aLOprnd = arg.getLeftOperand();
	    SqlExpression   sLOprnd = aLOprnd.accept(this);

        Expression      aROprnd = arg.getRightOperand();
        SqlExpression   sROprnd = aROprnd.accept(this);

		int oprtr = arg.getOperator();

		if (((SqlTypeAtom)aLOprnd.getType()).getType().getPrimitiveTypeCode() == java.sql.Types.VARCHAR ||
			((SqlTypeAtom)aROprnd.getType()).getType().getPrimitiveTypeCode() == java.sql.Types.VARCHAR ) {
			/* when one of the operands is of VARCHAR type, sufix spaces are significant in Oracle, which is
			   conform to XQuery */
			retVal = _factory.createBinOpCompare(oprtr,sLOprnd,sROprnd);
		}
		else {
			/* when two operands are of CHAR (including string canstanct), sufix space are not significant, which
			 is not conform to XQuery. Here, for string comparison, we add a proper comparison of string length to
			 go with it */
			SqlExpression comp = null;
			SqlExpression eq = null;
			SqlExpression lenLOprnd = null;
			SqlExpression lenROprnd = null;
			SqlExpression lenL_lenR = null;
			SqlExpression and =  null;
			SqlExpression or =  null;

			switch (oprtr) {
				case LT_COMPOP:
					/* lOprnd < rOprnd ==> lOprnd< rOprnd or ( lOprnd = rOprnd and length(lOprnd) < length(rOprnd) ) */
				case GT_COMPOP:
					/* lOprnd > rOprnd ==> lOprnd > rOprnd or ( lOprnd = rOprnd and length(lOprnd) > length(rOprnd) ) */
					comp = _factory.createBinOpCompare(oprtr,sLOprnd,sROprnd);

					eq = _factory.createBinOpCompare(EQ_COMPOP,sLOprnd,sROprnd);

					lenLOprnd = _factory.createSfStringLength(sLOprnd);
					lenROprnd = _factory.createSfStringLength(sROprnd);
					lenL_lenR = _factory.createBinOpCompare(oprtr,lenLOprnd,lenROprnd);

					and =  _factory.createBinOpBoolean(AND,eq,lenL_lenR);

					or =  _factory.createBinOpBoolean(OR, comp, and);

					retVal = or;
					break;
				case GEQ_COMPOP:
				case LEQ_COMPOP:
				case NEQ_COMPOP:
					retVal = _factory.createBinOpCompare(oprtr,sLOprnd,sROprnd);
					break;
				case EQ_COMPOP:
					/* lOprnd = rOprnd ==> lOprnd = rOprnd and length(lOprnd) = length(rOprnd) ) */
					eq = _factory.createBinOpCompare(oprtr,sLOprnd,sROprnd);

					lenLOprnd = _factory.createSfStringLength(sLOprnd);
					lenROprnd = _factory.createSfStringLength(sROprnd);
					lenL_lenR = _factory.createBinOpCompare(oprtr,lenLOprnd,lenROprnd);

					and = _factory.createBinOpBoolean(AND,eq,lenL_lenR);
					retVal = and;
					break;
				default:
					throw new SqlWrapperException(MessageLibrary.getMessage("T_N_SUP_OPTR",COMPOPSSTRINGS[oprtr]));
			}
		}

        Trace.exit(this,"visitCompareString(BinOpCompare arg)");
        return retVal;
    }


    protected org.xquark.extractor.sql.SqlSelect selectAddFromClause (org.xquark.extractor.sql.SqlSelect select, DummyTable operand)
    {
        Trace.enter(this,"selectAddFromClause (SqlSelect select, DummyTable operand)");

        org.xquark.extractor.sql.SqlTable newFrom = _factory.createTable ("dual") ;
        select.addFromClause(newFrom);

        Trace.exit(this,"selectAddFromClause (SqlSelect select, DummyTable operand)");
        return select;
    }
}
