/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.extractor.progress;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import org.xquark.extractor.algebra.*;
import org.xquark.extractor.common.*;
import org.xquark.extractor.sql.*;
import org.xquark.extractor.xfunctions.XfContains;

public final class ProgressGenSqlVisitor extends AbstractGenSqlVisitor {
	private static final String RCSRevision = "$Revision: 1.3 $";
	private static final String RCSName     = "$Name:  $";

	public ProgressGenSqlVisitor() {
		super((SqlFactory) ProgressFactory.getInstance());
	}


	public SqlExpression visit(AttributeExpression arg)
	{
		Trace.enter(this, "visit(AttributeExpression arg)");

		SqlExpression retVal = null;
		String attrName = arg.getName();

		String tableInstanceName ;
		RenameRelation  rr = arg.getTableInstance();
		if (null!=rr){
			tableInstanceName = rr.getUniqueName();
		}
		else {
			tableInstanceName = null;
		}

		if (attrName.equals("*")) {
			retVal = _factory.createAttributeExpression (null,attrName);
		}
		else {
			retVal = _factory.createAttributeExpression (tableInstanceName,attrName);
		}
		Trace.exit(this,"visit(AttributeExpression arg)");
		return retVal;
	}

	public SqlExpression visit(BinOpArithmetic arg)
	{
		Trace.enter(this,"visit(BinOpArithmetic arg)");

		SqlExpression  retVal = null;

		Expression      aLOprnd = arg.getLeftOperand();
		SqlExpression   sLOprnd = aLOprnd.accept(this);

		Expression      aROprnd = arg.getRightOperand();
		SqlExpression   sROprnd = aROprnd.accept(this);

		int oprtr = arg.getOperator();

		retVal = _factory.createBinOpArithmetic(oprtr, sLOprnd, sROprnd);

		if (false /** @todo   INT_DIVIDE_ARITHMETICS == oprtr &&
			aLOprnd.getType().isInteger() &&
			aROprnd.getType().isInteger()*/ ) {
			/** @todo pay attention that the round funcion on negative values */
			retVal = _factory.createSfRound(retVal);
		}

		Trace.exit(this,"visit(BinOpArithmetic arg)");
		return retVal;
	}

	protected SqlExpression visitCompareString(BinOpCompare arg)
	{
		Trace.enter(this,"visitCompareString(BinOpCompare arg)");
		SqlExpression  retVal = null;

		Expression      aLOprnd = arg.getLeftOperand();
		SqlExpression   sLOprnd = aLOprnd.accept(this);

		Expression      aROprnd = arg.getRightOperand();
		SqlExpression   sROprnd = aROprnd.accept(this);

		int oprtr = arg.getOperator();

		if (((SqlTypeAtom)aLOprnd.getType()).getType().getPrimitiveTypeCode() == java.sql.Types.VARCHAR ||
			((SqlTypeAtom)aROprnd.getType()).getType().getPrimitiveTypeCode() == java.sql.Types.VARCHAR ) {
			/* when one of the operands is of VARCHAR type, sufix spaces are significant in Oracle, which is
			   conform to XQuery */
			retVal = _factory.createBinOpCompare(oprtr,sLOprnd,sROprnd);
		}
		else {
			/* when two operands are of CHAR (including string canstanct), sufix space are not significant, which
			 is not conform to XQuery. Here, for string comparison, we add a proper comparison of string length to
			 go with it */
			SqlExpression comp = null;
			SqlExpression eq = null;
			SqlExpression lenLOprnd = null;
			SqlExpression lenROprnd = null;
			SqlExpression lenL_lenR = null;
			SqlExpression and =  null;
			SqlExpression or =  null;

			switch (oprtr) {
				case LT_COMPOP:
					/* lOprnd < rOprnd ==> lOprnd< rOprnd or ( lOprnd = rOprnd and length(lOprnd) < length(rOprnd) ) */
				case GT_COMPOP:
					/* lOprnd > rOprnd ==> lOprnd > rOprnd or ( lOprnd = rOprnd and length(lOprnd) > length(rOprnd) ) */
					comp = _factory.createBinOpCompare(oprtr,sLOprnd,sROprnd);

					eq = _factory.createBinOpCompare(EQ_COMPOP,sLOprnd,sROprnd);

					lenLOprnd = _factory.createSfStringLength(sLOprnd);
					lenROprnd = _factory.createSfStringLength(sROprnd);
					lenL_lenR = _factory.createBinOpCompare(oprtr,lenLOprnd,lenROprnd);

					and =  _factory.createBinOpBoolean(AND,eq,lenL_lenR);

					or =  _factory.createBinOpBoolean(OR, comp, and);

					retVal = or;
					break;
				case GEQ_COMPOP:
				case LEQ_COMPOP:
				case NEQ_COMPOP:
					retVal = _factory.createBinOpCompare(oprtr,sLOprnd,sROprnd);
					break;
				case EQ_COMPOP:
					/* lOprnd = rOprnd ==> lOprnd = rOprnd and length(lOprnd) = length(rOprnd) ) */
					eq = _factory.createBinOpCompare(oprtr,sLOprnd,sROprnd);

					lenLOprnd = _factory.createSfStringLength(sLOprnd);
					lenROprnd = _factory.createSfStringLength(sROprnd);
					lenL_lenR = _factory.createBinOpCompare(oprtr,lenLOprnd,lenROprnd);

					and = _factory.createBinOpBoolean(AND,eq,lenL_lenR);
					retVal = and;
					break;
				default:
					throw new SqlWrapperException(MessageLibrary.getMessage("T_N_SUP_OPTR",COMPOPSSTRINGS[oprtr]));
			}
		}

		Trace.exit(this,"visitCompareString(BinOpCompare arg)");
		return retVal;
	}

	public SqlExpression visit(SortSpecification arg)
	{
		Trace.enter(this,"visit(SortSpecification arg)", arg);
		SqlSortSpecification   retVal = null;
		Expression sortOperand = ((UnOpSort)arg.getFather()).getOperand();
		if (sortOperand instanceof BinaryAlgebra) {
			Expression sortExpr = arg.getSortExpression();

			SqlExpression   sqlSortExpr = sortExpr.accept(this);

			retVal = _factory.createSortSpecification(sqlSortExpr,arg.getSortDirection());
		}
		else {
			Expression sortExpr = arg.getSortExpression();
			Expression underlyingExpr = null;
			if (sortExpr instanceof AttributeExpression) {
				underlyingExpr = ((AttributeExpression)sortExpr).getUnderlyinExpr();
				if (underlyingExpr instanceof AttributeExpression) {
					sortExpr = underlyingExpr;
				}
				else if (underlyingExpr instanceof RenameItem ) {
					Expression operandOfUnderlyingExpr = ((RenameItem)underlyingExpr).getOperand();
					if (operandOfUnderlyingExpr instanceof AttributeExpression) {
						sortExpr = operandOfUnderlyingExpr;
					}
				}
			}
			SqlExpression   sqlSortExpr = sortExpr.accept(this);

			retVal = _factory.createSortSpecification(sqlSortExpr,arg.getSortDirection());
		}
		Trace.exit(this,"visit(SortSpecification arg)", arg);
		return retVal;
	}


	public SqlExpression visit(XfContains arg) throws SqlWrapperException
	{
		Trace.enter(this,"visit(XfContains arg)");

		Debug.assertTrue(2 == arg.getArgumentNumber(),"2 == arg.getArgumentNumber()");
		org.xquark.extractor.progress.sql.SfInStr inStr = new org.xquark.extractor.progress.sql.SfInStr();

		Expression aArg0 = (Expression)arg.getArgument(0);
		SqlExpression sArg0 = aArg0.accept(this);

		Expression aArg1 = (Expression)arg.getArgument(1);
		SqlExpression sArg1 = aArg1.accept(this);

		inStr.setArgument(0, sArg0);
		inStr.setArgument(1, sArg1);

		SqlBinOpCompare retVal = _factory.createBinOpCompare(GT_COMPOP, inStr,  _factory.createLitInteger(new Integer(0)));

		Trace.exit(this,"visit(XfContains arg)");
		return retVal;
	}

	protected org.xquark.extractor.sql.SqlSelect selectAddFromClause (org.xquark.extractor.sql.SqlSelect select, BinOpOuterJoin arg)
	{
		Trace.enter(this,"selectAddFromClause (SqlSelect select, BinOpOuterJoin arg)", arg);

		org.xquark.extractor.sql.SqlExpression newFrom = arg.accept(this);
		select.addFromClause(newFrom);

		Trace.exit(this,"selectAddFromClause (SqlSelect select, BinOpOuterJoin arg)", arg);
		return select;
	}

	public org.xquark.extractor.sql.SqlExpression visit(BinOpOuterJoin arg) {
		Trace.enter(this, "visit(BinOpOuterJoin arg)");

		Expression[] aOprnds =  {arg.getLeftOperand(), arg.getRightOperand()};
		org.xquark.extractor.sql.SqlExpression[] sOprnd = new SqlExpression[2];
		Set providedTableInstances;
		
		/* Generates operands */
        for (int i = 0; i < aOprnds.length; i++) {
            providedTableInstances =
                ((Relation) aOprnds[i]).providedTableInstances();
            if (1 == providedTableInstances.size()) {
                //			Debug.assertTrue(ti instanceof RenameRelation, "ti instanceof RenameRelation");
                RenameRelation ti =
                    (RenameRelation) providedTableInstances.toArray()[0];

                if (!(aOprnds[i] instanceof RenameRelation)) {
					aOprnds[i] = new RenameRelation(aOprnds[i], ti.getName());
                }

				sOprnd[i] = aOprnds[i].accept(this);
            } else if (aOprnds[i] instanceof BinOpOuterJoin) {
				sOprnd[i] = aOprnds[i].accept(this);
            } else {
                Debug.nyi("complex outerjoin");
            }
        }


		int outerJoinType = arg.getJoinType();

		/* Outer join predicate creation */
		List aPredicateList = arg.getPredicateList();
		List sPredicateList = null;
		if (null != aPredicateList) {
			/* translate predicates */
			sPredicateList = new ArrayList();
			Expression aPredicate = null;
			org.xquark.extractor.sql.SqlExpression sPredicate = null;

			for (int i = 0; i < aPredicateList.size(); i++) {
				aPredicate = (Expression)aPredicateList.get(i);
				sPredicate = aPredicate.accept(this);
				sPredicateList.add(sPredicate);
			}
		}

		/* Outer join clause creation */
		org.xquark.extractor.sql.SqlExpression retVal =
			new org.xquark.extractor.sql.SqlBinOpOuterJoin(sOprnd[0], sOprnd[1], outerJoinType, sPredicateList);

		Trace.enter(this, "visit(BinOpOuterJoin arg)");
		return retVal;
	}

	protected org.xquark.extractor.sql.SqlSelect selectAddFromClause (org.xquark.extractor.sql.SqlSelect select, DummyTable operand)
	{
		Trace.enter(this,"selectAddFromClause (SqlSelect select, DummyTable operand)");

		org.xquark.extractor.sql.SqlTable newFrom = _factory.createTable ("dual") ;
		select.addFromClause(newFrom);

		Trace.exit(this,"selectAddFromClause (SqlSelect select, DummyTable operand)");
		return select;
	}
}
