/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.fusion;

import java.io.*;
import java.net.URL;
import java.net.URLClassLoader;
import java.sql.Connection;
import java.util.*;

import javax.xml.parsers.SAXParserFactory;

import org.xquark.extractor.ExtractorReadOnlyConnection;
import org.xquark.extractor.common.Log;
import org.xquark.extractor.common.Trace;
import org.xquark.serialize.XMLSerializer;
import org.xquark.util.misc.Debug;
import org.xquark.xml.xdbc.*;

/**
 * Objects of this class encapsulate a JDBC connection and provide access to both
 * the XQuery and insertion capabilities of the XQBridge.<BR>
 * Objects of this class are thread-safe and can be reused multiple times.
 * However, as they encapsulate a JDBC connection, it is recommended to use
 * different objects with different JDBC connections for different threads.
 */
public class XQFusion {
	private static final String RCSRevision = "$Revision: 1.3 $";
	private static final String RCSName = "$Name:  $";

	private static final String HELP_OPT1 = "-help";
	private static final String HELP_OPT2 = "-?";
	private static final String FILE_OPT1 = "-file";
	private static final String FILE_OPT2 = "-f";
	private static final String DEBUG_OPT = "-debug";
	private static final String CLASSPATH_OPT1 = "-cp";
	private static final String CLASSPATH_OPT2 = "-classpath";

	private static final String[] DRIVER_LIST = { "com.mysql.jdbc.Driver", "com.microsoft.jdbc.sqlserver.SQLServerDriver", "com.sybase.jdbc2.jdbc.SybDriver", "oracle.jdbc.driver.OracleDriver" };

	private static SAXParserFactory parserFactory;

	static {
		parserFactory = SAXParserFactory.newInstance();
		parserFactory.setNamespaceAware(true);
	}

	private Connection jdbcConnection;
	private String configurationFile;
	private XMLConnection xmlConnection = null;
	private PrintStream log = null;
	private ClassLoader classLoader = null;

	/* The constructor for the XQFusion object. This constructor allocates its own JDBC connection.
	 * The close() method must be called on the object to free the JDBC resources.
	 * @param confFile the XQFusion configuration file
	 */
	public XQFusion(String confFile) {
		if (confFile == null)
			throw new NullPointerException("Illegal null value for configuration file");
		configurationFile = confFile;
	}

	/**
	 * Returns the JDBC connection used to access the underlying database. 
	 * Returns null when no connection to the database is available.
	 * @return a JDBC connection
	 */
	public Connection getConnection() {
		if (jdbcConnection == null) {
			try {
				ExtractorReadOnlyConnection extractorConn = (ExtractorReadOnlyConnection) getXMLConnection();
				jdbcConnection = extractorConn.getJdbcConnection();
			} catch (XMLDBCException ex) {
				// Empty on purpose
			}
		}
		return jdbcConnection;
	}

	/**
	 * Returns the XMLDBC connection used to execute XQuery statements over the database.
	 * @return an XMLConnection object
	 */
	public XMLConnection getXMLConnection() throws XMLDBCException {
		if (xmlConnection == null) {
			xmlConnection = new org.xquark.extractor.Extractor(jdbcConnection, configurationFile).getConnection();
		}
		return xmlConnection;
	}

	public void close() throws XMLDBCException {
		if (xmlConnection != null && !xmlConnection.isClosed())
			xmlConnection.close();
	}

	/* This part of the file deals with command-line execution */

	/**
	 * The method called from the command line to run the XQBridge as a standalone program.
	 * Use the -help option to display a list of available options.
	 * @param args the command line parameters
	 */
	public static void main(String[] args) {
		String configFile = null;
		String batchFile = null;
		String classpath = null;
		ClassLoader classloader = XQFusion.class.getClassLoader();
		ArrayList parameters = new ArrayList();

		//org.xquark.util.misc.Debug.ON = true;

		for (int i = 0; i < args.length; i++) {
			//System.out.println("args["+i+"] = " + args[i]);
			if (!args[i].startsWith("-")) {
				// not an option
				if (configFile == null)
					configFile = args[i];
				else
					parameters.add(args[i]);
				
			} else {
				// Check options are in first position
				if (parameters.size() > 0) {
					usage();
					System.err.println("Options must appear before the command parameters.");
					System.exit(-1);
				}
				// one-parameter option
				if (HELP_OPT1.equals(args[i]) || HELP_OPT2.equals(args[i])) {
					usage();
					System.exit(0);
				} else {
					// two-parameter option
					if (i + 1 == args.length) {
						System.err.println("Missing value for option " + args[i]);
						usage();
						System.exit(-1);
					} else if (DEBUG_OPT.equals(args[i])) {
						Debug.ON = true;
						Log.enable();
						Trace.traceClass("pippo");
					} else if (FILE_OPT1.equals(args[i]) || FILE_OPT2.equals(args[i])) {
						batchFile = args[++i];
					} else if (CLASSPATH_OPT1.equals(args[i]) || CLASSPATH_OPT2.equals(args[i])) {
						classpath = args[++i];
					} else {
						// Unknown option
						System.err.println("Invalid option " + args[i]);
						usage();
						System.exit(-1);
					}
				}
			}
		}

		if (classpath != null) {
			classloader = new URLClassLoader(parseClassPath(classpath), classloader);
		}

		ArrayList drivers = new ArrayList();
		for (int i = 0; i < DRIVER_LIST.length; i++) {
			try {
				Class c = Class.forName(DRIVER_LIST[i], true, classloader);
				drivers.add(c.newInstance());
			} catch (Exception e) {
			}
		}

		List parametersList = null;
		if (batchFile != null) {
			parametersList = parseBatchFile(batchFile);
		} else {
			checkParameters(parameters);
			parametersList = new ArrayList();
			parametersList.add(parameters);
		}
		if (configFile != null) {
			try {
				configFile = new File(configFile).toURL().toString();
			} catch (Exception ex) {
				System.err.println("Invalid configuration file specification.");
				System.exit(-1);
			}
		}
		
		XQFusion fusion = null;
		fusion = new XQFusion(configFile);

		try {
			fusion.executeQuery(configFile, parametersList);
		} catch (XMLDBCException ex) {
			System.err.println("Error while creating XQuery statement.");
			System.err.println(ex.getMessage());
			if (ex.getException() != null)
				System.err.println(ex.getException().getMessage());
			System.exit(ex.getCode());
		}
	}

	private static void checkParameters(List params) {
		int size = params.size();
		if (size < 1 || size > 2) {
			System.err.println("Wrong number of parameters.");
			System.exit(-1);
		}
	}

	private static List parseBatchFile(String batchFile) {
		ArrayList result = new ArrayList();
		BufferedReader reader = null;
		try {
			File f = new File(batchFile);
			reader = new BufferedReader(new FileReader(f));
		} catch (Exception ex) {
			System.err.println("Could not find specified batch file");
			System.exit(-1);
		}
		try {
			while (true) {
				String line = reader.readLine();
				if (line == null)
					break;
				if (line.startsWith("#") || line.trim().length() == 0)
					continue;
				java.util.StringTokenizer it = new java.util.StringTokenizer(line);
				ArrayList params = new ArrayList();
				while (it.hasMoreTokens()) {
					params.add(it.nextToken());
				}
				checkParameters(params);
				result.add(params);
			}
		} catch (IOException e) {
			System.err.println("Error while reading specified batch file");
			System.exit(-1);
		}
		return result;
	}

	private static String readQueryFile(String fileName) {
		try {
			StringBuffer retVal = new StringBuffer();
			String tmpStr;
			BufferedReader in = new BufferedReader(new FileReader(fileName));
			while ((tmpStr = in.readLine()) != null) {
				if (retVal.length() > 0)
					retVal.append('\n');
				retVal.append(tmpStr);
			}
			in.close();
			return retVal.toString();
		} catch (FileNotFoundException ex) {
			System.err.println("Could not find query file " + fileName);
		} catch (IOException ex) {
			System.err.println("Error while reading query file " + fileName);
		}
		return null;
	}

	private void executeQuery(String configFile, List parametersList) throws XMLDBCException {
		String driver = "org.xquark.mediator.MediatorDriver" ;
		String uri = "xdbc:xquark:mediator:" + configFile ;
		
//		Debug.ON = true;
//		Trace.traceClass("toto");
		
		try {
			Class.forName(driver) ;		
		} catch (ClassNotFoundException cnfe) {
			throw new XMLDBCException(cnfe.getMessage());
		}
		XMLConnection connection = XMLDriverManager.getConnection(uri);		
		XMLDataSourceMetaData meta = connection.getMetaData(true) ;
//		String strmeta = meta.getMetaData().getAsString() ;		
//		System.out.println (strmeta) ;
		XMLStatement stat = connection.createStatement();
				
		Iterator it = parametersList.iterator();
		while (it.hasNext()) {
			List params = (List) it.next();
			String queryFile = (String) params.get(0);
			String resultFile = null;
			if (params.size() > 1)
				resultFile = (String) params.get(1);
			String query = readQueryFile(queryFile);
			if (query != null) {
				OutputStream out = System.out;
				try {
					if (resultFile != null) {
						out = new BufferedOutputStream(new FileOutputStream(new File(resultFile)));
					}
				} catch (IOException ex) {
					out = null;
					System.err.println("Could not open result file " + resultFile);
				}
				if (out != null) {
					try {
						File file = new File(queryFile);
						if (file != null)
							file = file.getCanonicalFile();
						if (file != null)
							file = file.getParentFile();
						if (file != null)
							stat.setBaseURI(file.toURL().toString());
						XMLResultSet rs = stat.executeQuery(query);
						XMLSerializer handler = new XMLSerializer(out);
						rs.setContentHandler(handler);
						handler.startDocument();
						while (rs.hasNext())
							rs.nextAsSAX();
						handler.endDocument();
						rs.close();
						if (out != System.out)
							out.close();
					} catch (Exception ex) {
						System.err.println("Error while processing query file " + queryFile);
						System.err.println(ex.getMessage());
					}
				}
			}
		}
		stat.close();
	}

	private static URL[] parseClassPath(String classpath) {
		String pathSeparator = System.getProperty("path.separator");
		ArrayList urls = new ArrayList();
		if (classpath != null) {
			StringTokenizer tokenizer = new StringTokenizer(classpath, pathSeparator);
			while (tokenizer.hasMoreTokens()) {
				File file = new File(tokenizer.nextToken());
				try {
					urls.add(file.getCanonicalFile().toURL());
				} catch (Exception ex) {
				}
			}
		}
		return (URL[]) urls.toArray(new URL[0]);
	}

	private static void usage() {
		System.out.println("Usage: java org.xquark.fusion.XQFusion <common options> <file1> <file2> [<file3>]");
		System.out.println();
		System.out.println("file1:       Configuration file");
		System.out.println("file2:       XQuery file");
		System.out.println("filei:       optional output file, output to stdout by default");
		System.out.println();
		System.out.println("Common options:");
		System.out.println("-?, -help:       this message");
		System.out.println("-f, file:        file containing a list of files to be processed.");
		System.out.println("                 Each line of the file has the number of files described above.");
		System.out.println("                 Lines starting with # are comments");
	}

}
