/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator;

import java.net.URL;
import java.util.HashMap;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xquark.schema.SchemaManager;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.schema.validation.ValidatingSchemaFilter;
import org.xquark.xml.xdbc.XMLConnection;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xml.xdbc.XMLDataSource;

/** 
 * This class represent the mediator as a source that can be accessed by
 * xdbc.
 * 
 */
public class Mediator implements XMLDataSource {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.5 $";
	private static final String RCSName = "$Name:  $";
	// **********************************************************************
	// * CLASS VARIABLE
	// **********************************************************************
	/** 
	 * A hashmap for wrappers.
	 * Key (type : String) : xdbc String
	 * Value (type : SubAccessor) : a class encapsuling informations and 
	 * 				connection to a wrapper
	 */
	private HashMap wrappers = null;
	private URIParser urip = null;

	private static SAXParserFactory spf = null;
	private static SchemaManager schemamanager = new SchemaManager();
	public static final String MEDIATOR_XSD = "/org/xquark/mediator/resources/Mediator.xsd";

	static {
		// Create a JAXP SAXParserFactory and configure it
		spf = SAXParserFactory.newInstance();
		spf.setNamespaceAware(true);
		spf.setValidating(false);
		
		try {
			URL url = Mediator.class.getResource(MEDIATOR_XSD);
			InputSource source = new InputSource(url.toString());
			schemamanager.loadSchema(source);
		} catch (Exception e) {
			// do nothing !!!
		}
	}

	// **********************************************************************
	// * CONSTRUCTOR
	// **********************************************************************
	/**
	 * Initialize the mediator with a configuration file where wrappers
	 * location are defined.
	 */
	public Mediator(URIParser urip) {
		wrappers = new HashMap();
		this.urip = urip;
	}

	public URIParser getURIParser() {
		return urip;
	}
	
	// **********************************************************************
	// * XMLDATASOURCE IMPLEMENTATION
	// **********************************************************************
	/** 
	 * Attempts to establish a connection with the data source.
	 * @throws XMLDBCException If a data source access error occurs
	 * @return an XMLConnection object that represents a connection to the data source
	 */
	public XMLConnection getConnection() throws XMLDBCException {
		MediatorHandler handler = new MediatorHandler(this);

		try {
			// Create a JAXP SAXParserFactory and configure it
			//SAXParserFactory spf = SAXParserFactory.newInstance();

			// Create a JAXP SAXParser
			SAXParser saxParser = spf.newSAXParser();

			// Get the encapsulated SAX XMLReader
			XMLReader reader = saxParser.getXMLReader();
			ValidatingSchemaFilter filter = new ValidatingSchemaFilter(reader, new SchemaValidationContext(schemamanager));
			filter.setContentHandler(handler);
			filter.setErrorHandler(handler);

			// Because this *%@X of xerces/crimson parsers don't
			// match on default values...
			//reader.setFeature ("http://xml.org/sax/features/namespace-prefixes", false) ;
			filter.parse(urip.getConfFileName());
		} catch (Exception e) {
			//e.printStackTrace () ;
			throw new XMLDBCException("Mediator: getConnection : " + e.getMessage());
		}

		XMLConnection xmlc = new MediatorConnection(wrappers, urip);
		return xmlc;
	}

	/**
	 * Add a new Subaccessor in the mediator hashmap.
	 * @param name the name identifing the subaccessor
	 * @param driver the driver used for the subaccessor
	 * @param connection the connection string given to the accessor
	 */
	public void addSubAccessor(String name, String driver, String connection) throws MediatorException {
		wrappers.put(name, new SubAccessor(this, name, driver, connection));
	}

	/**
	 * Add a new Subaccessor in the mediator hashmap.
	 * @param name the name identifing the subaccessor
	 * @param driver the driver used for the subaccessor
	 * @param connection the connection string given to the accessor
	 */
	public void addSubAccessor(String name, String driver, String connection, String login, String passwd) throws MediatorException {
		wrappers.put(name, new SubAccessor(this, name, driver, connection, login, passwd));
	}

	/** 
	 *
	 */
	public HashMap getWrappers() throws MediatorException {
		return wrappers;
	}

	/** Attempts to establish a connection with the data source.
	 * @param user the data source registered user
	 * @param password the user's password
	 * @throws XMLDBCException If a data source access error occurs
	 * @return an XMLConnection object that represents a connection to the data source
	 */
	public XMLConnection getConnection(String user, String password) throws XMLDBCException {
		return getConnection();
	}

	/** Sets the logging/tracing java.io.PrintWriter object that is used by the XDataSource object.
	 * @param writer the new logging/tracing java.io.PrintWriter object, or null to disable logging and tracing
	 */
	public void setLogWriter(java.io.PrintWriter writer) {}

	/** Retrieves the log writer currently used by the XDataSource object to print messages.
	 * @return the java.io.PrintWriter object used to log data source messages, or null if none is defined
	 */
	public java.io.PrintWriter getLogWriter() {
		return null;
	}

	/** Sets the maximum time in seconds that the object will wait when attempting to log in to a data source. Setting the timeout to 0 will disable it.
	 * @param seconds the login time limit in seconds
	 */
	public void setLoginTimeout(int seconds) {}

	/** Gets the maximum time in seconds that the object can wait when attempting to log in to a data source
	 * @return the data source login timeout in seconds, 0 if no timeout is defined
	 */
	public int getLoginTimeout() {
		return -1;
	}
}
