/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator;

import java.util.HashMap;
import java.util.Iterator;

import org.xquark.xml.xdbc.*;
import org.xquark.xquery.metadata.*;

/**
 * This class implements the interface that represents a connection with
 * a specific XML data source.  Within the context of an XMLConnection,
 * XQuery or XPath statements are executed and XML results are returned.
 * <BR>
 * This interface is used by both read-only connections and storage
 * connections, which require access to a larger set of interfaces, such
 * as interfaces representing XML collections and documents.
 * Classes providing read-only access to XML data sources can avoid
 * implementing the optional interfaces for XML storage management,
 * by deriving from the abstract class DefautReadOnlyXMLConnection.
 *
 */
public class MediatorConnection extends DefaultReadOnlyXMLConnection implements XMLConnection {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.3 $";
	private static final String RCSName = "$Name:  $";	
	// **********************************************************************
	// * CLASS VARIABLE
	// **********************************************************************
	/**
	 * A hashmap for wrappers.
	 * Key (type : String) : xdbc String
	 * Value (type : SubAccessor) : a class encapsuling informations and
	 * 				connection to a wrapper
	 */
	private HashMap wrappers = null ;
	private MetaDataImpl metadata = null ;
	private boolean isClosed = false ;
	private boolean isConnected = false ;
	
	private URIParser urip ;
	
	
	// ************************************************************************
	// * INITIALISATION
	// ************************************************************************
	public MediatorConnection(HashMap wrappers, URIParser urip) {
		this.wrappers = wrappers ;
		this.urip = urip ;
		
	}
	
	// ************************************************************************
	// * XMLCONNECTION IMPLEMENTATION
	// ************************************************************************
	
	/*********************************************************************/
	/*************************** Connection management *******************/
	/*********************************************************************/
	/**
	 * To close the current data source connection.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public void close() throws XMLDBCException {
		boolean exception = false;
		if (wrappers!=null)
			for (Iterator it = wrappers.values().iterator() ; it.hasNext() ;) {
				SubAccessor subi = (SubAccessor) it.next() ;
				try {
					subi.close() ;
				}
				catch ( Exception e) { exception = true; }
			}
		if (exception) throw new XMLDBCException("MediatorConnection.close () : one or more wrapper couldn't be closed.") ;
		isClosed = true ;
	}
	
	/**
	 * To know the connection state (opened or closed).
	 * @return true if the current connection is closed, else false.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public boolean isClosed() throws XMLDBCException {
		return isClosed ;
	}
	
	/**
	 * Gets the metadata regarding this connection's data source. A
	 * connection's data source is able to provide information
	 * describing itself.
	 * @return an XMLDataSourceMetaData object for this connection.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public XMLDataSourceMetaData getMetaData() throws XMLDBCException {
		if (metadata == null) metadata = new MetaDataImpl(this);
		return metadata ;
	}
	
	/**
	 * Gets the metadata regarding this connection's data source. A
	 * connection's data source is able to provide information
	 * describing itself.
	 * @refresh when it is true, reload metadata.
	 * @return an XMLDataSourceMetaData object for this connection.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public XMLDataSourceMetaData getMetaData(boolean refresh) throws XMLDBCException {
		try {
			if (metadata == null) metadata = new MetaDataImpl(this);
			if (refresh && (wrappers != null)) {
				//metadata.clear() ;
				MetaParser parser = new MetaParser(metadata) ;
				
				for (Iterator it = wrappers.values().iterator() ; it.hasNext() ;) {
					SubAccessor subi = (SubAccessor) it.next() ;
					//System.out.println ("* Adding wrapper " + subi.getName ()) ;
					XMLConnection coni = null;
					XMLDataSourceMetaData metai = null;
					try {
						coni = subi.getConnection() ;
					}
					catch (Exception e) {
						//e.printStackTrace();
						throw new XMLDBCException("Could not get connection for sub element : " + subi.getName());
					}
					try {
						metai = coni.getMetaData(true) ;
					}
					catch (Exception e) {
						throw new XMLDBCException("Could not get medadata for sub element : " + subi.getName());
					}
					
					MetaWrapper metawrapper = parser.createWrapper(true) ;
				
					XMLDocument doc = metai.getMetaData() ;
					//org.xquark.xml.misc.Debug.print("metadata of wrapper = " + doc.getAsString());					
					doc.setContentHandler(parser) ;
					doc.getAsSAX() ;
					metadata.addMetaWrapper(metawrapper, subi.getName()) ;
				}
				isConnected = true;
			}
			return metadata ;
		}
		catch (MetadataException mde) {
			throw new XMLDBCException("MediatorConnection:getMetaData - metadata exception: " + mde.getMessage()) ;
		}
		catch (Exception e) {
			e.printStackTrace() ;
			throw new XMLDBCException("MediatorConnection:getMetaData - metadata exception: " + e.getMessage()) ;
		}
	}
	
	/**
	 * To get the current user name.
	 * @return the current user name.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public String getUserName() throws XMLDBCException {
		return null ;
	}
	
	/**
	 * To get the current URL of this connection.
	 * @return the current URL of this connection.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public String getURL() throws XMLDBCException {
		return urip.getURI() ;
	}
	
	/**
	 * To get the set of wrappers.
	 * @return the set of wrappers
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public HashMap getWrappers() throws XMLDBCException {
		return wrappers ;
	}
	
	/*********************************************************************/
	/*************************** Create/prepare statements ***************/
	/*********************************************************************/
	
	/**
	 * Creates a statement object for sending XQuery (query type by default)
	 * statements to the data source.
	 * @return a new statement.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public XMLStatement createStatement() throws XMLDBCException {
		XMLStatement state = createStatement(XQUERY_STRING_TYPE);
		return state ;
	}
	
	
	/**
	 * Creates a statement object for sending XQuery or XPath (according to the
	 * specified type) statements to the data source.
	 * @param queryType the type of query for this statement (see constants in this interface).
	 * @return a new statement.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public XMLStatement createStatement(short queryType) throws XMLDBCException {
		connect();
		XMLStatement state = new MediatorStatement(this);
		return state;
	}
	
	/**
	 * Creates a prepared statement object for sending parameterized XQuery
	 * (query type by default) statements to the data source.
	 * The advantage of prepared statements over statements is that the request
	 * may be precompiled once and executed several times.
	 * Note: this interface is optional.
	 * @param query a parameterized XQuery.
	 * @return a new prepared statement containing the pre-compiled statement.
	 * @throws XMLDBCException if a data source access error occurs.
	 * @throws XMLDBCNotSupportedException if the method is not supported.
	 */
	public PreparedXMLStatement prepareStatement(String query) throws XMLDBCException, XMLDBCNotSupportedException {
		throw new XMLDBCNotSupportedException("PreparedXMLStatement prepareStatement(String query, short queryType) is not supported.", null);
	}
	
	/**
	 * Creates a prepared statement object for sending parameterized XQuery or
	 * XPath (according to the specified type) statements to the data source.
	 * The advantage of prepared statements over statements is that the request
	 * may be precompiled once and executed several times.
	 * Note: this interface is optional.
	 * @param query a parameterized query of the specified type.
	 * @param queryType the type of query for this statement (see constants in this interface).
	 * @return a new Pprepared statement containing the pre-compiled statement.
	 * @throws XMLDBCException if a data source access error occurs.
	 * @throws XMLDBCNotSupportedException if the method is not supported.
	 */
	public PreparedXMLStatement prepareStatement(String query, short queryType) throws XMLDBCException, XMLDBCNotSupportedException {
		throw new XMLDBCNotSupportedException("PreparedXMLStatement prepareStatement(String query, short queryType) is not supported.", null);
	}
	
	// specific methods
	public SubAccessor getSubAccessorByName(String name) throws MediatorException {
		if (wrappers != null)
			for (Iterator it = wrappers.values().iterator() ; it.hasNext() ;) {
				SubAccessor sub = (SubAccessor) it.next() ;
				if (sub.getName().equalsIgnoreCase(name))
					return sub ;
			}
		return null ;
	}
	
	public XMLConnection getSubConnectionByName(String name) throws MediatorException {
		SubAccessor sub = getSubAccessorByName(name) ;
		return sub.getConnection() ;
	}
	
	private void connect() throws XMLDBCException { if (!isConnected) { getMetaData(true); isConnected = true; } }
	
}

