/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator.algebra;

import java.util.*;

import org.xquark.mediator.MediatorException;
import org.xquark.mediator.decomposer.*;
import org.xquark.mediator.plan.*;
import org.xquark.util.misc.Debug;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.parser.hinttree.HintTree;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.FunctionCOLLECTION;
import org.xquark.xquery.parser.util.Constants;

/**
 * Represent a node where nodes or leaves can fix to it.
 * Before atomization a DepMem binds only DepSource and DepVar.
 * After atomization, a DepMem binds only DepVar.
 *
 */
public class AlgFLWR extends Algebra {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.4 $";
	private static final String RCSName = "$Name:  $";
	// **********************************************************************
	// * CLASS VARIABLES
	// **********************************************************************
	private ArrayList subDepFLWRs = null;
	private AlgFLWR parentDepFLWR = null;
	//private ArrayList variables = null;
	private boolean monoSource = false;
	protected Set sources = null;
	protected HintTree hinttree = null;
	// telling if information has already been computed
	boolean hasInformation = false;
	// telling whether the subelements are done with nested loop
	int subhandling = -1;
	// relation between variables in where expression
	VariableDependencies vardependencies = null;
	//	private HashMap varVarWhereMap = null;
	//	private ArrayList allExprWhereList = null;
	//	private ArrayList complexExprWhereList = null;
	// maps for ordering purposes
	private HashMap varOrderMap = null;
	private HashMap orderVarMap = null;
	private ArrayList complexExprOrderList = null;
	private ArrayList varOrderList = null;
	private ArrayList varReOrderList = null;
	private ArrayList varHintList = null;
	// for compulsery hint
	private boolean hasCompulseryMerge = false;
	private boolean hasCompulseryNested = false;

	private IsolateWhereVisitor isolateVisitor = null;
	// ******************************************************************
	// * INITIALIZATION
	// ******************************************************************
	/**
	 *
	 * @param typevisitor
	 * @param expression
	 * @param depmanager
	 */
	public AlgFLWR(XQueryExpression expression, Variable var, AlgebraManager depmanager, boolean islet) {
		super(expression, var, depmanager, islet);
		// set HintTree
		hinttree = ((FLWRExpression) expression).getHintTree();
		// fill var lists
		fillVarLists();
	}
	// ******************************************************************
	// * GET/SET METHODS
	// ******************************************************************
	public void setDependency(AlgFLWR parentDepFLWR) {
		if (this.parentDepFLWR != null)
			this.parentDepFLWR.getDependencyList().remove(this);
		this.parentDepFLWR = parentDepFLWR;
		ArrayList tmpList = parentDepFLWR.getDependencyList();
		if (tmpList == null) {
			tmpList = new ArrayList();
			tmpList.add(this);
			parentDepFLWR.setDependencyList(tmpList);
		} else if (!tmpList.contains(this))
			tmpList.add(this);
	}

	public ArrayList getDependencyList() {
		return subDepFLWRs;
	}
	public void setDependencyList(ArrayList subDepFLWRs) {
		this.subDepFLWRs = subDepFLWRs;
	}

	public boolean isRootQDB() {
		return (parentDepFLWR == null && parent == null);
	}
	public boolean isQDB() {
		return (isRootQDB() || parent == null);
	}
	public boolean isMonoSource() {
		return monoSource;
	}

	public void setMonoSource(boolean monoSource) {
		this.monoSource = monoSource;
	}
	public HintTree getHintTree() {
		return hinttree;
	}
	public void setHintTree(HintTree hinttree) {
		this.hinttree = hinttree;
	}
	public int getSubHandling() {
		return subhandling;
	}
	public void setSubHandling(int subhandling) {
		this.subhandling = subhandling;
	}

	public boolean hasCompulseryMerge() {
		return hasCompulseryMerge;
	}

	public boolean hasCompulseryNested() {
		return hasCompulseryNested;
	}

	public void updateVarDependencyList(ArrayList dependVars) {
		if (parentDepFLWR == null || dependVars == null || dependVars.isEmpty())
			return;
		for (int i = 0; i < dependVars.size(); i++) {
			Variable tmpvar = (Variable) dependVars.get(i);
			if (!((FLWRExpression) parentDepFLWR.getExpression()).getVariables().contains(tmpvar))
				getParentDepFLWR().addDependSourceVariable(tmpvar);
			else
				dependVars.remove(tmpvar);
		}
		if (!dependVars.isEmpty())
			getParentDepFLWR().updateVarDependencyList(dependVars);
	}

	private AtomizeWhereVisitor awvisitor = null;

	private void computeVarOrderList() throws MediatorException {
		ArrayList orders = ((FLWRExpression) expression).getOrderBy();
		if (orders != null) {
			varOrderList = new ArrayList(orders.size());
			complexExprOrderList = new ArrayList(1);
			orderVarMap = new HashMap();
			varOrderMap = new HashMap();
			GetBoundVariablesVisitor gbvv = new GetBoundVariablesVisitor(true, false);
			for (int i = 0; i < orders.size(); i++) {
				XQueryExpression orderExpr = (XQueryExpression) orders.get(i);
				gbvv.reset();
				try {
					orderExpr.accept(gbvv);
				} catch (Exception e) {
					throw new MediatorException("Could not extract variables from expression : " + orderExpr, e);
				}
				ArrayList tmpList = gbvv.getUniqueVariables();
				orderVarMap.put(orderExpr, tmpList);
				if (tmpList.size() > 1) {
					complexExprOrderList.add(orderExpr);
					if (varOrderList != null && !varOrderList.containsAll(tmpList))
						varOrderList = null;
					ArrayList tmpOrders = (ArrayList) varOrderMap.get(tmpList);
					if (tmpOrders == null) {
						tmpOrders = new ArrayList(1);
						varOrderMap.put(tmpList, tmpOrders);
					}
					tmpOrders.add(orderExpr);
				} else {
					Variable tmpVar = (Variable) tmpList.get(0);
					if (((FLWRExpression) expression).getVariables().contains(tmpVar)) {
						int pos = varOrderList.indexOf(tmpVar);
						if (pos >= 0 && pos != varOrderList.size() - 1)
							varOrderList = null;
						else if (pos == -1)
							varOrderList.add(tmpVar);
						ArrayList tmpOrders = (ArrayList) varOrderMap.get(tmpVar);
						if (tmpOrders == null) {
							tmpOrders = new ArrayList(1);
							varOrderMap.put(tmpVar, tmpOrders);
						}
						tmpOrders.add(orderExpr);
					}
				}
			}
			if (complexExprOrderList.isEmpty())
				complexExprOrderList = null;
		}
	}

	/*
	private void updateVarListWithDependencies(ArrayList varlist) {
		updateVarListWithDependencies(varlist, null, false);
	}
	*/
	private void updateVarListWithDependencies(ArrayList varlist, ArrayList rangelist, boolean reorder) {
		if (varlist == null || varlist.isEmpty())
			return;
		Debug.print("START updateVarListWithDependencies -> varlist = " + varlist);
		int i = 0;
		while (i < varlist.size()) {
			Variable tmpVari = (Variable) varlist.get(i);
			Algebra tmpNode = algManager.getMapNode(tmpVari);
			ArrayList tmpList = tmpNode.getVarsDependingOn();
			if (tmpList != null) {
				for (int j = 0; j < tmpList.size(); j++) {
					Variable tmpVarj = (Variable) tmpList.get(j);
					if (rangelist != null && !rangelist.contains(tmpVarj))
						continue;
					int posj = varlist.indexOf(tmpVarj);
					if (posj > i) {
						varlist.remove(tmpVarj);
						varlist.add(i, tmpVarj);
						if (reorder) {
							if (!varReOrderList.contains(tmpVari))
								varReOrderList.add(tmpVari);
							if (!varReOrderList.contains(tmpVarj))
								varReOrderList.add(tmpVarj);
						}
						i--;
						break;
					} else if (posj == -1) {
						varlist.add(i, tmpVarj);
						i--;
						break;
					}
				}
			}
			i++;
		}
		Debug.print("STOP updateVarListWithDependencies -> varlist = " + varlist);
	}

	private void verifyVarOrderList() {
		if (varOrderList == null) {
			varOrderList = ((FLWRExpression) expression).getVariables();
			if (((FLWRExpression) expression).getOrderBy() != null)
				varReOrderList = ((FLWRExpression) expression).getVariables();
		} else {
			varReOrderList = new ArrayList();
			// manage depending on variables
			int i = 0;
			while (i < varOrderList.size()) {
				Variable tmpVari = (Variable) varOrderList.get(i);
				Algebra tmpNode = algManager.getMapNode(tmpVari);
				ArrayList tmpList = tmpNode.getVarsDependingOn();
				if (tmpList != null) {
					for (int j = 0; j < tmpList.size(); j++) {
						Variable tmpVarj = (Variable) tmpList.get(j);
						int posj = varOrderList.indexOf(tmpVarj);
						if (posj > i) {
							varOrderList.remove(tmpVarj);
							varOrderList.add(i, tmpVarj);
							if (!varReOrderList.contains(tmpVari))
								varReOrderList.add(tmpVari);
							if (!varReOrderList.contains(tmpVarj))
								varReOrderList.add(tmpVarj);
							i--;
							break;
						} else if (posj == -1) {
							varOrderList.add(i, tmpVarj);
							i--;
							break;
						}
					}
				}
				i++;
			}
			if (varReOrderList.isEmpty())
				varReOrderList = null;
			// manage other variables
			ArrayList varlist = ((FLWRExpression) expression).getVariables();
			if (varOrderList.size() < varlist.size()) {
				for (i = 0; i < varlist.size(); i++) {
					if (!varOrderList.contains(varlist.get(i))) {
						varOrderList.add(varlist.get(i));
					}
				}
			}
		}
	}

	private void fillList(ArrayList varlist, ArrayList allvars) {
		if (varlist.containsAll(allvars))
			return;
		for (int i = 0; i < allvars.size(); i++) {
			if (!varlist.contains(allvars.get(i))) {
				varlist.add(allvars.get(i));
			}
		}
	}

	/*
	private boolean verifyHintTree(HintTree hinttree) {
		return verifyHintTree(hinttree, new ArrayList());
	}
	private boolean verifyHintTree(HintTree hinttree, ArrayList hintVarList) {
		if (hinttree == null)
			return false;
		if (hinttree.getType() == HintTree.NESTED_LOOP_TYPE) {
			if (hinttree.getLeftHint() instanceof HintTree) {
				if (!verifyHintTree((HintTree) hinttree.getLeftHint(), hintVarList))
					return false;
			}
			if (hinttree.getLeftHint() instanceof Variable) {
				hintVarList.add(hinttree.getLeftHint());
			}
			if (hinttree.getRightHint() instanceof HintTree) {
				if (!verifyHintTree((HintTree) hinttree.getRightHint(), hintVarList))
					return false;
			}
			if (hinttree.getRightHint() instanceof Variable) {
				hintVarList.add(hinttree.getRightHint());
			}
		} else if (hinttree.getType() == HintTree.MERGE_TYPE) {
			ArrayList tmplistleft = (ArrayList) hintVarList.clone();
			if (hinttree.getLeftHint() instanceof HintTree) {
				if (!verifyHintTree((HintTree) hinttree.getLeftHint(), tmplistleft))
					return false;
				tmplistleft.removeAll(hintVarList);
			}
			ArrayList tmplistright = (ArrayList) hintVarList.clone();
			if (hinttree.getRightHint() instanceof HintTree) {
				if (!verifyHintTree((HintTree) hinttree.getRightHint(), tmplistright))
					return false;
				tmplistright.removeAll(hintVarList);
			}
			hintVarList.addAll(tmplistleft);
			hintVarList.addAll(tmplistright);
			if (hinttree.getLeftHint() instanceof Variable) {
				hintVarList.add(hinttree.getLeftHint());
			}
			if (hinttree.getRightHint() instanceof Variable) {
				hintVarList.add(hinttree.getRightHint());
			}
		}
		return true;
	}
	*/

	private boolean verifyVarHintList() {
		if (varHintList == null)
			return false;
		ArrayList vars = ((FLWRExpression) expression).getVariables();
		// verifying all variables of the hint tree are declared in flwr
		if (!vars.containsAll(varHintList))
			return false;
		// verifying that all variables are used only once
		for (int i = 0; i < varHintList.size(); i++) {
			if (varHintList.lastIndexOf(varHintList.get(i)) != i)
				return false;
		}
		// get all variables in vars but not in varlist
		ArrayList remainvars = (ArrayList) vars.clone();
		remainvars.removeAll(varHintList);
		if (!remainvars.isEmpty()) {
			updateVarListWithDependencies(remainvars, remainvars, false);
		}
		// verifying that variables in hint are bound to DepNodeSource on collections
		for (int i = 0; i < varHintList.size(); i++) {
			Variable vari = (Variable) vars.get(i);
			if (vari.getBindingType() != Constants.FOR_BINDINGTYPE)
				return false;
			XQueryExpression expri = vari.getExpression();
			if (expri instanceof LocatedExpression)
				expri = ((LocatedExpression) expri).getExpression();
			if (!(expri instanceof FunctionCOLLECTION))
				return false;
		}

		/*
		// verifying that all variables of the flwr are used in the hint tree
		for (int i = 0; i < vars.size(); i++) {
			Variable vari = (Variable) vars.get(i);
			if (vari.getBindingType() == Constants.FOR_BINDINGTYPE && !varHintList.contains(vari))
				return false;
			// for now no let
			if (vari.getBindingType() == Constants.LET_BINDINGTYPE)
				return false;
			//			if (vari.getBindingType() == Constants.LET_BINDINGTYPE && varHintList.contains(vari))
			//				return false;
		}
		*/
		return true;
	}

	private boolean updateVarHintList(HintTree hinttree) {
		if (hinttree == null)
			return false;
		ArrayList vars = ((FLWRExpression) expression).getVariables();
		/*
		Object obj = hinttree.getLeftHint();
		if (obj instanceof HintTree)
			updateVarListWithDependencies(((HintTree)obj).getVarList(),vars,false);
		obj = hinttree.getRightHint();
		if (obj instanceof HintTree)
			updateVarListWithDependencies(((HintTree)obj).getVarList(),vars,false);
		*/
		updateVarListWithDependencies(hinttree.getVarList(), vars, false);
		return true;
	}

	private void compeReOrdering(boolean hasMerge, ArrayList varList) {
		if (hasMerge)
			//varReOrderList = (ArrayList) hint.getVarList().clone();
			varReOrderList = (ArrayList) ((FLWRExpression) expression).getVariables().clone();
		else {
			if (varOrderList == null || varOrderList.isEmpty())
				return;
			ArrayList tmplist = (ArrayList) ((FLWRExpression) expression).getVariables().clone();
			tmplist.removeAll(varList);
			tmplist.addAll(0, varList);
			if (varReOrderList == null)
				varReOrderList = new ArrayList();
			else
				varReOrderList.clear();
			int pos = -1;
			for (int j = 0; j < tmplist.size(); j++) {
				Variable tmpVarj = (Variable) tmplist.get(j);
				int posj = varOrderList.indexOf(tmpVarj);
				if (posj > pos)
					pos = posj;
				else
					for (int i = 0; i < tmplist.size(); i++) {
						Variable tmpVari = (Variable) tmplist.get(i);
						if (!varReOrderList.contains(tmpVari))
							varReOrderList.add(tmpVari);
					}
			}
			if (varReOrderList.isEmpty())
				varReOrderList = null;
		}
	}

	/*
	private boolean computeVarHintList(HintTree hinttree) {
		if (hinttree == null)
			return false;
		Object obj = hinttree.getLeftHint();
		if (obj instanceof Variable) {
			if (varHintList.contains(obj))
				return false;
			else
				varHintList.add(obj);
		} else
			computeVarHintList((HintTree) obj);
		obj = hinttree.getRightHint();
		if (obj instanceof Variable) {
			if (varHintList.contains(obj))
				return false;
			else
				varHintList.add(obj);
		} else
			computeVarHintList((HintTree) obj);
		return true;
	}
	*/

	private void makeInformation() throws MediatorException {
		if (hasInformation)
			return;
		hasInformation = true;
		Debug.print("DepFLWR start making information");
		Debug.print("Scope = " + algManager.getScopeVariables());
		// verify that all children have only one variable
		if (!isMonoSource()) {
			if (algChildren != null) {
				for (int i = 0; i < algChildren.size(); i++)
					if (((Algebra) algChildren.get(i)).getVariables().size() != 1)
						throw new MediatorException("Found DepNode of more than one variable as child of DepFLWR");
			}
		} else {
			if (hinttree != null) {
				if (hinttree.getOuterType() != HintTree.NO_TYPE)
					hinttree.setType(HintTree.NO_TYPE);
			}
			/*
			if (depchildren.size() != 1)
				throw new MediatorException("Found DepFLWR mono source with more than one child");
			*/
		}
		// gather information on joins if any (where clause)

		XQueryExpression wexpr = ((FLWRExpression) expression).getWhereClause();
		if (wexpr != null) {
			vardependencies = new VariableDependencies(wexpr);
			//			if (awvisitor == null)
			//				awvisitor = new AtomizeWhereVisitor();
			//			try {
			//				wexpr.accept(awvisitor);
			//			} catch (Exception e) {
			//				throw new MediatorException("Could not atomize where clause : " + wexpr, e);
			//			}
			//			varVarWhereMap = awvisitor.getVarVarMap();
			//			allExprWhereList = awvisitor.getAllExprList();
			//			complexExprWhereList = awvisitor.getComplexExprList();
		}
		// gather inforamtion on hint tree if any
		if (hinttree != null && hinttree.getType() != HintTree.NO_TYPE) {
			Debug.print("hinttree = " + hinttree);
			varHintList = hinttree.getVarList();
			Debug.print("varHintList 0 = " + varHintList);
			/*
			if (!computeVarHintList(hinttree))
				throw new MediatorException("Incorrect hint tree, could not compute hint tree variable list!");
			Debug.print("varHintList 1 = " + varHintList);
			*/
			if (!updateVarHintList(hinttree))
				throw new MediatorException("Incorrect hint tree, could not verify hint tree variable list!");
			Debug.print("varHintList 1 = " + varHintList);
			if (!verifyVarHintList())
				throw new MediatorException("Incorrect hint tree, could not verify hint tree variable list!");
			Debug.print("varHintList 2 = " + varHintList);
			/*
			if (!verifyHintTree(hinttree))
				throw new MediatorException("Incorrect hint tree, could not verify hint tree!"); 
			Debug.print("varHintList 3 = " + varHintList);
			*/
		}
		computeVarOrderList();
		Debug.print("varOrderList = " + varOrderList);
		Debug.print("varReOrderList 1 = " + varReOrderList);
		if (hinttree != null)
			compeReOrdering(hinttree.hasMerge(), hinttree.getVarList());
		else
			compeReOrdering(algManager.getPlan().getDefaultSubHandling() == HintTree.MERGE_TYPE, ((FLWRExpression) expression).getVariables());
		Debug.print("varReOrderList 2 = " + varReOrderList);
		Debug.print("DepFLWR stop making information");
	}

	public XQueryExpression getWhereClause() {
		return ((FLWRExpression) expression).getWhereClause();
	}

	public VariableDependencies getVarDependencies() {
		return vardependencies;
	}

	// ******************************************************************
	// * CREATE EXECUTION PLAN
	// ******************************************************************
	// some documentation
	/*
	for now hint tree only applies on FLWR expressions (not nested for now)
	
	if (hint tree is correct)
		for now correct hint tree means that
			all (for) variables present in the FLWR expression are also present in the hint tree
		(let) variables are simply merged
		
	if (hint tree is not present or hint tree is incorrect)
		variables are ordered in the following order
		first get the order implied by the order by clauses if any and if possible
		insert all depending on variables
		add all variables not in this list
		make list of variables to be reordered (because of dependencies)	
	*/

	//##################################################################################################################################
	//##################################################################################################################################
	//##################################################################################################################################
	private Operator createVarsOperator(ExecutionPlan plan, Operator operator, List varlist, List vars) throws MediatorException {
		//##################################################################################################################################
		//##################################################################################################################################
		//##################################################################################################################################
		for (int i = 0; i < vars.size(); i++) {
			operator = createVarOperator(plan, operator, varlist, (Variable) vars.get(i));
		}
		return operator;
	}
	//##################################################################################################################################
	//##################################################################################################################################
	//##################################################################################################################################
	private Operator createVarOperator(ExecutionPlan plan, Operator operator, List varlist, Variable var) throws MediatorException {
		//##################################################################################################################################
		//##################################################################################################################################
		//##################################################################################################################################
		// getting child
		Algebra depchildi = algManager.getMapNode(var);
		XQueryExpression depchildiexpr = depchildi.getExpression();
		//ArrayList varidependingon = (ArrayList) depchildi.getVarsDependingOn().clone();
		Variable vari = (Variable) depchildi.getVariables().get(0);
		if (!isMonoSource())
			algManager.addVarToScope(vari);
		else
			algManager.addVarsToScope(depchildi.getVariables());
		boolean isleti = depchildi.isLet();
		varlist.add(vari);
		Debug.print("*********** DEPFLWR CHILD STUFF *************************");
		Debug.print("ClassName = " + depchildi.getClass().getName());
		Debug.print("is let = " + depchildi.isLet());
		Debug.print("depchildiexpr = " + depchildiexpr);
		Debug.print("vari = " + vari);
		//		Debug.print("value depends = " + depchildi.getValueDepends());
		//		Debug.print("where depends = " + depchildi.getWhereDepends());
		Debug.print("allVars = " + algManager.getScopeVariables());
		Debug.print("definition vars depending on = " + depchildi.getVarsDependingOn());
		Debug.print("where vars depending on = " + depchildi.getVarsWhereOn());
		Debug.print("vars referenced by = " + depchildi.getVarsReferencedBy());
		Debug.print("children = " + depchildi.getChildren());
		Debug.print("var dependencies = " + vardependencies);
		Debug.print("---------------------------------------------------------");
		//		if (varVarWhereMap != null && varVarWhereMap.get(vari) != null)
		//			Debug.print("join vars = " + ((varVarWhereMap.get(vari) == null) ? null : ((HashMap) varVarWhereMap.get(vari)).keySet()));
		Debug.print("---------------------------------------------------------");
		/*
		if (varVarWhereMap != null && varVarWhereMap.get(vari) != null) {
			for (Iterator it = ((HashMap) varVarWhereMap.get(vari)).keySet().iterator(); it.hasNext();) {
				Object obj = it.next();
				if (obj != vari && !varidependingon.contains(obj)) {
					if (depmanager.getScopeVariables().contains(obj))
						depchildi.getVarsDependingOn().add(obj);
					if (varlist.contains(obj))
						varidependingon.add(obj);
				}
			}
		}
		*/

		// add where dependencies
		if (vardependencies != null) {
			ArrayList tmplist = vardependencies.getAllAssociatedVariables(vari);
			if (tmplist != null) {
				for (int i = 0; i < tmplist.size(); i++) {
					Variable tmpvari = (Variable) tmplist.get(i);
					if (algManager.getScopeVariables().contains(tmpvari)) {
						if (depchildiexpr instanceof FLWRExpression && ((FLWRExpression) depchildiexpr).getVariables().contains(tmpvari)) {
							if (!depchildi.getVarsWhereOn().contains(tmpvari))
								depchildi.getVarsWhereOn().add(tmpvari);
						} else {
							if (!depchildi.getVarsDependingOn().contains(tmpvari))
								depchildi.getVarsDependingOn().add(tmpvari);
						}
					}

				}
			}
		}
		//		if (varVarWhereMap != null && varVarWhereMap.get(vari) != null) {
		//			for (Iterator it = ((HashMap) varVarWhereMap.get(vari)).keySet().iterator(); it.hasNext();) {
		//				Object obj = it.next();
		//				if (obj != vari && !depchildi.getVarsDependingOn().contains(obj) && depmanager.getScopeVariables().contains(obj))
		//					depchildi.getVarsDependingOn().add(obj);
		//			}
		//		}		

		//		if (!depchildi.getVarsDependingOn().isEmpty())
		//			depchildi.setDepends(true);
		//org.xquark.xml.misc.Debug.print("algebra = " + algebra.toCompleteString()) ;
		if (operator == null) {
			if (depchildi instanceof AlgLocated || depchildi instanceof AlgVar || depchildi instanceof AlgArithFunction)
				//algebra = new AlgNotSource(plan,depchildi.getExpression(),var,depchildi.isLet(),depchildi.hasIdentifier());
				operator = new OpNotSource(plan, depchildi);
			else
				operator = (Operator) depchildi.createOperator(plan);
		} else {
			if (!depchildi.getDepends()) {
				if (isleti) {
					operator = new OpMerge(plan, depchildiexpr, operator, (Operator) depchildi.createOperator(plan));
				} else {
					operator = new OpCartesian(plan, depchildiexpr, operator, (Operator) depchildi.createOperator(plan));
				}
			} else { // varidependingon is not empty
				if (depchildi instanceof AlgLocated || depchildi instanceof AlgVar || depchildi instanceof AlgArithFunction) {
					//algebra = new AlgSimpleMerge(plan, depchildiexpr, vari, algebra, depchildi.isLet());
					operator = new OpSimpleMerge(plan, depchildi, operator);
				} else if (depchildi instanceof AlgFunction) {
					if (depchildi.getChildren() == null)
						operator = new OpSimpleMerge(plan, depchildiexpr, vari, operator, depchildi.isLet());
					else {
						ArrayList tmplist = new ArrayList(2);
						tmplist.add(operator);
						Operator tmpAlg = depchildi.createOperator(plan);
						tmpAlg.setDepends(depchildi);
						tmplist.add(tmpAlg);
						operator = new OpSubQuery(plan, depchildiexpr, tmplist, true, false);
					}
				} else {
					ArrayList tmplist = new ArrayList(2);
					tmplist.add(operator);
					Operator tmpAlg = depchildi.createOperator(plan);
					tmpAlg.setDepends(depchildi);
					tmplist.add(tmpAlg);
					operator = new OpSubQuery(plan, depchildiexpr, tmplist, true, false);
				}

				//Debug.print("algebra = \n" + algebra.toCompleteString());
			}
		}
		// handle restriction if any

		XQueryExpression tmpExpr = ((FLWRExpression) expression).getWhereClause();
		//Debug.print("restriction = \n" + tmpExpr);
		if (tmpExpr != null) {
			if (operator instanceof OpSource || operator instanceof OpSubQuery) {
				//				depmanager.removeVarFromScope(vari);
				//				
				//				if (isolateVisitor == null)
				//					isolateVisitor = new IsolateWhereVisitor(depmanager.getScopeVariables());
				//				else 
				//					isolateVisitor.reset(depmanager.getScopeVariables());
				//				isolateVisitor.putVar(true);
				//				try {
				//					tmpExpr.accept(isolateVisitor);
				//				} catch (Exception e) {
				//					throw new MediatorException("Can't isolate where expression : " + tmpExpr, e);
				//				}
				//				
				//				depmanager.addVarToScope(vari);
				//				
				//				tmpExpr = isolateVisitor.getRemainingExpression();
				//				if (tmpExpr != null) {
				//					isolateVisitor.reset(varlist);
				//					//isolateVisitor = new IsolateWhereVisitor(depmanager.getScopeVariables());
				//					//isolateVisitor = new IsolateWhereVisitor(((FLWRExpression) expression).getVariables());
				//					try {
				//						tmpExpr.accept(isolateVisitor);
				//					} catch (Exception e) {
				//						throw new MediatorException("Can't isolate where expression : " + tmpExpr, e);
				//					}
				//					tmpExpr = isolateVisitor.getIsolatedExpression();
				//					if (tmpExpr != null)
				//						algebra = new AlgRestrict(plan, tmpExpr, algebra);
				//					tmpExpr = isolateVisitor.getRemainingExpression();
				//					((FLWRExpression) expression).setWhereClause(tmpExpr);
				//				}

				if (isolateVisitor == null)
					isolateVisitor = new IsolateWhereVisitor(algManager.getScopeVariables());
				else
					isolateVisitor.reset(algManager.getScopeVariables());
				isolateVisitor.putVar(true);
				try {
					tmpExpr.accept(isolateVisitor);
				} catch (Exception e) {
					throw new MediatorException("Can't isolate where expression : " + tmpExpr, e);
				}

				tmpExpr = isolateVisitor.getIsolatedExpression();
				if (tmpExpr != null) {
					operator = operator.addCondition(plan, tmpExpr);
					//operator = new OpRestrict(plan, tmpExpr, operator);
				}
				((FLWRExpression) expression).setWhereClause(isolateVisitor.getRemainingExpression());
			} else {
				//				depmanager.removeVarFromScope(vari);
				//				
				//				if (isolateVisitor == null)
				//					isolateVisitor = new IsolateWhereVisitor(depmanager.getScopeVariables());
				//				else 
				//					isolateVisitor.reset(depmanager.getScopeVariables());
				//				isolateVisitor.putVar(true);
				//				try {
				//					tmpExpr.accept(isolateVisitor);
				//				} catch (Exception e) {
				//					throw new MediatorException("Can't isolate where expression : " + tmpExpr, e);
				//				}
				//				
				//				depmanager.addVarToScope(vari);
				//
				//				tmpExpr = isolateVisitor.getRemainingExpression();
				//				if (tmpExpr != null) {
				//					isolateVisitor.reset(depmanager.getScopeVariables());
				//					//isolateVisitor = new IsolateWhereVisitor(depmanager.getScopeVariables());
				//					//isolateVisitor = new IsolateWhereVisitor(((FLWRExpression) expression).getVariables());
				//					try {
				//						tmpExpr.accept(isolateVisitor);
				//					} catch (Exception e) {
				//						throw new MediatorException("Can't isolate where expression : " + tmpExpr, e);
				//					}
				//					tmpExpr = isolateVisitor.getIsolatedExpression();
				//					if (tmpExpr != null)
				//						algebra = new AlgRestrict(plan, tmpExpr, algebra);
				//					tmpExpr = isolateVisitor.getRemainingExpression();
				//					((FLWRExpression) expression).setWhereClause(tmpExpr);
				//				}

				if (isolateVisitor == null)
					isolateVisitor = new IsolateWhereVisitor(algManager.getScopeVariables());
				else
					isolateVisitor.reset(algManager.getScopeVariables());
				isolateVisitor.putVar(true);
				try {
					tmpExpr.accept(isolateVisitor);
				} catch (Exception e) {
					throw new MediatorException("Can't isolate where expression : " + tmpExpr, e);
				}

				tmpExpr = isolateVisitor.getIsolatedExpression();
				if (tmpExpr != null) {
					operator = operator.addCondition(plan, tmpExpr);
					//algebra = new OpRestrict(plan, tmpExpr, algebra);
				}
				((FLWRExpression) expression).setWhereClause(isolateVisitor.getRemainingExpression());

			}
		}
		// test if sort clause and not source -> then sort in place !!!
		if (!(depchildi instanceof AlgSource) && varOrderMap != null) {
			ArrayList orders = (ArrayList) varOrderMap.get(vari);
			if (orders != null)
				operator = new OpSort(plan, operator, orders);
		}
		// add   28/03/2003 to remove variables from scope
		//		if (!isMonoSource())
		//			depmanager.removeVarFromScope(vari);
		//		else
		//			depmanager.removeVarsFromScope(depchildi.getVariables());
		return operator;
	}

	// utility
	private boolean extractConditionsfromMap(VariableDependencies dep, List varListLeft, List varListRight, ArrayList leftExpressions, ArrayList rightExpressions, ArrayList mergeRestrictions, ArrayList otherRestrictions) {
		boolean hasnoteq = false;
		for (int il = 0; il < varListLeft.size(); il++) {
			Variable leftvar = (Variable) varListLeft.get(il);
			if (vardependencies != null) {
				HashMap tmpmap = vardependencies.getJoinMap(leftvar);
				if (tmpmap != null) {
					// there is at least something because of hint tree verification
					for (int ir = 0; ir < varListRight.size(); ir++) {
						Variable rightvar = (Variable) varListRight.get(ir);
						ArrayList tmplist = (ArrayList) tmpmap.get(rightvar);
						if (tmplist != null) {
							for (int i = 0; i < tmplist.size(); i++) {
								XQueryExpression tmpexpr = (XQueryExpression) tmplist.get(i);
								if (tmpexpr instanceof ListOpCompExpression) {
									ListOpCompExpression listopcomp = (ListOpCompExpression) tmpexpr;
									XQueryExpression expr1 = listopcomp.getExpression1();
									XQueryExpression expr2 = listopcomp.getExpression2();
									if (expr1 instanceof LocatedExpression && expr2 instanceof LocatedExpression) {
										if (listopcomp.getOperator() == Constants.EQ_COMPOP) {
											if (((LocatedExpression) expr1).startsWith((Variable) rightvar) && ((LocatedExpression) expr2).startsWith((Variable) leftvar)) {
												XQueryExpression bufexpr = expr2;
												try {
													listopcomp.setExpression2(expr1);
													listopcomp.setExpression1(bufexpr);
												} catch (Exception e) {
												}
											}
											if (hasnoteq) {
												leftExpressions.add(mergeRestrictions.size() - 1, listopcomp.getExpression1());
												rightExpressions.add(mergeRestrictions.size() - 1, listopcomp.getExpression2());
												mergeRestrictions.add(mergeRestrictions.size() - 1, listopcomp);
											} else {
												leftExpressions.add(listopcomp.getExpression1());
												rightExpressions.add(listopcomp.getExpression2());
												mergeRestrictions.add(listopcomp);
											}
										} else if (!hasnoteq && (listopcomp.getOperator() == Constants.GEQ_COMPOP || listopcomp.getOperator() == Constants.GT_COMPOP || listopcomp.getOperator() == Constants.LEQ_COMPOP || listopcomp.getOperator() == Constants.LT_COMPOP)) {
											hasnoteq = true;
											if (((LocatedExpression) expr1).startsWith((Variable) rightvar) && ((LocatedExpression) expr2).startsWith((Variable) leftvar)) {
												XQueryExpression bufexpr = expr2;
												try {
													listopcomp.setExpression2(expr1);
													listopcomp.setExpression1(bufexpr);
													if (listopcomp.getOperator() == Constants.GEQ_COMPOP)
														listopcomp.setOperator(Constants.LEQ_COMPOP);
													else if (listopcomp.getOperator() == Constants.GT_COMPOP)
														listopcomp.setOperator(Constants.LT_COMPOP);
													else if (listopcomp.getOperator() == Constants.LEQ_COMPOP)
														listopcomp.setOperator(Constants.GEQ_COMPOP);
													else if (listopcomp.getOperator() == Constants.LT_COMPOP)
														listopcomp.setOperator(Constants.GT_COMPOP);
												} catch (Exception e) {
												}
											}
											leftExpressions.add(listopcomp.getExpression1());
											rightExpressions.add(listopcomp.getExpression2());
											mergeRestrictions.add(listopcomp);
										} else if (otherRestrictions != null)
											otherRestrictions.add(listopcomp);

									} else if (otherRestrictions != null)
										otherRestrictions.add(listopcomp);
								} else if (otherRestrictions != null)
									otherRestrictions.add(tmpexpr);
							}
						}
					}
				}
				ArrayList conds = vardependencies.getAllConditionsWithVar(leftvar);
				if (conds != null)
					for (int i = 0; i < conds.size(); i++)
						if (!otherRestrictions.contains(conds.get(i)))
							otherRestrictions.add(conds.get(i));
			}
		}
		return hasnoteq;
	}

	//##################################################################################################################################
	//##################################################################################################################################
	//##################################################################################################################################
	private Operator createOperator(ExecutionPlan plan, HintTree hint, List vars) throws MediatorException {
		//##################################################################################################################################
		//##################################################################################################################################
		//##################################################################################################################################

		int typ = getAlgebraManager().getPlan().getDefaultSubHandling();
		Object left = null;
		Object right = null;
		List varListLeft = null;
		List varListRight = null;
		if (hint != null && hint.getType() != HintTree.NO_TYPE) {
			typ = hint.getType();
			left = hint.getLeftHint();
			right = hint.getRightHint();

			if (left instanceof Variable) {
				varListLeft = new ArrayList(1);
				varListLeft.add(left);
			} else
				varListLeft = ((HintTree) left).getVarList();

			if (right instanceof Variable) {
				varListRight = new ArrayList(1);
				varListRight.add(right);
			} else
				varListRight = ((HintTree) right).getVarList();
		} else if (vars != null) {
			if (vars.size() < 2)
				return null;
			else {
				if (typ == HintTree.MERGE_TYPE) {
					// find first possible division
					int cutIndex = 1;
					for (int i = 1; i < vars.size(); i++) {
						ArrayList tmplist = algManager.getMapNode((Variable) vars.get(i)).getVarsDependingOn();
						if (tmplist != null)
							for (int j = 0; j < tmplist.size(); j++) {
								if (vars.indexOf(tmplist.get(j)) != -1) {
									cutIndex = i + 1;
									break;
								}
							}
					}
					if (cutIndex < vars.size()) {
						varListLeft = vars.subList(0, cutIndex);
						varListRight = vars.subList(cutIndex, vars.size());
					} else
						return null;
				} else if (typ == HintTree.NESTED_TYPE) {
					return null;
				}
			}
		} else
			return null;

		if (typ == HintTree.MERGE_TYPE) {
			ArrayList mergeRestrictions = new ArrayList(1);
			ArrayList otherRestrictions = new ArrayList(1);
			ArrayList leftExpressions = new ArrayList(1);
			ArrayList rightExpressions = new ArrayList(1);
			boolean hasnoteq = false;
			if (hint != null || vars.size() >= 2)
				hasnoteq = extractConditionsfromMap(vardependencies, varListLeft, varListRight, leftExpressions, rightExpressions, mergeRestrictions, otherRestrictions);

			if (mergeRestrictions.isEmpty() && otherRestrictions.isEmpty()) {
				if (hint != null)
					org.xquark.util.misc.Debug.print("WARNING non effective hint (-> " + hint.toString() + "), no restrictions are present, default plan will be used instead");
				/*
				ArrayList varlist = new ArrayList();
				Algebra leftAlg = createVarAlgebra(plan, null, varlist, (Variable) left);
				varlist.remove(left);
				depmanager.removeVarFromScope((Variable) left);
				Algebra rightAlg = createVarAlgebra(plan, null, varlist, (Variable) right);
				return new AlgCartesian(plan, expression, leftAlg, rightAlg);
				*/
				// change   23/06/2003
				ArrayList varlist = new ArrayList();
				Operator leftAlg = null;
				if (varListLeft.size() == 1)
					leftAlg = createVarOperator(plan, null, varlist, (Variable) varListLeft.get(0));
				else
					leftAlg = createOperator(plan, (HintTree) hint.getLeftHint(), null);
				varlist.removeAll(varListLeft);
				algManager.removeVarsFromScope(varListLeft);
				Operator rightAlg = null;
				if (varListRight.size() == 1)
					rightAlg = createVarOperator(plan, null, varlist, (Variable) varListRight.get(0));
				else
					rightAlg = createOperator(plan, null, varListRight);
				algManager.addVarsToScope(varListLeft);
				algManager.addVarsToScope(varListRight);
				return new OpCartesian(plan, expression, leftAlg, rightAlg);
				/*
				ArrayList varlist = new ArrayList();
				//updateVarListWithDependencies(varListLeft, ((FLWRExpression) expression).getVariables());
				//updateVarListWithDependencies(varListRight, ((FLWRExpression) expression).getVariables());
				ArrayList vars = new ArrayList();
				vars.addAll(varListLeft);
				vars.addAll(varListRight);
				return createVarsAlgebra(plan, null, varlist, vars);
				*/
			}

			// ATTENTION erase these conditions from dependencies to avoid having them used several times
			for (int i = 0; i < otherRestrictions.size(); i++) {
				XQueryExpression otheri = (XQueryExpression) otherRestrictions.get(i);
				ArrayList tmplist = vardependencies.getVarsOfCond(otheri);
				for (int j = 0; j < tmplist.size(); j++) {
					Variable varj = (Variable) tmplist.get(j);
					ArrayList varjlist = vardependencies.getAllConditionsWithVar(varj);
					if (varjlist != null)
						varjlist.remove(otheri);
				}
			}

			if (mergeRestrictions.isEmpty()) {
				if (hint != null)
					org.xquark.util.misc.Debug.print("WARNING non effective hint (-> " + hint.toString() + "), no interesting restrictions are present");
			}
			if (hasnoteq) {
				ListOpCompExpression listopcomp = (ListOpCompExpression) mergeRestrictions.get(mergeRestrictions.size() - 1);
				if (listopcomp.getOperator() == Constants.LT_COMPOP || listopcomp.getOperator() == Constants.LEQ_COMPOP) {
					((XQueryExpression) leftExpressions.get(mergeRestrictions.size() - 1)).setOrder(Constants.DESCENDING_ORDER);
					((XQueryExpression) rightExpressions.get(mergeRestrictions.size() - 1)).setOrder(Constants.DESCENDING_ORDER);
				}
			}

			ArrayList varlist = new ArrayList();
			Operator leftAlg = null;
			if (hint == null) {
				left = vars.get(0);
				leftAlg = createVarOperator(plan, null, varlist, (Variable) left);
				varlist.remove(left);
				algManager.removeVarFromScope((Variable) left);
			} else if (left instanceof Variable) {
				leftAlg = createVarOperator(plan, null, varlist, (Variable) left);
				varlist.remove(left);
				algManager.removeVarFromScope((Variable) left);
			} else {
				leftAlg = createOperator(plan, (HintTree) left, null);
				algManager.removeVarsFromScope(((HintTree) left).getVarList());
				//leftAlg = new AlgSort(plan, leftAlg, leftExpressions);
			}
			if (!leftExpressions.isEmpty())
				leftAlg.addIndexSpec(leftExpressions);
			Operator rightAlg = null;
			if (hint == null) {
				if (vars.size() == 2) {
					rightAlg = createVarOperator(plan, null, varlist, (Variable) vars.get(1));
				} else {
					rightAlg = createOperator(plan, null, vars.subList(1, vars.size()));
					algManager.removeVarsFromScope(vars.subList(1, vars.size()));
				}
			} else if (right instanceof Variable) {
				rightAlg = createVarOperator(plan, null, varlist, (Variable) right);
				//varlist.clear();
				//depmanager.removeVarFromScope((Variable) right);
			} else {
				rightAlg = createOperator(plan, (HintTree) right, null);
				algManager.removeVarsFromScope(((HintTree) right).getVarList());
				//rightAlg = new AlgSort(plan, rightAlg, rightExpressions);
			}
			if (!rightExpressions.isEmpty())
				rightAlg.addIndexSpec(rightExpressions);
			Operator operator = new OpSortMerge(plan, leftAlg, rightAlg, mergeRestrictions);
			// add restriction if necessary
			if (!otherRestrictions.isEmpty()) {
				XQueryExpression tmpexpr = null;
				if (otherRestrictions.size() == 1)
					tmpexpr = (XQueryExpression) otherRestrictions.get(0);
				else {
					try {
						tmpexpr = new BinOpANDExpression((XQueryExpression) otherRestrictions.get(0), (XQueryExpression) otherRestrictions.get(1), algManager.getTypeVisitor());
						for (int i = 2; i < otherRestrictions.size(); i++)
							tmpexpr = new BinOpANDExpression(tmpexpr, (XQueryExpression) otherRestrictions.get(i), algManager.getTypeVisitor());
					} catch (Exception e) {
						throw new MediatorException("Could not build where clause", e);
					}
				}
				operator = operator.addCondition(plan, tmpexpr);
				//operator = new OpRestrict(plan, tmpexpr, operator);
			}
			// reset scope
			algManager.addVarsToScope((hint != null) ? hint.getVarList() : vars);
			// reset order list
			//((FLWRExpression) expression).getOrderBy().removeAll(rightExpressions);
			//((FLWRExpression) expression).getOrderBy().removeAll(leftExpressions);
			return operator;
		}
		if (typ == HintTree.NESTED_LOOP_TYPE) {
			// create algebra
			Operator leftalg = null;
			ArrayList varlist = new ArrayList();
			if (left instanceof Variable) {
				//updateVarListWithDependencies(varListLeft, ((FLWRExpression) expression).getVariables());
				leftalg = createVarsOperator(plan, null, varlist, varListLeft);
				algManager.addVarToScope((Variable) left);
			} else {
				leftalg = createOperator(plan, (HintTree) left, null);
				algManager.addVarsToScope(((HintTree) left).getVarList());
				varlist.addAll(((HintTree) left).getVarList());
			}
			Operator rightalg = null;
			if (right instanceof Variable) {
				//updateVarListWithDependencies(varListRight, ((FLWRExpression) expression).getVariables());
				rightalg = createVarsOperator(plan, null, varlist, varListRight);
				algManager.addVarToScope((Variable) right);
			} else {
				rightalg = createOperator(plan, (HintTree) right, null);
				algManager.addVarsToScope(((HintTree) right).getVarList());
			}
			rightalg.setValueDepends(true);
			ArrayList tmplistalg = new ArrayList(2);
			tmplistalg.add(leftalg);
			tmplistalg.add(rightalg);
			return new OpSubQuery(plan, expression, tmplistalg, true, false);
		} else {
			throw new MediatorException("Unsupported type of hint : " + typ);
		}
	}

	private void verifySubElements() throws MediatorException {
		// rules of sub queries
		// if there is no outer hint -> take the default hint for all
		// if there is a single nested -> nested for all
		// if there is a merge and no nested -> merge for all
		if (subDepFLWRs == null || subDepFLWRs.isEmpty())
			return;
		boolean hasMerge = false;
		boolean hasNested = false;
		subhandling = HintTree.NO_TYPE; //getDepManager().getPlan().defaultsubqueryhandling;
		for (int i = 0; i < subDepFLWRs.size(); i++) {
			((AlgFLWR) subDepFLWRs.get(i)).verifySubElements();
		}
		if (subhandling != HintTree.NESTED_TYPE) {
			for (int i = 0; i < subDepFLWRs.size(); i++) {
				AlgFLWR subdepi = (AlgFLWR) subDepFLWRs.get(i);
				HintTree ht = ((FLWRExpression) subdepi.getExpression()).getHintTree();
				if (ht != null) {
					if (!hasNested && ht.getOuterType() == HintTree.NESTED_TYPE) {
						subhandling = HintTree.NESTED_TYPE;
						hasNested = true;
						break;
					}
					if (!hasMerge && ht.getOuterType() == HintTree.MERGE_TYPE)
						hasMerge = true;
				}
			}
		} else
			hasNested = true;

		// check that all sub elements have the same sort items for parent element (none is ok too)
		if (hasNested)
			subhandling = HintTree.NESTED_TYPE;
		else if (hasMerge)
			subhandling = HintTree.MERGE_TYPE;
		else
			subhandling = getAlgebraManager().getPlan().getDefaultSubHandling();

		algManager.addVarsToScope(getVariables());
		for (int i = 0; i < subDepFLWRs.size(); i++) {
			AlgFLWR depi = (AlgFLWR) subDepFLWRs.get(i);
			depi.makeInformation();
		}
		// verify that all sub elements can do merge (no condition outside)
		IsolateWhereVisitor visitor = new IsolateWhereVisitor();
		ArrayList varListLeft = ((FLWRExpression) expression).getVariables();
		for (int i = 0; i < subDepFLWRs.size(); i++) {
			AlgFLWR depi = (AlgFLWR) subDepFLWRs.get(i);
			varListLeft.addAll(((FLWRExpression) depi.getExpression()).getVariables());
			visitor.reset(varListLeft);
			if (((FLWRExpression) depi.getExpression()).getWhereClause() != null) {
				try {
					((FLWRExpression) depi.getExpression()).getWhereClause().accept(visitor);
				} catch (XQueryException e) {
					throw new MediatorException("Could not isolated expression : " + ((FLWRExpression) depi.getExpression()).getWhereClause());
				}
			}
			varListLeft.removeAll(((FLWRExpression) depi.getExpression()).getVariables());

			if (visitor.getRemainingExpression() != null) {
				org.xquark.util.misc.Debug.print("WARNING non effective hint (-> " + hinttree + "), outer hint is incorrect, default plan will be used instead");
				if (getParentDepFLWR() != null)
					getParentDepFLWR().setSubHandling(HintTree.NESTED_TYPE);
				break;
			}
		}

		algManager.removeVarsFromScope(getVariables());

		//		if (subDepFLWRs.size() == 1) {
		//			/*
		//			 * if hinttree is null no pb
		//			 * if hint tree is not null and no merge no pb
		//			 * if hinttree is not null and merge
		//			 * 	reorder algebra other elements
		//			 */
		//			DepFLWR dep0 = (DepFLWR) subDepFLWRs.get(0);
		//			FLWRExpression depexpr0 = (FLWRExpression) dep0.getExpression();
		//			ArrayList vars0 = depexpr0.getVariables();
		//			Debug.print("*********** DEPFLWR SUBDEPFLWR STUFF *************************");
		//			Debug.print("ClassName = " + dep0.getClass().getName());
		//			Debug.print("is let = " + dep0.isLet());
		//			Debug.print("depiexpr = " + dep0.getExpression());
		//			Debug.print("vari = " + dep0.getVariables());
		//			Debug.print("depends = " + dep0.getDepends());
		//			Debug.print("vars depending on = " + dep0.getVarsDependingOn());
		//			Debug.print("vars referenced by = " + dep0.getVarsReferencedBy());
		//			Debug.print("children = " + dep0.getChildren());
		//		}
	}

	//##################################################################################################################################
	//##################################################################################################################################
	//##################################################################################################################################
	private Operator createOperator(ExecutionPlan plan, ArrayList indexSpecs, ArrayList subRestrictions, ArrayList outsideRestrictions) throws MediatorException {
		//##################################################################################################################################
		//##################################################################################################################################
		//##################################################################################################################################
		Operator retval = null;
		if (dependDefinitionVars.isEmpty())
			retval = createOperator(plan);
		else if (((FLWRExpression) expression).getWhereClause() == null)
			retval = createOperator(plan);
		else if (vardependencies.hasJoins())
			retval = createOperator(plan);
		else {
			ArrayList varListLeft = ((FLWRExpression) getParentDepFLWR().getExpression()).getVariables();
			ArrayList varListRight = ((FLWRExpression) getExpression()).getVariables();
			ArrayList mergeRestrictions = new ArrayList();
			ArrayList otherRestrictions = new ArrayList();
			ArrayList leftExpressions = new ArrayList();
			ArrayList rightExpressions = new ArrayList();

			boolean hasnoteq = extractConditionsfromMap(vardependencies, varListLeft, varListRight, leftExpressions, rightExpressions, mergeRestrictions, otherRestrictions);
			if (mergeRestrictions.isEmpty()) {
				org.xquark.util.misc.Debug.print("WARNING non effective hint (-> " + hinttree + "), outer hint is incorrect, default plan will be used instead");
				retval = createOperator(plan);
			} else {
				if (hasnoteq) {
					ListOpCompExpression listopcomp = (ListOpCompExpression) mergeRestrictions.get(mergeRestrictions.size() - 1);
					if (listopcomp.getOperator() == Constants.LT_COMPOP || listopcomp.getOperator() == Constants.LEQ_COMPOP) {
						((XQueryExpression) leftExpressions.get(mergeRestrictions.size() - 1)).setOrder(Constants.DESCENDING_ORDER);
						((XQueryExpression) rightExpressions.get(mergeRestrictions.size() - 1)).setOrder(Constants.DESCENDING_ORDER);
					}
				}
				org.xquark.util.misc.Debug.print("mergeRestrictions = " + mergeRestrictions);
				org.xquark.util.misc.Debug.print("leftExpressions = " + leftExpressions);
				org.xquark.util.misc.Debug.print("rightExpressions = " + rightExpressions);

				// get all conditions only on subquery
				org.xquark.util.misc.Debug.print("varVarWhereMap = " + vardependencies.getJoinsMap());
				for (int i = 0; i < varListLeft.size(); i++) {
					Variable vari = (Variable) varListLeft.get(i);
					HashMap tmpmap = (HashMap) vardependencies.getJoinsMap().get(vari);
					org.xquark.util.misc.Debug.print("tmpmap = " + tmpmap);
					for (Iterator it = tmpmap.values().iterator(); it.hasNext();) {
						ArrayList tmplist = (ArrayList) it.next();
						for (int j = 0; j < tmplist.size(); j++) {
							if (!mergeRestrictions.contains(tmplist.get(j)) && !otherRestrictions.contains(tmplist.get(j)))
								otherRestrictions.add(tmplist.get(j));
						}
					}
				}
				org.xquark.util.misc.Debug.print("otherRestrictions = " + otherRestrictions);

				//				ArrayList localRestrictions = new ArrayList();
				//				// get all conditions only on subquery
				//				org.xquark.xml.misc.Debug.print("varVarWhereMap = " + varVarWhereMap);
				//				for (int i=0;i<varListRight.size();i++) {
				//					Variable vari = (Variable) varListRight.get(i);
				//					for (int j=0;j<varListLeft.size();j++) {					
				//						HashMap tmpmap = (HashMap)varVarWhereMap.get(varListRight.get(j));
				//						if (tmpmap != null) {
				//							ArrayList tmplist = (ArrayList)tmpmap.get(vari);
				//							org.xquark.xml.misc.Debug.print("tmplist = " + tmplist);
				//							if (tmplist != null)
				//								localRestrictions.addAll(tmplist);
				//						}
				//					}
				//				}
				//				org.xquark.xml.misc.Debug.print("localRestrictions = " + localRestrictions);

				// add index to parent algebra
				indexSpecs.add(leftExpressions);

				// modify where expression to contains only conditions on subquery
				// and add to ouside restrictions conditions either merge not allowed or all others
				XQueryExpression whereExpr = ((FLWRExpression) expression).getWhereClause();
				org.xquark.util.misc.Debug.print("whereExpr = " + whereExpr);
				try {
					for (int i = 0; i < mergeRestrictions.size(); i++) {
						//if (tmpwhere == null) break; // never happens!!!
						whereExpr = whereExpr.erasePart((XQueryExpression) mergeRestrictions.get(i));
					}
					for (int i = 0; i < otherRestrictions.size(); i++) {
						//if (tmpwhere == null) break; // never happens!!!
						whereExpr = whereExpr.erasePart((XQueryExpression) otherRestrictions.get(i));
					}
					//					for (int i=0;i < localRestrictions.size(); i++) {
					//						//if (tmpwhere == null) break; // never happens!!!
					//						whereExpr = whereExpr.erasePart((XQueryExpression)localRestrictions.get(i));
					//					}
				} catch (XQueryException e) {
					throw new MediatorException("Could not erase part from expression : " + whereExpr);
				}
				org.xquark.util.misc.Debug.print("whereExpr = " + whereExpr);
				((FLWRExpression) expression).setWhereClause(whereExpr);
				//outsideRestrictions.add(otherRestrictions);
				if (otherRestrictions.isEmpty()) {
					outsideRestrictions.add(null);
				} else {
					if (otherRestrictions.size() == 1)
						outsideRestrictions.add(otherRestrictions.get(0));
					else {
						try {
							BinOpANDExpression tmpBinOp = new BinOpANDExpression((XQueryExpression) otherRestrictions.get(0), (XQueryExpression) otherRestrictions.get(1), algManager.getTypeVisitor());
							for (int i = 2; i < otherRestrictions.size(); i++)
								tmpBinOp = new BinOpANDExpression(tmpBinOp, (XQueryExpression) otherRestrictions.get(i), algManager.getTypeVisitor());
							outsideRestrictions.add(tmpBinOp);
						} catch (Exception e) {
							throw new MediatorException("Could not aggregate list of expressions : " + otherRestrictions);
						}
					}
				}

				//				if (localRestrictions.isEmpty()) {
				//					((FLWRExpression)getExpression()).setWhereClause(null);
				//				}
				//				else {
				//					if (localRestrictions.size() == 1)
				//						((FLWRExpression)getExpression()).setWhereClause((XQueryExpression)localRestrictions.get(0));
				//					else {
				//						try {
				//							BinOpANDExpression tmpBinOp = new BinOpANDExpression((XQueryExpression) localRestrictions.get(0), (XQueryExpression) localRestrictions.get(1), depmanager.getTypeVisitor());
				//							for (int i = 2; i < localRestrictions.size(); i++)
				//								tmpBinOp = new BinOpANDExpression(tmpBinOp, (XQueryExpression) localRestrictions.get(i), depmanager.getTypeVisitor());
				//							((FLWRExpression)getExpression()).setWhereClause(tmpBinOp);
				//						} catch (Exception e) {
				//							throw new MediatorException("Could not aggregate list of expressions : " + localRestrictions);
				//						}
				//					}
				//				}

				retval = createOperator(plan);
				retval.addIndexSpec(rightExpressions);
				subRestrictions.add(mergeRestrictions);
			}
		}
		return retval;
	}

	public Operator createOperator(ExecutionPlan plan) throws MediatorException {
		//if (true) throw new MediatorException("createAlgebra not supported in DepFLWR"); 
		Debug.print("DepFLWR start creating Algebra : " + expression);

		// test child number (not necessary)
		/*
		if (depchildren.size() == 0)
			return null;
		*/
		// definition of return value
		Operator operator = null;
		// gather information concerning where expression if any
		makeInformation();
		// gather information on subelements if any
		if (getParentDepFLWR() == null)
			verifySubElements();

		ArrayList subElements = null;
		ArrayList subRestrictions = null;
		ArrayList outsideExpressions = null;
		ArrayList indexSpecs = null;

		if (subDepFLWRs != null && !subDepFLWRs.isEmpty()) {

			Debug.print("Scope = " + algManager.getScopeVariables());
			algManager.addVarsToScope(((FLWRExpression) expression).getVariables());
			Debug.print("Scope = " + algManager.getScopeVariables());

			if (subhandling == HintTree.NESTED_TYPE) {
				subElements = new ArrayList(subDepFLWRs.size() + 1);
				//subElements.add(algebra);
				//ArrayList subWhereExpressions = new ArrayList();
				for (int i = 0; i < subDepFLWRs.size(); i++) {
					AlgFLWR depi = (AlgFLWR) subDepFLWRs.get(i);
					ArrayList varsi = ((FLWRExpression) depi.getExpression()).getVariables();
					Debug.print("*********** DEPFLWR SUBDEPFLWR STUFF *************************");
					Debug.print("ClassName = " + depi.getClass().getName());
					Debug.print("is let = " + depi.isLet());
					Debug.print("depiexpr = " + depi.getExpression());
					Debug.print("vari = " + depi.getVariables());
					//					Debug.print("value depends = " + depi.getValueDepends());
					//					Debug.print("where depends = " + depi.getWhereDepends());
					Debug.print("definition vars depending on = " + depi.getVarsDependingOn());
					Debug.print("where vars depending on = " + depi.getVarsWhereOn());
					Debug.print("vars referenced by = " + depi.getVarsReferencedBy());
					Debug.print("children = " + depi.getChildren());

					Operator subAlg = depi.createOperator(plan);
					if (!depi.getVarsDependingOn().isEmpty()) // add   10/02/2003
						subAlg.setValueDepends(true);
					if (!depi.getVarsWhereOn().isEmpty()) // add   10/02/2003
						subAlg.setWhereDepends(true);
					subElements.add(subAlg);
					//subWhereExpressions.add(depi.getWhereClause());
				}
				//				algebra = new AlgSubQuery(plan, expression, subElements, true, true, false);
				//				if (!subWhereExpressions.isEmpty()) {
				//					try {
				//						for (int i = 0; i < subWhereExpressions.size(); i++)
				//							 ((FLWRExpression) expression).addWhereClause((XQueryExpression) subWhereExpressions.get(i), depmanager.getTypeVisitor());
				//					} catch (Exception e) {
				//						throw new MediatorException("Can't make external where expression");
				//					}
				//					XQueryExpression tmpExpr = ((FLWRExpression) expression).getWhereClause();
				//					if (tmpExpr != null) {
				//						isolateVisitor = new IsolateWhereVisitor(depmanager.getScopeVariables());
				//						try {
				//							tmpExpr.accept(isolateVisitor);
				//						} catch (Exception e) {
				//							throw new MediatorException("Can't isolate where expression : " + expression, e);
				//						}
				//						tmpExpr = isolateVisitor.getIsolatedExpression();
				//						if (tmpExpr != null)
				//							algebra = new AlgRestrict(plan, tmpExpr, algebra);
				//						tmpExpr = isolateVisitor.getRemainingExpression();
				//						((FLWRExpression) expression).setWhereClause(tmpExpr);
				//					}
				//				}
			}
			if (subhandling == HintTree.MERGE_TYPE) {
				subElements = new ArrayList(subDepFLWRs.size() + 1);
				//subElements.add(algebra);
				if (outsideExpressions == null)
					outsideExpressions = new ArrayList();
				if (indexSpecs == null)
					indexSpecs = new ArrayList();
				subRestrictions = new ArrayList();
				for (int i = 0; i < subDepFLWRs.size(); i++) {
					AlgFLWR depi = (AlgFLWR) subDepFLWRs.get(i);
					ArrayList varsi = ((FLWRExpression) depi.getExpression()).getVariables();
					Debug.print("*********** DEPFLWR SUBDEPFLWR STUFF *************************");
					Debug.print("ClassName = " + depi.getClass().getName());
					Debug.print("is let = " + depi.isLet());
					Debug.print("depiexpr = " + depi.getExpression());
					Debug.print("vari = " + depi.getVariables());
					//					Debug.print("value depends = " + depi.getValueDepends());
					//					Debug.print("where depends = " + depi.getWhereDepends());
					Debug.print("definition vars depending on = " + depi.getVarsDependingOn());
					Debug.print("where vars depending on = " + depi.getVarsWhereOn());
					Debug.print("vars referenced by = " + depi.getVarsReferencedBy());
					Debug.print("children = " + depi.getChildren());

					Operator subAlg = depi.createOperator(plan, indexSpecs, subRestrictions, outsideExpressions);
					Debug.print("subRestrictions = " + subRestrictions);
					Debug.print("outsideExpressions = " + outsideExpressions);
					if (!depi.getVarsDependingOn().isEmpty()) // add   10/02/2003
						subAlg.setValueDepends(true);
					if (!depi.getVarsWhereOn().isEmpty()) // add   10/02/2003
						subAlg.setWhereDepends(true);
					subElements.add(subAlg);
				}

				//				if (subRestrictions.isEmpty())
				//					algebra = new AlgSubQuery(plan, expression, subElements, true, true, false);
				//				else
				//					algebra = new AlgSubQuery(plan, expression, subElements, true, true, false, subRestrictions);
				//				if (!subWhereExpressions.isEmpty()) {
				//					try {
				//						for (int i = 0; i < subWhereExpressions.size(); i++)
				//							 ((FLWRExpression) expression).addWhereClause((XQueryExpression) subWhereExpressions.get(i), depmanager.getTypeVisitor());
				//					} catch (Exception e) {
				//						throw new MediatorException("Can't make external where expression");
				//					}
				//					XQueryExpression tmpExpr = ((FLWRExpression) expression).getWhereClause();
				//					if (tmpExpr != null) {
				//						isolateVisitor = new IsolateWhereVisitor(depmanager.getScopeVariables());
				//						try {
				//							tmpExpr.accept(isolateVisitor);
				//						} catch (Exception e) {
				//							throw new MediatorException("Can't isolate where expression : " + expression, e);
				//						}
				//						tmpExpr = isolateVisitor.getIsolatedExpression();
				//						if (tmpExpr != null)
				//							algebra = new AlgRestrict(plan, tmpExpr, algebra);
				//						tmpExpr = isolateVisitor.getRemainingExpression();
				//						((FLWRExpression) expression).setWhereClause(tmpExpr);
				//					}
				//				}
			}

			Debug.print("Scope = " + algManager.getScopeVariables());
			algManager.removeVarsFromScope(((FLWRExpression) expression).getVariables());
			Debug.print("Scope = " + algManager.getScopeVariables());

		}

		//if (hinttree != null && hinttree.getType() != HintTree.NO_TYPE) {
		operator = createOperator(plan, hinttree, ((FLWRExpression) expression).getVariables());
		if (operator != null) {
			// if flwr variables are not in modified hint tree create standard algebra
			if (hinttree != null) {
				ArrayList othervars = (ArrayList) ((FLWRExpression) expression).getVariables();
				if (!varHintList.containsAll(othervars)) {
					othervars.removeAll(varHintList);
					operator = createVarsOperator(plan, operator, varHintList, othervars);
				}
			}
		}
		if (operator == null) {
			// ==> NO HINT TREE 
			// change   27/03/2003 this was stupid!!!!
			//			// erase parts of where clause copied from parent dep flwrs if any
			//			if (((FLWRExpression) expression).getWhereClause() != null && !depmanager.getScopeVariables().isEmpty()) {
			//				isolateVisitor = new IsolateWhereVisitor(depmanager.getScopeVariables());
			//				try {
			//					((FLWRExpression) expression).getWhereClause().accept(isolateVisitor);
			//				} catch (Exception e) {
			//					throw new MediatorException("Can't isolate where expression : " + ((FLWRExpression) expression).getWhereClause(), e);
			//				}
			//				((FLWRExpression) expression).setWhereClause(isolateVisitor.getRemainingExpression());
			//			}
			// create order of variable depending on sort list, dependencies and declaration
			if (!isMonoSource()) {
				computeVarOrderList();
				verifyVarOrderList();
			} else {
				varOrderList = new ArrayList();
				varOrderList.add(((FLWRExpression) expression).getVariables().get(0));
			}

			Debug.print("********** WHERE STUFF **************************");
			Debug.print("where = " + ((FLWRExpression) expression).getWhereClause());
			Debug.print("vardependencies = " + vardependencies);
			//Debug.print("allExprWhereList = " + allExprWhereList);
			//Debug.print("complexExprWhereList = " + complexExprWhereList);
			Debug.print("varOrderMap = " + varOrderMap);
			Debug.print("orderVarMap = " + orderVarMap);
			Debug.print("varOrderList = " + varOrderList);
			Debug.print("varReOrderList = " + varReOrderList);
			//Debug.print("varHintList = " + varHintList);
			//Debug.print("complexExprOrderList = " + complexExprOrderList);
			Debug.print("children = " + algChildren);

			ArrayList varlist = new ArrayList(varOrderList.size());
			for (int i = 0; i < varOrderList.size(); i++) {
				operator = createVarOperator(plan, operator, varlist, (Variable) varOrderList.get(i));
				algManager.getScopeVariables().add(varOrderList.get(i));
			}
			algManager.getScopeVariables().removeAll(varOrderList);
		}

		if (subElements != null) {
			subElements.add(0, operator);
			if (subhandling == HintTree.NESTED_TYPE)
				operator = new OpSubQuery(plan, expression, subElements, true, true);
			if (subhandling == HintTree.MERGE_TYPE) {
				Debug.print("subRestrictions = " + subRestrictions);
				Debug.print("outsideExpressions = " + outsideExpressions);
				if (subRestrictions.isEmpty())
					operator = new OpSubQuery(plan, expression, subElements, true, true);
				else {
					operator.addIndexSpecs(indexSpecs);
					operator = new OpSubQuery(plan, expression, subElements, true, true, subRestrictions, outsideExpressions);
				}
			}
		}

		//		if (subDepFLWRs != null && !subDepFLWRs.isEmpty()) {
		//			if (subhandling == HintTree.NESTED_TYPE) {
		//				ArrayList subElements = new ArrayList(subDepFLWRs.size() + 1);
		//				subElements.add(algebra);
		//				//ArrayList subWhereExpressions = new ArrayList();
		//				for (int i = 0; i < subDepFLWRs.size(); i++) {
		//					DepFLWR depi = (DepFLWR) subDepFLWRs.get(i);
		//					ArrayList varsi = ((FLWRExpression) depi.getExpression()).getVariables();
		//					Debug.print("*********** DEPFLWR SUBDEPFLWR STUFF *************************");
		//					Debug.print("ClassName = " + depi.getClass().getName());
		//					Debug.print("is let = " + depi.isLet());
		//					Debug.print("depiexpr = " + depi.getExpression());
		//					Debug.print("vari = " + depi.getVariables());
		//					Debug.print("depends = " + depi.getDepends());
		//					Debug.print("vars depending on = " + depi.getVarsDependingOn());
		//					Debug.print("vars referenced by = " + depi.getVarsReferencedBy());
		//					Debug.print("children = " + depi.getChildren());
		//		
		//					Algebra subAlg = depi.createAlgebra(plan);
		//					if (!depi.getVarsDependingOn().isEmpty()) // add   10/02/2003
		//						subAlg.setDepends(true);
		//					subElements.add(subAlg);
		//					//subWhereExpressions.add(depi.getWhereClause());
		//				}
		//				algebra = new AlgSubQuery(plan, expression, subElements, true, true, false);
		////				if (!subWhereExpressions.isEmpty()) {
		////					try {
		////						for (int i = 0; i < subWhereExpressions.size(); i++)
		////							 ((FLWRExpression) expression).addWhereClause((XQueryExpression) subWhereExpressions.get(i), depmanager.getTypeVisitor());
		////					} catch (Exception e) {
		////						throw new MediatorException("Can't make external where expression");
		////					}
		////					XQueryExpression tmpExpr = ((FLWRExpression) expression).getWhereClause();
		////					if (tmpExpr != null) {
		////						isolateVisitor = new IsolateWhereVisitor(depmanager.getScopeVariables());
		////						try {
		////							tmpExpr.accept(isolateVisitor);
		////						} catch (Exception e) {
		////							throw new MediatorException("Can't isolate where expression : " + expression, e);
		////						}
		////						tmpExpr = isolateVisitor.getIsolatedExpression();
		////						if (tmpExpr != null)
		////							algebra = new AlgRestrict(plan, tmpExpr, algebra);
		////						tmpExpr = isolateVisitor.getRemainingExpression();
		////						((FLWRExpression) expression).setWhereClause(tmpExpr);
		////					}
		////				}
		//			}
		//			if (subhandling == HintTree.MERGE_TYPE) {
		//				ArrayList subElements = new ArrayList(subDepFLWRs.size() + 1);
		//				subElements.add(algebra);
		//				ArrayList subWhereExpressions = new ArrayList();
		//				ArrayList subRestrictions = new ArrayList();
		//				for (int i = 0; i < subDepFLWRs.size(); i++) {
		//					DepFLWR depi = (DepFLWR) subDepFLWRs.get(i);
		//					ArrayList varsi = ((FLWRExpression) depi.getExpression()).getVariables();
		//					Debug.print("*********** DEPFLWR SUBDEPFLWR STUFF *************************");
		//					Debug.print("ClassName = " + depi.getClass().getName());
		//					Debug.print("is let = " + depi.isLet());
		//					Debug.print("depiexpr = " + depi.getExpression());
		//					Debug.print("vari = " + depi.getVariables());
		//					Debug.print("depends = " + depi.getDepends());
		//					Debug.print("vars depending on = " + depi.getVarsDependingOn());
		//					Debug.print("vars referenced by = " + depi.getVarsReferencedBy());
		//					Debug.print("children = " + depi.getChildren());
		//		
		//					Algebra subAlg = depi.createAlgebra(plan,algebra,subRestrictions,subWhereExpressions);
		//					if (!depi.getVarsDependingOn().isEmpty()) // add   10/02/2003
		//						subAlg.setDepends(true);
		//					subElements.add(subAlg);
		//					//subWhereExpressions.add(depi.getWhereClause());
		//				}
		//				if (subRestrictions.isEmpty())
		//					algebra = new AlgSubQuery(plan, expression, subElements, true, true, false);
		//				else
		//					algebra = new AlgSubQuery(plan, expression, subElements, true, true, false, subRestrictions);
		////				if (!subWhereExpressions.isEmpty()) {
		////					try {
		////						for (int i = 0; i < subWhereExpressions.size(); i++)
		////							 ((FLWRExpression) expression).addWhereClause((XQueryExpression) subWhereExpressions.get(i), depmanager.getTypeVisitor());
		////					} catch (Exception e) {
		////						throw new MediatorException("Can't make external where expression");
		////					}
		////					XQueryExpression tmpExpr = ((FLWRExpression) expression).getWhereClause();
		////					if (tmpExpr != null) {
		////						isolateVisitor = new IsolateWhereVisitor(depmanager.getScopeVariables());
		////						try {
		////							tmpExpr.accept(isolateVisitor);
		////						} catch (Exception e) {
		////							throw new MediatorException("Can't isolate where expression : " + expression, e);
		////						}
		////						tmpExpr = isolateVisitor.getIsolatedExpression();
		////						if (tmpExpr != null)
		////							algebra = new AlgRestrict(plan, tmpExpr, algebra);
		////						tmpExpr = isolateVisitor.getRemainingExpression();
		////						((FLWRExpression) expression).setWhereClause(tmpExpr);
		////					}
		////				}
		//			}
		//		}

		// if it is a root node, than we must add final projection
		if (parentDepFLWR == null) {
			if (parent != null && (variables != null && !variables.isEmpty())) {
				// start add   30/05/2003
				boolean retvalue = false;
				XQueryExpression retexpr = ((FLWRExpression) expression).getReturnClause();
				while (!retvalue) {
					if (retexpr instanceof LocatedExpression || retexpr instanceof Variable)
						retvalue = true;
					else {
						if (retexpr instanceof XQueryExpressionSequence) {
							XQueryExpressionSequence retseq = (XQueryExpressionSequence) retexpr;
							if (retseq.getSubExpressions().size() == 1)
								retexpr = (XQueryExpression) retseq.getSubExpressions().get(0);
							else
								break;
						} else
							break;
					}
				}
				if (retvalue)
					operator = new OpNotSource(plan, retexpr, (Variable) variables.get(0), operator, islet);
				else
					// end add   30/05/2003
					operator = new OpNotSource(plan, null, (Variable) variables.get(0), operator, islet);
			}
		}
		if (varReOrderList != null && !varReOrderList.isEmpty()) {
			ArrayList neworders = new ArrayList();
			ArrayList orders = ((FLWRExpression) expression).getOrderBy();
			boolean doit = false;
			if (orders != null)
				for (int i = 0; i < orders.size(); i++) {
					/*
					if (orders.get(i) instanceof Variable) {
						Variable tmpvar = (Variable) orders.get(i);
						ArrayList tmpVarOrders = (ArrayList) varOrderMap.get(tmpvar);
						if (tmpVarOrders != null && tmpVarOrders.contains(tmpvar) && varReOrderList.contains(tmpvar))
							neworders.add(tmpvar);
					} else {
						LocatedExpression tmploc = (LocatedExpression) orders.get(i);
						Variable tmpvar = (Variable) tmploc.getStep(0).getExpression();
						ArrayList tmpVarOrders = (ArrayList) varOrderMap.get(tmpvar);
						if (tmpVarOrders != null && tmpVarOrders.contains(tmploc) && varReOrderList.contains(tmpvar))
							neworders.add(tmploc);
					}
					*/
					Variable tmpvar = null;
					XQueryExpression tmpexpr = null;
					if (orders.get(i) instanceof Variable) {
						tmpvar = (Variable) orders.get(i);
						tmpexpr = tmpvar;
					} else {
						tmpexpr = (XQueryExpression) orders.get(i);
						tmpvar = (Variable) ((LocatedExpression) tmpexpr).getExpression();
					}
					if (doit || varReOrderList.contains(tmpvar)) {
						if (!doit)
							doit = true;
						neworders.add(tmpexpr);
					}
				}
			if (!neworders.isEmpty())
				operator = new OpSort(plan, operator, neworders);
		}
		if (!isQDB())
			operator.isLet(islet);
		return operator;
	}

	public Object clone() throws CloneNotSupportedException {
		AlgFLWR newobj = (AlgFLWR) super.clone();
		newobj.parentDepFLWR = parentDepFLWR;
		if (subDepFLWRs == null)
			newobj.subDepFLWRs = null;
		else {
			newobj.subDepFLWRs = new ArrayList();
			for (int i = 0; i < subDepFLWRs.size(); i++)
				newobj.subDepFLWRs.add(((AlgFLWR) subDepFLWRs.get(i)).clone());
		}
		return newobj;
	} // ************************************************************************
	// * METHODS
	// ************************************************************************
	/**
	 *
	 * @return
	 */
	public ArrayList getPaths() {
		if (paths == null)
			paths = new ArrayList(); // CHANGE  
		/*
		if ((paths != null) && (paths.isEmpty())) {
			for (int i = 0 ; i < depchildren.size() ; i ++) {
			ArrayList al = ((DepLeaf) depchildren.get(i)).getPaths() ;
			for (int j = 0 ; j < al.size() ; j ++)
				paths.add(al.get(j)) ;
			}
		}
		 */
		if (paths.isEmpty()) {
			if (algChildren != null) {
				for (int i = 0; i < algChildren.size(); i++) {
					ArrayList tmplist = ((Algebra) algChildren.get(i)).getPaths();
					if (tmplist != null)
						paths.addAll(tmplist);
				}
			}
			if (subDepFLWRs != null) {
				for (int i = 0; i < subDepFLWRs.size(); i++) {
					ArrayList tmplist = ((AlgFLWR) subDepFLWRs.get(i)).getPaths();
					if (tmplist != null)
						paths.addAll(tmplist);
				}
			}
		}
		return paths;
	}

	/*
	public ArrayList getVariables() {
		return ((FLWRExpression) expression).getVariables();
	}
	*/

	public ArrayList getAllVariables() {
		ArrayList res = (ArrayList) ((FLWRExpression) expression).getVariables().clone();
		if (subDepFLWRs != null)
			for (int i = 0; i < subDepFLWRs.size(); i++)
				res.addAll(((AlgFLWR) subDepFLWRs.get(i)).getAllVariables());
		return res;
	}

	public AlgFLWR getParentDepFLWR() {
		return parentDepFLWR;
	}

	public ArrayList getAllOrderBys() {
		ArrayList parentres = ((FLWRExpression) expression).getOrderBy();
		if (subDepFLWRs == null)
			return parentres;
		ArrayList res = new ArrayList();
		for (int i = 0; i < subDepFLWRs.size(); i++) {
			ArrayList tmpres = ((AlgFLWR) subDepFLWRs.get(i)).getAllOrderBys();
			if (tmpres != null) {
				res.addAll(tmpres);
				if (parentres != null)
					res.removeAll(parentres);
			}
		}
		if (parentres != null)
			res.addAll(0, parentres);
		if (res.isEmpty())
			return null;
		return res;
	} // ******************************************************************
	// * DEBUGGING
	// ******************************************************************
	public String toCompleteString() {
		return toCompleteString(0);
	}
	public String toCompleteString(int indent) {
		StringBuffer buf = new StringBuffer();
		buf.append(Utils.makeIndent(indent) + "<" + getClass().getName() + ">\n");
		buf.append(Utils.makeIndent(indent + 1) + "<variables> " + variables + " </variables>\n");
		buf.append(Utils.makeIndent(indent + 1) + "<sources> " + getSources() + " </sources>\n");
		buf.append(Utils.makeIndent(indent + 1) + "<isLet> " + islet + " </isLet>\n");
		buf.append(Utils.makeIndent(indent + 1) + "<monosource> " + isMonoSource() + " </monosource>\n");
		buf.append(Utils.makeIndent(indent + 1) + "<hasIdentifier> " + hasIdentifier + " </hasIdentifier>\n");
		if (hinttree != null) {
			buf.append(Utils.makeIndent(indent + 1) + "<HintTree>\n");
			buf.append(Utils.makeIndent(indent + 2) + hinttree.toString() + "\n");
			buf.append(Utils.makeIndent(indent + 1) + "</HintTree>\n");
		}
		buf.append(Utils.makeIndent(indent + 1) + "<Expression type=\"" + expression.getClass() + "\">\n");
		buf.append(Utils.makeIndent(indent + 2) + expression + "\n");
		buf.append(Utils.makeIndent(indent + 1) + "</Expression>\n");
		if (algChildren != null && !algChildren.isEmpty()) {
			buf.append(Utils.makeIndent(indent + 1) + "<Leaves>\n");
			for (int i = 0; i < algChildren.size(); i++) {
				buf.append(Utils.makeIndent(indent + 2) + "<Leave>\n");
				Algebra depchild = (Algebra) algChildren.get(i);
				buf.append(depchild.toCompleteString(indent + 3));
				buf.append(Utils.makeIndent(indent + 2) + "</Leave>\n");
			}
			buf.append(Utils.makeIndent(indent + 1) + "</Leaves>\n");
		}
		if (referenceDefinitionVars != null && !referenceDefinitionVars.isEmpty()) {
			buf.append(Utils.makeIndent(indent + 1) + "<referenceDefinitionVars>\n");
			for (int i = 0; i < referenceDefinitionVars.size(); i++) { //				DepVar depvar = (DepVar) varsreferencedby.get(i);
				//				buf.append(Utils.makeIndent(indent + 2) + "<Variables>" + depvar.getVariables() + "</Variables>\n");
				buf.append(Utils.makeIndent(indent + 2) + "<Variables>" + referenceDefinitionVars.get(i) + "</Variables>\n");
			}
			buf.append(Utils.makeIndent(indent + 1) + "</referenceDefinitionVars>\n");
		}
		if (dependDefinitionVars != null && !dependDefinitionVars.isEmpty()) {
			buf.append(Utils.makeIndent(indent + 1) + "<dependDefinitionVars>\n");
			for (int i = 0; i < dependDefinitionVars.size(); i++) { //				DepVar depvar = (DepVar) varsdependingon.get(i);
				//				buf.append(Utils.makeIndent(indent + 2) + "<Variables>" + depvar.getVariables() + "</Variables>\n");
				buf.append(Utils.makeIndent(indent + 2) + "<Variables>" + dependDefinitionVars.get(i) + "</Variables>\n");
			}
			buf.append(Utils.makeIndent(indent + 1) + "</dependDefinitionVars>\n");
		}
		if (paths != null && !paths.isEmpty()) {
			buf.append(Utils.makeIndent(indent + 1) + "<Paths>\n");
			for (int i = 0; i < paths.size(); i++)
				buf.append(Utils.makeIndent(indent + 2) + "<Path>" + paths.get(i) + "</Path>\n");
			buf.append(Utils.makeIndent(indent + 1) + "</Paths>\n");
		}
		if (subDepFLWRs != null) {
			buf.append(Utils.makeIndent(indent + 1) + "<subDepFLWRs>\n");
			for (int i = 0; i < subDepFLWRs.size(); i++)
				buf.append(Utils.makeIndent(indent + 2) + "<depmem>" + ((AlgFLWR) subDepFLWRs.get(i)).getVariables() + "</depmem>\n");
			buf.append(Utils.makeIndent(indent + 1) + "</subDepFLWRs>\n");
		}
		buf.append(Utils.makeIndent(indent) + "</" + getClass().getName() + ">\n");
		return buf.toString();
	}

	public void execute(ExecutionPlan plan) throws MediatorException {
	}

}
