/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator.algebra;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Set;

import org.xquark.mediator.MediatorException;
import org.xquark.mediator.decomposer.GetUsedVariablesVisitor;
import org.xquark.mediator.decomposer.Utils;
import org.xquark.mediator.plan.ExecutionPlan;
import org.xquark.mediator.plan.Operator;
import org.xquark.xquery.parser.Variable;
import org.xquark.xquery.parser.XQueryExpression;

/**
 * This abstract class represent a NODE of the dependancy tree.
 * It will be extended either by DepMem either by DepNodeSource.
 *
 */
public abstract class Algebra extends Object implements Cloneable {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.4 $";
	private static final String RCSName = "$Name:  $";
	// **********************************************************************
	// * CLASS VARIABLES
	// **********************************************************************
	// variable associated to node
	// protected Variable var = null ;
	// variables associated to node
	protected ArrayList variables = null;
	// expression associated to variable
	protected XQueryExpression expression = null;
	// all paths associated to variable
	protected ArrayList paths = null;
	// type of variable binding
	protected boolean islet = false;
	// list of variables contained in variable definition
	protected ArrayList dependDefinitionVars = null;
	// list of variables contained in variable definition
	protected ArrayList dependWhereVars = null;
	// list of variables using this variable in their definition
	protected ArrayList referenceDefinitionVars = null;
	// list of child nodes
	protected ArrayList algChildren = null;
	// parent depnode
	protected Algebra parent = null;
	// stuff
	protected AlgebraManager algManager = null;
	//protected TypeVisitor typevisitor = null ;
	// array list of expression to be ordered
	//protected ArrayList orderbys = null;
	// tells whether the node is dependent on other node values
//	protected boolean valueDepends = false;
//	protected boolean whereDepends = false;
	// id size
	protected int idsize = -1;
	// tells whether the node is bound on id variable
	protected boolean hasIdentifier = false;

	// ******************************************************************
	// * INITIALIZATION
	// ******************************************************************
	/**
	 *
	 * @param typevisitor
	 * @param expression
	 * @param depmanager
	 */

	public Algebra(XQueryExpression expression, ArrayList variables, AlgebraManager depmanager, boolean islet) {
		this.expression = expression;
		setVariables(variables);
		//this.typevisitor = typevisitor ;
		this.islet = islet;
		this.paths = null;
		this.dependDefinitionVars = new ArrayList(0);
		this.dependWhereVars = new ArrayList(0);
		this.algChildren = null;
		this.referenceDefinitionVars = new ArrayList(0);
		this.algManager = depmanager;
		// verify identifier
		for (int i = 0; i < variables.size(); i++)
			if (depmanager.getIdVars().contains((Variable) variables.get(i))) {
				this.hasIdentifier = true;
				break;
			}
	}

	public Algebra(XQueryExpression expression, Variable var, AlgebraManager depmanager, boolean islet) {
		this.expression = expression;
		//this.var = var ;
		ArrayList varList = new ArrayList(0);
		if (var != null) varList.add(var);
		setVariables(varList);
		//this.typevisitor = typevisitor ;
		this.islet = islet;
		this.paths = null;
		this.dependDefinitionVars = new ArrayList(0);
		this.dependWhereVars = new ArrayList(0);
		this.algChildren = null;
		this.referenceDefinitionVars = new ArrayList(0);
		this.algManager = depmanager;
		// verify identifier
		if (depmanager.getIdVars().contains(var))
			this.hasIdentifier = true;
	}

	// ******************************************************************
	// * ATOMIZE
	// ******************************************************************
	/**
	 *
	 * @return
	 */
	// public abstract ArrayList atomize () throws MediatorException ;

	// ******************************************************************
	// * INFORMATION
	// ******************************************************************
	/**
	 * Identify dependencies of variables
	 */
	//public abstract void makeInformation () throws MediatorException ;

	// ******************************************************************
	// * METHODS
	// ******************************************************************

	public boolean isMonoSource() {
		//return this.parent.isMonoSource();
		int size = (expression.getSourceNames() == null || expression.getSourceNames().isEmpty())?0:expression.getSourceNames().size();
		size += (expression.getUrls() == null || expression.getUrls().isEmpty())?0:expression.getUrls().size();
		if (size == 1)
			return true;
		return false;
	}

	public Variable getVar(int i) {
		if (variables == null)
			return null; /** @todo  Should a Exception be throwed ? */
		return (Variable) variables.get(i);
	}
	public ArrayList getVariables() {
		return variables;
	}
	public void setVariables(ArrayList variables) {
		this.variables = variables;
		//		if (null != variables) {
		//			sources.clear();
		//			for (int i = 0; i < variables.size(); i++) {
		//				sources.addAll(((Variable)variables.get(i)).getSources());
		//			}
		//		}
	}
	public void setLet(boolean islet) {
		this.islet = islet;
	}
	public boolean isLet() {
		return islet;
	}
	public boolean hasIdentifier() {
		return hasIdentifier;
	}
	public boolean isRootQDB() {
		return false;
	}
	public boolean isQDB() {
		return false;
	}
	public XQueryExpression getExpression() {
		return expression;
	}
	public void setExpression(XQueryExpression expression) {
		this.expression = expression;
	}

	public int getIdSize() {
		if (idsize != -1)
			return idsize;
		idsize = 0;
		if (algChildren != null)
			for (int i = 0; i < algChildren.size(); i++) {
				if (algChildren.get(i) instanceof AlgSource)
					idsize++;
			} else if (this instanceof AlgSource)
			idsize++;
		if (isQDB()) {
			AlgFLWR depmem = (AlgFLWR) this;
			ArrayList subDepFLWRs = depmem.getDependencyList();
			if (subDepFLWRs != null)
				for (int i = 0; i < subDepFLWRs.size(); i++)
					idsize += ((AlgFLWR) subDepFLWRs.get(i)).getIdSize();
		}
		return idsize;
	}

	private void addReferenceVariable(Variable referencevar) {
		referenceDefinitionVars.add(referencevar);
	}
	private void addReferenceVariables(ArrayList referencevars) {
		referenceDefinitionVars.addAll(referencevars);
	}
	public void addDependSourceVariable(Variable dependvar) {
		dependDefinitionVars.add(dependvar);
		algManager.getMapNode(dependvar).addReferenceVariables(this.getVariables());
	}
	public void addDependWhereVariables(ArrayList dependvars) {
		if (dependvars == null || dependvars.isEmpty()) return;
		dependWhereVars.addAll(dependvars);
	}
	public void addDependSourceVariables(ArrayList dependvars) {
		if (dependvars == null || dependvars.isEmpty()) return;
		dependDefinitionVars.addAll(dependvars);
		for (int i = 0; i < dependvars.size(); i++) {
			Variable tmpvar = (Variable) dependvars.get(i);
			algManager.getMapNode(tmpvar).addReferenceVariables(this.getVariables());
		}
	}
	
	public ArrayList getPaths() {
		return paths;
	}
	public void setPaths() {
		if (paths != null)
			paths.clear();
		if (algChildren != null) {
			for (int i = 0; i < algChildren.size(); i++) {
				Algebra depnode = (Algebra) algChildren.get(i);
				depnode.setPaths();
				if (paths == null)
					paths = new ArrayList();
				if (!(depnode instanceof AlgSource)) {
//					if (i == 0)
						paths.addAll(depnode.getVariables());
				} else if (depnode.getPaths() != null)
					paths.addAll(depnode.getPaths());
			}
		} else {
			for (int i = 0; i < this.variables.size(); i++) {
				Variable var = (Variable) variables.get(i);
				ArrayList tmplist = (ArrayList) algManager.getVarPaths().get(var);
				if (tmplist == null) {
					if (!this.isMonoSource()) {
						if (paths == null) {
							paths = new ArrayList(1);
							paths.add(var);
						} else
							paths.add(var);
					}
				} else {
					if (paths == null)
						paths = (ArrayList) tmplist.clone();
					else
						paths.addAll(tmplist);
				}
			}
		}
		if (isQDB()) {
			AlgFLWR depmem = (AlgFLWR) this;
			if (depmem.getDependencyList() != null)
				for (int i = 0; i < depmem.getDependencyList().size(); i++) {
					AlgFLWR tmpdepmem = (AlgFLWR) depmem.getDependencyList().get(i);
					tmpdepmem.setPaths();
					if (paths == null)
						paths = new ArrayList();
					if (tmpdepmem.getPaths() != null)
						paths.addAll(tmpdepmem.getPaths());
				}
		}
	}
	//public void setOrderBys(ArrayList orderbys) { this.orderbys = orderbys; }
	//public ArrayList getOrderBys() { return orderbys ; }

	public ArrayList getVarsWhereOn() {
		return this.dependWhereVars;
	}
	public ArrayList getVarsDependingOn() {
		return this.dependDefinitionVars;
	}
//	public ArrayList getAllVarsDependingOn() {
//		//if (varsreferencedby.isEmpty()) return varsdependingon;
//		ArrayList reslist = (ArrayList) dependDefinitionVars.clone();
//		for (int i = 0; i < dependDefinitionVars.size(); i++) {
//			DepNode tmpnode = this.depmanager.getMapNode((Variable) dependDefinitionVars.get(i));
//			ArrayList tmplist = tmpnode.getAllVarsDependingOn();
//			for (int j = 0; j < tmplist.size(); j++) {
//				Variable tmpvar = (Variable) tmplist.get(j);
//				if (!reslist.contains(tmpvar))
//					reslist.add(tmpvar);
//			}
//		}
//		return reslist;
//	}

	public ArrayList getVarsReferencedBy() {
		return this.referenceDefinitionVars;
	}
	//public void setVarsReferencedBy(ArrayList list) { this.varsreferencedby = list; }

	public Algebra getParent() {
		return this.parent;
	}
	public AlgFLWR getParentDepFLWR() {
		return parent.getParentDepFLWR();
		//		if (this instanceof DepMem)
		//			return (DepMem) this;
		//		if (parent instanceof DepMem)
		//			return (DepMem) parent;
		//		else
		//			return parent.getParentDepMem();
	}

	public ArrayList getChildren() {
		return this.algChildren;
	}
	public AlgebraManager getAlgebraManager() {
		return this.algManager;
	}

	public boolean getDepends() {
		return ((dependWhereVars != null && !dependWhereVars.isEmpty()) || (dependDefinitionVars != null && !dependDefinitionVars.isEmpty()));
	}
//	public void setWhereDepends(boolean whereDepends) {
//		this.whereDepends = whereDepends;
//	}
//	public boolean getWhereDepends() {
//		return whereDepends;
//	}
//	public void setValueDepends(boolean valueDepends) {
//		this.valueDepends = valueDepends;
//	}
//	public boolean getValueDepends() {
//		return valueDepends;
//	}
//	public boolean getDepends() {
//		return valueDepends || whereDepends;
//	}

	public void setParent(Algebra parent) throws MediatorException {
		this.parent = parent;
		if (!this.hasIdentifier && parent.hasIdentifier)
			ParentHasIdentifier();
	}
	
	public void ParentHasIdentifier() {}
	
	public void addLeaf(Algebra leaf) throws MediatorException {
		if (algChildren == null)
			algChildren = new ArrayList(1);
		algChildren.add(leaf);
		leaf.setParent(this);
	}

//	public AlgSource getFirstSourceLeafDepNode() {
//		if (algChildren == null) {
//			if (this instanceof AlgSource)
//				return (AlgSource) this;
//			else
//				return null;
//		}
//		return ((Algebra) algChildren.get(0)).getFirstSourceLeafDepNode();
//	}

	public Object clone() throws CloneNotSupportedException {
		Algebra newobj = (Algebra) super.clone();
		/*
		HashMap vars = new HashMap(1+this.varsdependingon.size());
		vars.put(this.var.toString(),this.var);
		for (int i=0;i<this.varsdependingon.size();i++) vars.put(((Variable)this.varsdependingon.get(i)).toString(),this.varsdependingon.get(i));;
		 */
		newobj.expression = (XQueryExpression) this.expression.clone();
		newobj.algManager = this.algManager;
		newobj.islet = this.islet;
		newobj.paths = this.paths;
		// newobj.var = this.var;
		newobj.variables = (ArrayList) this.variables.clone();
		newobj.dependDefinitionVars = this.dependDefinitionVars;
		newobj.referenceDefinitionVars = this.referenceDefinitionVars;
		newobj.algChildren = null;
		if (this.algChildren != null) {
			newobj.algChildren = new ArrayList();
			for (int i = 0; i < this.algChildren.size(); i++)
				newobj.algChildren.add((Algebra) ((Algebra) this.algChildren.get(i)).clone());
		}
		return newobj;
	}

	public Set getSources() {
		Set retVal = null;
		if (null != variables && !variables.isEmpty()) {
			retVal = new HashSet();
			Variable var = null;
			for (int i = 0; i < variables.size(); i++) {
				var = (Variable) variables.get(i);
				if (null != var.getSourceNames()) {
					retVal.addAll(var.getSourceNames());
				}
			}
		}
		return retVal;
	}

	public Set getUrls() {
		Set retVal = null;
		if (null != variables && !variables.isEmpty()) {
			retVal = new HashSet();
			Variable var = null;
			for (int i = 0; i < variables.size(); i++) {
				var = (Variable) variables.get(i);
				if (null != var.getUrls()) {
					retVal.addAll(var.getUrls());
				}
			}
		}
		return retVal;
	}

	protected GetUsedVariablesVisitor guvvisitor = null;
	
	protected void fillVarLists() {
		if (guvvisitor == null)
			guvvisitor = new GetUsedVariablesVisitor();
		guvvisitor.reset();
		try {
			expression.accept(guvvisitor);
		} catch (Exception e) {
			return;
		}
		//addDependWhereVariables(guvvisitor.getWhereVariables());
		addDependSourceVariables(guvvisitor.getVariables());
	}
	
	public String toCompleteString(int indent) {
		StringBuffer buf = new StringBuffer();
		// buf.append(Utils.makeIndent(indent) + "<" + this.getClass().getName() + " varname=\"" + var + "\" let=" + isLet() + ">\n") ;
		buf.append(Utils.makeIndent(indent) + "<" + this.getClass().getName() + ">\n");
		buf.append(Utils.makeIndent(indent + 1) + "<variables> " + variables + " </variables>\n");
		buf.append(Utils.makeIndent(indent + 1) + "<sources> " + getSources() + " </sources>\n");
		buf.append(Utils.makeIndent(indent + 1) + "<urls> " + getUrls() + " </urls>\n");
		buf.append(Utils.makeIndent(indent + 1) + "<isLet> " + islet + " </isLet>\n");
		buf.append(Utils.makeIndent(indent + 1) + "<monosource> " + isMonoSource() + " </monosource>\n");
		buf.append(Utils.makeIndent(indent + 1) + "<hasIdentifier> " + hasIdentifier + " </hasIdentifier>\n");
		buf.append(Utils.makeIndent(indent + 1) + "<Expression type=\"" + expression.getClass() + "\">\n");
		buf.append(Utils.makeIndent(indent + 2) + expression + "\n");
		buf.append(Utils.makeIndent(indent + 1) + "</Expression>\n");
		if (algChildren != null) {
			buf.append(Utils.makeIndent(indent + 1) + "<Leaves>\n");
			for (int i = 0; i < algChildren.size(); i++) {
				buf.append(Utils.makeIndent(indent + 2) + "<Leave>\n");
				Algebra depchild = (Algebra) algChildren.get(i);
				buf.append(depchild.toCompleteString(indent + 3));
				buf.append(Utils.makeIndent(indent + 2) + "</Leave>\n");
			}
			buf.append(Utils.makeIndent(indent + 1) + "</Leaves>\n");
		}
		if (referenceDefinitionVars != null && !referenceDefinitionVars.isEmpty()) {
			buf.append(Utils.makeIndent(indent + 1) + "<referenceDefinitionVars>\n");
			for (int i = 0; i < referenceDefinitionVars.size(); i++) {
				Variable var = (Variable) referenceDefinitionVars.get(i);
				buf.append(Utils.makeIndent(indent + 2) + "<Var>" + var + "</Var>\n");
			}
			buf.append(Utils.makeIndent(indent + 1) + "</referenceDefinitionVars>\n");
		}
		if (dependDefinitionVars != null && !dependDefinitionVars.isEmpty()) {
			buf.append(Utils.makeIndent(indent + 1) + "<dependDefinitionVars>\n");
			for (int i = 0; i < dependDefinitionVars.size(); i++) {
				Variable var = (Variable) dependDefinitionVars.get(i);
				buf.append(Utils.makeIndent(indent + 2) + "<Var>" + var + "</Var>\n");
			}
			buf.append(Utils.makeIndent(indent + 1) + "</dependDefinitionVars>\n");
		}
		if (dependWhereVars != null && !dependWhereVars.isEmpty()) {
			buf.append(Utils.makeIndent(indent + 1) + "<dependWhereVars>\n");
			for (int i = 0; i < dependWhereVars.size(); i++) {
				Variable var = (Variable) dependWhereVars.get(i);
				buf.append(Utils.makeIndent(indent + 2) + "<Var>" + var + "</Var>\n");
			}
			buf.append(Utils.makeIndent(indent + 1) + "</dependWhereVars>\n");
		}
		if (paths != null && !paths.isEmpty()) {
			buf.append(Utils.makeIndent(indent + 1) + "<Paths>\n");
			for (int i = 0; i < paths.size(); i++)
				buf.append(Utils.makeIndent(indent + 2) + "<Path>" + paths.get(i) + "</Path>\n");
			buf.append(Utils.makeIndent(indent + 1) + "</Paths>\n");
		}
		buf.append(Utils.makeIndent(indent) + "</" + this.getClass().getName() + ">\n");
		return buf.toString();
	}

	public String toCompleteString() {
		return toCompleteString(0);
	}

	public abstract Operator createOperator(ExecutionPlan plan) throws MediatorException;

}
