/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator.algebra;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.xquark.mediator.MediatorException;
import org.xquark.mediator.decomposer.AlgebraStructureVisitor;
import org.xquark.mediator.decomposer.GetVariablePathsVisitor;
import org.xquark.mediator.plan.ExecutionPlan;
import org.xquark.mediator.plan.OpSubQuery;
import org.xquark.mediator.plan.Operator;
import org.xquark.xquery.metadata.MetaDataImpl;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.parser.util.Constants;

/**
 * This class is designed to handle dependancies and reduction
 * of requests. It identifies sources and is the pre-transformation before
 * generating the execution plan.
 * <p>The structure is as follows at initialization :</p>
 * <ul>
 *    <li> There is an array of DepNodes where each DepNode is associated to
 *      one QDB.</li>
 *    <li> for each DepNode, there is a DepLeaf (that can be a DepVar or
 *      a DepSource) that is associated to the variables of the
 *      FLWR expression of the QDB.</li>
 * </ul>
 *
 */
public class AlgebraManager {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.4 $";
	private static final String RCSName = "$Name:  $";

	// **********************************************************************
	// * CLASS VARIABLES
	// **********************************************************************

	private ArrayList depnodes = null;
	private TypeVisitor typevisitor = null;
	private ExecutionPlan plan = null;
	// make correspondance on variable and DepNode that handle it
	private HashMap mapnode = null;
	private XQueryExpression qmem = null;
	private boolean isQmemRoot = false;
	private ArrayList mainvars = null;
	private ArrayList idvars = null;
	private HashMap mappath = null;
	private ArrayList pathlist = null;
	private MetaDataImpl metadata = null;
	//protected HashMap aliases = null;
	//protected HashMap dependvarmap = null;
	private AlgFLWR currentDepFLWR = null;
	private GetVariablePathsVisitor gvpvisitor = null;
	private ArrayList scopeVariables = new ArrayList();

	// ******************************************************************
	// * INITIALISATION
	// ******************************************************************
	/**
	 *
	 * @param metadata
	 * @param typevisitor
	 * @param qmem
	 */
	public AlgebraManager(ExecutionPlan plan, MetaDataImpl metadata, TypeVisitor typevisitor, XQueryExpression qmem, ArrayList mainvars, ArrayList idvars) throws MediatorException, Exception {
		this.typevisitor = typevisitor;
		this.metadata = metadata;
		this.depnodes = new ArrayList();
		this.mapnode = new HashMap();
		this.plan = plan;
		this.qmem = qmem;
		this.isQmemRoot = qmem.getRoot();
		this.mainvars = mainvars;
		this.idvars = idvars;
		mappath = new HashMap();
		pathlist = new ArrayList();
	}

	// ************************************************************************
	// * GET/SET METHODS
	// ************************************************************************
	public MetaDataImpl getMetaData() {
		return metadata;
	}
	//public XQueryExpression getQMem() { return qmem ; }
	//public ArrayList getNodes() { return depnodes ; }
	public TypeVisitor getTypeVisitor() {
		return typevisitor;
	}
	public ArrayList getIdVars() {
		return idvars;
	}
	public HashMap getVarPaths() {
		return mappath;
	}
	public ExecutionPlan getPlan() {
		return plan;
	}

	public void addVarsToScope(List vars) {
		if (vars == null)
			return;
		for (int i = 0; i < vars.size(); i++)
			addVarToScope((Variable) vars.get(i));
	}

	public void removeVarsFromScope(List vars) {
		scopeVariables.removeAll(vars);
	}

	public void addVarToScope(Variable var) {
		if (!this.scopeVariables.contains(var))
			scopeVariables.add(var);
	}

	public void removeVarFromScope(Variable var) {
		scopeVariables.remove(var);
	}

	public ArrayList getScopeVariables() {
		return scopeVariables;
	}
	// ************************************************************************
	// * CONSTRUCT THE DEPENDENCY TREE (accessed by MediatorStatement)
	// ************************************************************************

	/** start <code>AlgebraStructureVisitor</code> to visit all the QDBs
	 */
	public void generate() throws MediatorException {
		try {
			AlgebraStructureVisitor asv = new AlgebraStructureVisitor(this);
			for (int k = 0; k < mainvars.size(); k++)
				 ((Variable) mainvars.get(k)).accept(asv);
		} catch (Exception e) {
			e.printStackTrace();
			throw new MediatorException("Could not build structure to construct algebra", e);
		}
	}
	/** to be commented ... : fills mappath , modify depnodes */

	public void computeVarPaths() throws MediatorException, XQueryException {
		org.xquark.util.misc.Debug.print("mappath = " + mappath);
		if (gvpvisitor == null)
			gvpvisitor = new GetVariablePathsVisitor(true, this.mappath, this.pathlist);
		/*
		for (int i = 0; i < depnodes.size(); i++) {
			DepNode depnode = (DepNode) depnodes.get(i);
			if (depnode instanceof DepNodeSource) {
				org.xquark.xml.misc.Debug.print("depnode = " + depnode.toCompleteString());
				if (depnode.getParent() instanceof DepFLWR && ((DepFLWR) depnode.getParent()).isQDB())
					gvpvisitor.setRoot(true);
				else
					gvpvisitor.setRoot(false);
				gvpvisitor.setVarList(depnode.getVariables());
				org.xquark.xml.misc.Debug.print("variables = " + depnode.getVariables());
				org.xquark.xml.misc.Debug.print("expression = " + depnode.getExpression());
				depnode.getExpression().accept(gvpvisitor);
				org.xquark.xml.misc.Debug.print("mappath = " + gvpvisitor.getVarPaths());
				org.xquark.xml.misc.Debug.print("pathlist = " + gvpvisitor.getPathStrList());
			}
			else if (!(depnode instanceof DepFLWR)) {
				org.xquark.xml.misc.Debug.print("depnode = " + depnode.toCompleteString());
				if (depnode.getParent() instanceof DepFLWR && ((DepFLWR) depnode.getParent()).isQDB())
					gvpvisitor.setRoot(true);
				else
					gvpvisitor.setRoot(false);
				gvpvisitor.setVarList(depnode.getVariables());
				org.xquark.xml.misc.Debug.print("variables = " + depnode.getVariables());
				org.xquark.xml.misc.Debug.print("expression = " + depnode.getExpression());
				depnode.getExpression().accept(gvpvisitor);
				org.xquark.xml.misc.Debug.print("mappath = " + gvpvisitor.getVarPaths());
				org.xquark.xml.misc.Debug.print("pathlist = " + gvpvisitor.getPathStrList());
			}
		}
		*/

		for (int i = 0; i < depnodes.size(); i++) {
			Algebra depnode = (Algebra) depnodes.get(i);
			if (depnode instanceof AlgFLWR) {
				if (((AlgFLWR) depnode).isQDB()) {
					org.xquark.util.misc.Debug.print("depnode = " + depnode.toCompleteString());
					gvpvisitor.setRootQDB(depnode.isRootQDB());
					gvpvisitor.setQDB(depnode.isQDB());
					gvpvisitor.setMonoSource(depnode.isMonoSource());
					if (depnode.isMonoSource())
						gvpvisitor.setCurrentList(((FLWRExpression) depnode.getExpression()).getVariables());
					else
						gvpvisitor.resetCurrentVarList();
					//gvpvisitor.addVarList(((FLWRExpression)depnode.getExpression()).getVariables());
					org.xquark.util.misc.Debug.print("variables = " + depnode.getVariables());
					org.xquark.util.misc.Debug.print("expression = " + depnode.getExpression());
					depnode.getExpression().accept(gvpvisitor);
					org.xquark.util.misc.Debug.print("mappath = " + gvpvisitor.getVarPaths());
					org.xquark.util.misc.Debug.print("pathlist = " + gvpvisitor.getPathStrList());
				}
			}
		}

		//gvpvisitor.resetVarList();
		gvpvisitor.setRootQDB(false);
		gvpvisitor.setQDB(false);
		gvpvisitor.setMonoSource(false);
		gvpvisitor.resetCurrentVarList();
		qmem.accept(gvpvisitor);
		mappath = gvpvisitor.getVarPaths();
		pathlist = gvpvisitor.getPathStrList();
		org.xquark.util.misc.Debug.print("mappath = " + mappath);
		org.xquark.util.misc.Debug.print("pathlist = " + pathlist);
		//		Map map = gvpvisitor.getAliasMap();
		//		Set set = map.entrySet();
		//		for (Iterator it = set.iterator();it.hasNext();) {
		//			Entry entry = (Entry)it.next();
		//			plan.makeAlias((Variable)entry.getKey(),(XQueryExpression)entry.getValue());
		//		}

		for (int i = 0; i < depnodes.size(); i++) {
			Algebra depnode = (Algebra) depnodes.get(i);
			if (depnode instanceof AlgFLWR && ((AlgFLWR) depnode).isRootQDB())
				depnode.setPaths();
		}
	}

	/**
	 * Add a new QDB in the dependency tree. (create a AlgMem)
	 *
	 * @param expression
	 * @param var
	 */
	public AlgFLWR addQDB(Variable var, boolean islet) throws MediatorException {
		return addQDB(var.getExpression(), var, islet);
	}
	public AlgFLWR addQDB(XQueryExpression expression, Variable var, boolean islet) throws MediatorException {
		currentDepFLWR = new AlgFLWR(expression, var, this, islet);
		depnodes.add(currentDepFLWR);
		addMapNode(var, currentDepFLWR);
		return currentDepFLWR;
	}

	/**
	 * Add a new Variable in the dependency tree. (create a AlgVar)
	 *
	 * @param expression
	 * @param var
	 */
	public AlgVar addVar(Variable addvar, Variable var, boolean islet) throws MediatorException {
		AlgVar depvar = new AlgVar(addvar, var, this, islet);
		depnodes.add(depvar);
		addMapNode(var, depvar);
		return depvar;
	}

	/**
	 * Add a new Source in the dependency tree. (create a AlgSource)
	 *
	 * @param expression
	 * @param var
	 */
	public AlgSource addSource(XQueryExpression expression, Variable var, boolean islet) throws MediatorException {
		ArrayList variables = null;
		if (var != null) {
			variables = new ArrayList(1);
			variables.add(var);
		}
		if (var == null)
			return addSource(expression, variables, islet);
		return addSource(var.getParentFLWRExpression(), variables, islet);
	}
	public AlgSource addSource(XQueryExpression expression, ArrayList variables, boolean islet) throws MediatorException {
		AlgSource source = null;
		Variable var = null;
		if (variables == null || variables.isEmpty()) {
			try {
				var = this.createVariable(Constants.FOR_BINDINGTYPE, expression, this.typevisitor);
				ArrayList vars = new ArrayList(1);
				vars.add(var);
				mappath.put(var, vars);
				pathlist.add(var.toString());
				FLWRExpression tmpFlwr = new FLWRExpression(vars, null, var, null);
				tmpFlwr.setParentUnit(expression.getParentUnit());
				source = new AlgSource(tmpFlwr, var, this, islet);
			} catch (Exception e) {
				throw new MediatorException("Could not create variable", e);
			}
		} else
			source = new AlgSource(expression, variables, this, islet);
		depnodes.add(source);
		if (variables == null || variables.isEmpty())
			addMapNode(var, source);
		else
			for (int i = 0; i < variables.size(); i++)
				addMapNode((Variable) variables.get(i), source);
		return source;
	}

	/**
	 * Add a new Eval in the dependency tree. (create a AlgEval)
	 *
	 * @param expression
	 * @param var
	 */
	public AlgSourceEval addSourceEval(XQueryExpression expression, Variable var, boolean islet) throws MediatorException {
		ArrayList variables = null;
		if (var != null) {
			variables = new ArrayList(1);
			variables.add(var);
		}
		if (var == null)
			return addSourceEval(expression, variables, islet);
		return addSourceEval(var.getParentFLWRExpression(), variables, islet);
	}
	public AlgSourceEval addSourceEval(XQueryExpression expression, ArrayList variables, boolean islet) throws MediatorException {
		AlgSourceEval source = null;
		Variable var = null;
		if (variables == null || variables.isEmpty()) {
			try {
				var = this.createVariable(Constants.FOR_BINDINGTYPE, expression, this.typevisitor);
				ArrayList vars = new ArrayList(1);
				vars.add(var);
				mappath.put(var, vars);
				pathlist.add(var.toString());
				FLWRExpression tmpFlwr = new FLWRExpression(vars, null, var, null);
				tmpFlwr.setParentUnit(expression.getParentUnit());
				source = new AlgSourceEval(tmpFlwr, var, this, islet);
			} catch (Exception e) {
				throw new MediatorException("Could not create variable", e);
			}
		} else
			source = new AlgSourceEval(expression, variables, this, islet);
		depnodes.add(source);
		if (variables == null || variables.isEmpty())
			addMapNode(var, source);
		else
			for (int i = 0; i < variables.size(); i++)
				addMapNode((Variable) variables.get(i), source);
		return source;
	}

	/**
	 * Add a new Sequence in the dependency tree. (create a AlgSequence)
	 *
	 * @param expression
	 * @param var
	 */
	public AlgSequence addSequence(XQueryExpressionSequence expression, Variable var, boolean islet) throws MediatorException {
		AlgSequence depsequence = new AlgSequence(expression, var, this, islet);
		//depfunction.generate(metadata) ;
		depnodes.add(depsequence);
		addMapNode(var, depsequence);
		return depsequence;
	}

	/**
	 * Add a new union in the dependency tree. (create a AlgUnion)
	 *
	 * @param expression
	 * @param var
	 */
	public AlgUnion addUnion(ListOpUNIONExpression expression, Variable var, boolean islet) throws MediatorException {
		AlgUnion depunion = new AlgUnion(expression, var, this, islet);
		//depfunction.generate(metadata) ;
		depnodes.add(depunion);
		addMapNode(var, depunion);
		return depunion;
	}

	/**
	 * Add a new Function-Source in the dependency tree. (create a AlgFunction)
	 *
	 * @param expression
	 * @param var
	 */
	public AlgFunction addFunction(XQueryExpression expression, Variable var, boolean islet) throws MediatorException {
		AlgFunction depfunction = new AlgFunction(expression, var, this, islet);
		//depfunction.generate(metadata) ;
		depnodes.add(depfunction);
		addMapNode(var, depfunction);
		return depfunction;
	}

	/**
	 * Add a new Arith Function-Source in the dependency tree. (create a AlgArithFunction)
	 *
	 * @param expression
	 * @param var
	 */
	public AlgArithFunction addArithFunction(ListOpArithExpression expression, Variable var, boolean islet) throws MediatorException {
		AlgArithFunction deparithfunction = new AlgArithFunction(expression, var, this, islet);
		//depfunction.generate(metadata) ;
		depnodes.add(deparithfunction);
		addMapNode(var, deparithfunction);
		return deparithfunction;
	}

	/**
	 * Add a new Value in the dependency tree. (create a AlgValue)
	 *
	 * @param expression
	 * @param var
	 */
	public AlgValue addValue(XQueryExpression expression, Variable var, boolean islet) throws MediatorException { // fillMapPath(expression) ;
		AlgValue depvalue = new AlgValue(expression, var, this, islet);
		depnodes.add(depvalue);
		addMapNode(var, depvalue);
		return depvalue;
	}

	/**
	 * Add a new Value in the dependency tree. (create a OpValue)
	 *
	 * @param expression
	 * @param var
	 */
	public AlgEval addEval(XQueryExpression expression, Variable var, boolean islet) throws MediatorException { // fillMapPath(expression) ;
		AlgEval depeval = new AlgEval(expression, var, this, islet);
		depnodes.add(depeval);
		addMapNode(var, depeval);
		return depeval;
	}

	/**
	 * Add a new Value in the dependency tree. (create a AlgFLWR)
	 *
	 * @param expression
	 * @param var
	 */
	public AlgFLWR addFlwr(XQueryExpression expression, Variable var, boolean islet) throws MediatorException {
		AlgFLWR depflwr = new AlgFLWR(expression, var, this, islet);
		depnodes.add(depflwr);
		addMapNode(var, depflwr);
		return depflwr;
	}

	/**
	 * Add a new Value in the dependency tree. (create a AlgValue)
	 *
	 * @param expression
	 * @param var
	 */
	public AlgLocated addLocated(LocatedExpression expression, Variable addvar, Variable var, boolean islet) throws MediatorException {
		AlgLocated deplocated = new AlgLocated(expression, addvar, var, this, islet);
		depnodes.add(deplocated);
		addMapNode(var, deplocated);
		return deplocated;
	}

	/**
	 *
	 * @param variable
	 * @param node
	 */
	private void addMapNode(Variable variable, Algebra node) throws MediatorException {
		//System.out.println ("PUTTING " + variable + " node " + node) ;
		if (variable != null)
			mapnode.put(variable, node);
	}

	/**
	 *
	 * @param variable
	 */
	public Algebra getMapNode(Variable variable) {
		Algebra newnode = (Algebra) mapnode.get(variable);
		//if (newnode == null) throw new MediatorException("DepManager.getMapNode:  no matching node to " + variable) ;
		return newnode;
	}

	/**
	 *
	 * @param var
	 * @return
	 */
	public ArrayList getPathsfromVar(Variable var) {
		return (ArrayList) mappath.get(var);
	}

	public ArrayList getPathsList() {
		return this.pathlist;
	}

	/**
	 *
	 * @param var
	 * @return
	 */
	public GetVariablePathsVisitor getGVPVisitor() {
		return gvpvisitor;
	}

	// ******************************************************************
	// * OPTIMISATIONS
	// ******************************************************************
	/**
	 * Optimising by regrouping by sources if useful.
	 */
	/*
	public void optimize() {
	    // TODO start optimising by regrouping by sources if useful.
	}
	*/
	// ******************************************************************
	// MAKING EXECUTION PLAN
	// ******************************************************************
	/**
	 * The execution plan is done by identifying the root node(s).
	 * The whole tree come with it (with help of variable referencedBy
	 * defined in DepNode).
	 * @param connection
	 * @return
	 */

	public ExecutionPlan FillPlan() throws MediatorException {
		ArrayList roots = new ArrayList();
		for (int i = 0; i < depnodes.size(); i++) {
			Algebra depnode = (Algebra) depnodes.get(i);
			if (depnode instanceof AlgFLWR && ((AlgFLWR) depnode).isRootQDB()) {
				Operator tmpalgebra = (Operator) depnode.createOperator(plan);
				if (tmpalgebra != null)
					if (this.isQmemRoot)
						roots.add(tmpalgebra);
					else
						plan.setRoot(tmpalgebra);
			}
		}
		if (roots.size() == 1)
			plan.setRoot((Operator) roots.get(0));
		else if (roots.size() > 1)
			plan.setRoot(new OpSubQuery(plan, null, roots, false, true));
		return plan;
	}

	private int varIndex = 0;
	private Variable createVariable(int bindingType, XQueryExpression expr, TypeVisitor typeVisitor) throws XQueryException {
		if (typeVisitor != null && expr != null)
			expr.accept(typeVisitor);
		Variable var = new Variable(new QName(null, null, "var_depman" + (++varIndex), typeVisitor), typeVisitor);
		var.setBindingType(bindingType);
		var.setExpression(expr, typeVisitor);
		return var;
	}

	// ******************************************************************
	// * DEBUGGING
	// ******************************************************************
	/**
	 *
	 * @return
	 */
	public String toString() {
		StringBuffer buf = new StringBuffer();
		buf.append("<DepManager>\n");
		for (int i = 0; i < mainvars.size(); i++) {
			Algebra depnode = (Algebra) mapnode.get(mainvars.get(i));
			if (depnode instanceof AlgFLWR && ((AlgFLWR) depnode).isQDB())
				buf.append(depnode.toCompleteString(1));
		}
		//		for (int i = 0; i < depnodes.size(); i++) {
		//			DepNode depnode = (DepNode) depnodes.get(i);
		//			if (depnode instanceof DepFLWR && ((DepFLWR) depnode).isQDB())
		//				buf.append(depnode.toCompleteString(1));
		//		}
		buf.append("</DepManager>\n");
		return buf.toString();
	}
}
