/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator.plan ;

import java.util.ArrayList;

import org.xquark.mediator.MediatorException;
import org.xquark.xquery.parser.XQueryExpression;


/**
 * This class is a binary node that is a component of the execution tree.
 * Execution can be serialized or parallelized.
 *
 */
public abstract class BinOp extends MultOp implements Runnable {
    // **********************************************************************
    // * VERSIONING
    // **********************************************************************
    private static final String RCSRevision = "$Revision: 1.1 $";
    private static final String RCSName = "$Name:  $";
    // **********************************************************************
    // * CLASS VARIABLES
    // **********************************************************************
        
    // ***********************************************************************
    // * INITIALIZATION
    // ***********************************************************************
    /**
     * Construct an operator of the execution plan handling deux childs
     * expressions. This method must be overrided by descendant, but a
     * super (...) on this should always be called first by it's descendant
     * constructor in order to avoid problems.
     *
     * @param plan a reference on the general execution plan.
     * @param expression the expression handled by _this_ operator.
     * @param leftexp a reference on the left child operator.
     * @param rightexp a refrence on the right child operator.
     *
     * @param plan
     * @param expression
     * @param leftexp
     * @param rightexp
     * @return
     */
    public BinOp(ExecutionPlan plan, XQueryExpression expression, Operator left, Operator right) throws MediatorException {
        super(plan, expression, left, right) ;
    }
    
    public BinOp(ExecutionPlan plan, XQueryExpression expression, ArrayList algebralist) throws MediatorException {
        super(plan, expression,algebralist) ;
        if (algebralist.size() != 2) throw new MediatorException("AlgBinOp must have two elements");
    }
    
	// #############################################################################
	// VISITOR STUFF
	// #############################################################################

	public void accept(OperatorVisitor visitor) throws MediatorException {
		visitor.visit(this);
	}

    // ***********************************************************************
    // * EXECUTE QUERY
    // ***********************************************************************
    /**
     * Start the execution, as a binary node, it start the execution
     * of the two children it refers. The execution is parallelized or
     * serialized depending the value of 'ordertype' field that can be
     * ORDER_SERIALIZATION or ORDER_PARALLELIZATION.
     */
    //public void executeQuery() throws MediatorException {
    
    // ***********************************************************************
    // * GET/SET METHODS
    // ***********************************************************************
    /**
     * Return a reference on left child expression.
     *
     * @return a reference on left child expression.
     */
    public Operator getLeftAlgebra() { return getSubElement(0); }
    
    /**
     * Return a reference on right child expression.
     *
     * @return a reference on right child expression.
     */
    public Operator getRightAlgebra() { return getSubElement(1); }
    
    // **********************************************************************
    // * OPTIMIZE
    // **********************************************************************
    /**
     * If the *exactly* same set of source is used for an operation,
     * then the subtree can be passed entierely to the source.
     * Beware of processing this from the leaf to the node.
     *
     * <p>Note: If this method is not overloaded or if it returns null,
     * it means that _this_ node will not be optimized.</p>
     * @return
     */
    //public Algebra optimizeSource() throws MediatorException {
    
    /**
     * Add projections on attributes the soonest as possible in order to
     * manipulate the less data as possible. As soon an attribute isn't
     * necessary, it must be unselected (by a projection on the complementary
     * variables).
     *
     * <p>Note: If this method is not overloaded or if it returns null,
     * it means that _this_ node will not be optimized.</p>
     * @return
     */
    //public Algebra optimizeProjection() throws MediatorException {
    
    /**
     *
     *
     * <p>Note: If this method is not overloaded or if it returns null,
     * it means that _this_ node will not be optimized.</p>
     * @return
     */
    //public Algebra optimizeRestriction() throws MediatorException {
    
    /**
     * If the set of sources handled by this node are exactly the same (order
     * not important) then return true.
     * An AlgBinValue is true if both of its child are true and IF the
     * set of sources handled by its children are exactly the same.
     * @return
     */
    //protected boolean shareSameSources() {
    
    /**
     * Compare set of sources.
     * @param algebra1
     * @param algebra2
     * @return
     */
    //private boolean compareSource(Algebra algebra1, Algebra algebra2) {
    
    /**
     * Return true if the node is of type source.
     * AlgBinOp is not a source.
     *
     * @return false ;
     */
    //public boolean isSource() {
    
    /**
     * Return all the sources depending of this node.
     * @return
     */
    //public ArrayList getSources() {
}
