/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator.plan;

import java.util.ArrayList;

import org.xquark.mediator.MediatorException;
import org.xquark.mediator.decomposer.Utils;
import org.xquark.xquery.parser.XQueryExpression;

/**
 * This class is a binary node that is a component of the execution tree.
 * Execution can be serialized or parallelized.
 *
 */
public abstract class MultOp extends Operator {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.1 $";
	private static final String RCSName = "$Name:  $";

	// **********************************************************************
	// * CLASS VARIABLES
	// **********************************************************************

	protected ArrayList algebralist = null;
	protected int ordertype = ORDER_SERIALIZATION;

	// ***********************************************************************
	// * INITIALIZATION
	// ***********************************************************************
	/**
	 * Construct an operator of the execution plan handling deux childs
	 * expressions. This method must be overrided by descendant, but a
	 * super (...) on this should always be called first by it's descendant
	 * constructor in order to avoid problems.
	 *
	 * @param plan a reference on the general execution plan.
	 * @param expression the expression handled by _this_ operator.
	 * @param leftexp a reference on the left child operator.
	 * @param rightexp a refrence on the right child operator.
	 *
	 * @param plan
	 * @param expression
	 * @param leftexp
	 * @param rightexp
	 * @return
	 */
	public MultOp(ExecutionPlan plan, XQueryExpression expression, Operator left, Operator right) throws MediatorException {
		super(plan, expression);
		this.algebralist = new ArrayList(2);
		algebralist.add(left);
		algebralist.add(right);
		init();
	}

	public MultOp(ExecutionPlan plan, XQueryExpression expression, ArrayList algebralist) throws MediatorException {
		super(plan, expression);
		this.algebralist = algebralist;
		//if (algebralist == null) throw new MediatorException("Cannot build AlgMultOp without elements");
		init();
	}

	private void init() throws MediatorException {
		// Order is important
		if (algebralist != null)
			for (int i = 0; i < algebralist.size(); i++) {
				Operator algebrai = (Operator) algebralist.get(i);
				addPaths(algebrai.getPaths());
				algebrai.incParentCount();
			}
	}

	// #############################################################################
	// VISITOR STUFF
	// #############################################################################

	public void accept(OperatorVisitor visitor) throws MediatorException {
		visitor.visit(this);
	}

	// ***********************************************************************
	// * EXECUTE QUERY
	// ***********************************************************************
	/**
	 * Start the execution, as a binary node, it start the execution
	 * of the two children it refers. The execution is parallelized or
	 * serialized depending the value of 'ordertype' field that can be
	 * ORDER_SERIALIZATION or ORDER_PARALLELIZATION.
	 */
	public void executeQuery() throws MediatorException {
		///*DEBUG*/System.out.println ("****** " + getClass () + ".executeQuery ()") ;
		Thread[] threads = null;
		try {

			if (ordertype == ORDER_SERIALIZATION) {
				for (int i = 0; i < algebralist.size(); i++)
					 ((Operator) algebralist.get(i)).run();
			} else if (ordertype == ORDER_PARALLELIZATION) {
				threads = new Thread[algebralist.size()];
				for (int i = 0; i < algebralist.size(); i++) {
					threads[i] = new Thread((Runnable) algebralist.get(i));
					threads[i].start();
				}
				// wait for all threads to exit
				for (int i = 0; i < threads.length; i++) {
					Thread thread = threads[i];
					if (thread.isAlive())
						thread.join();
				}
			}
		} catch (Exception e) {
			throw new MediatorException("AlgMultOp.executeQuery: " + e.getMessage(), e);
		}
	}

	// ***********************************************************************
	// * GET/SET METHODS
	// ***********************************************************************
	/**
	 * Return a reference on the child at given position expression.
	 *
	 * @return a reference on the child at given position expression.
	 */
	public Operator getSubElement(int pos) {
		if (pos < 0 || pos >= algebralist.size())
			return null;
		return (Operator) algebralist.get(pos);
	}

	public int getSubElementsSize() {
		return this.algebralist.size();
	}
	// **********************************************************************
	// * OPTIMIZE
	// **********************************************************************
	/**
	 * If the *exactly* same set of source is used for an operation,
	 * then the subtree can be passed entierely to the source.
	 * Beware of processing this from the leaf to the node.
	 *
	 * <p>Note: If this method is not overloaded or if it returns null,
	 * it means that _this_ node will not be optimized.</p>
	 * @return
	 */
	public Operator optimizeSource() throws MediatorException {
		for (int i = 0; i < algebralist.size(); i++)
			 ((Operator) algebralist.get(i)).optimizeSource();
		/*
		if (shareSameSources()) {
		    Algebra node = mergeWithSub() ;
		    return node ;
		}
		*/
		return null;
	}

	/**
	 * Add projections on attributes the soonest as possible in order to
	 * manipulate the less data as possible. As soon an attribute isn't
	 * necessary, it must be unselected (by a projection on the complementary
	 * variables).
	 *
	 * <p>Note: If this method is not overloaded or if it returns null,
	 * it means that _this_ node will not be optimized.</p>
	 * @return
	 */
	public Operator optimizeProjection() throws MediatorException {
		return null;
	}

	/**
	 *
	 *
	 * <p>Note: If this method is not overloaded or if it returns null,
	 * it means that _this_ node will not be optimized.</p>
	 * @return
	 */
	public Operator optimizeRestriction() throws MediatorException {
		return null;
	}

	/**
	 * If the set of sources handled by this node are exactly the same (order
	 * not important) then return true.
	 * An AlgBinValue is true if both of its child are true and IF the
	 * set of sources handled by its children are exactly the same.
	 * @return
	 */
	protected boolean shareSameSources() {
		for (int i = 0; i < algebralist.size(); i++)
			if (((Operator) algebralist.get(i)).shareSameSources() == false)
				return false;
		if (compareSource())
			return true;
		return false;
	}

	/**
	 * Compare set of sources.
	 * @param algebra1
	 * @param algebra2
	 * @return
	 */
	private boolean compareSource() {
		ArrayList reflist = null;
		for (int i = 0; i < algebralist.size(); i++) {
			Operator algebrai = (Operator) algebralist.get(i);
			if (i == 0)
				reflist = algebrai.getSources();
			else {
				ArrayList tmplist = algebrai.getSources();
				if (tmplist.size() != reflist.size() || !reflist.containsAll(tmplist))
					return false;
			}
		}
		return true;
	}

	/**
	 * Return true if the node is of type source.
	 * AlgBinOp is not a source.
	 *
	 * @return false ;
	 */
	public boolean isSource() {
		return false;
	}

	/**
	 * Return all the sources depending of this node.
	 * @return
	 */
	public ArrayList getSources() {
		if (sources == null) {
			sources = new ArrayList();
			for (int i = 0; i < algebralist.size(); i++) {
				ArrayList tmplist = ((Operator) algebralist.get(i)).getSources();
				for (int j = 0; j < tmplist.size(); j++) {
					Object sourcej = tmplist.get(j);
					if (!sources.contains(sourcej))
						sources.add(sourcej);
				}
			}
		}
		return sources;
	}

	public String toCompleteString(int indent) {
		StringBuffer buf = new StringBuffer();
		buf.append(Utils.makeIndent(indent) + "<" + getClass().getName() + " size=" + size + " id size=" + idsize + " isLet=\"" + islet + "\" valueDepends=\"" + valueDepends + "\" whereDepends=\"" + whereDepends + "\">\n");
		
		buf.append(Utils.makeIndent(indent + 1) + "<Expression>\n") ;
		buf.append(Utils.makeIndent(indent + 2) + expression + "\n") ;
		buf.append(Utils.makeIndent(indent + 1) + "</Expression>\n") ;
//		if (variables != null) {
//		    buf.append(Utils.makeIndent(indent + 1) + "<Variables>\n") ;
//		    for (int i = 0 ; i < variables.size() ; i ++) {
//		        buf.append(Utils.makeIndent(indent + 2) + "<Variable>" + variables.get(i) + "</Variable>\n") ;
//		    }
//		    buf.append(Utils.makeIndent(indent + 1) + "</Variables>\n") ;
//		}
		if (paths != null) {
		    buf.append(Utils.makeIndent(indent + 1) + "<Paths>\n") ;
		    for (int i = 0 ; i < paths.size() ; i ++) {
		        buf.append(Utils.makeIndent(indent + 2) + "<Path>" + paths.get(i) + "</Path>\n") ;
		    }
		    buf.append(Utils.makeIndent(indent + 1) + "</Paths>\n") ;
		}
		
		for (int i = 0; i < algebralist.size(); i++) {
			buf.append(Utils.makeIndent(indent + 1) + "<Element" + (i + 1) + ">\n");
			buf.append(((Operator) algebralist.get(i)).toCompleteString(indent + 2));
			buf.append(Utils.makeIndent(indent + 1) + "</Element" + (i + 1) + ">\n");
		}
		buf.append(Utils.makeIndent(indent) + "</" + getClass().getName() + ">\n");
		return buf.toString();
	}
}
