/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.mediator.plan;

import java.util.ArrayList;

import org.xml.sax.ContentHandler;
import org.xquark.mediator.DOMUtils.BufferTuple;
import org.xquark.mediator.DOMUtils.SortedKeyMap;
import org.xquark.mediator.DOMUtils.Tuple;
import org.xquark.util.DOMReader;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xquery.parser.XQueryExpression;
import org.xquark.xquery.xdbc.XDBCResultSetInterface;

public abstract class ResultSet {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.1 $";
	private static final String RCSName = "$Name:  $";
	// **********************************************************************
	// * CLASS VARIABLES
	// **********************************************************************
	protected static DOMReader dom2sax = null;
	static {
		dom2sax = new DOMReader();
	}

	protected XQueryExpression expression = null;
	protected Operator algebra;
	private boolean first_time = true;
	private boolean finished_when_empty = false;

	protected BufferTuple buftuples = null;
	private Tuple cur_tuple = null;
	private Tuple next_tuple = null;
	private Object identifiers[] = null;

	/**
	 * If fork is set to true, then resultset are kept in memory in order to
	 * be used if called again without re-asking upper for resultset.
	 */
	protected boolean fork = false;
	// tells if the resultset is ordered
	protected boolean ordered = false;

	protected ArrayList paths;
	protected ArrayList steps;

	// ***********************************************************************
	// * INITIALIZATION
	// ***********************************************************************
	/**
	 *
	 */
	public ResultSet(Operator algebra) {
		this.algebra = algebra;
		init(algebra);
	}

	public void init(Operator algebra) {
		if (algebra != null) {
			expression = algebra.getExpression();
			paths = algebra.getPaths();

			buftuples = new BufferTuple(paths, algebra.getIdSize());
			//this.buftuples = new BufferTuple(this.paths,this.algebra.getIdSize()) ;
			steps = buftuples.getPathList();
			buftuples.setIndexesSpec(algebra.getIndexesSpec());
		}
	}

	/**
	 *
	 */
	/*
	protected void fillSteps() {
	    steps = new ArrayList() ;
	    for (int i = 0 ; i < paths.size() ; i ++) {
	        ArrayList step = new ArrayList() ;
	        String pathi = (String) paths.get(i) ;
	        
	        StringTokenizer st = new StringTokenizer(pathi, "/") ;
	        while (st.hasMoreTokens()) {
	            String token = (String) st.nextToken() ;
	            step.add(token) ;
	        }
	        steps.add(step) ;
	    }
	}
	*/
	/**
	 * Set the variable "fork". If fork is set to true, then resultset are
	 * kept in memory in order to be used if called again without re-asking
	 * upper for resultset.
	 */
	public void setFork(boolean fork) {
		this.fork = fork;
	}

	public void setOrdered(boolean ordered) {
		this.ordered = ordered;
	}

	protected void generateTuple() throws XMLDBCException {
		// if the buffer is not empty, use it before generating the next elements
		if (!buftuples.isEmpty())
			return;

		// a LET is a blocking case.
		evaluate(!algebra.isLet());

	}

	/**
	 * Fill the buffer buftuples with tuple.
	 *
	 * @param non_blocking if the evaluation is blocking, we must
	 *        generate all the results before passing.
	 */
	protected abstract void evaluate(boolean non_blocking) throws XMLDBCException;

	protected void setFinishedWhenEmpty() {
		this.finished_when_empty = true;
	}

	public ArrayList getPaths() {
		return paths;
	}

	public Operator getAlgebra() {
		return algebra;
	}

	public SortedKeyMap getIndex(int index) {
		return buftuples.getIndex(index);
	}

	//public Tuple newTuple()  { return buftuples.newTuple() ; }   

	/**
	 * Give the new current result set (the next one).
	 * @return true if the new current result set is valid, false otherwise
	 * @throws XDBCException if a database access error occurs.
	 */
	public Tuple next() throws XMLDBCException {
		if (next_tuple != null) {
			cur_tuple = next_tuple;
			next_tuple = null;
			return cur_tuple;
		}

		if (first_time) {
			hasNext();
			if (buftuples.isEmpty())
				throw new XMLDBCException("No more results");
		}

		// CHANGE   04/02/2003
		if (buftuples.isEmpty()) {
			generateTuple();
			if (buftuples.isEmpty())
				throw new XMLDBCException("No more results");
		}

		if (!buftuples.isEmpty()) {
			cur_tuple = (Tuple) buftuples.remove(0);
			return cur_tuple;
		}

		return cur_tuple;
	}

	/**
	 * Returns true if the result set has more elements.
	 * @return true if the result set has more elements, false otherwise
	 */
	public boolean hasNext() throws XMLDBCException {

		if (next_tuple != null)
			return true;

		if (first_time) {
			generateTuple();
			first_time = false;
		}

		if ((finished_when_empty) && (buftuples == null || buftuples.isEmpty()))
			return false;

		if (buftuples.isEmpty())
			generateTuple();

		return (!((finished_when_empty) && (buftuples == null || buftuples.isEmpty())));
	}

	/**
	 * Releases this result set object's resources immediately instead of waiting
	 * for this to happen when it is automatically closed.<BR>
	 * <B>Note:</B> A XResultSet object is automatically closed by the XStatement
	 * object that generated it when that XStatement object is closed,
	 * re-executed, or is used to retrieve the next result from a sequence of
	 * multiple results.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public void close() {
		buftuples = null;
	}

	/**
	 * Return the value of the element specified by the path.
	 * @param path
	 * @param nodeAccessor
	 * @param loopID the ID of the loop concerned if any
	 * @return value of the element
	 * @throws XDBCException if a database access error occurs.
	 */
	public String fetch(String path, int nodeAccessor, String loopID) throws XMLDBCException {
		String alias = path;
		while ((alias = algebra.getPlan().getAlias(path)) != null) {
			path = alias;
		}
		//if (alias != null) path = alias;
		int pos = buftuples.getPos(path);
		if (pos != -1)
			return cur_tuple.makeStringPath(pos);
		switch (nodeAccessor) {
			case XDBCResultSetInterface.NODE_ACCESSOR_TEXT :
				pos = buftuples.getPos(path + "/child::text()");
				break;
				/*
				case XDBCResultSetInterface.NODE_ACCESSOR_PROCESSING_INSTRUCTION : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_COMMENT : break;
				*/
			case XDBCResultSetInterface.NODE_ACCESSOR_NODE :
				pos = buftuples.getPos(path + "/child::node()");
				break;
				/*
				case XDBCResultSetInterface.NODE_ACCESSOR_NAME : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_LOCAL_NAME : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_NAMESPACE_URI : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_PARENT : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_STRING_VALUE : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_BASE_URI : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_CHILDREN : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_NAMESPACES : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_ATTRIBUTES : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_DECLARATION : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_TYPE : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_TYPE_VALUE : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_PREFIX : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_URI : break;
				*/
			case XDBCResultSetInterface.NODE_ACCESSOR_SELF_TEXT :
				pos = buftuples.getPos(path + "/self::text()");
				break;
			case XDBCResultSetInterface.NODE_ACCESSOR_SELF_NODE :
				pos = buftuples.getPos(path + "/self::node()");
				break;
		}
		if (pos == -1)
			pos = buftuples.getPos(path);
		if (pos == -1)
			throw new XMLDBCException("No column " + path + " in tuple.");
		return cur_tuple.makeStringPath(pos);
		//throw new XMLDBCException("Shouldn't be called: a tree can't be cast to a String") ;
	}

	/**
	 * Generate the value of the result set specified by the path in the handler.
	 * @param path
	 * @param nodeAccessor the type of node to be generated
	 * @param loopID the ID of the loop concerned if any
	 * @throws XDBCException if a database access error occurs.
	 */
	public void generate(ContentHandler handler, String path, int nodeAccessor, String loopID) throws XMLDBCException {
		String alias = path;
		while ((alias = algebra.getPlan().getAlias(path)) != null) {
			path = alias;
		}
		//if (alias != null) path = alias;
		int pos = -1;
		switch (nodeAccessor) {
			case XDBCResultSetInterface.NODE_ACCESSOR_TEXT :
				pos = buftuples.getPos(path + "/child::text()");
				break;
				/*
				case XDBCResultSetInterface.NODE_ACCESSOR_PROCESSING_INSTRUCTION : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_COMMENT : break;
				*/
			case XDBCResultSetInterface.NODE_ACCESSOR_NODE :
				pos = buftuples.getPos(path + "/child::node()");
				break;
				/*
				case XDBCResultSetInterface.NODE_ACCESSOR_NAME : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_LOCAL_NAME : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_NAMESPACE_URI : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_PARENT : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_STRING_VALUE : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_BASE_URI : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_CHILDREN : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_NAMESPACES : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_ATTRIBUTES : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_DECLARATION : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_TYPE : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_TYPE_VALUE : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_PREFIX : break;
				case XDBCResultSetInterface.NODE_ACCESSOR_URI : break;
				*/
			case XDBCResultSetInterface.NODE_ACCESSOR_SELF_TEXT :
				pos = buftuples.getPos(path + "/self::text()");
				break;
			case XDBCResultSetInterface.NODE_ACCESSOR_SELF_NODE :
				pos = buftuples.getPos(path + "/self::node()");
				break;
		}
		if (pos == -1)
			pos = buftuples.getPos(path);
		if (pos == -1)
			throw new XMLDBCException("No column " + path + " in tuple.");
		try {
			cur_tuple.makeEventPath(dom2sax, handler, pos);
		} catch (Exception e) {
			throw new XMLDBCException(e.getMessage());
		}
	}

	/**
	 * Recover the identifiers of the current result set and stores them in the table <I>identifiers</I>.
	 * @param identifiers the table of identifiers
	 * @throws XDBCException if a database access error occurs.
	 */
	public void getIdentifiers(Object[] identifiers) {
		if (cur_tuple == null)
			return;
		cur_tuple.getIdentifiers(identifiers);
	}

	/**
	 * Recover the identifiers of the next result set and stores them in the tab
	 * le <I>identifiers</I>.
	 * @param identifiers the table of identifiers
	 * @throws XDBCException if a database access error occurs.
	 */
	public void getNextIdentifiers(Object[] identifiers) throws XMLDBCException {
		// because 'getNextIdentifiers' is very special (comprendre
		// very chiant). We must get identifier of next tuple, but keep
		// the cursor on the current one.
		// we solve this by using a "cache" tuple
		if (next_tuple != null) {
			next_tuple.getIdentifiers(identifiers);
			return;
		}
		if (!hasNext()) {
			next_tuple = null;
			return;
		}
		Tuple old_tuple = cur_tuple;
		next();
		next_tuple = cur_tuple;
		next_tuple.getIdentifiers(identifiers);
		cur_tuple = old_tuple;
	}
}
