/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xml.xqueryevaluator.eval;

import java.net.URL;
import java.util.ArrayList;

import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.*;
import org.xquark.mediator.DOMUtils.BufferTuple;
import org.xquark.mediator.plan.ExecutionPlan;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.schema.validation.ValidatingSchemaFilter;
import org.xquark.schema.validation.ValidatingSchemaHandler;
import org.xquark.xml.xdbc.XMLDBCException;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.FunctionDOC;

/**
 * 
 */
public class EvalFilter {

	// **********************************************************************
	// * VERSIONING
	// **********************************************************************
	
	private static final String RCSRevision = "$Revision: 1.3 $";
	private static final String RCSName = "$Name:  $";
	
	// **********************************************************************
	// * CLASS VARIABLES
	// **********************************************************************
	
	private static SAXParserFactory saxParserFactory = null;
	static
	{
		// Create a JAXP SAXParserFactory and configure it
		saxParserFactory = SAXParserFactory.newInstance();
		saxParserFactory.setNamespaceAware(true);
		saxParserFactory.setValidating(false);
	}

    private XMLReader saxReader = null;
    private InputSource inputSource = null;
    private XPPReader xppReader = null;
    private BufferTuple buftuples = null;
    	
	// **********************************************************************
	// * INITIALIZATION
	// **********************************************************************
	
	public EvalFilter(BufferTuple buftuples, ExecutionPlan plan, ArrayList paths, FLWRExpression flwr) throws XMLDBCException {
		this(buftuples, plan, paths, flwr, true);
	}
	
	public EvalFilter(BufferTuple buftuples, ExecutionPlan plan, ArrayList paths, FLWRExpression flwr, boolean useXPP) throws XMLDBCException {
		this.buftuples = buftuples;
		try {
			SchemaValidationContext svc = new SchemaValidationContext(plan.getTypeVisitor().getSchemaManager());
			SAX2TypedDOM handler = new SAX2TypedDOM(flwr,buftuples,plan,paths,svc);
			Variable var = (Variable)flwr.getVariables().get(0);
			String doc = null;
			if (var.getExpression() instanceof LocatedExpression)
				doc = ((ValueString)((FunctionDOC)((Step)var.getExpression().getSteps().get(0)).getExpression()).getArgument(0)).getStringValue();
			else if (var.getExpression() instanceof FunctionDOC)
				doc = ((ValueString)((FunctionDOC)var.getExpression()).getArgument(0)).getStringValue();
			if (doc.startsWith("*:")) doc = doc.substring(2);
			URL url;
			if (plan.getBaseURI() != null) url = new URL(new URL(plan.getBaseURI()),doc);
			else url = new URL(doc);
			if (useXPP) {
				xppReader = new XPPReader();
				ValidatingSchemaHandler filter = new ValidatingSchemaHandler(svc, false);
				filter.setContentHandler(handler);
				filter.setLexicalHandler(handler);
				filter.setErrorHandler(
					new ErrorHandler() {
						public void warning(SAXParseException ex) throws SAXException {
						}

						public void error(SAXParseException ex) throws SAXException {
							throw ex;
						}

						public void fatalError(SAXParseException ex) throws SAXException {
							throw ex;
						}
					});
				xppReader.setContentHandler(filter);
				xppReader.setLexicalHandler(filter);
				xppReader.setErrorHandler(filter);
				xppReader.setInput(url.toString(), url.openStream());
			} else {
				XMLReader reader = saxParserFactory.newSAXParser().getXMLReader();
				ValidatingSchemaFilter filter = new ValidatingSchemaFilter(reader, svc, false);
				filter.setContentHandler(handler);
				filter.setLexicalHandler(handler);
				filter.setErrorHandler(
					new ErrorHandler() {
						public void warning(SAXParseException ex) throws SAXException {
						}

						public void error(SAXParseException ex) throws SAXException {
							throw ex;
						}

						public void fatalError(SAXParseException ex) throws SAXException {
							throw ex;
						}
					});
				saxReader = filter;
				inputSource = new InputSource(url.toString());
			}
		} catch (Exception e) {
			throw new XMLDBCException(e.getMessage());
		}
	}
	// **********************************************************************
	// * ACCESS
	// **********************************************************************
	
	public boolean evaluate() throws XQueryException {
		if (xppReader != null) return evaluateXPP();
		else return evaluateSAX();
	}

	public boolean evaluateSAX() throws XQueryException {
		try {
			saxReader.parse(inputSource);
			return true;
		}
		catch (Exception e) {
			throw new XQueryException(e.getMessage());
		}
	}
		
	public boolean evaluateXPP() throws XQueryException {
		try {
			boolean isDone = false;
			if (xppReader.isAtStart()) xppReader.getContentHandler().startDocument();
			while (buftuples.isEmpty() && !isDone) {
				isDone = xppReader.parseNextElement(); 
			}
			if (isDone) xppReader.getContentHandler().endDocument();
			return isDone;
		} catch (Exception e) {
			throw new XQueryException(e.getMessage());
		}
	}
		
	// **********************************************************************
	// * DEBUGGING
	// **********************************************************************
	
//	private void printTuples() {
//		try {
//			for (int i = 0; i < tuples.size(); i++) {
//				Tuple tuple = (Tuple)tuples.get(i);
//				for (int j = 0; j < tuple.size(); j++) {
//					System.out.println("tuple["+i+", "+j+"]");
//					TypedNode tnode = (TypedNode)tuple.getNodesAtIndex(j).get(0);
//					if (tnode.getFirstChild() != null) {
//						org.w3c.dom.Node n = tnode.getFirstChild().getParentNode();
//						org.xquark.xml.util.DOMReader domReader = new org.xquark.xml.util.DOMReader((org.w3c.dom.Element)n);
//						org.xquark.xml.serialize.XMLSerializer seria = new org.xquark.xml.serialize.XMLSerializer(new java.io.FileOutputStream(new java.io.File("tuple("+i+","+j+").xml")));
//						domReader.setContentHandler(seria);
//						domReader.parse("");
//					}
//				}
//			}
//		}
//		catch (Exception e) {
//			System.out.println(e.getMessage());
//			e.printStackTrace();
//		}
//	}
}
