/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xml.xqueryevaluator.eval;

import java.util.*;

import org.w3c.dom.Attr;
import org.w3c.dom.Comment;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.ext.LexicalHandler;
import org.xml.sax.helpers.DefaultHandler;
import org.xquark.mediator.DOMUtils.BufferTuple;
import org.xquark.mediator.DOMUtils.EvaluationVisitor;
import org.xquark.mediator.DOMUtils.Tuple;
import org.xquark.mediator.plan.ExecutionPlan;
import org.xquark.schema.Type;
import org.xquark.schema.validation.ElementPSVInfoset;
import org.xquark.schema.validation.PSVInfoset;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.util.NamespaceContextStack;
import org.xquark.util.SAXConstants;
import org.xquark.xpath.Axis;
import org.xquark.xpath.NodeKind;
import org.xquark.xpath.datamodel.*;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.FunctionDOC;
import org.xquark.xquery.parser.util.Constants;

/**
 * 
 */
public class SAX2TypedDOM extends DefaultHandler implements LexicalHandler {
	// **********************************************************************
	// * VERSIONING
	// **********************************************************************

	private static final String RCSRevision = "$Revision: 1.2 $";
	private static final String RCSName = "$Name:  $";

	private SchemaValidationContext svc = null;
	private BufferTuple buffer = null;
	private ExecutionPlan plan = null;
	private XQueryExpression whereClause = null;
	private ArrayList computedVars = new ArrayList();
	private EvaluationVisitor evaluator = null;
	private List rootFilters = new ArrayList();
	private NamespaceContextStack nsStack = null;

	private ArrayList contexts = new ArrayList(32);
	private int level = 0;
	private List bindings = new ArrayList();	

	private TypedDocumentImpl doc = new TypedDocumentImpl();
	private TypedNode currentNode = null;
	private HashSet usedURI = new HashSet();
	private HashSet declaredURI = new HashSet();
	private StringBuffer charsRead = new StringBuffer();

	public SAX2TypedDOM(FLWRExpression flwr, BufferTuple buffer, ExecutionPlan plan, ArrayList paths, SchemaValidationContext svc)
		throws XQueryException {
		this.buffer = buffer;
		this.plan = plan;
		this.svc = svc;
		this.nsStack = flwr.getParentUnit().getDeclarations();
		this.evaluator = new EvaluationVisitor(svc.getSchemaManager());
		this.whereClause = flwr.getWhereClause();
		List variables = flwr.getVariables();
		for (int i = 0; i < variables.size(); i++) {
			Variable v = (Variable) variables.get(i);
			List steps = v.getExpression().getSteps();
			XQueryExpression rootExpr;
			List prefix = null;
			if (steps == null) {
				rootExpr = v.getExpression();
			} else {
				rootExpr = ((Step) steps.get(0)).getExpression();
			}
			if (rootExpr instanceof FunctionDOC) {
				if (v.getBindingType() == Constants.FOR_BINDINGTYPE) {
					XPathExpr[] pathExprs = buildXPath(steps);
					for (int k = 0; k < pathExprs.length; k++) {
					    VariableFilter rootFilter = new VariableFilter(v.print(), pathExprs[k]);
					    rootFilters.add(rootFilter);
					}
				} else {
					VariableFilter rootFilter = new VariableFilter("", new XPathExpr());
					rootFilters.add(rootFilter);
					int index = buffer.getPathIndex(v.print());
					if (index != -1) {
						XPathExpr[] pathExprs = buildXPath(steps);
						for (int k = 0; k < pathExprs.length; k++) {
							rootFilter.addSubFilter(index, pathExprs[k]);
						}
					}
					prefix = steps;
				}
				for (int j = 0; j < paths.size(); j++) {
					XQueryExpression expr = (XQueryExpression) paths.get(j);
					// Remove the variable from paths list: it can be a Variable or a LocExpr of length 1
					if (expr instanceof LocatedExpression && expr.getSteps().size() > 1) {
						int index = buffer.getPathIndex(((LocatedExpression)expr).print());
						XPathExpr[] pathExprs = buildXPath(expr.getSteps(), prefix);
						for (int k = 0; k < pathExprs.length; k++) {
							for (int l = 0; l < rootFilters.size(); l++) {
								((VariableFilter) rootFilters.get(l)).addSubFilter(index, pathExprs[k]);
							}
						}
					}
				}
			} else if (steps == null) {
				computedVars.add(v);
			} else {
				// TODO: What do we do here??
				throw new RuntimeException("TODO");
			}
		}
	}

	private XPathExpr[] buildXPath(List steps, List prefix) throws XQueryException {
		XPathExpr[] paths = { new XPathExpr() };
		if (prefix != null) {
			for (int i = 1; i < prefix.size(); i++) {
				paths = addStep(paths, (Step) prefix.get(i));
			}
		}
		if (steps != null) {
			for (int i = 1; i < steps.size(); i++) {
				paths = addStep(paths, (Step) steps.get(i));
			}
		}
		for (int i = 0; i < paths.length; i++) {
			paths[i].removeLastSelfNodeStep();
		}
		return paths;
	}

	private XPathExpr[] buildXPath(List steps) throws XQueryException {
		return buildXPath(steps, null);
	}

    private XPathExpr[] addStep(XPathExpr[] paths, Step step) throws XQueryException {
		QName stepName = null;
		boolean isSelf = false;
		boolean isAnyLevel = false;
		XQueryExpression stepExpr = step.getExpression();
		switch (step.getAxis()) {
			case Axis.DESCENDANT_OR_SELF :
				isAnyLevel = true;
				// pass through
			case Axis.SELF :
				isSelf = true;
				// pass through
			case Axis.CHILD :
				if (stepExpr instanceof QName) {
					stepName = (QName) stepExpr;
					for (int i = 0; i < paths.length; i++) {
						XPathExpr pathExpr = paths[i];
						pathExpr.addElementStep(stepName.getNameSpace(), stepName.getLocalName(), isAnyLevel, isSelf);
					}
				} else if (stepExpr instanceof NodeType) {
					switch (((NodeType) stepExpr).getKind()) {
						case NodeKind.NODE :
							for (int i = 0; i < paths.length; i++) {
								XPathExpr pathExpr = paths[i];
								pathExpr.addNodeStep(isAnyLevel, isSelf);
							}
							break;
						case NodeKind.TEXT :
							for (int i = 0; i < paths.length; i++) {
								XPathExpr pathExpr = paths[i];
								pathExpr.addTextStep(isAnyLevel, isSelf);
							}
							break;
						default :
							throw new XQueryException("Unsupported XPath expression");
					}
				} else {
					throw new XQueryException("Unsupported XPath expression");
				}
				break;
			case Axis.ATTRIBUTE :
				stepName = (QName) step.getExpression();
				for (int i = 0; i < paths.length; i++) {
					XPathExpr pathExpr = paths[i];
					pathExpr.addAttributeStep(stepName.getNameSpace(), stepName.getLocalName(), isAnyLevel);
				}
				break;
			case Axis.NONE :
				if (stepExpr instanceof ListOpUNIONExpression) {
					ArrayList steps1 = ((ListOpUNIONExpression)stepExpr).getExpression1().getSteps();
					ArrayList steps2 = ((ListOpUNIONExpression)stepExpr).getExpression2().getSteps();
					if (steps1 == null || steps2 == null)
						throw new XQueryException("Unsupported XPath expression");
					XPathExpr[] paths1 = paths;
					XPathExpr[] paths2 = new XPathExpr[paths.length];
					for (int i = 0; i < paths.length; i++) {
						paths2[i] = (XPathExpr)paths[i].clone();
					}
					for (int i = 0; i < steps1.size(); i++) paths1 = addStep(paths1, (Step)steps1.get(i));
					for (int i = 0; i < steps2.size(); i++) paths2 = addStep(paths2, (Step)steps2.get(i));
					paths = new XPathExpr[paths1.length+paths2.length];
					System.arraycopy(paths1, 0, paths, 0, paths1.length);
					System.arraycopy(paths2, 0, paths, paths1.length, paths1.length);
				} else {
					throw new XQueryException("Unsupported XPath expression");
				}
				break;
			default :
				throw new XQueryException("Unsupported XPath expression");
		}
    	return paths;
    }
    
	private void reset() {
		level = 0;
		contexts.clear();
		bindings.clear();
		contexts.add(new FilterContext(rootFilters));
	}

	private FilterContext pushContext(String uri, String localName) {
		FilterContext parent = getCurrentContext();
		FilterContext current;
		level++;
		if (level == contexts.size())
			contexts.add(new FilterContext());
		current = getCurrentContext();
		current.initialize(uri, localName, parent);
		return current;
	}

	private void popContext() throws XQueryException {
		getCurrentContext().cleanup();
		level--;
	}

	private FilterContext getCurrentContext() {
		return getContext(level);
	}

	private FilterContext getContext(int i) {
		return (FilterContext) contexts.get(i);
	}

	private void filterTuple(Tuple tuple) throws XQueryException {
		for (int i = 0; i < computedVars.size(); i++) {
			Variable var = (Variable) computedVars.get(i);
			int index = buffer.getPathIndex(var.print());
			evaluator.reset(tuple);
			var.getExpression().accept(evaluator);
			tuple.addNodesAtIndex(index, evaluator.getResNodes());
		}
		tuple.fillIdentifiers();
		if (whereClause != null) {
			plan.addCurrentTuple(tuple);
			evaluator.reset(plan.getCurrentTuples());
			evaluator.setReturnType(EvaluationVisitor.BOOLEAN_TYPE);
			whereClause.accept(evaluator);
			plan.deleteCurrentTuple(tuple);
			if (evaluator.getVerdict())
				buffer.add(tuple);
		} else buffer.add(tuple);
	}

	private TypedAttribute buildAttribute(PSVInfoset infoset, String value) {
		TypedAttribute attr = (TypedAttribute) doc.createAttributeNS(infoset.getNamespaceURI(), infoset.getLocalName());
		if (infoset.getNormalizedValue() != null)
			attr.setNodeValue(infoset.getNormalizedValue());
		else
			attr.setNodeValue(value);
		if (infoset.getActualValue() != null)
			attr.setTypedValue(infoset.getActualValue());
		else
			attr.setTypedValue(value);
		if (infoset.getType() != null)
			attr.setType(infoset.getType());
		else
			attr.setType(svc.getSchemaManager().getAnySimpleType());
		attr.setDeclaration(infoset.getDeclaration());
		return attr;
	}

	private TypedValue buildValue(String value, PSVInfoset infoset) {
		TypedValue tv = (TypedValue) doc.createTextNode(value);
		Type type = null;
		if (infoset != null)
			type = infoset.getType();
		if (type != null && type.getValueType() != null) {
			tv.setType(type.getValueType());
			tv.setTypedValue(infoset.getActualValue());
			tv.setNodeValue(infoset.getNormalizedValue());
		} else {
			// TODO: Check if typing is needed here. A Text node might be enough
			tv.setType(svc.getSchemaManager().getAnySimpleType());
			tv.setTypedValue(value);
			tv.setNodeValue(value);
		}
		return tv;
	}

	private Attr buildPrefixAttribute(String prefix, String uri) {
		if (prefix == null || prefix.length() == 0) {
			prefix = "xmlns";
		} else {
			prefix = "xmlns:" + prefix;
		}
		Attr attr = doc.createAttributeNS(SAXConstants.XMLNS_URI, prefix);
		attr.setValue(uri);
		return attr;
	}

	private Comment buildComment(char[] chars, int start, int length) {
		return doc.createComment(new String(chars, start, length));
	}

	public void startDocument() throws SAXException {
		reset();
		FilterContext context = getCurrentContext();
		if (context.isNodeNeeded()) {
			context.filterDocument(doc);
			currentNode = doc;
		}
	}

	public void endDocument() throws SAXException {
		try {
            popContext();
        } catch (XQueryException e) {
            throw new SAXException(e);
        }
	}
	
	public void startElement(String uri, String local, String raw, Attributes atts) throws SAXException {
		FilterContext context = getCurrentContext();
		if (charsRead.length() > 0) {
			TypedValue tv = null;
			String value = charsRead.toString();
			charsRead.setLength(0);
			tv = context.filterValue(null, value);
			if (currentNode != null) {
				if (tv == null) {
					tv = buildValue(value, null);
				}
				currentNode.appendChild(tv);
			}
		}
		context = pushContext(uri, local);
		TypedElement el = null;
		ElementPSVInfoset epsvis = svc.getCurrentInfoset();
		if (currentNode != null || context.isNodeNeeded()) {
			el = (TypedElement) doc.createElementNS(uri, local);
			el.setType(epsvis.getType());
			el.setDeclaration(epsvis.getDeclaration());
			context.filterNode(el);
			// Handle prefix declarations
			if (currentNode == null) {
				// New root: 
				usedURI.clear();
				usedURI.add(uri);
				declaredURI.clear();
				declaredURI.addAll(epsvis.getNamespaceURIs());
			}
			Iterator it = epsvis.getDeclaredPrefixes().iterator();
			while (it.hasNext()) {
				String prefix = (String) it.next();
				String nsURI = epsvis.getNamespaceURI(prefix);
				// Keep only previous declarations in declaredURI
				if (currentNode == null)
					declaredURI.remove(nsURI);
				// Prefixes declared in the query
				Collection knownPrefixes = nsStack.getPrefixes(nsURI);
				if (knownPrefixes.isEmpty()) {
					el.setAttributeNode(buildPrefixAttribute(prefix, nsURI));
				}
			}
		}
		for (int i = 0; i < atts.getLength(); i++) {
			PSVInfoset infoset = epsvis.getAttributePSVInfoset(i);

			/* XXX: The schema implementation guarantees that SAX
			 * and PSVI attributes have the same order, so this works
			 */
			String attrValue = atts.getValue(i);
			TypedAttribute attr = null;
			attr = context.filterAttribute(infoset, attrValue);
			if (el != null) {
				if (attr == null)
					attr = buildAttribute(infoset, attrValue);
				el.setAttributeNode(attr);
				usedURI.add(atts.getURI(i));
			}
		}
		if (currentNode != null) {
			currentNode.appendChild(el);
		}
		currentNode = el;
	}

	public void characters(char[] ch, int start, int length) throws SAXException {
		charsRead.append(ch, start, length);
	}

	public void endElement(String uri, String local, String raw) throws SAXException {
		FilterContext context = getCurrentContext();
		ElementPSVInfoset epsvis = svc.getCurrentInfoset();
		if (charsRead.length() > 0) {
			TypedValue tv = null;
			String value = charsRead.toString();
			charsRead.setLength(0);
			tv = context.filterValue(epsvis, value);
			if (currentNode != null) {
				if (tv == null) {
					tv = buildValue(value, epsvis);
				}
				currentNode.appendChild(tv);
			}
		}
		if (currentNode != null) {
			if (currentNode.getParentNode() == null) {
				Iterator it = usedURI.iterator();
				while (it.hasNext()) {
					String used = (String) it.next();
					if (declaredURI.contains(used) && nsStack.getPrefixes(used).isEmpty()) {
						String prefix = epsvis.getPrefix(used);
						((TypedElement) currentNode).setAttributeNode(buildPrefixAttribute(prefix, used));
					}
				}
			}
			currentNode = (TypedNode) currentNode.getParentNode();
		}
		try {
            popContext();
        } catch (XQueryException e) {
            throw new SAXException(e);
        }
	}

	/* (non-Javadoc)
	 * @see org.xml.sax.ext.LexicalHandler#startDTD(java.lang.String, java.lang.String, java.lang.String)
	 */
	public void startDTD(String arg0, String arg1, String arg2) throws SAXException {
	}

	/* (non-Javadoc)
	 * @see org.xml.sax.ext.LexicalHandler#endDTD()
	 */
	public void endDTD() throws SAXException {
	}

	/* (non-Javadoc)
	 * @see org.xml.sax.ext.LexicalHandler#startEntity(java.lang.String)
	 */
	public void startEntity(String arg0) throws SAXException {
	}

	/* (non-Javadoc)
	 * @see org.xml.sax.ext.LexicalHandler#endEntity(java.lang.String)
	 */
	public void endEntity(String arg0) throws SAXException {
	}

	/* (non-Javadoc)
	 * @see org.xml.sax.ext.LexicalHandler#startCDATA()
	 */
	public void startCDATA() throws SAXException {
	}

	/* (non-Javadoc)
	 * @see org.xml.sax.ext.LexicalHandler#endCDATA()
	 */
	public void endCDATA() throws SAXException {
	}

	/* (non-Javadoc)
	 * @see org.xml.sax.ext.LexicalHandler#comment(char[], int, int)
	 */
	public void comment(char[] chars, int start, int length) throws SAXException {
		FilterContext context = getCurrentContext();
		ElementPSVInfoset epsvis = svc.getCurrentInfoset();
		if (charsRead.length() > 0) {
			TypedValue tv = null;
			String value = charsRead.toString();
			charsRead.setLength(0);
			tv = context.filterValue(epsvis, value);
			if (currentNode != null) {
				if (tv == null) {
					tv = buildValue(value, epsvis);
				}
				currentNode.appendChild(tv);
			}
		}
		if (currentNode != null)
			currentNode.appendChild(buildComment(chars, start, length));
	}

	class FilterContext {

		private boolean nodeNeeded = false;
		private List filters = new ArrayList();
		private boolean bindingRoot = false;

		FilterContext(List fList) {
			Tuple tuple = null;
			for (int i = 0; i < fList.size(); i++) {
                VariableFilter xf = (VariableFilter)fList.get(i);
    			filters.add(xf);
	    		if (xf.isDone()) {
		    		if (xf.isBinding()) {
						bindingRoot = bindingRoot || bindings.isEmpty();
		    			if (tuple == null) {
		    				tuple = buffer.newTuple();
		    				bindings.add(tuple);
		    			} 
				    	xf.setTuple(tuple);
    				}
	    			nodeNeeded = nodeNeeded || xf.isNodeNeeded();
		    		List subFilters = xf.getSubFilters();
					if (subFilters != null) {
						Iterator itsub = subFilters.iterator();
						while (itsub.hasNext()) {
							NodeFilter sub = (NodeFilter) itsub.next();
							filters.add(sub);
							if (sub.isDone() || sub.matchSelfNode()) 
								nodeNeeded = nodeNeeded || sub.isNodeNeeded();
						}
					}
		        }
			}
		}

		FilterContext() {}

		/**
		 * @param uri
		 * @param localName
		 * @param parent
		 */
		void initialize(String uri, String localName, FilterContext parent) {
			if (uri != null && uri.length() == 0)
				uri = null;
			filters.clear();
			nodeNeeded = parent.isNodeNeeded();
			Tuple tuple = null;
			Iterator it = parent.filters.iterator();
			while (it.hasNext()) {
				NodeFilter xf = (NodeFilter) it.next();
				if (xf.pushElement(uri, localName)) {
					filters.add(xf);
					/*
					 * Special case: the filter is forked when there is a
					 * descendant-or-self step.
					 */
					if (xf.matchDescendantOrSelfStep(uri, localName)) {
						xf = (NodeFilter) xf.fork(uri, localName);
						filters.add(xf);
					}
					if (xf.isDone() || xf.matchSelf(uri, localName)) {
						if (xf.isBinding()) {
							bindingRoot = bindingRoot || bindings.isEmpty();
							if (tuple == null) {
								tuple = buffer.newTuple();
								bindings.add(tuple);
							} 
							xf.setTuple(tuple);
						}
						nodeNeeded = nodeNeeded || xf.isNodeNeeded();
						List subFilters = xf.getSubFilters();
						if (subFilters != null) {
							Iterator itsub = subFilters.iterator();
							while (itsub.hasNext()) {
								NodeFilter sub = (NodeFilter) itsub.next();
                                sub.skipSelfStep(uri, localName);
								filters.add(sub);
								if (sub.isDone() || sub.matchSelf(uri, localName)) 
									nodeNeeded = nodeNeeded || sub.isNodeNeeded();
							}
						}
					}
				}
			}
		}

		/**
		 * @param node
		 */
		public void filterDocument(TypedDocument doc) {
			Iterator it = filters.iterator();
			while (it.hasNext()) {
				NodeFilter xf = (NodeFilter) it.next();
				if (xf.isDone() || xf.matchSelfNode()) {
					xf.setDOMNode(doc);
				}
			}
		}

		/**
		 * @param node
		 */
		public void filterNode(TypedElement node) {
			Iterator it = filters.iterator();
			while (it.hasNext()) {
				NodeFilter xf = (NodeFilter) it.next();
				String uri = node.getNamespaceURI();
				if (uri.length() == 0) uri = null;
				if (xf.isDone() || xf.matchSelf(uri, node.getLocalName())) {
					xf.setDOMNode(node);
				}
			}
		}

		/**
		 * @param infoset
		 * @param value
		 */
		public TypedAttribute filterAttribute(PSVInfoset infoset, String value) {
			String uri = infoset.getNamespaceURI();
			String name = infoset.getLocalName();
			TypedAttribute attr = null;
			if (uri != null && uri.length() == 0)
				uri = null;
            Tuple tuple = null;
			Iterator it = filters.iterator();
			while (it.hasNext()) {
				NodeFilter xf = (NodeFilter) it.next();
				if (xf.matchAttribute(uri, name)) {
					if (attr == null) {
						attr = buildAttribute(infoset, value);
					}
					if (xf.isBinding()) {
						bindingRoot = bindingRoot || bindings.isEmpty();
						if (tuple == null) {
							tuple = buffer.newTuple();
							bindings.add(tuple);
						} 
						xf = (NodeFilter) xf.clone();
						xf.setTuple(tuple);
					}
					xf.setDOMNode(attr);
				}
			}
			return attr;
		}

		public TypedValue filterValue(PSVInfoset infoset, String value) {
			TypedValue tv = null;
            Tuple tuple = null;
			Iterator it = filters.iterator();
			while (it.hasNext()) {
				NodeFilter xf = (NodeFilter) it.next();
				if (xf.matchText()) {
					if (tv == null) {
						tv = buildValue(value, infoset);
					}
					if (xf.isBinding()) {
						bindingRoot = bindingRoot || bindings.isEmpty();
						if (tuple == null) {
							tuple = buffer.newTuple();
							bindings.add(tuple);
						} 
						xf = (NodeFilter) xf.clone();
						xf.setTuple(tuple);
					}
					xf.setDOMNode(tv);
					List subFilters = xf.getSubFilters();
					if (subFilters != null) {
						Iterator itsub = subFilters.iterator();
						while (itsub.hasNext()) {
							NodeFilter sub = (NodeFilter) itsub.next();
							if (sub.matchSelfText()) 
								sub.setDOMNode(tv);
						}
					}
				}
			}
			return tv;
		}
		
		void flushBindings() throws XQueryException {
			if (bindingRoot) {
				Iterator it = bindings.iterator();
				while (it.hasNext()) {
					filterTuple((Tuple) it.next());
				}
				bindings.clear();
				bindingRoot = false;
			}
		}

		/**
		 * @return
		 */
		boolean isNodeNeeded() {
			return nodeNeeded;
		}

		void cleanup() throws XQueryException {
			Iterator it = filters.iterator();
			while (it.hasNext()) {
				NodeFilter xf = (NodeFilter) it.next();
				xf.popElement();
				xf.setTuple(null);
			}
			filters.clear();
			flushBindings();
		}
	}

	static abstract class NodeFilter extends XPathFilter {

		protected List subFilters = null;
		int colIndex;

		/**
		 * @param xpath
		 */
		public NodeFilter(XPathExpr xpath, int colIndex) {
			super(xpath);
			this.colIndex = colIndex;
		}

		List getSubFilters() {
			return subFilters;
		}

		NodeFilter addSubFilter(int index, XPathExpr xpath) {
			SubPathFilter sub = new SubPathFilter(this, index, xpath);
			if (subFilters == null)
				subFilters = new ArrayList();
			subFilters.add(sub);
			return sub;
		}

		abstract boolean isNodeNeeded();

		/* (non-Javadoc)
		 * @see org.xquark.xml.xqueryevaluator.eval.SAX2TypedDOM.NodeFilter#setDOMNode(org.xquark.dom.typed.TypedNode)
		 */
		void setDOMNode(TypedNode node) {
			if (colIndex != -1) {
				Tuple t = getTuple();
				if (t != null) {
					List nodes = t.getNodesAtIndex(colIndex);
					if (nodes == null || node != nodes.get(nodes.size()-1))
					    t.addNodeAtIndex(colIndex, node);
				}
			}
		}

		abstract boolean isBinding();
		
		abstract void setTuple(Tuple tuple);

		abstract Tuple getTuple();
	}

	class VariableFilter extends NodeFilter {

		String name;
		Tuple tuple = null;

		/**
		 * @param xpath
		 */
		VariableFilter(String name, XPathExpr xpath) {
			super(xpath, buffer.getPathIndex(name));
		}

		protected Object clone() {
			VariableFilter result = (VariableFilter) super.clone();
			result.tuple = null;
			if (subFilters != null) {
				result.subFilters = new ArrayList(subFilters.size());
				for (int i = 0; i < subFilters.size(); i++) {
					SubPathFilter sub = (SubPathFilter) subFilters.get(i);
					result.addSubFilter(sub.colIndex, sub.xpath);
				}
			}
			return result;
		}

		boolean isNodeNeeded() {
			return colIndex != -1;
		}

        boolean isBinding() {
        	return true;
        }
        
		Tuple getTuple() {
			return tuple;
		}
		
		void setTuple(Tuple tuple) {
			this.tuple = tuple;
		}
		
	}

	static class SubPathFilter extends NodeFilter {

		NodeFilter parent;

		/**
		 * @param xpath
		 */
		public SubPathFilter(NodeFilter parent, int index, XPathExpr xpath) {
			super(xpath, index);
			this.parent = parent;
		}

		/* (non-Javadoc)
		 * @see org.xquark.xml.xqueryevaluator.eval.SAX2TypedDOM.NodeFilter#isNodeNeeded()
		 */
		boolean isNodeNeeded() {
			return true;
		}

		boolean isBinding() {
			return false;
		}
        
		/* (non-Javadoc)
		 * @see org.xquark.xml.xqueryevaluator.eval.SAX2TypedDOM.NodeFilter#getTuple()
		 */
		 
		Tuple getTuple() {
			return parent.getTuple();
		}
		
		void setTuple(Tuple tuple) {
		}

	}

}
