/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema;

import org.xquark.schema.validation.ValidationContextProvider;
import org.xquark.schema.validation.ValidationInfo;

/**
 * Abstract class for AttributeDeclaration and ElementDeclaration.
 *
 * <p>This is the implementation of the common propertoes of Attribute Declaration
 * and Element Declaration. The following properties are defined :</p>
 * <p> 1. type defintion : a simple type or a complex type defintion.
 * if a declaration is an instance of Attribute Declaration, 
 * the type of this declaration must be a simple type</p>
 * <p> 2. scope : either null (means gloabal) or a complex type defintion. </p>
 * <p> 3. value constraint : either a defaultValue or a fixedValue. </p>
 *
 * <p>See XML Schema Part 1: Structures 3.2, 3.3
 * W3C Recommendation 2 May 2001</p>
 *
 * @see org.xquark.xml.schema.SchemaComponent
 * @see org.xquark.xml.schema.AttributeDeclaration
 * @see org.xquark.xml.schema.ElementDeclaration
 */
public abstract class Declaration extends SchemaComponent
  implements Cloneable 
{
private static final String RCSRevision = "$Revision: 1.1 $"; 
private static final String RCSName     = "$Name:  $";
  protected SchemaScope scope;
  private Type type = null;
  protected String defaultValue = null;
  protected String fixedValue = null;
  private boolean qualified = false;
  
  /**
   * Create a new Declaration.
   *
   * @param schema The schema which contains this declaration.
   * @param name The declaration name, a NCName.
   * @param scope The scope, null if the declaration is in a group
   * which contains this declaration.
   */
  protected Declaration(Schema schema, String name, SchemaScope scope) {
    super(schema, name);
    this.scope = scope;
  }

  public Object clone() throws CloneNotSupportedException {
    Object result = null;
    result = super.clone();
    return result;
  }
  
  public void accept(SchemaVisitor visitor) throws SchemaException {
    visitor.visit(this);
  }

  /**
   * Return the declaration's target namespace.
   *
   * @return The target namespace if the declaration is qualified, 
   * or null if the declaration is unqualified.
   */
  public String getNamespace() { 
    if (qualified) return super.getNamespace();
    else return null;
  }

  /**
   * Return the declaration's data type.
   *
   * @return The data type, if the declaration is an instance of Attribute Declaration, 
   * a simple type must be returned.
   */
  public Type getType() {
    return type;
  }

  /**
   * Return the declaration's scope. Can be null is the declaration is part of a group definition
   *
   * @return The scope, if the declaration is a top-level declaration, return the schema,
   * otherwise return a complex type which contains the declaration.
   */
  public SchemaScope getScope() {
    return scope;
  }

  public String getValueConstraint() {
    if (defaultValue != null) return defaultValue;
    else if (fixedValue != null) return fixedValue;
    else return null;
  }
  
  public void setValueConstraint(String value) {
    if (defaultValue != null) defaultValue = value;
    else if (fixedValue != null) fixedValue = value;
  }
  
  /**
   * Return the declaration's default value, if any.
   *
   * @return The default value, or null if there is none.
   */
  public String getDefaultValue() {
    return defaultValue;
  }
  
  public void setDefaultValue(String defaultValue) {
    this.defaultValue = defaultValue;
  }
  
  /**
   * Return the declaration's fixed value, if any.
   *
   * @return The fixed value, or null if there is none.
   */
  public String getFixedValue() {
    return fixedValue;
  }
  
  public void setFixedValue(String fixedValue) {
    this.fixedValue = fixedValue;
  }

  public void setType(Type type) {
    this.type = type;
  }

  public void setScope(SchemaScope scope) {
    this.scope = scope;
  }

  public void setQualified(boolean qualified) {
    this.qualified = qualified;
  }

  /*
une nouvelle mthode est ajoute dans la classe Declaration. 

pourquoi cette mthode est ajoute dans Declaration, pas dans SimpleType ?
La fixed valeur et la default valeur sont definie dans Declaration.

  public int validate(String value, ValidationContextProvider context) throws SchemaException

valeurs d'entre:
value : la valeur que l'on veut valider
context : conext provider, c'est un fournisseur de prefix et document base.
          Si le type de cette declaration n'est pas QName or AnyURI,metter contex == null

valeur de retour:
0: value est valid
-1: le type de cette Declaration est null ou ComplexType, c'est  dir que cette mthode ne amarche que avec SimpleType
SchemaException: value n'est pas valid ou value n'gale pas la fixed valeur de cette Declaration.
*/
  public int validate(String value, ValidationContextProvider context) throws SchemaException {
        if ( type == null ) return -1;      
        SimpleType valueType = type.getValueType();
        if (valueType == null ) return -1;
        ValidationInfo info = valueType.validate(value, true, context);
        SimpleType realType = info.getValidationType();
        Object actualValue = info.getActualValue();
        if ( getFixedValue() != null) {
	      if (!getFixedValue().equals(realType.toXMLString(actualValue))) {
              if ( this instanceof AttributeDeclaration )
                  throw new SchemaException("cvc-attribute.4", this);
              else
                  throw new SchemaException("cvc-elt.5.2.2.2.2", this);
          }
        }
        
        return 0;
  }	
      
}
