/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema;

import java.util.Iterator;

public class IterationVisitor extends DefaultSchemaVisitor {
  private static final String RCSRevision = "$Revision: 1.1 $"; 
  private static final String RCSName     = "$Name:  $";

  protected SchemaVisitor pre;
  protected SchemaVisitor post;
  protected boolean inTopLevelDeclaration = true;
  protected boolean inAttributeWildcard = false;

  public IterationVisitor(SchemaVisitor pre, SchemaVisitor post) {
    this.pre = pre;
    this.post = post;
  }
  
  public IterationVisitor(SchemaVisitor visitor) {
    this(visitor, null);
  }
  
  public boolean isTopLevel() {
    return inTopLevelDeclaration;
  }

  public boolean isAttributeWildcard() {
    return inAttributeWildcard;
  }

  public void visit(Schema schema) throws SchemaException {
		
    if (pre != null) schema.accept(pre);
		
    Iterator it = schema.getAttributeDeclarations().iterator();
    while (it.hasNext()) { 
      inTopLevelDeclaration = true;
      ((SchemaVisitable)it.next()).accept(this);
    }

    it = schema.getElementDeclarations().iterator();
    while (it.hasNext()) {
      inTopLevelDeclaration = true;
      ((SchemaVisitable)it.next()).accept(this);
    }

    it = schema.getTypes().iterator();
    while (it.hasNext()) ((SchemaVisitable)it.next()).accept(this);

    it = schema.getModelGroupDefinitions().iterator();
    while (it.hasNext()) ((SchemaVisitable)it.next()).accept(this);

    it = schema.getAttributeGroupDefinitions().iterator();
    while (it.hasNext()) ((SchemaVisitable)it.next()).accept(this);
		
    if (post != null) schema.accept(post);
  }
  
  public void visit(SchemaComponent comp) throws SchemaException {
    if (pre != null) comp.accept(pre);
    if (post != null) comp.accept(post);
  }
  
  public void visit(ComplexType type) throws SchemaException {
    inTopLevelDeclaration = false;
    if (pre != null) type.accept(pre);
		
    type.getContentModel().accept(this);
		
    Iterator it = type.getAttributeDeclarations().iterator();
    while (it.hasNext()) ((SchemaVisitable)it.next()).accept(this);
				
    if (type.getAttributeWildcard() != null) {
      inAttributeWildcard = true;
      type.getAttributeWildcard().accept(this);
      inAttributeWildcard = false;
    }
		
    if (post != null) type.accept(post);
  }

  public void visit(Declaration decl) throws SchemaException {
    if (pre != null) decl.accept(pre);
    if (inTopLevelDeclaration || !decl.getScope().isGlobalScope()) {
      Type type = decl.getType();
      if (type != null && type.getName() == null) type.accept(this);
    }
    if (post != null) decl.accept(post);
  }

  public void visit(Wildcard wildcard) throws SchemaException {
    if (pre != null) wildcard.accept(pre);
    if (post != null) wildcard.accept(post);
  }
  
  public void visit(ModelGroupDefinition def) throws SchemaException {
    if (pre != null) def.accept(pre);
    ModelGroup model = def.getModelGroup();
    model.accept(this);
    if (post != null) def.accept(post);
  }
    
  public void visit(AttributeGroupDefinition def) throws SchemaException {
    if (pre != null) def.accept(pre);
    Iterator it = def.getContents().iterator();
    while (it.hasNext()) ((SchemaVisitable)it.next()).accept(this);
    if (post != null) def.accept(post);
  }
  
  public void visit(ContentModel model) throws SchemaException {
    if (pre != null) model.accept(pre);
    if ( model.getModel() != null ) 
      ((SchemaVisitable)model.getModel()).accept(this);
    if (post != null) model.accept(post);
  }
  
  public void visit(Particle particle) throws SchemaException {
    if (pre != null) particle.accept(pre);
    ((SchemaVisitable)particle.getTerm()).accept(this);
    if (post != null) particle.accept(post);
  }
  
  public void visit(ModelGroup group) throws SchemaException {
    if (pre != null) group.accept(pre);
    Iterator it = group.iterator();
    while (it.hasNext()) ((SchemaVisitable)it.next()).accept(this);
    if (post != null) group.accept(post);
  }
  
}
