/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema;

/**
 * Abstract class for schema component such as Declaration and Type.
 *
 * <p>This is the implementation of the common propertoes of Declaration,
 * Type, and IdentityConstraint. The following properties are defined :</p>
 * <p> 1. name : a NCName as defined by [XML-Namespaces].</p>
 * <p> 2. target namespace : represented by a schema deinition. </p>
 *
 * <p>See XML Schema Part 1: Structures 3.2, 3.3, 3.4, 3.11, 3.14
 * W3C Recommendation 2 May 2001</p>
 *
 * @see org.xquark.xml.schema.Declaration
 * @see org.xquark.xml.schema.Type
 * @see org.xquark.xml.schema.AttributeGroupDefinition
 */
public abstract class SchemaComponent 
  implements SchemaConstants, SchemaVisitable {
private static final String RCSRevision = "$Revision: 1.1 $"; 
private static final String RCSName     = "$Name:  $";
  
  protected Schema schema;
  protected String name;

  /**
   * Create a new SchemaComponent.
   *
   * @param schema The schema which contains this component.
   * @param name The component name, a NCName.
   */
  protected SchemaComponent(Schema schema, String name) {
    this.schema = schema;
    this.name = name;
  }

  public void accept(SchemaVisitor visitor) throws SchemaException {
    visitor.visit(this);
  }
/*
  public String toString() {
    return "#<"+getClass().getName()+" "+getNamespace()+":"+getName()+">";
  }
*/
  public Schema getSchema() { 
    return schema; 
  }
    
  /**
   * Return the component's name, if any.
   *
   * @return The component's name, or null if there is none.
   */
  public String getName() { 
    return name; 
  }
  
  /**
   * Set the component's name.
   */
  public void setName(String name) { this.name = name; }

  /**
   * Return the component's name, if any.
   *
   * @return The component's name, or null if there is none.
   */
   public String toString() {
    return name;
  }
  
  /**
   * Return the component's target namespace, if any.
   *
   * @return The target namespace, or null if the schema is null.
   */
  public String getNamespace() { 
    if (schema == null) return null;
    return schema.getNamespace();
  }

  /**
   * Returns the component's schema manager
   *
   * @return The schema manager, or null is the schema is null
   */
  public SchemaManager getManager() {
    if (schema == null) return null;
    return schema.getManager();
  }

  /**
   * @return true if component is named and has the requested name
   */
  
  public boolean hasName(String namespace, String localName) {
    if (namespace != null && namespace.length() == 0) namespace = null;
    return localName != null && localName.equals(getName()) &&
      ((namespace == null && getNamespace() == null) || 
       (namespace != null && namespace.equals(getNamespace())));
  }

}
