/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema;

import java.io.IOException;
import java.util.Iterator;
import java.util.Map;

import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;
import org.xquark.schema.loader.AttributeDeclarationRef;
import org.xquark.schema.loader.ElementDeclarationRef;
import org.xquark.schema.loader.SchemaComponentRef;

public class SchemaReader extends org.xquark.util.DefaultXMLReader implements SchemaConstants {
private static final String RCSRevision = "$Revision: 1.1 $"; 
private static final String RCSName     = "$Name:  $";

  private SchemaManager manager;
  private ComplexType extensionComplexType = null;
  private int min = -1;
  private int max = -1;
  private boolean unbounded = false;
  private java.util.HashMap namespaceMap = new java.util.HashMap();
  private boolean topLevelGroup = false;
  private int depthCount = 0;
  private AttributesImpl atts = new AttributesImpl();
  private AttributesImpl emptyAtts = new AttributesImpl();
  private IterationVisitor visitor = null;

  public SchemaReader(SchemaManager manager) {
    this.manager = manager;
  }

  public void parse(org.xml.sax.InputSource input) throws SAXException, IOException
  {
    parse(input.getSystemId());
  }
  
  public void parse(String systemId) throws SAXException, IOException
  {
    contentHandler.startDocument();
    Schema schema = manager.getSchema(systemId);
    parse (schema, true) ;
  }
  
  // add from dntt
  public void parse(Schema schema) throws SAXException, java.io.IOException
  {
    parse (schema, true) ;
  }

  public void parse(Schema schema, boolean withStart) 
    throws SAXException, java.io.IOException
  {
    if (withStart)
        contentHandler.startDocument();
    
    try {
      visitor = new SerializingIterationVisitor(new StartElementVisitor(), new EndElementVisitor());
      schema.accept(visitor);
    }
    catch ( SchemaException se ) {
      throw new SAXException(se);
    }
      
    if (withStart)
        contentHandler.endDocument();
  }
  
  public void parse(SchemaComponent component) throws SAXException, IOException
  {
    try {
      visitor = new SerializingIterationVisitor(new StartElementVisitor(), new EndElementVisitor());
      component.accept(visitor);
    }
    catch ( SchemaException se ) {
      throw new SAXException(se);
    }
  }
  
  public void setSchemaPrefixMap(String prefix, String namespace) {
    namespaceMap.put(namespace, prefix);
  }
  
  private String getRef(String uri, String localName) {
    String prefix = (String)namespaceMap.get(uri);
    if ( prefix == null ) prefix = "";
    if ( !"".equals(prefix) ) prefix += ":";		
    return prefix+localName;
  }

  private String getRef(SchemaComponent comp) {
    return getRef(comp.getNamespace(), comp.getName());
  }

  private void addOccurrence(AttributesImpl atts) {
    if (!visitor.isTopLevel()) {
      if (min != 1) 
	atts.addAttribute("", MIN_OCCURS_ATTR, null, "CDATA", String.valueOf(min));
      if (unbounded) 
	atts.addAttribute("", MAX_OCCURS_ATTR, null, "CDATA", UNBOUNDED_VALUE);
      else if (max != 1) 
	atts.addAttribute("", MAX_OCCURS_ATTR, null, "CDATA", String.valueOf(max));
    }
  }

  private void addForm(Declaration decl, AttributesImpl atts, String defForm) {
    if (decl.getScope() != null && decl.getSchema() != null ) {
      String defaultValue = decl.getSchema().getDefault(defForm);
      if (defaultValue == null || defaultValue.equals(UNQUALIFIED_VALUE)) {
	if (decl.getNamespace() != null) 
	 atts.addAttribute("", FORM_ATTR, null, "CDATA", QUALIFIED_VALUE);
      } 
      else if (decl.getNamespace() == null) 
      atts.addAttribute("", FORM_ATTR, null, "CDATA", UNQUALIFIED_VALUE);
    }
  }
  
  class StartElementVisitor extends DefaultSchemaVisitor {
	private static final String RCSRevision = "$Revision: 1.1 $"; 
	private static final String RCSName     = "$Name:  $";

    public void visit(Schema schema) throws SchemaException {
      Iterator it = schema.getPrefixMap().entrySet().iterator();
      while (it.hasNext()) {
				Map.Entry entry = (Map.Entry)it.next();
        if ( ! XML_PREFIX.equals(entry.getKey()) ) {
          try {
            getContentHandler().startPrefixMapping((String)entry.getKey(), (String)entry.getValue());
          }
          catch ( SAXException se ) {
            throw new SchemaException(se.getMessage());
          }
          namespaceMap.put(entry.getValue(), entry.getKey());
        }
      }
      atts.clear();
      atts.addAttribute("", TARGET_NAMESPACE_ATTR, null, "CDATA", schema.getNamespace());
			
      String value = schema.getDefault(ATTRIBUTE_FORM_DEFAULT_ATTR);
      if (value != null) 
				atts.addAttribute("", ATTRIBUTE_FORM_DEFAULT_ATTR, null, "CDATA", value);
			
      value = schema.getDefault(BLOCK_DEFAULT_ATTR);
      if (value != null) 
				atts.addAttribute("", BLOCK_DEFAULT_ATTR, null, "CDATA", value);
			
      value = schema.getDefault(ELEMENT_FORM_DEFAULT_ATTR);
      if (value != null) 
				atts.addAttribute("", ELEMENT_FORM_DEFAULT_ATTR, null, "CDATA", value);
			
      value = schema.getDefault(FINAL_DEFAULT_ATTR);
      if (value != null) 
				atts.addAttribute("", FINAL_DEFAULT_ATTR, null, "CDATA", value);
			
      try {
        getContentHandler().startElement(XMLSCHEMA_URI, SCHEMA_TAG, "", atts);
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
      generateImports(schema);
    }
    
    private void generateImports(Schema schema) throws SchemaException {
      Iterator it = schema.getPrefixMap().values().iterator();
      while (it.hasNext()) {
				String ns = (String)it.next();
				if (!ns.equals(XMLSCHEMA_URI) && schema.isImported(ns)) {
					atts.clear();
					atts.addAttribute("", NAMESPACE_ATTR, null, "CDATA", ns);
          try {
            getContentHandler().startElement(XMLSCHEMA_URI, IMPORT_TAG, "", atts);
            getContentHandler().endElement(XMLSCHEMA_URI, IMPORT_TAG, "");
          }
          catch ( SAXException se ) {
            throw new SchemaException(se.getMessage());
          }
				}
      }
    }
    
		//             Created 
		//             schema 2000-10-24
		//
    public void visit(SimpleType type) throws SchemaException {
      try {
        
			// get type name
			atts.clear();
			if ( type.getName() != null ) {
				atts.addAttribute("", NAME_ATTR, null, "CDATA", type.getName());
				getContentHandler().startElement(XMLSCHEMA_URI, SIMPLE_TYPE_TAG, "", atts);
				atts.clear();
			}
			else {
				getContentHandler().startElement(XMLSCHEMA_URI, SIMPLE_TYPE_TAG, "", atts);
			}
						
			// restriction, list, or union
			switch ( type.getVariety() ) {
				
				case SimpleType.VARIETY_ATOMIC :
					// restriction
					Type base = type.getBaseType();
					if ( base != null ) {
						if ( base.getName() != null ) {
							atts.addAttribute("", BASE_ATTR, null, "CDATA", getRef(base));
							getContentHandler().startElement(XMLSCHEMA_URI, RESTRICTION_TAG, "", atts);
							atts.clear();
						}
					}
					
					// facets
					Iterator it = type.getAllFacets().iterator();
					while (it.hasNext()) {
						Facet facet = (Facet)it.next();
            if ( facet.getEnumFacets().size() != 0 ) {
              String name = facet.getName();
              Iterator fIt = facet.getEnumFacets().iterator();
              while ( fIt.hasNext() ) {
                String value = (String)fIt.next();
                atts.addAttribute("", VALUE_ATTR, null, "CDATA", value);
                getContentHandler().startElement(XMLSCHEMA_URI, name, "", atts);
                atts.clear();
                getContentHandler().endElement(XMLSCHEMA_URI, name, "");  
              }
            }
            else {
              String name = facet.getName();
              String value = facet.getValue();
              atts.addAttribute("", VALUE_ATTR, null, "CDATA", value);
              getContentHandler().startElement(XMLSCHEMA_URI, name, "", atts);
              atts.clear();
              getContentHandler().endElement(XMLSCHEMA_URI, name, "");  
            }
          }
 
					break;
					
				case SimpleType.VARIETY_LIST :
          
          if ( type.getItemType().getName() != null ) {
            atts.addAttribute("", ITEM_TYPE_ATTR, null, "CDATA", getRef(type.getItemType()));
            getContentHandler().startElement(XMLSCHEMA_URI, LIST_TAG, "", atts);
            atts.clear();
          }
          else {
            getContentHandler().startElement(XMLSCHEMA_URI, LIST_TAG, "", atts);
            parse(type.getItemType());
          }
                    
					break;
					
				case SimpleType.VARIETY_UNION :
          
          if ( type.getMemberTypes() != null ) {
            String attr = null;
            for ( int i = 0; i < type.getMemberTypes().size(); i++ ) {
              SimpleType member = (SimpleType)type.getMemberTypes().get(i);
              if ( member.getName() != null ) {
                if ( i == 0 ) attr = getRef(member) + " ";
                else attr += getRef(member) + " ";
              }
            }
            atts.addAttribute("", MEMBER_TYPES_ATTR, null, "CDATA", attr);
            if ( attr != null )
              getContentHandler().startElement(XMLSCHEMA_URI, UNION_TAG, "", atts);
            else {
              atts.clear();
              getContentHandler().startElement(XMLSCHEMA_URI, UNION_TAG, "", atts);
            }
            atts.clear();
            for ( int i = 0; i < type.getMemberTypes().size(); i++ ) {
              SimpleType member = (SimpleType)type.getMemberTypes().get(i);
              if ( member.getName() == null ) {
                 parse(member);
              }
            }
          }
          break;
			}
      
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
      catch ( java.io.IOException se ) {
        throw new SchemaException(se.getMessage());
      }
		}
  
		//             schema 2000-10-24
		//
    public void visit(ComplexType type) throws SchemaException {

      try {
        //  get name 
        atts.clear();
        
        if ( type.getName() != null ) 
          atts.addAttribute("", NAME_ATTR, null, "CDATA", type.getName());

        if ( type.isAbstract() ) 
          atts.addAttribute("", ABSTRACT_ATTR, null, "CDATA", TRUE_VALUE);

        if ( type.getContentModel().getContentType() == MIXED )
          atts.addAttribute("", MIXED_VALUE, null, "CDATA", TRUE_VALUE);

        getContentHandler().startElement(XMLSCHEMA_URI, COMPLEX_TYPE_TAG, "", atts);

        // simpleContex
        Object model = type.getContentModel().getModel();
        if ( model instanceof SimpleType ) {
            getContentHandler().startElement(XMLSCHEMA_URI, SIMPLECONTENT_TAG, "", atts);
            atts.clear();
            atts.addAttribute("", BASE_ATTR, null, "CDATA", getRef((SimpleType)model));
            getContentHandler().startElement(XMLSCHEMA_URI, EXTENSION_TAG, "", atts);          
        }

        // complexContent, extension
        else if ( type.isExtension() && type.getBaseType() != null ) {
          getContentHandler().startElement(XMLSCHEMA_URI, COMPLEXCONTENT_TAG, "", emptyAtts);          
          atts.clear();
          extensionComplexType = type;
          atts.addAttribute("", BASE_ATTR, null, "CDATA", getRef(type.getBaseType()));
          getContentHandler().startElement(XMLSCHEMA_URI, EXTENSION_TAG, "", atts);
        }

        // complexContent, restriction
        else if ( type.isRestriction() && type.getBaseType() != null ) {
          getContentHandler().startElement(XMLSCHEMA_URI, COMPLEXCONTENT_TAG, "", emptyAtts);          
          atts.clear();
          atts.addAttribute("", BASE_ATTR, null, "CDATA", getRef(type.getBaseType()));
          getContentHandler().startElement(XMLSCHEMA_URI, RESTRICTION_TAG, "", atts);          
        }

        topLevelGroup = true;      
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
    }

    public void visit(ElementDeclaration decl) throws SchemaException {
      // check if this element is alread visited in base type
      if ( extensionComplexType != null ) {
        ComplexType base = (ComplexType)extensionComplexType.getBaseType();
        if ( base.getElementDeclaration(decl.getNamespace(), decl.getName()) != null )
          return;
      }
      
      atts.clear();
      
      if ( !visitor.isTopLevel() && decl.getScope().isGlobalScope() ) {
        atts.addAttribute("", REF_ATTR, null, "CDATA", getRef(decl));
      }
      else {
        Type type = decl.getType();			
        if (decl.getName() != null) 
          atts.addAttribute("", NAME_ATTR, null, "CDATA", decl.getName());
        if (type != null && type.getName() != null)
          atts.addAttribute("", TYPE_ATTR, null, "CDATA", getRef(type));
        if (decl.isAbstract()) 
          atts.addAttribute("", ABSTRACT_ATTR, null, "CDATA", TRUE_VALUE);
        if (decl.isNillable()) 
          atts.addAttribute("", NILLABLE_ATTR, null, "CDATA", TRUE_VALUE);
        
        if ( decl.getDefaultValue() != null )
          atts.addAttribute("", DEFAULT_ATTR, null, "CDATA", decl.getDefaultValue());
        if ( decl.getFixedValue() != null )
          atts.addAttribute("", FIXED_ATTR, null, "CDATA", decl.getFixedValue());
//      addForm(decl, atts, ELEMENT_FORM_DEFAULT_ATTR);
      }			
			
      addOccurrence(atts);
				
      try {
        getContentHandler().startElement(XMLSCHEMA_URI, ELEMENT_TAG, "", atts);
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
    }

    public void visit(ElementDeclarationRef decl) throws SchemaException {
			atts.clear();
      atts.addAttribute("", REF_ATTR, null, "CDATA", getRef(decl));
      if (decl.isNillable()) 
				atts.addAttribute("", NILLABLE_ATTR, null, "CDATA", TRUE_VALUE);
      if ( decl.getDefaultValue() != null )
   			atts.addAttribute("", DEFAULT_ATTR, null, "CDATA", decl.getDefaultValue());
      if ( decl.getFixedValue() != null )
   			atts.addAttribute("", FIXED_ATTR, null, "CDATA", decl.getFixedValue());
      addOccurrence(atts);
      try {
        getContentHandler().startElement(XMLSCHEMA_URI, ELEMENT_TAG, "", atts);
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
    }

    public void visit(AttributeDeclaration decl) throws SchemaException {
			atts.clear();
      Type type = decl.getType();
      if (decl.getName() != null) 
				atts.addAttribute("", NAME_ATTR, null, "CDATA", decl.getName());
      if (type != null && type.getName() != null)
				atts.addAttribute("", TYPE_ATTR, null, "CDATA", getRef(type));
			
//      addForm(decl, atts, ATTRIBUTE_FORM_DEFAULT_ATTR);
      if ( decl.getDefaultValue() != null )
   			atts.addAttribute("", DEFAULT_ATTR, null, "CDATA", decl.getDefaultValue());
      if ( decl.getFixedValue() != null )
   			atts.addAttribute("", FIXED_ATTR, null, "CDATA", decl.getFixedValue());
			if ( decl.getUseString() != null && !OPTIONAL_VALUE.equals(decl.getUseString()) ) 
				atts.addAttribute("", USE_ATTR, null, "CDATA", decl.getUseString());
			
      try {
        getContentHandler().startElement(XMLSCHEMA_URI, ATTRIBUTE_TAG, "", atts);
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
    }
  
    public void visit(AttributeDeclarationRef decl) throws SchemaException {
			atts.clear();
      atts.addAttribute("", REF_ATTR, null, "CDATA", getRef(decl));
      if ( decl.getDefaultValue() != null )
   			atts.addAttribute("", DEFAULT_ATTR, null, "CDATA", decl.getDefaultValue());
      if ( decl.getFixedValue() != null )
   			atts.addAttribute("", FIXED_ATTR, null, "CDATA", decl.getFixedValue());
			if ( decl.getUseString() !=null ) 
				atts.addAttribute("", USE_ATTR, null, "CDATA", decl.getUseString());
			
      try {
        getContentHandler().startElement(XMLSCHEMA_URI, ATTRIBUTE_TAG, "", atts);
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
    }
  
    public void visit(SchemaComponentRef ref) throws SchemaException {
      try {
			atts.clear();
      atts.addAttribute("", REF_ATTR, null, "CDATA", getRef(ref));
      if (ref.getType() == Schema.MODEL_GROUP_DEFINITION) {
				addOccurrence(atts);
				getContentHandler().startElement(XMLSCHEMA_URI, GROUP_TAG, "", atts);
      } 
			else if (ref.getType() == Schema.ATTRIBUTE_GROUP_DEFINITION) {
				getContentHandler().startElement(XMLSCHEMA_URI, ATTRIBUTE_GROUP_TAG, "", atts);
      }	
      
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
    }
    
    public void visit(ModelGroupDefinition def) throws SchemaException {
			atts.clear();
      atts.addAttribute("", NAME_ATTR, null, "CDATA", def.getName());
      try {
        getContentHandler().startElement(XMLSCHEMA_URI, GROUP_TAG, "", atts);
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
      topLevelGroup = true;
    }
    
    public void visit(AttributeGroupDefinition def) throws SchemaException {
			atts.clear();
      atts.addAttribute("", NAME_ATTR, null, "CDATA", def.getName());
      try {
        getContentHandler().startElement(XMLSCHEMA_URI, ATTRIBUTE_GROUP_TAG, "", atts);
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
    }
  
    public void visit(Wildcard wildcard) throws SchemaException {
      try {
			atts.clear();
      if (!visitor.isAttributeWildcard()) {
				addOccurrence(atts);
				getContentHandler().startElement(XMLSCHEMA_URI, ANY_TAG, "", atts);
      } 
			else {
				getContentHandler().startElement(XMLSCHEMA_URI, ANY_ATTRIBUTE_TAG, "", atts);
      }
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
    }
  
    public void visit(ContentModel model) throws SchemaException {
			atts.clear();
      Object p = model.getModel();
      if ( p != null && !(p instanceof SimpleType) ) {
        if ( ((Particle)p).getTerm() != null && ((Particle)p).getTerm() instanceof ElementDeclaration )
          try {
            getContentHandler().startElement(XMLSCHEMA_URI, SEQUENCE_TAG, "", atts);
          }
          catch ( SAXException se ) {
            throw new SchemaException(se.getMessage());
          }
      }
    }
    
    public void visit(Particle particle) throws SchemaException {
      min = particle.getMinOccurs();
      max = particle.getMaxOccurs();
      unbounded = particle.isMaxOccursUnbounded();
    }
    
    public void visit(AllModelGroup group) throws SchemaException {
			atts.clear();
      addOccurrence(atts);
      try {
        getContentHandler().startElement(XMLSCHEMA_URI, ALL_TAG, "", atts);
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
      depthCount++;
      topLevelGroup = false;
    }
    
    public void visit(ChoiceModelGroup group) throws SchemaException {
			atts.clear();
      addOccurrence(atts);
      try {
        getContentHandler().startElement(XMLSCHEMA_URI, CHOICE_TAG, "", atts);
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
      depthCount++;
      topLevelGroup = false;
    }
  
    public void visit(SequenceModelGroup group) throws SchemaException {
			atts.clear();
      addOccurrence(atts);
      try {
        getContentHandler().startElement(XMLSCHEMA_URI, SEQUENCE_TAG, "", atts);
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
      topLevelGroup = false;
    }
  }

  class EndElementVisitor extends DefaultSchemaVisitor {
	private static final String RCSRevision = "$Revision: 1.1 $"; 
	private static final String RCSName     = "$Name:  $";

    public void visit(Schema schema) throws SchemaException {
      try {
        getContentHandler().endElement(XMLSCHEMA_URI, SCHEMA_TAG, "");
        Iterator it = schema.getPrefixMap().entrySet().iterator();
        while (it.hasNext()) {
          Map.Entry entry = (Map.Entry)it.next();
          getContentHandler().endPrefixMapping((String)entry.getKey());
        }
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
    }
    
    public void visit(SimpleType type) throws SchemaException {
      try {
			// restriction, list, or union
			switch ( type.getVariety() ) {
				
				case SimpleType.VARIETY_ATOMIC :
          Type base = type.getBaseType();
          if ( base != null ) {
            getContentHandler().endElement(XMLSCHEMA_URI, RESTRICTION_TAG, "");
          }
          break;
          
 				case SimpleType.VARIETY_LIST :
          getContentHandler().endElement(XMLSCHEMA_URI, LIST_TAG, "");
          break;
          
 				case SimpleType.VARIETY_UNION :
          if ( type.getMemberTypes() != null )
            getContentHandler().endElement(XMLSCHEMA_URI, UNION_TAG, "");
          break;
      }

      getContentHandler().endElement(XMLSCHEMA_URI, SIMPLE_TYPE_TAG, "");
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }

    }
  
    public void visit(ComplexType type) throws SchemaException {
      try {
      Object model = type.getContentModel().getModel();
      // simpleContent
      if ( model instanceof SimpleType ) {
        getContentHandler().endElement(XMLSCHEMA_URI, EXTENSION_TAG, "");
        getContentHandler().endElement(XMLSCHEMA_URI, SIMPLECONTENT_TAG, "");
      }
      
      // complexContent, extension
      else if ( type.isExtension() && type.getBaseType() != null ) {
        extensionComplexType = null;
        getContentHandler().endElement(XMLSCHEMA_URI, EXTENSION_TAG, "");          
        getContentHandler().endElement(XMLSCHEMA_URI, COMPLEXCONTENT_TAG, "");          
      }
      
      // complexContent, restriction
      else if ( type.isRestriction() && type.getBaseType() != null ) {
        getContentHandler().endElement(XMLSCHEMA_URI, RESTRICTION_TAG, "");          
        getContentHandler().endElement(XMLSCHEMA_URI, COMPLEXCONTENT_TAG, "");          
      }
      
			getContentHandler().endElement(XMLSCHEMA_URI, COMPLEX_TYPE_TAG, "");
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
    }

    public void visit(ElementDeclaration decl) throws SchemaException {
      // check if this element is alread visited in base type
      if ( extensionComplexType != null ) {
        ComplexType base = (ComplexType)extensionComplexType.getBaseType();
        if ( base.getElementDeclaration(decl.getNamespace(), decl.getName()) != null )
          return;
      }
      
      try {
        getContentHandler().endElement(XMLSCHEMA_URI, ELEMENT_TAG, "");
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
    }

    public void visit(AttributeDeclaration decl) throws SchemaException {
      try {
        getContentHandler().endElement(XMLSCHEMA_URI, ATTRIBUTE_TAG, "");
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
    }
  
    public void visit(SchemaComponentRef ref) throws SchemaException {
      try {
        if (ref.getType() == Schema.MODEL_GROUP_DEFINITION) {
          getContentHandler().endElement(XMLSCHEMA_URI, GROUP_TAG, "");
        } 
        else if (ref.getType() == Schema.ATTRIBUTE_GROUP_DEFINITION) {
          getContentHandler().endElement(XMLSCHEMA_URI, ATTRIBUTE_GROUP_TAG, "");
        }	
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
    }
    
    public void visit(ModelGroupDefinition def) throws SchemaException {
      try {
        getContentHandler().endElement(XMLSCHEMA_URI, GROUP_TAG, "");
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
    }
    
    public void visit(AttributeGroupDefinition def) throws SchemaException {
      try {
        getContentHandler().endElement(XMLSCHEMA_URI, ATTRIBUTE_GROUP_TAG, "");
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
    }
  
    public void visit(Wildcard wildcard) throws SchemaException {
      try {
        if (!visitor.isAttributeWildcard()) getContentHandler().endElement(XMLSCHEMA_URI, ANY_TAG, "");
        else getContentHandler().endElement(XMLSCHEMA_URI, ANY_ATTRIBUTE_TAG, "");
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
    }
  
    public void visit(ContentModel model) throws SchemaException {
      try {
        Object p = model.getModel();
        if ( p != null && !(p instanceof SimpleType) ) {
          if ( ((Particle)p).getTerm() != null && ((Particle)p).getTerm() instanceof ElementDeclaration )
            getContentHandler().endElement(XMLSCHEMA_URI, SEQUENCE_TAG, "");
        }
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
    }

    public void visit(Particle particle) throws SchemaException {
      min = -1;
      max = -1;
      unbounded = false;
    }
    
    public void visit(AllModelGroup group) throws SchemaException {
      try {
        getContentHandler().endElement(XMLSCHEMA_URI, ALL_TAG, "");
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
      depthCount--;
    }
    
    public void visit(ChoiceModelGroup group) throws SchemaException {
      try {
        getContentHandler().endElement(XMLSCHEMA_URI, CHOICE_TAG, "");
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
    }
  
		// 26/01/2001  
    public void visit(SequenceModelGroup group) throws SchemaException {
      try {
        getContentHandler().endElement(XMLSCHEMA_URI, SEQUENCE_TAG, "");
      }
      catch ( SAXException se ) {
        throw new SchemaException(se.getMessage());
      }
    }
    
  }
  
}
