/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema.datatypes;

import java.text.ParseException;

import org.xquark.schema.SchemaException;
import org.xquark.schema.validation.ValidationContextProvider;

class Base64BinaryType extends MesureableType {
    private static final String RCSRevision = "$Revision: 1.1 $";
    private static final String RCSName = "$Name:  $";
    private static final byte PAD = (byte) '=';
    private static byte[] decodeMap = new byte[256];
    private static byte[] encodeMap = new byte[64];

    static {
        for (int i = 0; i < 256; i++) {
            decodeMap[i] = -1;
        }

        for (int i = 'A'; i <= 'Z'; i++)
            decodeMap[i] = (byte) (i - 'A');
        for (int i = 'a'; i <= 'z'; i++)
            decodeMap[i] = (byte) (i - 'a' + 26);
        for (int i = '0'; i <= '9'; i++)
            decodeMap[i] = (byte) (i - '0' + 52);
        decodeMap['+'] = 62;
        decodeMap['/'] = 63;

        for (int i = 0; i < 26; i++)
            encodeMap[i] = (byte) ('A' + i);
        for (int i = 26; i < 52; i++)
            encodeMap[i] = (byte) ('a' + i - 26);
        for (int i = 52; i < 62; i++)
            encodeMap[i] = (byte) ('0' + i - 52);
        encodeMap[62] = (byte) '+';
        encodeMap[63] = (byte) '/';
    }

    Base64BinaryType() {
        super("base64Binary", PrimitiveType.BASE64_BINARY);
    }

    public void checkFacets(Object valueSpace) throws SchemaException {
        super.checkFacets(valueSpace);
        super.checkLength(((ByteArray) valueSpace).getLength(), valueSpace);
    }

    protected Object toValueSpace(String value, ValidationContextProvider context) throws SchemaException {
        try {
            return new Value(decode(value));
        } catch (Exception ee) {
            super.invalidValue(value.toString());
            return null;
        }
    }

	private boolean isBase64(char c) {
	  return (decodeMap[c] != -1); 
	}

	private boolean isBase64Final2(char c) {
		return (decodeMap[c] & 0xF) == 0;
	}
	
	private boolean isBase64Final1(char c) {
		return (decodeMap[c] & 0x3) == 0;
	}

    private int computeSize(String value) throws ParseException {
        int count = 0;
        int charLen = value.length();
        char c;
        char prev = 0;
        int prevIndex = 0;
        int padCount = 0;
        for (int i = 0; i < charLen; i++) {
			c = value.charAt(i);
            if (!Character.isWhitespace(c)) {
            	if (padCount > 0 && (c != PAD || padCount > 1)) {
					throw new ParseException("Illegal character after padding in base64 encoded string", i);
            	} else if (c == PAD) {
            		padCount++;
                } else if (!isBase64(c)) {
					throw new ParseException("Invalid characted in base64 encoded string", i);
                } else {
					prev = c;
					prevIndex = i;
					count++;
                }
            }
        }
        if ((count+padCount) % 4 > 0)
			throw new ParseException("base64 encoded string has invalid length", charLen);
        switch (padCount) {
        	case 0:
				return 3*(count/4);
			case 1:
				if (!isBase64Final1(prev))
					throw new ParseException("Invalid characted in base64 encoded string", prevIndex);
				return 3*(count/4)+2;
			case 2:
			default:
				if (!isBase64Final2(prev))
					throw new ParseException("Invalid characted in base64 encoded string", prevIndex);
				return 3*(count/4)+1;
        }
    }

    public byte[] decode(String value) throws ParseException {
    	int decLen = computeSize(value);
		byte[] decodedData = new byte[decLen];
		int charIndex = 0;
		for (int i = 0; i < decLen/3; i++) {
			int tmp = 0;
			for (int j = 0; j < 4; j++) {
				char c = value.charAt(charIndex++);
				while (Character.isWhitespace(c)) 
					c = value.charAt(charIndex++);
				tmp |= (decodeMap[c] << (6*(3-j)));
			}
			decodedData[3*i] = (byte)((tmp >> 16) & 0xFF);
			decodedData[3*i+1] = (byte)((tmp >> 8) & 0xFF);
			decodedData[3*i+2] = (byte)(tmp & 0xFF);
		}
		if (decLen%3 == 1) {
			int tmp = 0;
			for (int j = 0; j < 2; j++) {
				char c = value.charAt(charIndex++);
				while (Character.isWhitespace(c)) 
					c = value.charAt(charIndex++);
				tmp |= (decodeMap[c] << (6*(3-j)));
			}
			decodedData[decLen-1] = (byte)((tmp >> 16) & 0xFF);
		} else if (decLen%3 == 2) {
			int tmp = 0;
			for (int j = 0; j < 3; j++) {
				char c = value.charAt(charIndex++);
				while (Character.isWhitespace(c)) 
					c = value.charAt(charIndex++);
				tmp |= (decodeMap[c] << (6*(3-j)));
			}
			decodedData[decLen-2] = (byte)((tmp >> 16) & 0xFF);
			decodedData[decLen-1] = (byte)((tmp >> 8) & 0xFF);
		}
		return decodedData;
    }

    public String encode(byte[] data) {
    	int decLen = data.length;
    	int encLen = (decLen % 3 == 0) ? 4*(decLen/3) : 4*(decLen/3+1);
    	// The encoded string is split into lines of 76 characters
    	encLen += (encLen-1)/76;
    	byte[] encodedData = new byte[encLen];
    	int nLines = 0;
    	for (int i = 0; i < decLen/3; i++) {
    		int tmp = ((0xFF&data[3*i]) << 16) | ((0xFF&data[3*i+1]) << 8) | (0xFF&data[3*i+2]);
    		int encIndex = 4*i+nLines;
    		encodedData[encIndex] = encodeMap[(byte)((tmp >> 18) & 0x3F)];
			encodedData[encIndex+1] = encodeMap[(byte)((tmp >> 12) & 0x3F)];
			encodedData[encIndex+2] = encodeMap[(byte)((tmp >> 6) & 0x3F)];
			encodedData[encIndex+3] = encodeMap[(byte)(tmp & 0x3F)];
			if ((i+1) % 19 == 0 && encIndex + 4 < encLen) {
				encodedData[encIndex+4] = (byte)0xA;
				nLines++;
			}
    	}
    	if (decLen % 3 == 1) {
			int tmp = (0xFF&data[decLen-1]) << 16;
    		encodedData[encLen - 4] = encodeMap[(byte)((tmp >> 18) & 0x3F)];
			encodedData[encLen - 3] = encodeMap[(byte)((tmp >> 12) & 0x3F)];
    		encodedData[encLen - 2] = encodedData[encLen - 1] = PAD;
    	} else if (decLen %3 == 2) {
			int tmp = ((0xFF&data[decLen-2]) << 16) | ((0xFF&data[decLen-1]) << 8);
			encodedData[encLen - 4] = encodeMap[(byte)((tmp >> 18) & 0x3F)];
			encodedData[encLen - 3] = encodeMap[(byte)((tmp >> 12) & 0x3F)];
			encodedData[encLen - 2] = encodeMap[(byte)((tmp >> 6) & 0x3F)];
			encodedData[encLen - 1] = PAD;
    	}
		return new String(encodedData);    	
    }

    class Value extends ByteArray {

        Value(byte[] value) {
            super(value);
        }

        protected boolean checkClass(Object obj) {
            return obj instanceof Value;
        }

        public String toString() {
            return encode(getData());
        }

    }
}
