/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema.datatypes;

import java.util.Date;

/**
 * This class represents the actual value of a date- and time-related XML type. It extends the 
 * standard java.util.Date class with optional time zone information. When specified, the time zone
 * information allows the DateTime object to represents an absolute date
 */
public class DateTime extends Date {
	
	public static final int MAX_TIMEZONE_OFFSET = 14 * 60 * 60 * 1000;
	/**
	 * tz is the difference with GMT in milliseconds.
	 */
	private long tz = 0;
	/**
	 * tzSpecified is true when timezone is known.
	 */
	private boolean tzSpecified = false;

    /**
     * @param millis milliseconds since "epoch" 
     */
    public DateTime(long millis) {
        this(millis, 0, false);
    }
	
	/**
	 * 
	 * @param millis milliseconds since epoch
	 * @param tz timezone difference with GMT in milliseconds
	 */
	public DateTime(long millis, long tz) {
		this(millis, tz, true);
	}
	
	/**
	 * 
	 * @param millis milliseconds since epoch
	 * @param tz timezone difference with GMT in milliseconds
	 * @param tzSpecified whether tz is significant or not
	 */
	public DateTime(long millis, long tz, boolean tzSpecified) {
		super(millis-tz);
		this.tz = tz;
		this.tzSpecified = tzSpecified;
	}
	
	public boolean equals(Object obj) {
		if (this == obj) return true;
		if (!checkClass(obj)) return false;
		return this.tzSpecified == ((DateTime)obj).tzSpecified && this.getTime() == ((DateTime)obj).getTime();
	}

	protected boolean checkClass(Object obj) {
		return obj instanceof DateTime;	
	}
	
	public boolean hasTimeZone() {
		return tzSpecified;
	}
	
	public long getTimeZone() {
		return tz;
	}
	/**
	 * Implements the compareTo method of interface Comparable
	 * @param obj The object to be compared to this object. It must be of class
	 * java.util.Date or of a subclass. Objects which are not of class DateTime or of a
	 * subclass are considered as being without timezone information.
	 * @return -1 if this object is smaller than obj, 0 if it is equal, and 1 if it is greater
	 * @exception ClassCastException if the object to be compared is not castable to a Date
	 * @exception IllegalArgumentException if the comparison between the two Date objects is indeterminate
	 * (see XML Schema Datatypes specification for details)
	 */
	public int compareTo(Object obj) throws ClassCastException, IllegalArgumentException {
		return compareTo((Date) obj);
	}

	public int compareTo(Date other) throws IllegalArgumentException {
		long min1, min2, max1, max2;
		min1 = max1 = getTime();
		min2 = max2 = other.getTime();
		boolean tz1 = tzSpecified;
		boolean tz2 = (other instanceof DateTime) && ((DateTime)other).tzSpecified;
		if (tz1 == tz2) {
			if (min1 < min2) return -1;
			else if (min1 > min2) return 1;
			else return 0;						
		} else if (!tz1) {
			min1 -= MAX_TIMEZONE_OFFSET;
			max1 += MAX_TIMEZONE_OFFSET;
		} else {
			min2 -= MAX_TIMEZONE_OFFSET;
			max2 += MAX_TIMEZONE_OFFSET;
		}
		if (max1 < min2) return -1;
		else if (max2 < min1) return 1;
		else throw new IllegalArgumentException("Indeterminate comparison result for these arguments");
	}
}
