/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema.datatypes;

import java.text.ParseException;

public class Duration implements Comparable {

    private static final String RCSRevision = "$Revision: 1.1 $";
    private static final String RCSName = "$Name:  $";

    private static long MILLISINMIN = 60000L;
    private static long MILLISINHOUR = MILLISINMIN * 60;
    private static long MILLISINDAY = MILLISINHOUR * 24;
    private static int MONTHSINYEAR = 12;
    private static int[] MINDAYSINMONTHS = { 0, 28, 59, 89, 120, 150, 181, 212, 242, 273, 303, 334 };
    private static int[] MAXDAYSINMONTHS = { 0, 31, 62, 92, 123, 153, 184, 215, 245, 276, 306, 337 };
    private static int[] POWERSOFTEN = { 1, 10, 100, 1000 };
  	private static int MILLISECOND_FIELD = 1 << 0;
	private static int MINUTE_FIELD = 1 << 1;
	private static int HOUR_FIELD = 1 << 2;
	private static int DAY_FIELD = 1 << 3;
	private static int MONTH_FIELD = 1 << 4;
	private static int YEAR_FIELD = 1 << 5;
	private static int SIGN_FIELD = 1 << 6;
  	
  	private int fields = 0;
    private int sign = 1;
    private int year = 0;
    private int month = 0;
    private int day = 0;
    private int hour = 0;
    private int minute = 0;
    private long millisecond = 0;

    public Duration(String value) throws ParseException {
        parseDuration(value); // get YY, MM, DD, HH, MM, SS value from string value
    }
    
    public Duration(int sign, int year, int month, int day, int hour, int minute, long millisecond) {
    	this.sign = sign;
    	this.year = year;
    	this.month = month;
    	this.day = day;
    	this.hour = hour;
    	this.minute = minute;
    	this.millisecond = millisecond;
    }

	/**
	 * Implements the compareTo method of interface Comparable
	 * @param obj The object to be compared to this object. It must be of the same
	 * class or of a subclass.
	 * @return -1 if this object is smaller than obj, 0 if it is equal, and 1 if it is greater
	 * @exception ClassCastException if the object to be compared is not castable to a Duration
	 * @exception IllegalArgumentException if the comparison between the two Duration objects is indeterminate
	 * (see XML Schema Datatypes specification for details)
	 */
    public int compareTo(Object obj) throws ClassCastException, IllegalArgumentException {
        return compareTo((Duration) obj);
    }

    private int compareTo(Duration other) throws IllegalArgumentException {
        if (this.sign < other.sign) {
            return -1;
        } else if (this.sign > other.sign) {
            return 1;
        } else {
			long deltaMillis = this.getDayTime()-other.getDayTime();
			int deltaMonths = this.getYearMonth()-other.getYearMonth();
        	if (deltaMonths == 0) {
            	return (deltaMillis == 0) ? 0 : ((deltaMillis < 0) ? -this.sign : this.sign);
	        } else if (deltaMillis == 0) {
	            return (deltaMonths < 0) ? -this.sign : this.sign;
        	} else {
	        	long min = 0;
				long max = 0;
        		if (deltaMonths > 0) {
	        		min = getMinDaysFromMonths(deltaMonths) * MILLISINDAY;
	        		max = getMaxDaysFromMonths(deltaMonths) * MILLISINDAY;
        		} else {
					max = -getMinDaysFromMonths(-deltaMonths) * MILLISINDAY;
					min = -getMaxDaysFromMonths(-deltaMonths) * MILLISINDAY;
        		}
		        if (max+deltaMillis < 0)
    	            return -this.sign;
        	    else if (min+deltaMillis > 0)
            	   	return this.sign;
           		else
               		throw new IllegalArgumentException("Indeterminate comparison result for these arguments");
        	}
        }
    }

    private int getYearMonth() {
        return year * MONTHSINYEAR + month;
    }

    private long getDayTime() {
        return day * MILLISINDAY + hour * MILLISINHOUR + minute * MILLISINMIN + millisecond;
    }

    private static int getMinDaysFromMonths(int nMonths) {
        int m = nMonths % MONTHSINYEAR;
        int y = nMonths / MONTHSINYEAR;
        return y * 365 + MINDAYSINMONTHS[m] + getMinLeapYears(y + ((m > 0) ? 1 : 0));
    }

    private static int getMaxDaysFromMonths(int nMonths) {
        int m = nMonths % MONTHSINYEAR;
        int y = nMonths / MONTHSINYEAR;
        return y * 365 + MAXDAYSINMONTHS[m] + getMaxLeapYears(y);
    }

    private static int getMinLeapYears(int y) {
        if (y < 8)
            return 0;
        else if (y < 300)
            return y / 4 - (y / 100 + 1);
        else if (y < 400)
            return y / 4 - y / 100;
        else
            return y / 400 * 97 + getMinLeapYears(y % 400);
    }

    private static int getMaxLeapYears(int y) {
        int part = (y % 4 == 0) ? 0 : 1;
        if (y < 200)
            return y / 4 + part;
        else if (y < 400)
            return y / 4 + part - (y / 100 - 1);
        else
            return y / 400 * 97 + getMaxLeapYears(y % 400);
    }

    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (!(obj instanceof Duration))
            return false;
        Duration other = (Duration) obj;
        return this.sign == other.sign
            && this.getYearMonth() == other.getYearMonth()
            && this.getDayTime() == other.getDayTime();
    }

    public int hashCode() {
        long m = getDayTime();
        int highMillis = (int) (m >> 32);
        int lowMillis = (int) (m & 0x00000000ffffffffL);
        return sign ^ getYearMonth() ^ highMillis ^ lowMillis;
    }

	private void setField(int field) {
		fields |= field;
	}
	
	private boolean hasField(int field) {
		return (fields & field) != 0;
	}
	
    public String toString() {
        StringBuffer print = new StringBuffer(20);
        if (sign == -1)
            print.append('-');
        print.append('P');
        if (hasField(YEAR_FIELD)) {
            print.append(year);
            print.append('Y');
        }
        if (hasField(MONTH_FIELD)) {
            print.append(month);
            print.append('M');
        }
        if (hasField(DAY_FIELD)) {
            print.append(day);
            print.append('D');
        }
        boolean inTime = false;
        if (hasField(HOUR_FIELD)) {
            print.append('T');
            print.append(hour);
            print.append('H');
            inTime = true;
        }
        if (hasField(MINUTE_FIELD)) {
            if (!inTime)
                print.append('T');
            print.append(minute);
            print.append('M');
            inTime = true;
        }
        if (hasField(MILLISECOND_FIELD)) {
            if (!inTime)
                print.append('T');
            print.append(millisecond / 1000);
            int millis = (int) millisecond % 1000;
            if (millis > 0) {
                print.append('.');
                if (millis < 10) {
                    print.append("00");
                    print.append(millis);
                } else if (millis < 100) {
                    print.append("0");
                    print.append(millis);
                } else {
                    print.append(millis);
                }
                int len = print.length();
                while (print.charAt(len-1) == '0') {
                	print.setLength(len-1);
                	len--;
                }
            }
            print.append('S');
        }
        return print.toString();
    }

    private void parseDuration(String value) throws ParseException {
        int len = value.length();
        int index = 0;
        int fieldIndex = 0;
        if (len <= 2)
            throw new ParseException("Invalid duration specification: " + value, index);
        if (value.charAt(index) == '-') {
            sign = -1;
            index++;
			setField(SIGN_FIELD);
        }
        if (value.charAt(index) != 'P')
            throw new ParseException("Invalid duration specification: " + value, index);
        index++;
        while (index < len && value.charAt(index) != 'T') {
			int val = 0;
			int second = -1;
			int count = 0;
			char c = 0;
			while (index < len) {
				c = value.charAt(index++);
				if (!Character.isDigit(c)) break;
				val *= 10;
				val += ((int) c - (int) '0');
				count++;
			}
            if (c == 'Y') {
                if (count == 0 || fieldIndex > 0)
                    throw new ParseException("Invalid duration specification: " + value, index);
                year = val;
                fieldIndex = 1;
                setField(YEAR_FIELD);
            } else if (c == 'M') {
                if (count == 0 || fieldIndex > 1)
                    throw new ParseException("Invalid duration specification: " + value, index);
                month = val;
                fieldIndex = 2;
				setField(MONTH_FIELD);
            } else if (count == 0 || c == 'D') {
                if (fieldIndex > 2)
                    throw new ParseException("Invalid duration specification: " + value, index);
                day = val;
                fieldIndex = 3;
				setField(DAY_FIELD);
            } else {
                throw new ParseException("Invalid duration specification: " + value, index);
            }
        }
        if (index < len) {
        	index++;
            while (index < len) {
                int val = 0;
                int second = -1;
                int count = 0;
                char c = 0;
                while (index < len) {
					c = value.charAt(index++);
					if (!Character.isDigit(c)) break;
					val *= 10;
					val += ((int) c - (int) '0');
					count++;
                }
                if (c == 'H') {
                    if (count == 0 || fieldIndex > 3)
                        throw new ParseException("Invalid duration specification: " + value, index);
                    hour = val;
                    fieldIndex = 4;
					setField(HOUR_FIELD);
                } else if (c == 'M') {
                    if (count == 0 || fieldIndex > 4)
                        throw new ParseException("Invalid duration specification: " + value, index);
                    minute = val;
                    fieldIndex = 5;
					setField(MINUTE_FIELD);
                } else if (c == '.') {
                    if (count == 0 || fieldIndex > 5)
                        throw new ParseException("Invalid duration specification: " + value, index);
                    second = val;
                    fieldIndex = 6;
				} else if (c == 'S') {
					if (count == 0 || fieldIndex > 6)
						throw new ParseException("Invalid duration specification: " + value, index);
					if (second != -1) {
						if (count > 3) {
							// truncate milliseconds digits after the third
							val = (int)(val/Math.pow(10, count-3));
						} else {
							val *= POWERSOFTEN[3 - count];
						}
						millisecond = 1000*second+val;
					} else {
						millisecond = 1000*val;
					}
					fieldIndex = 7;
					setField(MILLISECOND_FIELD);
                } else {
                    throw new ParseException("Invalid duration specification: " + value, index);
                }
            }
            if (fieldIndex < 4 || fieldIndex == 6)
				throw new ParseException("Invalid duration specification: " + value, index);
        } else if (fieldIndex == 0) {
			throw new ParseException("Invalid duration specification: " + value, index);
        }
    }
}
