/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema.datatypes;

import java.text.ParseException;

import org.xquark.schema.SchemaException;
import org.xquark.schema.validation.ValidationContextProvider;


class HexBinaryType extends MesureableType {
    private static final String RCSRevision = "$Revision: 1.1 $";
    private static final String RCSName = "$Name:  $";
	private static byte[] decodeMap = new byte[256]; 
	private static byte[] encodeMap = new byte[16];

	static {
	  for (int i = 0; i < 256; i++) {
		decodeMap[i] = -1; 
	  }

	  for (int i = '0'; i <= '9'; i++)
		decodeMap[i] = (byte)(i-'0');
	  for (int i = 'A'; i <= 'F'; i++)
		decodeMap[i] = (byte)(i-'A'+10);
	  for (int i = 'a'; i <= 'f'; i++)
		decodeMap[i] = (byte)(i-'a'+10);
    
	  for (int i = 0; i < 10; i++)
		encodeMap[i] = (byte)('0'+i);
	  for (int i = 10; i < 16; i++)
		encodeMap[i] = (byte)('A'+i-10);
	}

    HexBinaryType() {
        super("hexBinary", PrimitiveType.HEX_BINARY);
    }

	public void checkFacets(Object valueSpace) throws SchemaException {
		super.checkFacets(valueSpace);
		super.checkLength(((ByteArray) valueSpace).getLength(), valueSpace);
	}

	protected Object toValueSpace(String value, ValidationContextProvider context) throws SchemaException {
		try {
			return new Value(decode(value));
		} catch (Exception ee) {
			super.invalidValue(value.toString());
			return null;
		}
	}

	private boolean isHexChar(char c) {
	  return (decodeMap[c] != -1); 
	}

	private int computeSize(String value) throws ParseException {
		int count = 0;
		int charLen = value.length();
		for (int i = 0; i < charLen; i++) {
			char c = value.charAt(i);
			if (!Character.isWhitespace(c)) {
				if (!isHexChar(c)) {
					throw new ParseException("Invalid characted in hexadeciaml encoded string", i);
				}
				count++;
			}
		}
		if (count % 2 > 0)
			throw new ParseException("Hexadecimal encoded string has invalid length", charLen);
		return count/2;
	}

	public byte[] decode(String value) throws ParseException {
		int decLen = computeSize(value);
		byte[] decodedData = new byte[decLen];
		int charIndex = 0;
		for (int i = 0; i < decLen; i++) {
			int tmp = 0;
			for (int j = 0; j < 2; j++) {
				char c = value.charAt(charIndex++);
				while (Character.isWhitespace(c)) 
					c = value.charAt(charIndex++);
				tmp |= (decodeMap[c] << (4*(1-j)));
			}
			decodedData[i] = (byte)(tmp & 0xFF);
		}
		return decodedData;
	}

	public String encode(byte[] data) {
		int decLen = data.length;
		int encLen = 2*decLen;
		// The encoded string is split into lines of 76 characters
		encLen += (encLen-1)/76;
		byte[] encodedData = new byte[encLen];
		int nLines = 0;
		for (int i = 0; i < decLen; i++) {
			int encIndex = 2*i+nLines;
			encodedData[encIndex] = encodeMap[(byte)((data[i] >> 4) & 0xF)];
			encodedData[encIndex+1] = encodeMap[(byte)(data[i] & 0xF)];
			if ((i+1) % 19 == 0 && encIndex + 2 < encLen) {
				encodedData[encIndex+2] = (byte)0xA;
				nLines++;
			}
		}
		return new String(encodedData);    	
	}

	class Value extends ByteArray {

		Value(byte[] value) {
			super(value);
		}

		protected boolean checkClass(Object obj) {
			return obj instanceof Value;
		}

		public String toString() {
			return encode(getData());
		}

	}
}
