/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema.loader;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.StringTokenizer;

import org.xml.sax.*;
import org.xquark.schema.*;

public class Loader extends org.xquark.util.SAXLoader
implements SchemaConstants, ErrorHandler {
private static final String RCSRevision = "$Revision: 1.1 $"; 
private static final String RCSName     = "$Name:  $";
  
  private Schema schema = null;
  private boolean topLevel = false;
  private SchemaManager manager;
  private SchemaLocator locator;
  private boolean noIncludeTargetNamespace = false;
  
  private HashMap componentContextLocators = new HashMap();
  
  public Loader(SchemaManager manager) {
    this(manager, manager, null);
  }
  
  public Loader(SchemaManager manager, SchemaLocator locator) {
    this(manager, locator, null);
  }
  
  public Loader(SchemaManager manager, SchemaLocator locator, Schema schema) {
    super(null);
    this.manager = manager;
    this.locator = locator;
    this.schema = schema;
    if (schema == null) topLevel = true;
    setElementHandler(new SchemaHandler());
  }
  
  public Schema getSchema() throws SAXException {
    return schema;
  }
  
  public String getDocumentBase() {
    if (getDocumentLocator() != null)
      return getDocumentLocator().getSystemId();
    return null;
  }
    
  public Schema load(URL url) throws java.io.IOException, SAXException {
    org.xml.sax.InputSource source = new org.xml.sax.InputSource(url.openStream());
    source.setSystemId(url.toString());
    return load(source);
  }
  
  public Schema load(org.xml.sax.InputSource source) throws SAXException, java.io.IOException {
    org.xml.sax.XMLReader reader = locator.getSchemaReader();
    reader.setContentHandler(this);
    reader.setErrorHandler(this);
    reader.parse(source);    
    return schema;
  }
  
  private void setComponentContextLocators(HashMap map) {
    componentContextLocators = map;
  }
  
  private Schema resolvePrefix(String qname)
  throws SchemaException {
    String prefix = "";
    int index = qname.indexOf(':');
    if (index > 0) prefix = qname.substring(0, index);
    String ns = getPrefixMapping(prefix);
    if (ns == null && !prefix.equals("")) {
//    "Unknown namespace prefix: " + prefix;
      throw new SchemaException("src-qname.1.1");
    }
    
    if ((ns == null && schema.getNamespace() == null)
    || (ns != null && ns.equals(schema.getNamespace())))
      return schema;   
    if ( ns == null && noIncludeTargetNamespace ) // it's a inlcuded schema without namespace
      return schema;
    if (schema.isImported(ns))
      return manager.getSchema(ns);
    
//  "No schema for namespace: " + ns;
    throw new SchemaException("src-resolve.4");
  }
  
  // initilization schema
  private void initialize() throws SAXException {
    try {
      InitializationVisitor initializer = new InitializationVisitor(schema, componentContextLocators, this);

      java.util.Iterator it = schema.getTypes().iterator();
      while (it.hasNext()) ((SchemaVisitable)it.next()).accept(initializer);

      it = schema.getElementDeclarations().iterator();
      while (it.hasNext()) ((SchemaVisitable)it.next()).accept(initializer);

      it = schema.getAttributeDeclarations().iterator();
      while (it.hasNext()) ((SchemaVisitable)it.next()).accept(initializer);
      
      it = schema.getAttributeGroupDefinitions().iterator();
      while (it.hasNext()) ((SchemaVisitable)it.next()).accept(initializer);
    }
    catch ( SchemaException se ) {
      Object invalidObject = se.getInvalidObject();
      Locator componentLocator = (Locator)componentContextLocators.get(invalidObject);
      throw new SAXParseException(se.getMessage(), componentLocator, se);
    }
  }
  
  public void addComp2ComponentContextLocators(Object newComp, Object oldComp) {
      Object componentLocator = componentContextLocators.get(oldComp);
      if ( componentLocator != null )
        componentContextLocators.put(newComp, componentLocator);
  }
  
  public org.xml.sax.Locator getComponentLocaltor(Object obj) {
    return (org.xml.sax.Locator)componentContextLocators.get(obj);
  }
    
  public boolean isSchemaNamespace(String uri) {
    return uri.equals(XMLSCHEMA_URI);
  }
    
  private String getLocalName(String qname) {
    int index = qname.indexOf(':');
    if (index > 0) return qname.substring(index+1);
    else return qname;
  }
  
  private Loader getLoader() {
    return this;
  }
  
  public Type getType(String ref) throws SchemaException {
    if (ref == null) return null;
    
    TypeRef typeRef = new TypeRef(resolvePrefix(ref), getLocalName(ref));
    
    // store namespace context and locator
    ContextLocatorImpl contextLoc = new ContextLocatorImpl(super.getNamespaceContext(), getDocumentBase(), super.getDocumentLocator());
    componentContextLocators.put(typeRef, contextLoc);
    
    return typeRef;
  }
  
  public SchemaComponentRef getRef(String ref, int type) throws SchemaException {
    if (ref == null) return null;
    
    SchemaComponentRef compRef = new SchemaComponentRef(resolvePrefix(ref), getLocalName(ref), type);
        
    // store namespace context and locator
    ContextLocatorImpl contextLoc = new ContextLocatorImpl(super.getNamespaceContext(), getDocumentBase(), super.getDocumentLocator());
    componentContextLocators.put(compRef, contextLoc);
    
    return compRef;
  }
    
  public Facet buildFacet(String localName, Attributes atts) {
    Facet result = null;
    for (int i = 0; i < FACETS.length; i++) {
      if (FACETS[i].equals(localName)) {
        result = new Facet(localName, atts.getValue("", VALUE_ATTR), atts.getValue("", FIXED_ATTR));
        break;
      }
    }
    
    // store namespace context and locator
    ContextLocatorImpl contextLoc = new ContextLocatorImpl(super.getNamespaceContext(), getDocumentBase(), super.getDocumentLocator());
    componentContextLocators.put(result, contextLoc);
    
    return result;
  }
  
  public NotationDeclaration buildNotation(Attributes atts) {
    String name = atts.getValue("", NAME_ATTR);
    String publicId = atts.getValue("", PUBLIC_ATTR);
    String systemId = atts.getValue("", SYSTEM_ATTR);
    NotationDeclaration result = new NotationDeclaration(schema, name, publicId, systemId);
    
    // store namespace context and locator
    ContextLocatorImpl contextLoc = new ContextLocatorImpl(super.getNamespaceContext(), getDocumentBase(), super.getDocumentLocator());
    componentContextLocators.put(result, contextLoc);

    return result;
  }    
  
  private int getMethod(String method, int mask) {
    StringTokenizer st = new StringTokenizer(method);
    int result = 0;
    while (st.hasMoreTokens()) {
      String tok = st.nextToken();
      if ((LIST & mask) != 0 && LIST_VALUE.equals(tok))
        result |= LIST;
      else if ((UNION & mask) != 0 && UNION_VALUE.equals(tok))
        result |= UNION;
      else if ((RESTRICTION & mask) != 0 && RESTRICTION_VALUE.equals(tok))
        result |= RESTRICTION;
      else if ((EXTENSION & mask) != 0 && EXTENSION_VALUE.equals(tok))
        result |= EXTENSION;
      else if ((SUBSTITUTION & mask) != 0 && SUBSTITUTION_VALUE.equals(tok))
        result |= SUBSTITUTION;
      else if (ALL_VALUE.equals(tok))
        return mask;
    }
    return result;
  }
  
  private void setOccurrence(Particle particle, Attributes atts)
  throws SchemaException {
    String minOccurs = atts.getValue("", MIN_OCCURS_ATTR);
    if (minOccurs != null) {
      try {
        particle.setMinOccurs(Integer.parseInt(minOccurs));
      }
      catch (NumberFormatException ex) {
          // @@ should not occur (schema)
      }
    }
    
    String maxOccurs = atts.getValue("", MAX_OCCURS_ATTR);
    if (maxOccurs != null) {
      if (maxOccurs.equals(UNBOUNDED_VALUE)) {
        particle.setMaxOccursUnbounded();
      }
      else {
        try {
          particle.setMaxOccurs(Integer.parseInt(maxOccurs));
        }
        catch (NumberFormatException ex) {
          // @@ should not occur (schema)
        }
      }
    }
    
    // check minOccurs and maxOccurs
    int nMin = particle.getMinOccurs();
    int nMax = particle.getMaxOccurs();
    
    if ( nMax < nMin ) {
//    "minOccurs must not be greater than maxOccurs.";
      throw new SchemaException("p-props-correct.2.1");
    }
  }
  
  public Particle buildParticle(String localName, SchemaScope scope, Attributes atts)
  throws SchemaException {
    Particle result = null;
    if (localName.equals(ELEMENT_TAG)) {
      result = new Particle(buildElementDeclaration(atts, scope));
    }
    else if (localName.equals(GROUP_TAG)) {
      String refName = atts.getValue("", REF_ATTR);
      result = new Particle(getRef(refName, Schema.MODEL_GROUP_DEFINITION));
    }
    else if (localName.equals(SEQUENCE_TAG)) {
      result = new Particle(new SequenceModelGroup());
    }
    else if (localName.equals(CHOICE_TAG)) {
      result = new Particle(new ChoiceModelGroup());
    }
    else if (localName.equals(ALL_TAG)) {
      result = new Particle(new AllModelGroup());
    }
    else if (localName.equals(ANY_TAG)) {
      result = new Particle(buildWildcard(atts));
    }
    
    if (result != null) {
      setOccurrence(result, atts);
    }
    
    // store namespace context and locator
    ContextLocatorImpl contextLoc = new ContextLocatorImpl(super.getNamespaceContext(), getDocumentBase(), super.getDocumentLocator());
    componentContextLocators.put(result, contextLoc);
    
    return result;
  }
  
  public Wildcard buildWildcard(Attributes atts)
  throws SchemaException {
    Wildcard result = new Wildcard(schema, true);
    String namespaces = atts.getValue("", NAMESPACE_ATTR);
    if (namespaces != null) {
      if (OTHER_VALUE.equals(namespaces)) {
        result.exclude(schema.getNamespace());
      } else if (!ANY_VALUE.equals(namespaces)) {
        StringTokenizer st = new StringTokenizer(namespaces);
        while (st.hasMoreTokens()) {
          String tok = st.nextToken();
          if (TARGET_NAMESPACE_VALUE.equals(tok)) {
            result.add(schema.getNamespace());
          } else if (LOCAL_VALUE.equals(tok)) {
            result.add(null);
          } else {
            result.add(tok);
          }
        }
      }
    }
    String process = atts.getValue("", PROCESS_CONTENTS_ATTR);
    if (process != null) {
      if (LAX_VALUE.equals(process)) result.setProcessContents(LAX);
      else if (SKIP_VALUE.equals(process)) result.setProcessContents(SKIP);
      else if (STRICT_VALUE.equals(process)) result.setProcessContents(STRICT);
    }
    
    // store namespace context and locator
    ContextLocatorImpl contextLoc = new ContextLocatorImpl(super.getNamespaceContext(), getDocumentBase(), super.getDocumentLocator());
    componentContextLocators.put(result, contextLoc);

    return result;
  }
  
  // Schema creation
  private void initSchema(Attributes atts)
  throws SchemaException {
    String targetNamespace = atts.getValue("", TARGET_NAMESPACE_ATTR); 
    if ( XSI_URI.equals(targetNamespace) && 
         getDocumentBase() != null &&
         !getDocumentBase().endsWith("XMLSchema-instance.xsd")) {
        throw new SchemaException("no-xsi");
    }
    if (schema == null) {
      String schemaLocation = null;
      if (getDocumentLocator() != null)
        schemaLocation = getDocumentLocator().getSystemId();
      schema = new Schema(targetNamespace, schemaLocation, manager);
    } else if (targetNamespace != null && !targetNamespace.equals(schema.getNamespace())) {
//    "Included target namespace must be equal to including namespace";
      throw new SchemaException("src-include.2.1");
    } else if ( targetNamespace == null ) {
      noIncludeTargetNamespace = true;
    }
    
    schema.setDefault(ATTRIBUTE_FORM_DEFAULT_ATTR,
    atts.getValue("", ATTRIBUTE_FORM_DEFAULT_ATTR));
    schema.setDefault(BLOCK_DEFAULT_ATTR,
    atts.getValue("", BLOCK_DEFAULT_ATTR));
    schema.setDefault(ELEMENT_FORM_DEFAULT_ATTR,
    atts.getValue("", ELEMENT_FORM_DEFAULT_ATTR));
    schema.setDefault(FINAL_DEFAULT_ATTR,
    atts.getValue("", FINAL_DEFAULT_ATTR));
    
    schema.setPrefixMap(getElementPrefixMap());
  }
  
  private URL resolveLocation(String location, String base) throws SAXException {
    URL context = null;
    if (base != null) {
      try {
        context = new URL(base);
      }
      catch (MalformedURLException ex0) {
	//@commented out to support invalid URL returned by a locator, in the case of import
        //reportLoadingError(new SchemaException("schema_reference"));
      }
    }
    
    try {
      return new URL(context, location);
    }
    catch (MalformedURLException ex1) {
      try {
        java.io.File file = new java.io.File(location);
        return file.toURL();
      }
      catch (MalformedURLException ex2) {
	//@commented out to support invalid URL returned by a locator, in the case of import
        //reportLoadingError(new SchemaException("schema_reference"));
      }
    }
    
    return null;
  }
    
  private void includeSchema(Attributes atts)
  throws SAXException {
    String location = atts.getValue("", SCHEMA_LOCATION_ATTR);
    location = locator.resolveInclude(location);
    String schemaLocation = null;
    if (getDocumentLocator() != null)
      schemaLocation = getDocumentLocator().getSystemId();
    
    URL url = null;
    url = resolveLocation(location, schemaLocation);
    if (url == null) 
      reportLoadingError("Could not resolve location "+location, new SchemaException("schema_reference"));
    location = url.toString();
    if (!schema.isIncluded(location)) {
      schema.addIncludedLocation(location);
      schema.pushDefaults();
      Loader loader = new Loader(manager, locator, schema);
      loader.setComponentContextLocators(componentContextLocators);
      try {
        loader.load(url);
      } catch ( java.io.IOException e) {
        // should be a warning : invalid included schema
//      "Invalid included schema "+schemaLocation;
        reportLoadingError(new SchemaException("src-include.1"));
      }
      schema.popDefaults();
    }
  }
  
  private void importSchema(Attributes atts) throws SAXException {
    String location = atts.getValue("", SCHEMA_LOCATION_ATTR);
    String schemaLocation = null;
    if (getDocumentLocator() != null)
      schemaLocation = getDocumentLocator().getSystemId();
    String ns = atts.getValue("", NAMESPACE_ATTR);
    if (!schema.isImported(ns)) {
      Schema imported = null;
      imported = manager.getSchema(ns);
      if (imported == null) {
          // catch the SAXException throwed by manager.loadSchema
          if (location != null) {
            URL url = null;
            url = resolveLocation(location, schemaLocation);
            if (url != null) {
              imported = manager.loadSchema(locator, ns, url);
            }
            if (imported == null) imported = manager.loadSchema(locator, ns);
          }
          else {
            URL url = null;
            url = resolveLocation(ns, schemaLocation);
            if (url != null) {
              imported = manager.loadSchema(locator, ns, url);
            }
            if (imported == null) imported = manager.loadSchema(locator, ns);
          }
      }
      if (imported == null) {
        // It is not an error for the application schema reference strategy to fail
        reportLoadingError(new SchemaException("src-import"));
      }
      schema.addImportedNamespace(ns);
    }
  }
  
  // history
  //              Simple type creation
  //              Schema 2000-10-24
  //              locator is added
  //
  public SimpleType buildSimpleType(Attributes atts)
  throws SchemaException {
    String strName = atts.getValue("", NAME_ATTR);
    String strId = atts.getValue("", ID_ATTR);
    
    SimpleType aType = new SimpleType(schema, strName, null);
    String fin = atts.getValue("", FINAL_ATTR);
    if ( fin == null ) fin = schema.getDefault(FINAL_DEFAULT_ATTR);
    if ( fin != null )
      aType.setFinal(getMethod(fin, RESTRICTION | EXTENSION | LIST | UNION));
    
    // store namespace context and locator
    ContextLocatorImpl contextLoc = new ContextLocatorImpl(super.getNamespaceContext(), getDocumentBase(), super.getDocumentLocator());
    componentContextLocators.put(aType, contextLoc);

    return aType;
  }
  
  // history
  //           Complex type creation
  //
  //           Schema 2000-10-24
  //           locator is added
  //
  //           Schema 2001-05
  public ComplexType buildComplexType(Attributes atts)
  throws SchemaException {
    String name = atts.getValue("", NAME_ATTR);
    ComplexType type = new ComplexType(schema, name, null);
    
    String booleanValue = atts.getValue("", ABSTRACT_ATTR);
    if ( TRUE_VALUE.equals(booleanValue) || ONE_VALUE.equals(booleanValue) )
      type.setAbstract(true);
    
    // schema 2000-10, only 2 content type :
    // 1. mixed = true, mixed content type
    // 2. mixed = false, or default, element only type
    // noted   15/01/01
    booleanValue = atts.getValue("", MIXED_VALUE);
    if ( TRUE_VALUE.equals(booleanValue) || ONE_VALUE.equals(booleanValue) )
      type.setContentModel(new ContentModel(MIXED));
    
    String block = atts.getValue("", BLOCK_ATTR);
    if ( block == null ) block = schema.getDefault(BLOCK_DEFAULT_ATTR);
    if (block != null ) type.setBlock(getMethod(block, RESTRICTION | EXTENSION));
    
    String fin = atts.getValue("", FINAL_ATTR);
    if ( fin == null ) fin = schema.getDefault(FINAL_DEFAULT_ATTR);
    if ( fin != null )
      type.setFinal(getMethod(fin, RESTRICTION | EXTENSION));
    
    // store namespace context and locator
    ContextLocatorImpl contextLoc = new ContextLocatorImpl(super.getNamespaceContext(), getDocumentBase(), super.getDocumentLocator());
    componentContextLocators.put(type, contextLoc);
    
    return type;
  }
  
  // Element declaration
  // schema 2001-05
  // changed by ZL, May 17 2001
  public ElementDeclaration buildElementDeclaration(Attributes atts, SchemaScope scope)
  throws SchemaException {
    
    boolean global = (scope != null && scope.isGlobalScope());
    String name = atts.getValue("", NAME_ATTR);
    String ref = atts.getValue("", REF_ATTR);
    ElementDeclaration decl = null;
    if (name != null) {
      if ( ref != null ) {
        throw new SchemaException("src-element.2.1");
      }
      decl = new ElementDeclaration(schema, name, scope);
      decl.setType(getType(atts.getValue("", TYPE_ATTR)));
      if (global) {
        decl.setQualified(true);
        decl.setSubstitutionGroup(getRef(atts.getValue("", SUBSTITUTION_GROUP_ATTR), Schema.ELEMENT_DECLARATION));
        String fin = atts.getValue("", FINAL_ATTR);
        if (fin == null) fin = schema.getDefault(FINAL_DEFAULT_ATTR);
        if (fin != null) decl.setFinal(getMethod(fin, RESTRICTION | EXTENSION));        
      }
      else {
        String form = atts.getValue("", FORM_ATTR);
        if (form == null) form = schema.getDefault(ELEMENT_FORM_DEFAULT_ATTR);
        if (QUALIFIED_VALUE.equals(form)) decl.setQualified(true);
      }
      
    String booleanValue = atts.getValue("", ABSTRACT_ATTR);
    if ( TRUE_VALUE.equals(booleanValue) || ONE_VALUE.equals(booleanValue) )
        decl.setAbstract(true);
      
      String block = atts.getValue("", BLOCK_ATTR);
      if (block == null) block = schema.getDefault(BLOCK_DEFAULT_ATTR);
      if (block != null) decl.setBlock(getMethod(block, RESTRICTION | EXTENSION | SUBSTITUTION));
      
    } else {
      if ( global ) {
        // @@ known from schema
      }
      else if ( ref == null ) {
//      "one of <name> or <ref> must be present in element declaration";
        throw new SchemaException("src-element.2.1");
      }
      else {
        if ( atts.getValue("", NILLABLE_ATTR) != null ||
        atts.getValue("", DEFAULT_ATTR) != null ||
        atts.getValue("", FIXED_ATTR) != null ||
        atts.getValue("", FORM_ATTR) != null ||
        atts.getValue("", BLOCK_ATTR) != null ||
        atts.getValue("", TYPE_ATTR) != null ) {
//        "if <ref> is present, all of <nillable, default, fixed, form, block, final, type> must be absent";
          throw new SchemaException("src-element.2.2");
        }
        decl = new ElementDeclarationRef(resolvePrefix(ref), getLocalName(ref), scope);
        decl.setQualified(true);
      }
    }
    
    String booleanValue = atts.getValue("", NILLABLE_ATTR);
    if ( TRUE_VALUE.equals(booleanValue) || ONE_VALUE.equals(booleanValue) )
      decl.setNillable(true);
    
    // schema 2001-05, value contrraint
    String defaultValue = atts.getValue("", DEFAULT_ATTR);
    String fixedValue = atts.getValue("", FIXED_ATTR);
    
    if ( defaultValue != null ) {
      decl.setDefaultValue(defaultValue);
      if ( fixedValue != null ) {
//       "<default> and <fixed> must not both be present in attribute declaration";
        throw new SchemaException("src-element.1");
      }
    }
    else if ( fixedValue != null ) {
      decl.setFixedValue(fixedValue);
    }
    
    // store namespace context and locator
    ContextLocatorImpl contextLoc = new ContextLocatorImpl(super.getNamespaceContext(), getDocumentBase(), super.getDocumentLocator());
    componentContextLocators.put(decl, contextLoc);
    
    return decl;
  }
  
  // schema 2001-05
  // changed by ZL, May 16 2001
  public AttributeDeclaration buildAttributeDeclaration(Attributes atts, SchemaScope scope)
  throws SchemaException {
    
    boolean global = (scope != null && scope.isGlobalScope());
    String name = atts.getValue("", NAME_ATTR);
    String ref = atts.getValue("", REF_ATTR);
    AttributeDeclaration decl = null;
    
    // one of name or ref must be present, but not both
    if ( name != null ) {
      if ( XMLNS_PREFIX.equals(name) )
        throw new SchemaException("no-xmlns");
      if ( ref != null ) {
//      "one of <name> or <ref> must be present, but not both in attribute declaration";
        throw new SchemaException("src-attribute.3.1");
      }
      decl = new AttributeDeclaration(schema, name, scope);
      decl.setType(getType(atts.getValue("", TYPE_ATTR)));
      if ( global ) {
        // attribute element information item has schema as its parent
        decl.setQualified(true);
      } else {
        // local attribute
        String form = atts.getValue("", FORM_ATTR);
        if ( form == null )
          form = schema.getDefault(ATTRIBUTE_FORM_DEFAULT_ATTR);
        if ( QUALIFIED_VALUE.equals(form) ) decl.setQualified(true);
      }
    } else {
      // name is null, ref must not be null, and it's a local attribute
      if (ref == null) {
//      "one of name or ref must be present in attribute declaration";
        throw new SchemaException("src-attribute.3.1");
      }
      else {
        if (atts.getValue("", FORM_ATTR) != null ||
        atts.getValue("", TYPE_ATTR) != null ) {
//        "If <ref> is present, then all of <simpleType, form, type> must be absent";
          throw new SchemaException("src-attribute.3.2");
        }
        decl = new AttributeDeclarationRef(resolvePrefix(ref), getLocalName(ref), scope);
        decl.setQualified(true);
      }
    }
    
    // schema 2001-05, use defintion is different from that of schema 2000-10
    String use = atts.getValue("", USE_ATTR);
    if ( use != null ) {
      if ( OPTIONAL_VALUE.equals(use) )
        decl.setUse(OPTIONAL);
      else if ( PROHIBITED_VALUE.equals(use) )
        decl.setUse(PROHIBITED);
      else if ( REQUIRED_VALUE.equals(use) )
        decl.setUse(REQUIRED);
    }
    
    // schema 2001-05, value contrraint
    String defaultValue = atts.getValue("", DEFAULT_ATTR);
    String fixedValue = atts.getValue("", FIXED_ATTR);
    
    if ( defaultValue != null ) {
      if ( fixedValue != null ) {
//      "<default> and <fixed> must not both be present";
        throw new SchemaException("src-attribute.1");
      }
      if (use != null && !OPTIONAL_VALUE.equals(use)) {
//      "<use> must have the value <optional>";
        throw new SchemaException("src-attribute.2");
      }
      decl.setDefaultValue(defaultValue);
    }
    else if ( fixedValue != null ) {
      decl.setFixedValue(fixedValue);
    }
    
    // store namespace context and locator
    ContextLocatorImpl contextLoc = new ContextLocatorImpl(super.getNamespaceContext(), getDocumentBase(), super.getDocumentLocator());
    componentContextLocators.put(decl, contextLoc);
    
    return decl;
  }
  
  public ModelGroupDefinition buildModelGroupDefinition(Attributes atts, Schema schema) {
    String name = atts.getValue("", NAME_ATTR);
    ModelGroupDefinition def = new ModelGroupDefinition(schema, name);
    ContextLocatorImpl contextLoc = new ContextLocatorImpl(super.getNamespaceContext(), getDocumentBase(), super.getDocumentLocator());
    componentContextLocators.put(def, contextLoc);
    return def;
  }
  
  public AttributeGroupDefinition buildAttributeGroupDefinition(Attributes atts, Schema schema) {
    String name = atts.getValue("", NAME_ATTR);
    AttributeGroupDefinition def = new AttributeGroupDefinition(schema, name);
    ContextLocatorImpl contextLoc = new ContextLocatorImpl(super.getNamespaceContext(), getDocumentBase(), super.getDocumentLocator());
    componentContextLocators.put(def, contextLoc);
    return def;
  }
  
  // create Identiy-constraint
  // changed by ZL, May 16 2001
  public IdentityConstraint buildIdentityConstraint(Attributes atts,
              String localName, ElementDeclaration context)
  throws SchemaException {
     
    String name = atts.getValue("", NAME_ATTR);
    String refer = atts.getValue("", REFER_ATTR);
     
    IdentityConstraint identityConstraint = new IdentityConstraint(schema, name, context);
    if (localName.equals(KEY_TAG))
      identityConstraint.setCategory(IdentityConstraint.CATEGORY_KEY);
    else if (localName.equals(KEYREF_TAG))
      identityConstraint.setCategory(IdentityConstraint.CATEGORY_KEYREF);
    else if (localName.equals(UNIQUE_TAG))
      identityConstraint.setCategory(IdentityConstraint.CATEGORY_UNIQUE);
    
    // store namespace context and locator
    ContextLocatorImpl contextLoc = new ContextLocatorImpl(super.getNamespaceContext(), getDocumentBase(), super.getDocumentLocator());
    componentContextLocators.put(identityConstraint, contextLoc);
    
    if (identityConstraint.getCategory() ==  IdentityConstraint.CATEGORY_KEYREF) {
      IdentityConstraintRef icRef = new IdentityConstraintRef(resolvePrefix(refer), getLocalName(refer));
      componentContextLocators.put(icRef, contextLoc);
      identityConstraint.setReferencedKey(icRef);
    }
    
    
    return identityConstraint;
  }
  
  public java.util.List buildXPaths(String xpath) throws SchemaException {
    java.util.ArrayList result = new java.util.ArrayList(4);
    while (xpath != null) {
      int i = xpath.indexOf('|');
      String current = null;
      if (i == -1) {
        current = xpath;
        xpath = null;
      } else {
        current = xpath.substring(0, i);
        xpath = xpath.substring(i+1);
      }
      result.add(buildXPath(current));
    }
    return result;
  }
  
  public XPathExpr buildXPath(String xpath) throws SchemaException {
    boolean isAnyLevel = false;
    boolean isAttribute = false;
    if (xpath.startsWith(".//")) {
      isAnyLevel = true;
      xpath = xpath.substring(3);
    }
    XPathExpr result = new XPathExpr(isAnyLevel);
    while (xpath != null) {
      int i = xpath.indexOf('/');
      String qName = null;
      if (i == -1) {
        qName = xpath;
        xpath = null;
      } else {
        qName = xpath.substring(0, i);
        xpath = xpath.substring(i+1);
      }
      if (qName.startsWith("child::")) {
        qName = qName.substring(7);
      } else if (qName.startsWith("@")) {
        isAttribute = true;
        qName = qName.substring(1);
      } else if (qName.startsWith("attribute::")) {
        isAttribute = true;
        qName = qName.substring(11);
      }
      result.setAttribute(isAttribute);
      if (!qName.equals(".")) {
        String ns = null;
        String localName = null;
        int j = qName.indexOf(':');
        if (j == -1) {
          localName = qName;
          // use default prefix if defined, but only for named elements 
          if (!isAttribute && !"*".equals(qName))
            ns = getPrefixMapping("");
        } else {
          ns = getPrefixMapping(qName.substring(0, j));
          if ( ns == null ) 
            throw new SchemaException("src-identity-constraint");
          localName = qName.substring(j+1);
        }
        result.addStep(ns, localName);
      }
    }
    
    // store namespace context and locator
    ContextLocatorImpl contextLoc = new ContextLocatorImpl(super.getNamespaceContext(), getDocumentBase(), super.getDocumentLocator());
    componentContextLocators.put(result, contextLoc);
    
    return result;
  }
  
  public void warning(SAXParseException ex) throws SAXException {
    throw ex;
  }

  public void error(SAXParseException ex) throws SAXException {
    throw ex;
  }

  public void fatalError(SAXParseException ex) throws SAXException {
    throw ex;
  }
  
  public void notifyUnknownElement(String uriNamespace, String localName ) 
  throws SAXException {
    // should not be happened
    // all is verificated by Schema of schemas
    String errMsg = "Unknown element : {" + uriNamespace + "}" + localName; 
    reportLoadingError(errMsg, null);
  }
  
  public void reportLoadingError(SchemaException se) throws SAXException {
    reportLoadingError(null, se);
  }

  public void reportLoadingError(String message, SchemaException se) throws SAXException {
    if ( message != null && se != null ) message += "\n"+se.getMessage();
    throw new SAXParseException(message, super.getDocumentLocator(), se);
  }
  
  class SchemaHandler extends org.xquark.util.DefaultElementHandler {
	private static final String RCSRevision = "$Revision: 1.1 $"; 
	private static final String RCSName     = "$Name:  $";
    public org.xquark.util.ElementHandler startElement(String strNamespaceURI, String localName, Attributes atts) 
    throws SAXException {
      
    if (isSchemaNamespace(strNamespaceURI)) {

      if (localName.equals(SCHEMA_TAG)) {
        try {
          initSchema(atts);
        } 
        catch ( SchemaException se ) {
          String errMsg = "Error while processing localName -> " + localName; 
          reportLoadingError(errMsg, se);
        }
        return this;
      }

      else if (localName.equals(INCLUDE_TAG)) {
        includeSchema(atts);
        return this;
      }

      else if (localName.equals(IMPORT_TAG)) {
        importSchema(atts);
        return this;
      }

      else if (localName.equals(REDEFINE_TAG)) {
        includeSchema(atts);
        return new RedefineHandler(getLoader());
      }

      else if ( localName.equals(SIMPLE_TYPE_TAG) ) {
        SimpleType type = null;
        try {
          type = buildSimpleType(atts);
          schema.register(type);
        } 
        catch ( SchemaException se ) {
          String errMsg = "Error while processing localName -> " + localName; 
          reportLoadingError(errMsg, se);
        }
        return new SimpleTypeHandler(getLoader(), type, false, false);
      }

      else if (localName.equals(COMPLEX_TYPE_TAG)) {
        ComplexType type = null;
        try {
          type = buildComplexType(atts);
          schema.register(type);
        } 
        catch ( SchemaException se ) {
          String errMsg = "Error while processing localName -> " + localName; 
          reportLoadingError(errMsg, se);
        }
        return new ComplexTypeHandler(getLoader(), type, false, false);
      }

      else if (localName.equals(ELEMENT_TAG)) {
        ElementDeclaration decl = null;
        try {
          decl = buildElementDeclaration(atts, schema);
          schema.register(decl);
        } 
        catch ( SchemaException se ) {
          String errMsg = "Error while processing localName -> " + localName; 
          reportLoadingError(errMsg, se);
        }
        return new ElementTypeHandler(getLoader(), decl);
      }

      else if (localName.equals(ATTRIBUTE_TAG)) {
        AttributeDeclaration decl = null;
        try {
          decl = buildAttributeDeclaration(atts, schema);
          schema.register(decl);
        } 
        catch ( SchemaException se ) {
          String errMsg = "Error while processing localName -> " + localName; 
          reportLoadingError(errMsg, se);
        }
        return new AttributeTypeHandler(getLoader(), decl);
      }

      else if (localName.equals(GROUP_TAG)) {
        ModelGroupDefinition def = buildModelGroupDefinition(atts, schema);
        try {
          schema.register(def);
        } 
        catch ( SchemaException se ) {
          String errMsg = "Error while processing localName -> " + localName; 
          reportLoadingError(errMsg, se);
        }
        return new ContentModelHandler(getLoader(), def, false);
      }

      else if (localName.equals(ATTRIBUTE_GROUP_TAG)) {
        AttributeGroupDefinition def = buildAttributeGroupDefinition(atts, schema);
        try {
          schema.register(def);
        } 
        catch ( SchemaException se ) {
          String errMsg = "Error while processing localName -> " + localName; 
          reportLoadingError(errMsg, se);
        }
        return new AttributeGroupHandler(getLoader(), def);
      }

      else if (localName.equals(NOTATION_TAG)) {
        NotationDeclaration decl = buildNotation(atts);
        try {
          schema.register(decl);
        } 
        catch ( SchemaException se ) {
          String errMsg = "Error while processing localName -> " + localName; 
          reportLoadingError(errMsg, se);
        }
        return this;
      }

      else if (localName.equals(ANNOTATION_TAG)) {
        return new AnnotationHandler();
      }
    }

    notifyUnknownElement(strNamespaceURI, localName);
    return this;
    }
    
    public void endElement(String namespace, String localName)
    throws SAXException {
      super.endElement(namespace, localName);
      if ( isSchemaNamespace(namespace) && localName.equals(SCHEMA_TAG) && topLevel ) {
          initialize();
      }
    }
  }
  
}
