/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.schema.validation;

import java.util.Iterator;

import org.xquark.schema.*;

public class SequenceContentIterator extends ContentIterator {
private static final String RCSRevision = "$Revision: 1.1 $"; 
private static final String RCSName     = "$Name:  $";

  private SequenceModelGroup group;
  private Iterator groupIterator;
  private Particle currentParticle = null;
  private int count = 0;
  private ContentIterator iterator = null;
  
  public SequenceContentIterator(Particle particle)
  {
    super(particle);
    group = (SequenceModelGroup)particle.getTerm();
    groupIterator = group.iterator();
  }

  public ElementDeclaration getMatchedDeclaration() {
    if (iterator != null) 
      return iterator.getMatchedDeclaration();
    else return null;
  }

  public ElementDeclaration getModelDeclaration() {
    if (iterator != null) 
      return iterator.getModelDeclaration();
    else return null;
  }

  public int getProcessContents() {
    if (iterator != null) 
      return iterator.getProcessContents();
    else return LAX;
  }
  
  private ContentIterator selectContentIterator(String namespace, String localName)
  {
    // nTested is used in case all particles are optional, 
    // as this could lead to an infinite loop
    int nTested = 0;
    while (nTested < group.size()) {
      if (groupIterator.hasNext()) {
	Particle particle = (Particle)groupIterator.next();
	// first, check if local is directly found in particle
	if (particle.hasNextElement(namespace, localName)) {
	  if (currentParticle == null && count < maxOccurs && reporter != null) {
	    reporter.startGroup(parentParticle, ModelGroup.SEQUENCE, count);	
          }
	  currentParticle = particle;
          return ContentIteratorFactory.createIterator(currentParticle, reporter);
	}				
	else if (particle.getMinExtent() > 0) {
	  return null;
	} 
        
        // it's necessary to test if the parentParticle has this element
        // if not, too many particle will be skiped
        // noted  22, nov 2001
	else if ( reporter != null && parentParticle.hasNextElement(namespace, localName) ) {
	  reporter.skipped(particle);
	}
	nTested++;
      } 
      else {
	if (reporter != null) 
	  reporter.endGroup(parentParticle, ModelGroup.SEQUENCE);
	currentParticle = null;
	count++;
	groupIterator = group.iterator();
      }
    }
    return null;
  }
  
  protected int nextElement(String namespace, String localName)
  {
    // On-going iterator
    java.util.ArrayList previousExceptions = null;
    if (iterator != null) {
      switch (iterator.nextElement(namespace, localName)) {
      case MATCHED:
	return MATCHED;
      case UNKNOWN:
        return UNKNOWN;
      case INVALID:
	if (iterator.getMatchedDeclaration() != null) 
	  return invalidElement(iterator.getExceptions());
	else
	  return invalidElement("cvc-particle.3.3", parentParticle, iterator.getExceptions());
      case TOO_MANY:
        previousExceptions = iterator.getExceptions();
	// try next particle
      case DONE:
	// try next particle
      }
    }

    iterator = selectContentIterator(namespace, localName);
    if (iterator == null) {
      if (previousExceptions != null) {
	// Previous iterator had too many occurrences, and no new iterator matched
	// Report previous error
	return invalidElement("cvc-particle.3.3", parentParticle, previousExceptions);
      } else if (count == 0 && currentParticle == null && parentParticle.getMinExtent() > 0) {
	SchemaException se = SchemaException.computeMinExtentException(parentParticle);
	return invalidElement(se);
      } else {
	// Iterator is finished
	return stopProcessing();
      }
    }
    if (count >= maxOccurs) {
      if (previousExceptions != null)
	invalidElement("cvc-particle.3.2", parentParticle, previousExceptions);
      else 
	invalidElement("cvc-particle.3.2", parentParticle);
      return TOO_MANY;
    } 
    switch (iterator.nextElement(namespace, localName)) {
    case MATCHED:
      return MATCHED;
    case UNKNOWN:
      return UNKNOWN;
    case INVALID:
    case TOO_MANY:
    case DONE:
    default:
      if (iterator.getMatchedDeclaration() != null) 
	return invalidElement(iterator.getExceptions());
      else
	return invalidElement("cvc-particle.3.3", parentParticle, iterator.getExceptions());
    }
  }

  public int stopProcessing() {
    // Stop current iterator
    if (iterator != null) {
      switch (iterator.stopProcessing()) {
      case INVALID:
	return invalidElement("cvc-particle.3.3", parentParticle, iterator.getExceptions());
      case DONE:
	// Current iterator can be stopped. Continue
      }
    }
    int c;
    if (currentParticle == null) {
      // at beginning of sequence
      c = count;
    } else {
      // Can skip the rest of the group (all elements at the end are optional)
      Iterator it = group.iterator();
      // Move to current position in the group
      while (it.hasNext() && it.next() != currentParticle);
      while (it.hasNext()) {
	Particle particle = (Particle)it.next();
	// One particle at least is not optional
	if (particle.getMinExtent() > 0) {
	  SchemaException se = SchemaException.computeMinExtentException(particle);
	  return invalidElement("cvc-particle.3.3", parentParticle, se);
	} else if (reporter != null) {
	  reporter.skipped(particle);
	}
      }
      // end of group reached
      if (reporter != null)
	reporter.endGroup(parentParticle, ModelGroup.SEQUENCE);
      // increment the current count
      c = count+1;
    }
    if (c < minOccurs && parentParticle.getMinExtent() != 0) {
      return invalidElement("cvc-particle.3.1", parentParticle);
    } else if (reporter != null && c < maxOccurs) {
      // it's better to skip the parentParticle than the children
      // noted  22, Nov 2001
      reporter.skipped(parentParticle);
    }
    return DONE;
  }
 
  protected boolean nextValidElements(java.util.List elements) {
    // Iterator in use
    if (iterator != null) {
      boolean canStop = iterator.nextValidElements(elements);
      // If current iterator cannot be terminated immediately
      // return false
      if (!canStop) return false;
    }
    int c;
    if (currentParticle == null) {
      // at beginning of sequence
      c = count;
    } else {
      // else look for other possible elements in the rest of the group
      Iterator it = group.iterator();
      while (it.hasNext() && it.next() != currentParticle);
      while (it.hasNext()) {
	Particle particle = (Particle)it.next();
	if (elements != null) elements.addAll(particle.firstValidElements());
	// Particle is not optional
	if (particle.getMinExtent() > 0) return false;
      }
      c = count+1;
    }
    // If all previous elements are optional and loop is possible,
    // restart from the beginning of the group
    if (c < maxOccurs && elements != null) {
      Iterator it = group.iterator();
      while (it.hasNext()) {
	Particle particle = (Particle)it.next();
	elements.addAll(particle.firstValidElements());
	if (particle == currentParticle || particle.getMinExtent() > 0)
	  break;
      }
    }
    if (c >= minOccurs || parentParticle.getMinExtent() == 0) return true;
    else return false;
  }
}
