/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

/*
 * CompactXMLSerializer.java
 *
 * Created on 5 novembre 2001, 14:47
 */

package org.xquark.serialize;

import java.io.*;
import java.util.Arrays;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xquark.util.DefaultXMLFilter;
import org.xquark.util.SAXConstants;

/**
 * A simple SAX2 XML serializer implementing SAX ContentHandler and
 * LexicalHandler.
 *
 * IMPORTANT: SAX2 XMLReader
 * <code>http://xml.org/sax/features/namespaces</code> and
 * <code>http://xml.org/sax/features/namespace-prefixes</code> features must be
 * set to true.
 */
public class BasicXMLSerializer extends DefaultXMLFilter
{
    private static final String RCSRevision = "$Revision: 1.1 $";
    private static final String RCSName     = "$Name:  $";
    
    private final static int INDENT_CHAR_ARRAY_SIZE = 10;
    private final static char[] INDENT_CHAR_ARRAY = {'\t','\t','\t','\t','\t',
                                                     '\t','\t','\t','\t','\t'};
    private PrintWriter writer;
    private boolean indent = true;
    private boolean orderAttributes = false;
    private SortableAttributes orderedAtts;
    private boolean useIgnorableWhitespaces = true;
    private String encoding = null;
    private String userSetEncoding = null;
    private boolean generateXMLDeclaration = true;
    private boolean closeStreamAutomatically = false;
    
    
    /** element hierarchy depth*/
    private int depth = -1;
    /** max depth where indentation is performed */
    private int level = 0;
    
    private boolean cdata = false;
    private boolean elementPending = false;
    private boolean lastStartIsStart = false;
    private boolean charDataEncountered = false;
    private boolean extraDataEncountered = false;
    private boolean prologGenerated = false;
    private StringBuffer buf = new StringBuffer();
    

    BasicXMLSerializer() {}
    
    /** Creates new BasicXMLSerializer
     *
     * @param out the OutputStream where the serializer writes bytes. For
     * instance a FileOutputStream.
     */
    public BasicXMLSerializer(OutputStream out)
    throws UnsupportedEncodingException
    {
        setOutputStream(out);
    }
    
    /** Creates new BasicXMLSerializer
     *
     * @param out the writer where the serializer writes characters. For
     * instance a FileOutputStream.
     */
    public BasicXMLSerializer(Writer out)
    {
        setWriter(out);
    }
    
    /** Creates new BasicXMLSerializer
     *
     * @param out the writer where the serializer writes characters. For
     * instance a FileOutputStream.
     * @param encoding a string for the MIME encoding (must be supported by the
     * JDK used)
     * @see java.io.OutputStream
     */
    public BasicXMLSerializer(OutputStream out, String encoding)
    throws UnsupportedEncodingException
    {
        setDefaultEncoding(encoding);
        setOutputStream(out);
    }
    
    /** Creates new BasicXMLSerializer
     *
     * @param out the OutputStream where the serializer writes bytes. For
     * instance a FileOutputStream.
     * @param encoding a string for the MIME encoding (must be compatible with
     * the encoding used by the writer)
     */
    public BasicXMLSerializer(Writer out, String encoding)
    {
        this();
        setDefaultEncoding(encoding);
        setWriter(out);
    }
    
    /** Set the OutputStream where the serializer writes bytes.
     *
     * <B>Once this methods is called output Writer is created with the current
     * encoding and will not change.</B>
     *
     * @param out the OutputStream where the serializer writes bytes. For
     * instance a FileOutputStream.
     */
    public void setOutputStream(OutputStream out) throws UnsupportedEncodingException
    {
        Writer outputWriter;
        if (userSetEncoding == null)
        {
            if (out.equals(System.out) || out.equals(System.err))
            {
                encoding = null;
                outputWriter = new OutputStreamWriter(out);
            }
            else
            {
//                encoding = "UTF-8";
//                outputWriter = new OutputStreamWriter(out, encoding);
                encoding = null;
                outputWriter = new OutputStreamWriter(out, "UTF-8");
            }
        }
        else
        {
            encoding = userSetEncoding;
            outputWriter = new OutputStreamWriter(out, encoding);
        }
        
	writer = new PrintWriter(new BufferedWriter(outputWriter), true);
        
    }
    
    /** Set the Writer where the serializer writes characters.
     *
     * <B>Once this methods is called output encoding is set with the current
     * value and will not change.</B>
     *
     * @param out the writer where the serializer writes characters. For
     * instance a FileOutputStream.
     */
    public void setWriter(Writer out)
    {
      writer = new PrintWriter(out, true);
      
      encoding = userSetEncoding;
    }
    
    /** Set the indenting mode.
     * <B>Warning:</B>Default behavior is to perform automatic indentation (
     * ignorable whitespaces generated by the XML processor are ignored).
     *
     * @param indent if true, tabs are used at the beginning of lines to indent the output file.
     */
    public void setIndent(boolean indent)
    {
        this.indent = indent;
    }

    /** turn on or off the use of ignorable whitespace for indentation.
     * <B>Warning:</B>This flag is ignored if automatic indentation is on.
     *
     * @param indent if true (the default), ignorable whitespace are used for
     * indentation unless automatic indentation is on.
     */
    public void setUseIgnorableWhitespaces(boolean use)
    {
        useIgnorableWhitespaces = use;
    }
    
    /** Set the ordering mode for attributes.
     * The default behavior is to perform no ordering.
     *
     * @param order If true, ordering is performed following the
     * {@link <a href="http://www.w3.org/TR/xml-c14n">Canonical XML<a>} W3C
     * recommendation. If false, attributes are serialized as passed by the
     * SAX2 XMLReader.
     */
    public void setCanonicalOutput(boolean mode)
    {
        orderAttributes = mode;
    }
    
  /** Set the character encoding.
   *
   * @param encoding a string specifying the character encoding, which must be supported by the
   * JDK. Example: UTF-8, ISO-8859-1
   */
    public void setDefaultEncoding(String encoding)
    {
        this.userSetEncoding = encoding;
    }

    /** Enable XML declaration generation on startDocument() event.
     *
     * @param enable Declaration is geenrated if true. Default is true.
     */
    public void setGenerateXMLDeclaration(boolean enable)
    {
        generateXMLDeclaration = enable;
    }
    
    /** Enable automatic close of the output user stream on endDocument() event.
     *
     * @param close Output stream will be closed if true. Default is false.
     */
    public void setAutoStreamClose(boolean close)
    {
        closeStreamAutomatically = close;
    }
    
  /**
   * @return the indenting mode.
   */
    public boolean getIndent()
    {
        return indent;
    }

    /**
     * @return the status of the use of gnorable whitespace by the serializer.
     */
    public boolean getUseIgnorableWhitespaces()
    {
        return useIgnorableWhitespaces;
    }
    
    /** Accessor to the current ordering mode for attributes.
     * @return the current ordering mode for attributes.
     */
    public boolean getCanonicalOutput()
    {
        return orderAttributes;
    }
    
  /**
   * @return the character encoding in use, if previously specified.
   */
    public String getDefaultEncoding()
    {
        return userSetEncoding;
    }
    
    /** Accessor to the current flag for XML declaration generation.
     * @return if true, the XML declaration will be generated.
     */
    public boolean getGenerateXMLDeclaration()
    {
        return generateXMLDeclaration;
    }
    
    /** Enable automatic close of the output user stream on endDocument() event.
     *
     * @param close Output stream will be closed if true. Default is false.
     */
    public boolean getAutoStreamClose()
    {
        return closeStreamAutomatically;
    }
    
  /** Reset internal state before reuse.
   */
    public void reset()
    {
//        writer = null;
        level = 0;
        cdata = false;
        elementPending = false;
        lastStartIsStart = false;
        charDataEncountered = false;
        prologGenerated = false;
        depth = -1;
    }
    
  /** Closes the user stream.
   */
    public void close()
    {
        writer.close();
    }
    
    ////////////////////////////////////////////////////////////////////////////
    //                      Handlers Implementation
    ////////////////////////////////////////////////////////////////////////////
    
    public void startDocument() throws SAXException
    {
        if (generateXMLDeclaration)
        {
            if (encoding == null)
                writer.write("<?xml version=\"1.0\"?>");
            else
                writer.write("<?xml version=\"1.0\" encoding=\"" + encoding + "\"?>");
            prologGenerated = true;
        }
    }
    
    public void characters(char[] ch, int start, int length) throws SAXException
    {
        if (length <= 0)
            return;
        
        // Complete potential start markup
        completeStartTag();
        
        if (cdata)
            writer.write(ch, start, length);
        else
        {
            int max =  start + length;
            int i, last = start;
            for (i = start; i < max; i++)
            {
                if (isCharDataForbidden(ch[i]))
                {
                    writer.write(ch, last, i - last);
                    writer.write(getPredefinedEntityRef(ch[i]));
                    last = i + 1;
                }
                else if (ch[i] == 0xA)
                {
                    writer.write(ch, last, i - last);
                    writer.println();
                    last = i + 1;
                }
            }
            if ((last == start) || (last < max))
                writer.write(ch, last, i - last);
        }
            
        charDataEncountered = true;
    }
    
    public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException
    {
        if (indent || !useIgnorableWhitespaces || (length <= 0))
            return;
        
        // Complete potential start markup
        completeStartTag();
        
        int max =  start + length;
        int i, last = start;
        for (i = start; i < max; i++)
        {
            if (ch[i] == 0xA)
            {
                writer.write(ch, last, i - last);
                writer.println();
                last = i + 1;
            }
        }
        if ((last == start) || (last < max))
            writer.write(ch, last, i - last);
    }
        
    private boolean isCharDataForbidden(int ch)
    {
        return (ch == '<') || (ch == '&') || (ch == '>'); // last is not forbidden but more elegant
    }

    private boolean isAttributeForbidden(int ch)
    {
        return (ch == '<') || (ch == '&') || (ch == '"');
    }
    
    private String getPredefinedEntityRef(int ch)
    {
        // Encode special XML characters into the equivalent character references.
        // These five are defined by default for all XML documents.
        switch ( ch ) {
        case '<':
            return "&lt;";
        case '>':
            return "&gt;";
        case '"':
            return "&quot;";
        case '\'':
            return "&apos;";
        case '&':
            return "&amp;";
        }
        return null;
    }
    
    public void startElement(String namespaceURI, String localName, String qName, Attributes atts) throws SAXException
    {
        // Complete potential start markup
        completeStartTag();
        
        depth++;
        if (!charDataEncountered && carriageReturn(depth))
            level = depth + 1;
        
        // begin starting markup for element
        buf.append('<');
        buf.append(qName);
        
        // Process attributes (xmlns are supposed to be included)
        if (orderAttributes)
            atts = canonicalizeAttributes(atts);
        
        int nbAtts = atts.getLength();
        String attValue;
        for (int i = 0; i < nbAtts; i++)
        {
            buf.append(' ');
            buf.append(atts.getQName(i));
            buf.append("=\"");
            
            attValue = atts.getValue(i);
            
            int max =  attValue.length();
            int j, last = 0;
            for (j = 0; j < max; j++)
            {
                if (isAttributeForbidden(attValue.charAt(j)))
                {
                    buf.append(attValue.substring(last, j));
                    buf.append(getPredefinedEntityRef(attValue.charAt(j)));
                    last = j + 1;
                }
            }
            if ((last == 0) || (last < max))
                buf.append(attValue.substring(last, j));
            
            buf.append('"');
        }
        
        extraDataEncountered = false;
        charDataEncountered = false;
        elementPending = true;
        lastStartIsStart = true;
    }

    private Attributes canonicalizeAttributes(Attributes atts)
    {
        int nbAtts = atts.getLength();
        if (nbAtts <= 1)
            return atts;
        
        if (orderedAtts == null)
            orderedAtts = new SortableAttributes(atts);
        else
            orderedAtts.setAttributes(atts);
            
        orderedAtts.sort();
        return orderedAtts;
    }
    
    private void completeStartTag()
    {
        // Complete potential start markup
        if (elementPending)
        {
            buf.append('>');
            writer.write(buf.toString());
            buf.setLength(0);
            elementPending = false;
        }
    }
    public void processingInstruction(String target, String data) throws SAXException
    {
        completeStartTag();
        
        writer.write("<?");
        writer.write(target);
        writer.write(' ');
        
        char[] ch = data.toCharArray();
        int max =  data.length();
        int i, last = 0;
        for (i = 0; i < max; i++)
        {
            if (ch[i] == 0xA)
            {
                writer.write(ch, last, i - last);
                writer.println();
                last = i + 1;
            }
        }
        if ((last == 0) || (last < max))
            writer.write(ch, last, i - last);
        
        writer.write("?>");
        extraDataEncountered = true;
    }
    
    public void endElement(String namespaceURI, String localName, String qName) throws SAXException
    {
        // Complete potential start markup
        if (elementPending)
        {
            buf.append("/>");
            writer.write(buf.toString());
            buf.setLength(0);
        }
        else
        {
            if (!charDataEncountered && (!lastStartIsStart || (depth == 0))  && carriageReturn(depth))
                level = depth;
            
            // process ending markup for element
            buf.append("</");
            buf.append(qName);
            buf.append('>');
            writer.write(buf.toString());
            buf.setLength(0);
        }
        charDataEncountered = false;
        extraDataEncountered = false;
        elementPending = false;
        lastStartIsStart = false;
        depth--;
        writer.flush();
    }
    
    public void endDocument() throws SAXException
    {
        writer.println();
        if (closeStreamAutomatically)
            writer.close();
        else
            writer.flush();
        reset();
    }
    
    public void startCDATA() throws SAXException
    {
        // Complete potential start markup
        completeStartTag();
        writer.print("<![CDATA[");
        cdata = true;
    }
    
    public void endCDATA() throws SAXException
    {
        writer.print("]]>");
        cdata = false;
    }
    
    public void comment(char[] ch, int start, int length) throws SAXException
    {
        completeStartTag();
        
        writer.write("<!--");
        
        int max =  start + length;
        int i, last = start;
        for (i = start; i < max; i++)
        {
            if (ch[i] == 0xA)
            {
                writer.write(ch, last, i - last);
                writer.println();
                last = i + 1;
            }
        }
        if ((last == start) || (last < max))
            writer.write(ch, last, i - last);
        
        writer.write("-->");
        extraDataEncountered = true;
    }
    
    private class SortableAttributes implements Attributes
    {
        SortableAttribute[] attTab;
        int nbAtts = 0;
        
        SortableAttributes(Attributes atts)
        {
            setAttributes(atts);
        }
        
        void setAttributes(Attributes atts)
        {
            nbAtts = atts.getLength();
            
            if ((attTab == null) || (attTab.length < nbAtts))
            {
                attTab = new SortableAttribute[nbAtts];
                for (int i=0; i < nbAtts; i++)
                    attTab[i] = new SortableAttribute(atts.getURI(i), atts.getLocalName(i),
                    atts.getQName(i), atts.getValue(i));
            }
            else
                for (int i=0; i < nbAtts; i++)
                     attTab[i].set(atts.getURI(i), atts.getLocalName(i),
                                            atts.getQName(i), atts.getValue(i));
        }
        
        void sort()
        {
            Arrays.sort(attTab, 0, nbAtts);
        }
        
        public int getIndex(String str)
        {
            return -1;  // not Implemented
        }
        
        public int getIndex(String str, String str1)
        {
            return -1;  // not Implemented
        }
        
        public int getLength()
        {
            return nbAtts;
        }
        
        public String getLocalName(int param)
        {
            return attTab[param].localName;
        }
        
        public String getQName(int param)
        {
            return attTab[param].QName;
        }
        
        public String getType(int param)
        {
            return "";  // not Implemented
        }
        
        public String getType(String str)
        {
            return "";  // not Implemented
        }
        
        public String getType(String str, String str1)
        {
            return "";  // not Implemented
        }
        
        public String getURI(int param)
        {
            return attTab[param].namespaceURI;
        }
        
        public String getValue(String str)
        {
            return "";  // not Implemented
        }
        
        public String getValue(int param)
        {
            return attTab[param].value;
        }
        
        public String getValue(String str, String str1)
        {
            return "";  // not Implemented
        }
        
    }
    
    ////////////////////////////////////////////////////////////////////
    //                          PRIVATE
    ////////////////////////////////////////////////////////////////////
    private boolean carriageReturn(int indent)
    {
        boolean ret = false;
        if (this.indent || (level == 0))
        {
            ret = (indent <= level);
            if (ret)   // not for mixed
            {
                if (indent == 0)
                {
                    if (prologGenerated)
                        writer.println();
                }
                else
                {
                    writer.println();
                    indent(indent);
                }
            }
        }
        return ret;
    }
    
    private void indent(int indent)
    {
        if (indent < 0)
            return;
        int quotient = indent/INDENT_CHAR_ARRAY_SIZE;
        
        for (int i = 1; i  <= quotient; i++)
        {
            writer.write(INDENT_CHAR_ARRAY, 0, INDENT_CHAR_ARRAY_SIZE);
        }
        writer.write(INDENT_CHAR_ARRAY, 0, indent%INDENT_CHAR_ARRAY_SIZE);
    }
    
    /**
     *  Canonical XML rules are:
     *<ul>
     * <li>An element's namespace and attribute nodes have a document order
     *     position greater than the element but less than any child node of the
     *     element.</li>
     * <li>Namespace nodes have a lesser document order position than attribute
     *     nodes.</li>
     * <li>An element's namespace nodes are sorted lexicographically by local name
     *     (the default namespace node, if one exists, has no local name and is
     *     therefore lexicographically least).</li>
     * <li>An element's attribute nodes are sorted lexicographically with namespace
     *     URI as the primary key and local name as the secondary key (an empty
     *     namespace URI is lexicographically least).</li>
     * </ul>
     */
    private class SortableAttribute implements Comparable, SAXConstants
    {
        String namespaceURI;
        String localName;
        String QName;
        String value;
        
        SortableAttribute(String namespaceURI, String localName, String QName, String value)
        {
            set(namespaceURI, localName, QName, value);
        }

        void set(String namespaceURI, String localName, String QName, String value)
        {
            this.namespaceURI = namespaceURI;
            this.localName = localName;
            this.QName = QName;
            this.value = value;
        }

        public int compareTo(Object o)
        {
            SortableAttribute obj = (SortableAttribute)o;
            
            if (isNamespace())
            {
                if (obj.isNamespace())
                    return localName.compareTo(obj.localName);
                else
                    return -1;
            }
            else if (obj.isNamespace())
                return 1;
            else    // both are normal attributes
            {
                int comparison = namespaceURI.compareTo(obj.namespaceURI);
                if (comparison == 0)  // then check the local name
                    return localName.compareTo(obj.localName);
                else
                    return comparison;
            }
        }
        
        boolean isNamespace()
        {
            return (namespaceURI.equals(XMLNS_URI));
        }
    }
   
}
