/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.util;

import java.util.Iterator;
import java.util.NoSuchElementException;

import org.w3c.dom.Attr;
import org.w3c.dom.NamedNodeMap;
import org.xml.sax.Attributes;

/** Provides a dynamic (no memory structure is built) wrapping of DOM2
 * to SAX2 attributes (<B>assuming xmlns attributes are removed</B>).
 */
public class DOM2SAXAttrs implements Attributes
{
private static final String RCSRevision = "$Revision: 1.1 $"; 
private static final String RCSName     = "$Name:  $";
    
    private NamedNodeMap atts;
    private int length = 0;
    private boolean hasXMLNSAttrs = false;  // used to avoid list processing when no xmlms is present
                                            // Could not remove because source cannot be changed and
                                            // Clone() is to expensive for a dynamic wrapping
    private boolean namespacePrefixes = false;
    
    public DOM2SAXAttrs()
    {
    }
    public DOM2SAXAttrs(NamedNodeMap atts)
    {
        setDOMAttributes(atts);
    }
    
    public DOM2SAXAttrs(NamedNodeMap atts, boolean namespacePrefixes)
    {
        setNamespacePrefixes(namespacePrefixes);    // Must be first !!!
        setDOMAttributes(atts);
    }
    
    public DOM2SAXAttrs(boolean namespacePrefixes)
    {
        setNamespacePrefixes(namespacePrefixes);
    }
    
    public void setDOMAttributes(NamedNodeMap atts)
    {
        this.atts = atts;
        if (namespacePrefixes)  // process xmlns attribute like any other
            length = atts.getLength();
        else    // they must be ignored
        {
            length = 0;
            for (int i = 0; i < atts.getLength(); i++)
            {
                Attr a = (Attr)atts.item(i);
                if (isPrefixDef(a))
                    hasXMLNSAttrs = true;
                else
                    length++;
            }
        }
    }
    
    public void setNamespacePrefixes(boolean namespacePrefixes)
    {
        this.namespacePrefixes = namespacePrefixes;
    }
        
    /*
     * returns an iteartor on DOM Attr nodes corresponding to xmlns attributes.
     * <p><B>WARNING ! If Namespace-prefixes feature is on or no xmlms attributes
     * : null is returned.<p>
     */

    public Iterator getPrefixIterator()
    {
    	if (!hasXMLNSAttrs || namespacePrefixes) // NOTE : should return an iterator even if namespacePrefixes
            return null;
        else
            return new Iterator() {
                int i = 0;  // next index in atts to read
                int remainingPrefixes = atts.getLength() - length; // number of prefixes remaining to browse
                public boolean hasNext()
                {
                    return (remainingPrefixes > 0);
                }
                public Object next()
                {
                    if (remainingPrefixes == 0)
                        throw new NoSuchElementException("No more xmlns attributes");
                    Attr a;
                    do
                    {
                        a = (Attr)atts.item(i);
                        i++;
                    }
                    while (!isPrefixDef(a));

                    remainingPrefixes--;
                    return a;
                }
                public void remove()
                {
                    throw new UnsupportedOperationException();
                }
            };
    }
    
    ///////////////////////////////////////////////////////////////////////////
    //                      PRIVATE UTILITIES
    ///////////////////////////////////////////////////////////////////////////
    private Attr getAttr(int index)
    {
        if (namespacePrefixes || !hasXMLNSAttrs)
            return (Attr)atts.item(index);
        else
        {
            for (int i = 0; i < atts.getLength(); i++)
            {
                Attr a = (Attr)atts.item(i);
                if (!isPrefixDef(a))
                {
                    if (index == 0)
                        return a;
                    index--;
                }
            }
        }
        return null;
    }
    
    private int getIndex(Attr attr)
    {
        if (attr != null && !isPrefixDef(attr))
        {
            int result = 0;
            for (int i = 0; i < atts.getLength(); i++)
            {
                Attr a = (Attr)atts.item(i);
                if (!hasXMLNSAttrs || !isPrefixDef(a))
                {
                    if (attr == atts.item(i))
                        return result;
                    result++;
                }
            }
        }
        return -1;
    }
    
    private String getLocalName(Attr attr)
    {
        if (attr != null && !isPrefixDef(attr))
            return attr.getLocalName() == null ? "" : attr.getLocalName();
        else
            return null;
    }
    
    private String getQName(Attr attr)
    {
        if (attr != null && !isPrefixDef(attr))
            return attr.getName() == null ? "" : attr.getName();
        else
            return null;
    }
    
    private String getURI(Attr attr)
    {
        if (attr != null && !isPrefixDef(attr))
            return attr.getNamespaceURI() == null ? "" : attr.getNamespaceURI();
        else
            return null;
    }
    
    private String getValue(Attr attr)
    {
        if (attr != null && !isPrefixDef(attr))
            return attr.getValue();
        else
            return null;
    }
    
    private boolean isPrefixDef(Attr attr)
    {
        return (attr.getNamespaceURI() != null)
            && (attr.getNamespaceURI().equals(SAXConstants.XMLNS_URI));
    }
    
    ///////////////////////////////////////////////////////////////////////////
    //                      Attributes IMPLEMENTATION
    ///////////////////////////////////////////////////////////////////////////
    public int getLength()
    {
        return length;
    }
    
    public String getLocalName(int index)
    {
        return getLocalName(getAttr(index));
    }
    
    public String getQName(int index)
    {
        return getQName(getAttr(index));
    }
    
    public String getType(int index)
    {
        return "CDATA";
    }
    
    public String getURI(int index)
    {
        return getURI(getAttr(index));
    }
    
    public String getValue(int index) {
        return getValue(getAttr(index));
    }
    
    public int getIndex(String uri, String localPart)
    {
        if ((uri == null) || (uri.length() == 0))
            return getIndex((Attr)atts.getNamedItem(localPart));
        else
            return getIndex((Attr)atts.getNamedItemNS(uri, localPart));
    }
    
    public String getType(String uri, String localPart)
    {
        return "CDATA";
    }
    
    public String getValue(String uri, String localPart)
    {
        if ((uri == null) || (uri.length() == 0))
            return getValue((Attr)atts.getNamedItem(localPart));
        else
            return getValue((Attr)atts.getNamedItemNS(uri, localPart));
    }
    
    public int getIndex(String qName)
    {
        return getIndex((Attr)atts.getNamedItem(qName));
    }
    
    public String getType(String qName)
    {
        return "CDATA";
    }
    
    public String getValue(String qName)
    {
        return getValue((Attr)atts.getNamedItem(qName));
    }
    
}

