/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.util.Stack;

import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.schema.validation.ValidatingSchemaFilter;

/**
 * This class allows hiding the SAX specificities when constructing java objects
 * from an XML document with an XML schema. This abstract has to be subclassed
 * to make a custom builder for java objects
 *
 */
public abstract class SaxMapper extends DefaultHandler
{
	private static final String RCSRevision = "$Revision: 1.1 $";
	private static final String RCSName = "$Name:  $";

	public abstract Object getMappedObject();
	public abstract TagTracker createTagTrackerNetwork();

	// A stack for the tag trackers to
	// coordinate on.
	//
	private Stack tagStack = new Stack();

	// The SAX 2 parser...
	private XMLReader xr;

	private SchemaValidationContext context;

	public SaxMapper()
	{

		try
		{
			SAXParserFactory factory = SAXParserFactory.newInstance();
			factory.setNamespaceAware(true);
			xr = factory.newSAXParser().getXMLReader();
		}
		catch (Exception e)
		{
			throw new RuntimeException(
				"JAXP exception while intantiating a SAX parser for internal use ("
					+ e.getMessage()
					+ ").");
		}

		context = new SchemaValidationContext();
		xr = new ValidatingSchemaFilter(xr, context);

		// Create the tag tracker network 
		// and initialize the stack with 
		// it.
		//
		// This constructor anchors the tag
		// tracking network to the begining
		// of the XML document. ( before the
		// first tag name is located ).
		//
		// By placing it first on the stack
		// all future tag tracking will follow
		// the network anchored by this
		// root tag tracker.
		//
		// The createTagTrackerNetwork() method
		// is abstract.  All sub classes are 
		// responsible for reacting to this 
		// request with the creation of a 
		// tag tracking network that will 
		// perform the mapping for the sub class.
		//
		tagStack.push(createTagTrackerNetwork());

	}

	public Object fromXML(String url)
	{

		try
		{
			return fromXML(new InputSource(url));

		}
		catch (Exception e)
		{
			e.printStackTrace();
			return null;
		}
	}

	public Object fromXML(InputStream in)
	{
		try
		{
			return fromXML(new InputSource(in));

		}
		catch (Exception e)
		{
			e.printStackTrace();
			return null;
		}
	}

	public Object fromXML(Reader in)
	{
		try
		{
			return fromXML(new InputSource(in));

		}
		catch (Exception e)
		{
			e.printStackTrace();
			return null;
		}
	}

	private synchronized Object fromXML(InputSource in)
	throws SAXException, IOException
	{

		// notes, 
		// 1.  The calling "fromXML" methods catch 
		//     any parsing exceptions.
		// 2.  The method is synchronized to keep 
		//     multiple threads from accessing the XML parser
		//     at once.  This is a limitation imposed by SAX.

		// Set the ContentHandler...
		xr.setContentHandler(this);

		// Parse the file...
		xr.parse(in);

		return getMappedObject();
	}

	// Implement the content hander methods that 
	// will delegate SAX events to the tag tracker network.

	public void startElement(
		String namespaceURI,
		String localName,
		String qName,
		Attributes attr)
		throws SAXException
	{

		// delegate the event handling to the tag tracker
		// network.
		TagTracker activeTracker = (TagTracker) tagStack.peek();
		activeTracker.startElement(context.getCurrentInfoset(), tagStack);

	}

	public void endElement(String namespaceURI, String localName, String qName)
		throws SAXException
	{

		// delegate the event handling to the tag tracker
		// network.
		TagTracker activeTracker = (TagTracker) tagStack.peek();
		activeTracker.endElement(context.getCurrentInfoset(), tagStack);

	}
}
