/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.util.misc;

import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.*;


/**
 * A facility class for printing tree-structured objects. This class is <em>NOT</em>
 * thread safe.
 * 
 * Comments and feature requests are welcome; changes are welcome; commit
 * changes to cvs is not.
 * 
 */

public final class Printing
{
	private static final String RCSRevision = "$Revision: 1.1 $"; 
	private static final String RCSName     = "$Name:  $";

	private static final byte PUBLIC = 0x00;
	private static final byte PROTECTED = 0x01;
	private static final byte PRIVATE = 0x02;
	
	private static boolean linefed = false;
	
	public static char BOTTOM_LEFT_ANGLE = '+';  // 192;
	public static char VERTICAL_LINE = '|'; // 179;
	public static char HORIZONTAL_LINE = '-'; // 196;
	public static char RIGHT_T = '+'; // 195;
	
	static
	{
		setHideStaticMembers(true);
		setHideFinalMembers(true);
	}
	
	
	/**
	 * Customization method. 
	 * 
	 * @param h_line horizontal line
	 * @param v_line vertical line
	 * @param right_t vertical line conjuncts with a half horizontal at its right side  
	 * @param bl_angle bottom left angle
	 */
	public static void init (
		char h_line, 
		char v_line,
		char right_t,
		char bl_angle
		)
	{
		HORIZONTAL_LINE = h_line;
		VERTICAL_LINE = v_line;
		RIGHT_T = right_t;
		BOTTOM_LEFT_ANGLE = bl_angle;
	}

	private static String newline = System.getProperty("line.separator");  
	
	private static final String INDENT = "                                                                     ";

	
	/**
	 * Returns an indentation string for a given indent length. Actually the
	 * returned indent string is twice as long as the argument <code>len</code>.
	 */
	
	public static String getIndent(int len)
	{
		if(len==0)
			return "";
		return System.getProperty("line.separator")+INDENT.substring(0, (len-1)*2)+"  ";
	}
	
	
	
	/**
	 * This function prints an object as a tree using reflection.
	 */
	
	public static void print(Object o)
	{
		reinit();
		System.out.println();		
		printHeader();
		System.out.println();
		printTree("this", o, newline);
		System.out.println();
	
	}

	
	private static void printTree(String name, Object o, String indent)
	{
		printObject(name, o, indent, true);	
	}
	
	
	private static void printObject(String name, Object o, String indent, boolean last)
	{
		if(isHidden(name))
			return;
		//StringBuffer sb = new StringBuffer();
		
		//System.out.println("O = "+o);
		indent = indent.replace(BOTTOM_LEFT_ANGLE, VERTICAL_LINE);
		indent = indent.replace(RIGHT_T, VERTICAL_LINE);
		
		if(last && indent.length()>2)
		{
			int len = indent.length();
		
			
			indent = indent.substring(0, len-3)+BOTTOM_LEFT_ANGLE+indent.substring(len-2, len);
		}
		else if(last==false && indent.length()>2)
		{
			int len = indent.length();
			indent = indent.substring(0, len-3)+RIGHT_T+indent.substring(len-2, len);
		}
					
		// null object
		if(o==null)
		{	
			printNullObject(name, o, indent, last);
			return;
		}
		
		// primitive values
		else if(o.getClass().isPrimitive())
		{
			printPrimitiveValue(name, o, indent, last);
		}
		
		// system base classes (Boolean, Interger, String ...)
		
		else if(o.getClass().getName().startsWith("java.lang"))
		{
			/*System.out.println("1="+o);
			System.out.println("2="+o.getClass());
			System.out.println("3="+o.getClass().getPackage());
			System.out.println("4="+o.getClass().getPackage().getName());
			System.out.println("5="+o.getClass().getPackage().getName().startsWith("java.lang"));*/
			printPrimitiveObject(name, o, indent, last);
		}
		
		
		else if(registered(o))
		{
			printReference(name, o, indent, last);
		}
	
		else
		{
			//push(name);
			register(name, o);
			
			// Array		
			if(o.getClass().isArray())
			{
				printArray(name, o, indent, last);
			}
		
			// Dictionary (Hashtable ...)
			else if(o instanceof Dictionary)
			{
				printDictionary(name, o, indent, last);	
			}
			
			// Map (HashMap ...)
			else if(o instanceof Map)
			{
				printMap(name, o, indent, last);	
			}
		
			// Collection (List, Set, Vector ...)	
			else if(o instanceof Collection)
			{
				printCollection(name, o, indent , last);
			}
					
			else
			{
				printCompositeObject(name, o, indent, last);	
			}
		}
	}
	
	
	private static String getShortClassName(Class c)
	{
		String longName = c.getName();
		int i = longName.lastIndexOf('.');
		String shortName = longName.substring(i+1, longName.length());
		//System.out.println("short name = "+shortName);
		if(c.isArray())
		{
			shortName = longName.substring(i+1, longName.length()-1)+"[]";
		}
		
		return shortName.toUpperCase();
	}
	
	
	private static void printArray(String name, Object o, String indent, boolean last)
	{
		Class c = o.getClass();	
		if(c.isArray());

		print(indent+getShortClassName(c)+" "+name);
		for(int i=0; i<Array.getLength(o); i++)
		{
			//sb.classToString(indent+1);
			printObject(name+"["+i+"]", Array.get(o, i), getNextIndent(indent, last), i==Array.getLength(o)-1);
			
		}	
		print(getLineFeed(indent, last));
	}
	
	
	/**
	 * 
	 */
	private static void printDictionary(String name, Object o, String indent, boolean last)
	{
		push(name);
		Class c = o.getClass();	
	
		print(indent+getShortClassName(c)+" "+name);
			
		Enumeration enum = ((Dictionary)o).keys();
		while(enum.hasMoreElements())
		{		
			Object key = enum.nextElement();
			//sb.append(getIndent(indent+1)+key+" = ");
			Object value = ((Dictionary)o).get(key);
			String nextIndent = getNextIndent(indent, !enum.hasMoreElements());
			printObject(key.toString(), value, getNextIndent(indent, last), !enum.hasMoreElements());
		}
		print(getLineFeed(indent, last));
		pop();
	}
	
	
		
	/**
	 * 
	 */
	private static void printMap(String name, Object o, String indent, boolean last)
	{
		push(name);
		Class c = o.getClass();	
	
		print(indent+getShortClassName(c)+" "+name);
		
		Iterator it = ((Map)o).keySet().iterator();
		while(it.hasNext())
		{		
			Object key = it.next();
			//sb.append(getIndent(indent+1)+key+" = ");
			Object value = ((Map)o).get(key);
			String nextIndent = getNextIndent(indent, !it.hasNext());
			try
			{
				printObject(((key==null)? "null":key.toString()), value, getNextIndent(indent, last), !it.hasNext());
			}
			catch(Exception x)
			{
				//System.out.println("KEY="+key);
				//throw new RuntimeException("Hmm.");
			}
			//push();
		}	
		print(getLineFeed(indent, last));
		pop();
	}
	
	
	private static void printHashValue(String name, Object value, String indent, boolean last)
	{
		print(indent+"VALUE");
		printObject("value", value, getNextIndent(indent, last), last);
	}
	
	
	private static String getThisIndent(String indent, boolean last)
	{
		if(true)
			return indent;
		
		if(indent==null || indent.equals(newline))
			return newline;
		   
		else if(indent.length()<3)
		{
			if(last)
				return newline+BOTTOM_LEFT_ANGLE+HORIZONTAL_LINE+HORIZONTAL_LINE;
			else
				return newline+RIGHT_T+HORIZONTAL_LINE+HORIZONTAL_LINE;
		}
		else
		{
			if(last)
			{
				return indent.substring(0, indent.length()-3).replace(BOTTOM_LEFT_ANGLE, RIGHT_T)+"   "+BOTTOM_LEFT_ANGLE+HORIZONTAL_LINE+HORIZONTAL_LINE;
			}
			else
			{
				return indent.substring(0, indent.length()-2).replace(BOTTOM_LEFT_ANGLE, RIGHT_T)+"  "+VERTICAL_LINE+HORIZONTAL_LINE+HORIZONTAL_LINE;
			}
		}

	}
		
	public static String getNextIndent(String indent, boolean last)
	{
		if(indent.length()<=3)
		{
			if(last)
				return newline+BOTTOM_LEFT_ANGLE+HORIZONTAL_LINE+HORIZONTAL_LINE;
			else
				return newline+RIGHT_T+HORIZONTAL_LINE+HORIZONTAL_LINE;
		}
							
		try
		{
			if(last)
			{
				return indent.substring(0, indent.length()-3).replace(BOTTOM_LEFT_ANGLE, RIGHT_T)+"   "+BOTTOM_LEFT_ANGLE+HORIZONTAL_LINE+HORIZONTAL_LINE;
			}
			else
			{
				return indent.substring(0, indent.length()-2).replace(BOTTOM_LEFT_ANGLE, RIGHT_T)+"  "+RIGHT_T+HORIZONTAL_LINE+HORIZONTAL_LINE;
			}
		}
		catch(IndexOutOfBoundsException x)
		{
			return newline+BOTTOM_LEFT_ANGLE+HORIZONTAL_LINE+HORIZONTAL_LINE;
		}
	}
	
	
	public static String getLineFeed(String line, boolean last)
	{
		
		if(!last)
		{
			linefed = false;
			return "";
		}
		
		try
		{
			if(linefed==false)
			{
				linefed = true;
				return line.substring(0, line.length()-3);	
			}
			else return "";
		}
		catch(IndexOutOfBoundsException x)
		{
			return "";	
		}
	}
	
	
	public static Stack path = new Stack();
	
	public static HashMap olist = new HashMap();
	
	public static boolean registered(Object o)
	{
		return olist.containsKey(o);
	}
	
	public static void register(String name, Object o)
	{
		olist.put(o, generatePath()+name);		
		// System.out.println("\nREG : "+o+" at "+generatePath()+name);
	}
	
	
	public static String generatePath()
	{
		StringBuffer sb = new StringBuffer(32);
		for(int i=0; i<path.size(); i++)
		{
			sb.append(path.get(i).toString()+".")	;
		}

		return sb.toString();
	}
	
	private static String getPath(Object o)
	{
		return (String) olist.get(o);
	}
	
	private static void print(String s)
	{
		System.out.print(s);	
	}
	
	public static void reinit()
	{
		olist.clear();
		path.clear();
	}

	
	private static boolean hideStaticMembers = true;
	private static boolean hideFinalMembers = true;
	
	/**
	 * Specify whether or not hide the static members.
	 */
	public static void setHideStaticMembers(boolean b)
	{
		//System.out.println("Hide static members = " + b);
		hideStaticMembers = b;
	}
	
	
	/**
	 * Specify whether or not hide the final members.
	 */
	public static void setHideFinalMembers(boolean b)
	{
		//System.out.println("Hide final members = " + b);
		hideFinalMembers = b;
	}
	
	
	private static void push(String name)
	{
		path.push(name);
		//System.out.println("\nPUSH: "+generatePath());
	}
	
	
	private static void pop()
	{
		path.pop();
		//System.out.println("\nPOP: "+generatePath());
	}
	
	
	private static void printNullObject(String name, Object o, String indent, boolean last) 
	{
		push(name);
		print(getThisIndent(indent, last)+name+" = null");
		print(getLineFeed(indent, last));
		pop();
	}
	
	
	private static void printPrimitiveValue(String name, Object o, String indent, boolean last) 
	{
		push(name);
		print(getThisIndent(indent, last)+getShortClassName(o.getClass())+" "+name+" = "+o.toString());
		print(getLineFeed(indent, last));
		pop();
	}

	
	private static void printPrimitiveObject(String name, Object o, String indent, boolean last) 
	{
		push(name);
		Class c = o.getClass();
			String className = getShortClassName(c);
			try
			{
				if(o.getClass().getPackage().getName().startsWith("java.lang"))
				{
					if(o instanceof String)
					{
						String s = o.toString();
						s = s.trim();
						if(s.length()>40)
						{
							s = s.substring(0, 40)+"...";	
						}
						s.replace('\r', ' ');
						s.replace('\n', ' ');	
						print(getThisIndent(indent, last)+className+" "+name+" = \""+s+"\"");
					}
					else
						print(getThisIndent(indent, last)+className+" "+name+" = "+o);
					
					print(getLineFeed(indent, last));
					pop();
					return ;
				}
			}
			catch(NullPointerException x)
			{
				
			}
	}
	
	
	private static void printCompositeObject(String name, Object o, String indent, boolean last) 
	{	
		push(name);	
		// display current class name
		print(getThisIndent(indent, last)+getShortClassName(o.getClass()));
		
	
		if(name!=null)
			print(" "+name);
	
		if(isAtomic(o.getClass().getName()))
		{
			pop();
			return;
		}
		
		// Build superclass list
		ArrayList classes = new ArrayList();	
		Class c = o.getClass();
		for(int i=0; i<10; i++)
		{
			if(c.getSuperclass()==null)
				break;
			classes.add(0, c);
			c = c.getSuperclass();
		}
		
		// display inhereted fields
		ArrayList fields = new ArrayList();

		for(int i=0; i<classes.size(); i++)
		{
			fields.addAll(Arrays.asList(((Class)classes.get(i)).getDeclaredFields()));
		}	
		
		for(int i=0; i<fields.size(); i++)
		{
			Field field = (Field) fields.get(i);
			field.setAccessible(true);
			
			
			int modifiers = field.getModifiers();
			
			//System.out.println("member is static: "+Modifier.isStatic(modifiers));
			//System.out.println("hide static member : "+hideStaticMembers);
			
			if(hideStaticMembers && Modifier.isStatic(modifiers))
			{
				//System.out.println("Skip static member: "+field.getName());
				continue;
			}
			
			if(hideFinalMembers && Modifier.isFinal(modifiers))
			{
				//System.out.println("Skip final member: "+field.getName());
				continue;
			}
			
			// skip hidden fields
			if(isHidden(field.getName()))
				continue;
			
			try
			{
				Object fo = field.get(o);
				if(fo==null)
					printObject(getShortClassName(field.getType())+" "+field.getName(), fo, getNextIndent(indent, last), (i==fields.size()-1));
				else
					printObject(field.getName(), fo, getNextIndent(indent, last), (i==fields.size()-1));
				//sb.append(objectToString(field.getName(), field.get(o), getIndent(indent, last), false));
			}
			catch(IllegalAccessException x)
			{
				
			}
		}
	
		print(getLineFeed(indent, last));
		pop();
		
		return;	
	}
	
	
	private static void printReference(String name, Object o, String indent, boolean last) 
	{	
		print(getThisIndent(indent, last)+getShortClassName(o.getClass())+" "+name+" -> "+getPath(o));
		print(getLineFeed(indent, last));
	}

	
	private static void printCollection(String name, Object o, String indent, boolean last) 
	{	
		//push(name);
		int e = 0;
		Iterator it = ((Collection)o).iterator(); 	
		print(getThisIndent(indent, last)+getShortClassName(o.getClass())+" "+name);
		while(it.hasNext())
		{
			Object item = it.next();
			printObject(name+"["+(e++)+"]", item, getNextIndent(indent, last), !it.hasNext());
		}
			
		print(getLineFeed(indent, last));		
		//pop();
	}
	
	private static HashMap hiddenFields = new HashMap();
	
	
	/**
	 * Do not display the fields with the given name.
	 * 
	 * @deprecated please use setVisible().
	 */
	public static void hide(String name)
	{
		hiddenFields.put(name, null);
	}
	
	
	private static boolean isHidden(String name)
	{
		return hiddenFields.containsKey(name);
	}
	
	
	private static HashSet _noexpand = new HashSet();
	
	
	/**
	 * Do not expand the given class and all its sub classes. Please provide full class name.
	 */
	public static void dontExpand(String name)
	{
		try
		{
			Class klass = Class.forName(name);
			_noexpand.add(klass);
		}
		catch(ClassNotFoundException x)
		{
			x.printStackTrace();	
		}
	}
	
	
	/**
	 * Show/hide members of a given name.
	 * 
	 * @param name, the name of the member.
	 * @param visible true for visible false for invisible.
	 */
	public static void setVisible(String name, boolean visible)
	{
		if(visible)
		{
			hiddenFields.remove(name);	
		}
		else
		{
			hiddenFields.put(name, null);
		}
	}

	
	private static boolean isAtomic(String name)
	{
		Class klass = null;
		try
		{
			klass = Class.forName(name);
		}
		catch(ClassNotFoundException x)
		{
			x.printStackTrace();
		}
			
		Iterator it = _noexpand.iterator();
		while(it.hasNext())
		{
			Class KLASS = (Class) it.next();
			if(KLASS.isAssignableFrom(klass))
				return true;
		}
		return false;	
	}
	
	
	private static void printHeader()
	{
		// print no expand classes
		System.out.println("NOT EXPANDED CLASSES: ");
		Iterator cit = _noexpand.iterator();
		while(cit.hasNext())
		{
			System.out.println("-- "+cit.next());
		}
		System.out.println();
		
		
		// hidden fields
		System.out.println("HIDDEN FIELDS:");
		Iterator fit = hiddenFields.keySet().iterator();
		while(fit.hasNext())
		{
			System.out.println("-- "+fit.next());
		}
	}
	
	
	
	
	
	
}
