/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.util.misc;


import java.util.ArrayList;
import java.util.List;


/**
 * 
 */
public class Timers
{
	
	ArrayList timers;
	
	public Timers()
	{
		timers = new ArrayList();
	}
	
	
	/**
	 * Start the timer identified by the given name.
	 * If the timer is not found, no action is taken.
	 * If the timer is already started, restart.
	 * If the timer is stopped, restart. 
	 */
	public void start(String name)
	{
		Timer timer = getTimer(name);
		if(timer==null)
		{
			return;
		}
		else
		{
			timer.start();
		}
	}
	
	
	public void resume(String name)
	{
		Timer timer = getTimer(name);
		if(timer==null)
		{
			return;
		}
		else
		{
			timer.resume();
		}
	}
	
	
	/**
	 * Stop the timer identified by the given name.
	 * If the timer is not started, no action is taken. 
	 */
	public void stop(String name)
	{		
		Timer timer = getTimer(name);
		if(timer==null)
		{
			return;
		}
		else
		{
			timer.stop();
		}
	}
	
	
	/**
	 * Pause the timer identified by the given name.
	 * If the timer is not started, no action is taken. 
	 */
	public void pause(String name)
	{		
		Timer timer = getTimer(name);
		if(timer==null)
		{
			return;
		}
		else
		{
			timer.pause();
		}
	}
	
	
	/**
	 * Returns the duration of the timer. 
	 * 
	 * If the timer is not found, returns -1.
	 * If the timer is not started, returns -1.
	 * If the timer is started but not stopped yet, returns the current time - the timer start time.
	 * If the timer is stopped, returns the stop time - start time. 
	 */
	public double get(String name)
	{
		Timer timer = getTimer(name);
		if(timer==null)
		{
			return -1;
		}
		else
		{
			return timer.get();
		}
	}
	
	
	/**
	 * Returns the list of all existing timers.
	 */
	public List list()
	{
		return timers;
	}
	
	
	/**
	 * Add a new timer. No action is taken if the timer is alreday in the
	 * timer list.
	 */
	public void add(String name)
	{
		Timer timer = getTimer(name);
		if(timer==null)
		{
			timers.add(new Timer(name));
		}
		else
		{
			return;
		}
	}
	
	
	/**
	 * Remove an existing timer. No action is taken if the required timer
	 * is not found. 
	 */
	public void remove(String name)
	{
		timers.remove(name);
	}
	
	
	private Timer getTimer(String name)
	{
		for(int i=0; i<timers.size(); i++)
		{
			Timer timer = (Timer) timers.get(i);	
			if(timer.name.equals(name))
			{
				return timer;
			}
		}
		return null;
	}
	
	
	public String toString()
	{
		StringBuffer sb = new StringBuffer();
		
		for(int i=0; i<timers.size(); i++)
		{
			sb.append(((Timer)timers.get(i)).get()+ "\t");	
		}
		
		return sb.toString();
	}
	
		
	///////////////////////////////////////////////////////////////////////////
	
	
	private static class Timer
	{
		String name;
		long start;
		long stop;
		
		Timer(String name)
		{
			this.name = name;
			start = 0;
			stop = 0;
		}
		
		void start()
		{
			start = System.currentTimeMillis();
			stop = 0;
		}
		
		void stop()
		{
			if(start==0)
			{
				return;
			}
			else
			{
				stop = System.currentTimeMillis();
			}
		}
		
		void pause()
		{
			stop();	
		}
		
		
		void resume()
		{
			double diff = stop - start;
			if(diff==0)
			{
				start();	
			}
			else
			{
				start();
				start -= diff; 	
			}
		}
		
		
		double get()
		{
			if(start==0)
			{
				return 0;	
			}
			else 
			{
				if(stop==0)
				{
					return (System.currentTimeMillis()-start)/1000.0	;
				}
				else
				{
					return (stop-start)/1000.0;
				}
			}
		}
	}
}
