/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.util.misc;


import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;


/**
 * Tracing class.
 * 
 * Functionality available (for now):
 * <ol>
 * <li>Printing function call trace with indentation.
 * <li>Printing tracing message with source code line number.
 * </ol>
 * 
 * Sample code:<p>
 * 
 * <pre>
 * package x.y.z;
 * class Dummy
 * {
 *	    public void dummy()
 *     {
 *         Trace.begin();
 *         ...
 *         Trace.print("o = " + o);
 *         ...
 *         Trace.end();
 *     }
 * }
 * </pre>
 * 
 * Sample output:<p>
 * <pre>
 * 
 * ....[03] CallingClass.CallingMethod >>
 * ......[04] Dummy.dummy >>
 * :::: x.y.z.Dummy.dummy(Dummy.java:18)
 * o = o
 * ......[04] Dummy.dummy >>
 * ....[03] CallingClass.CallingMethod <<
 * </pre>
 * 
 * 
 */
public class Trace
{
	private static final String RCSRevision = "$Revision: 1.1 $"; 
	private static final String RCSName     = "$Name:  $";
	private static StringWriter sw = new StringWriter();	
	
	private static PrintWriter ps = new PrintWriter(sw, true);
	
	private static final String PREFIX = ":::: ";
	
	private static PrintStream out = System.out;
	
	private static final String LINE = System.getProperty("line.separator");
	
	private static boolean trace = false;
	
	
	private static String ignoredPackageNamePrefix = "org.xquark.";
	
	private static int indent = 0;
								   
	private static String indentString = System.getProperty("line.separator")+"..................................................";
	
	private static boolean lineNumberOn;
	
	private static ArrayList excludedPackageList = new ArrayList();
																 
	
	/**
	 * Set the package name prefix that you don't want to print.
	 * 
	 * If the IgnoredPackageNamePrefix is set to "org.xquark.",
	 * then the class org.xquark.xml.misc.Trace" will be printed as:
	 * "xml.misc.Trace".
	 * 
	 * Not to confuse with addExcludedPackage().
	 */
	public static void setIgnoredPackageNamePrefix(String s)
	{
		ignoredPackageNamePrefix = s;	
	}

	

	/**
	 * Set the trace mode to true/false.
	 */
	public static void setTraceMode(boolean bool)
	{
		trace = bool;
	}


	/**
	 * 
	 */
	private static String getLocation()
	{
		StringBuffer sb = sw.getBuffer();
		sb.setLength(0);
		new Throwable().printStackTrace(ps);
		
		int b = 0;
		int e = 0;
		for(int i=0;;i++)
		{
			if(sb.charAt(i)=='\n')
			{
				if(b==0)
				{
					b = 1;
				}
				else if(b==1)
				{
					b = 2;
				}
				else if(b==2)
				{
					b = i;
				}
				else 
				{
					e = i;
					break;	
				}
			}
		}

		return PREFIX+sb.substring(b+4+ignoredPackageNamePrefix.length()+1, e);
	}
	
	
	/**
	 * Print an object and the source code line number.
	 * If the line number is 
	 */
	public static void print(Object o)
	{	
		if(trace==false) return;
		String location = getLocation();
		if(tracing(location))
		{
			out.println(LINE+location);
			out.println(o+LINE);		
		}
	}
	
	/**
	 * Print an object and optionaly the source code line number.
	 * 
	 * @param o the object to print. 
	 * @param bool whether to print the source code line number.
	 */
	public static void print(Object o, boolean bool)
	{	
		if(trace==false) return;
		String location = getLocation();
		if(tracing(location))
		{
			if(bool)
			{
				out.println(LINE+location);
			}
			out.println(o+LINE);		
		}
	}
	
	
	private static String getFunctionName()
	{
		StringBuffer sb = sw.getBuffer();
		sb.setLength(0);
		new Throwable().printStackTrace(ps);
		
		int b = 0;
		int e = 0;
		for(int i=0;;i++)
		{
			if(sb.charAt(i)=='\n')
			{
				if(b==0)
				{
					b = 1;
				}
				else if(b==1)
				{
					b = 2;
				}
				else if(b==2)
				{
					b = i;
				}
				else 
				{
					e = i;
					break;	
				}
			}
		}

		int x = 0, y = 0;
		for(int i=e; i>=b; i--)
		{
			if(sb.charAt(i)=='(')	
			{
				y = i;	
			}
		}
		
		for(int i=y; i>=b; i--)
		{
			if(sb.charAt(i)=='.')	
			{
				if(x==0)
				{
					x = 1;
				}
				else //if( x==1)
				{
					x += i;	
					break;
				}				
			}
		}
		return sb.substring(x, y);		
	}
	
	
	private static String getIndentation()
	{
		return indentString.substring(0, indent*2);
	}
	
	
	private static String getNestingLevel()
	{
		if(indent<=9)
		{
			return "0" + indent;	
		}
		else
		{
			return Integer.toString(indent);
		}
	}
	
	
	/**
	 * Print a trace when entering a function.
	 */
	public static void begin()
	{	
		if(trace==false) return;
		String location = getLocation();
		if(tracing(location))
		{
			indent++;
			out.print(getIndentation()+"["+getNestingLevel()+"] "+getFunctionName() + " >>");
		}
	}
	
	
	/**
	 * Print a trace when leaving a function.
	 */
	public static void end()
	{		
		if(trace==false) return;
		String location = getLocation();
		if(tracing(location))
		{
			out.print(getIndentation()+"["+getNestingLevel()+"] "+getFunctionName() + " <<");
			indent--;
		}
	}
	
	
	/**
	 * Print a trace when leaving a function. The integer allows you
	 * to distinguish the different exits of a function.
	 */
	public static void end(int exit)
	{	
		if(trace==false) return;
		
		String location = getLocation();
		if(tracing(location))
		{
			out.print(getIndentation()+"["+getNestingLevel()+"] "+getFunctionName() + " << ("+exit+")" );
			indent--;
		}
	}
	
	
	/**
	 * Add a package to the exclusion list. 
	 */
	public static void addExcludedPackage(String s)
	{
		excludedPackageList.add(s);
	}
	
	
	/**
	 * Remove a package from the exclusion list. 
	 */
	public static void removeExcludedPackage(String s)
	{
		excludedPackageList.remove(s);
	}
	
	
	/**
	 * 
	 */
	private static boolean tracing(String s)
	{	
		for(int i=0; i<excludedPackageList.size(); i++)
		{
			String p = (String) excludedPackageList.get(i);	
			if(s.startsWith(p))
			{
				return false;
			}
		}
		return true;
	}
}
