/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xml.xdbc;

/**
 * This interface provides setters and getters for metadata (features or properties) associated to an object.
 * The set of supported features and properties depends on vendors implementation.<BR>
 * For some objects, some configuration data may not be changed after the object creation.<BR>
 * An XMLDBCNotRecognizedException or XMLDBCNotSupportedException will be thrown if an object does not
 * support a metadata method.<BR><BR>
 */
public interface Configurable {
 


    //*************************** Features/properties management:

    /**
     * Set the state of a configuration feature.
     * @param featureId The unique identifier (URI) of the feature.
     * @param state The requested state of the feature (true or false).
     * @throws XMLDBCNotRecognizedException if feature does not exist for this object.
     * @throws XMLDBCNotSupportedException if state for the feature is not supported by this object.
     */
    public void setFeature(String featureId, boolean state) throws XMLDBCNotRecognizedException,
        XMLDBCNotSupportedException;

    /**
     * Get the state of a configuration feature.
     * @param featureId The unique identifier (URI) of the feature to get.
     * @return the state of a configuration feature.
     * @throws XMLDBCNotRecognizedException if feature does not exist for this object.
     */
    public boolean getFeature(String featureId) throws XMLDBCNotRecognizedException;

    /**
     * Get the list of supported configuration features
     * @ return the array of supported feature names
     */
    
    public String[] getFeatureList(); 
    
    /**
     * Set the state of a configuration property.
     * @param propertyId The unique identifier (URI) of the property.
     * @param value The requested value of the property (can be null).
     * @throws XMLDBCNotRecognizedException if property does not exist for this driver.
     * @throws XMLDBCNotSupportedException if value for the property is not supported by this object.
     */
    public void setProperty(String propertyId, Object value) throws XMLDBCNotRecognizedException,
        XMLDBCNotSupportedException;

    /**
     * Get the value of a configuration property (only properties set in this configuration).
     * @param propertyId The unique identifier (URI) of the feature to get.
     * @return the value of the specified property (can be null).
     * @throws XMLDBCNotRecognizedException if property does not exist for this object.
     */
    public Object getProperty(String propertyId)  throws XMLDBCNotRecognizedException;

    /**
     * Get the list of supported properties
     * @ return the array of supported property names
     */
    
    public String[] getPropertyList(); 
    
}

