/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xml.xdbc;

/**
 * A default implementation of XMLConnection for read-only connections
 */
public abstract class DefaultReadOnlyXMLConnection implements XMLConnection {
 


  /** Creates a new instance of DefaultReadOnlyXMLConnection */
    public DefaultReadOnlyXMLConnection() {
    }

    /**
     * Makes all changes made since the previous commit/rollback
     * permanent and releases any data source locks currently
     * held by the connection. This method has no effect when
     * auto-commit mode has been enabled.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the data source does not support transactions
     */
    public void commit() throws XMLDBCException, XMLDBCNotSupportedException {
      throw new XMLDBCNotSupportedException("Read-only connection does not support transactions");
    }
    
    /**
     * To create a new XML collection. Each XML collections can be configured.
     * @param name the unique name of the new XML collection.
     * @param description the description of the new XML collection (optional).
     * @param config a XML collection configuration. If null, the XML collection will be created with default configuration.
     * @return an instance of XML collection.
     * @throws XMLDBCException if a data source access error occurs or if the XML collection name already exists.
     * @throws XMLDBCNotSupportedException if the data source does not support this operation (typically, when the data source is read-only)
     */
    public XMLCollection createCollection(String name, String description, Configurable config) throws XMLDBCException, XMLDBCNotSupportedException {
      throw new XMLDBCNotSupportedException("Read-only connection does not support collection management");
    }
    
    /**
     * To create a new XML Collection configuration for set features and properties before create a XML Collection.
     * @return a XML Collection config.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the data source does not support this operation (typically, when the data source is read-only)
     */
    public Configurable createCollectionConfig() throws XMLDBCException, XMLDBCNotSupportedException {
      throw new XMLDBCNotSupportedException("Read-only connection does not support collection management");
    }
    
    /**
     * To delete all XML collections of this connection.
     * @return the number of XML collections deleted.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the data source does not support this operation (typically, when the data source is read-only)
     */
    public int deleteAllCollections() throws XMLDBCException, XMLDBCNotSupportedException {
      throw new XMLDBCNotSupportedException("Read-only connection does not support collection management");
    }
    
    /**
     * To delete a XML collection by its name.
     * @param name the unique name of an existing XML collection.
     * @throws XMLDBCException if a data source access error occurs or if the XML collection does not exist.
     * @throws XMLDBCNotSupportedException if the data source does not support this operation (typically, when the data source is read-only)
     */
    public void deleteCollection(String name) throws XMLDBCException, XMLDBCNotSupportedException {
      throw new XMLDBCNotSupportedException("Read-only connection does not support collection management");
    }
    
    /**
     * To know the current auto-commit state.
     * @return true if the current connection is in auto-commit mode, else false.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the data source does not support transactions
     */
    public boolean getAutoCommit() throws XMLDBCException, XMLDBCNotSupportedException {
      throw new XMLDBCNotSupportedException("Read-only connection does not support transactions");
    }
    
    /**
     * Gets this connection's current transaction isolation level.
     * @return the current TRANSACTION_* mode value (see fields summary).
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the data source does not support transactions
     */
    public short getTransactionIsolation() throws XMLDBCException, XMLDBCNotSupportedException {
      throw new XMLDBCNotSupportedException("Read-only connection does not support transactions");
    }
    
    /**
     * To retrieve a XMLcollection by its name.
     * @param name a name of XML collection.
     * @return the XML collection that has the specified name or 'null' if there is no XML collection with this name.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the data source does not support this operation (typically, when the data source is read-only)
     */
    public XMLCollection getCollection(String name) throws XMLDBCException, XMLDBCNotSupportedException {
      throw new XMLDBCNotSupportedException("Read-only connection does not support collection management");
    }
    
    /**
     * To know the read/write mode of the current connection.
     * @return true if connection is read-only, else false.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public boolean isReadOnly() throws XMLDBCException {
      return true;
    }
    
    /**
     * To rename an XML collection by its name.
     * @param oldname the unique name of an existing XML collection.
     * @param newname the new unique name of an existing XML collection.
     * @throws XMLDBCException if a data source access error errors, if the collection does not exist, or if the new name is not unique.
     * @throws XMLDBCNotSupportedException if the data source does not support this operation (typically, when the data source is read-only)
     */
    public void renameCollection(String oldname, String newname) throws XMLDBCException, XMLDBCNotSupportedException {
      throw new XMLDBCNotSupportedException("Read-only connection does not support collection management");
    }
    
    /**
     * Drops all changes made since the previous commit/rollback and releases any data source locks currently held by
     * this connection. This method should be used only when auto-commit has been disabled.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the data source does not support transactions
     */
    public void rollback() throws XMLDBCException, XMLDBCNotSupportedException {
      throw new XMLDBCNotSupportedException("Read-only connection does not support transactions");
    }
    
    /**
     * Sets this connection's auto-commit mode. If the parameter is true, a commit will be done after each statement execution, else
     * otherwise a commit will have to be made explicitly by calling the commit method.
     * @param autoCommitMode if true, the connection will pass in auto-commit mode, else in manual commit mode.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the data source does not support transactions
     */
    public void setAutoCommit(boolean autoCommitMode) throws XMLDBCException, XMLDBCNotSupportedException {
      throw new XMLDBCNotSupportedException("Read-only connection does not support transactions");
    }
    
    /**
     * To change the read/write mode of the connection. Some data source could make optimisations if the connection
     * is in read-only mode.
     * @param readOnlyMode if true, the connection will pass in read-only mode, else in read/write mode.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the data source is read-only and the parameter is false.
     */
    public void setReadOnly(boolean readOnlyMode) throws XMLDBCException, XMLDBCNotSupportedException {
      if (!readOnlyMode) 
        throw new XMLDBCNotSupportedException("Read-only mode cannot be changed for this connection");
    }
    
    /**
     * Attempts to change the transaction isolation level to the one given. The constants defined in the interface
     * Connection are the possible transaction isolation levels.
     * This method cannot be called while in the middle of a transaction.
     * @param level one of the TRANSACTION_* (see field summary) isolation values with the exception of
     * TRANSACTION_NONE; some data sources may not support other values.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the data source does not support transactions
     */
    public void setTransactionIsolation(short level) throws XMLDBCException, XMLDBCNotSupportedException {
      throw new XMLDBCNotSupportedException("Read-only connection does not support transactions");
    }
    
}
