/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xml.xdbc;

/**
 * This prepared statement interface is used for executing parameterized XQuery or XPath prepared
 * statements, setting parameters and obtaining the results (update statements are not supported yet).<BR>
 * Prepared statements allow queries precompilation by the driver.<BR>
 */
public interface PreparedXMLStatement extends XMLStatement {
 


    /*********************************************************************/
    /*************************** Parameters management *******************/
    /*********************************************************************/

    /**
     * To set the specified parameter whose index is parameterIndex with a value
     * (as a String) without using the Java/XML type conversion.
     * @param parameterIndex index of parameter (first parameter has index 1).
     * @param value a string that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setParameter(int parameterIndex, java.lang.String value) throws XMLDBCException;

    /**
     * To set the specified parameter whose index is parameterIndex with a value (inside an object)
     * using the Java/XML type conversion.
     * @param parameterIndex index of parameter (first parameter has index 1).
     * @param value an object that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setObject(int parameterIndex, java.lang.Object value) throws XMLDBCException;

    /**
     * To set the specified parameter whose index is parameterIndex with a java.math.BigDecimal value.
     * @param parameterIndex index of parameter (first parameter has index 1).
     * @param value a java.math.BigDecimal that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setBigDecimal(int parameterIndex, java.math.BigDecimal value) throws XMLDBCException;

    /**
     * To set the specified parameter whose index is parameterIndex with a Base64Binary value.
     * @param parameterIndex index of parameter (first parameter has index 1).
     * @param value a Base64Binary (in an input stream) that is the value to set for the parameter.
     * @param length the length of given input stream.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setBase64Binary(int parameterIndex, java.io.InputStream value, int length) throws XMLDBCException;

    /**
     * To set the specified parameter whose index is parameterIndex with a HexBinary value.
     * @param parameterIndex index of parameter (first parameter has index 1).
     * @param value a HexBinary (in an input stream) that is the value to set for the parameter.
     * @param length the length of given input stream.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setHexBinary(int parameterIndex, java.io.InputStream value, int length) throws XMLDBCException;

    /**
     * To set the specified parameter whose index is parameterIndex with a boolean value.
     * @param parameterIndex index of parameter (first parameter has index 1).
     * @param value a boolean that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setBoolean(int parameterIndex, boolean value) throws XMLDBCException;

    /**
     * To set the specified parameter whose index is parameterIndex with a byte value.
     * @param parameterIndex index of parameter (first parameter has index 1).
     * @param value a byte that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setByte(int parameterIndex, byte value) throws XMLDBCException;

    /**
     * To set the specified parameter whose index is parameterIndex with a byte array value.
     * @param parameterIndex index of parameter (first parameter has index 1).
     * @param value a byte array that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setBytes(int parameterIndex, byte[] value) throws XMLDBCException;

    /**
     * To set the specified parameter whose index is parameterIndex with a java.sql.Date value.
     * @param parameterIndex index of parameter (first parameter has index 1).
     * @param value a java.sql.Date that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setDate(int parameterIndex, java.sql.Date value) throws XMLDBCException;

    /**
     * To set the specified parameter whose index is parameterIndex with an universal date and time value.
     * @param parameterIndex index of parameter (first parameter has index 1).
     * @param value a java.util.GregorianCalendar (an universal date and time) that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setDateTime(int parameterIndex, java.sql.Timestamp value) throws XMLDBCException;

    /**
     * To set the specified parameter whose index is parameterIndex with a double value.
     * @param parameterIndex index of parameter (first parameter has index 1).
     * @param value a double that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setDouble(int parameterIndex, double value) throws XMLDBCException;

    /**
     * To set the specified parameter whose index is parameterIndex with a float value.
     * @param parameterIndex index of parameter (first parameter has index 1).
     * @param value a float that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setFloat(int parameterIndex, float value) throws XMLDBCException;

    /**
     * To set the specified parameter whose index is parameterIndex with an int value.
     * @param parameterIndex index of parameter (first parameter has index 1).
     * @param value an int that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setInt(int parameterIndex, int value) throws XMLDBCException;

    /**
     * To set the specified parameter whose index is parameterIndex with a long value.
     * @param parameterIndex index of parameter (first parameter has index 1).
     * @param value a long that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setLong(int parameterIndex, long value) throws XMLDBCException;

    /**
     * To set the specified parameter whose index is parameterIndex with a short value.
     * @param parameterIndex index of parameter (first parameter has index 1).
     * @param value a short that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setShort(int parameterIndex, short value) throws XMLDBCException;

    /**
     * To set the specified parameter whose index is parameterIndex with a string value.
     * @param parameterIndex index of parameter (first parameter has index 1).
     * @param value a string that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setString(int parameterIndex, java.lang.String value) throws XMLDBCException;

    /**
     * To set the specified parameter whose index is parameterIndex with a java.sql.Time value.
     * @param parameterIndex index of parameter (first parameter has index 1).
     * @param value a java.sql.Time that is the value to set for the parameter.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void setTime(int parameterIndex, java.sql.Time value) throws XMLDBCException;

    /**
     * Clears the current parameters values immediately.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void clearParameters() throws XMLDBCException;
    
    /**
     * Executes the query statement specified when this object was built with
     * the prepareStatement method (see XMLConnection interface) that may
     * return results.
     * If the query generates results, use the <I>getResultSet()</I> or
     * <I>getDocumentSet()</I> methods to get the set of results.
     * @return true if there are available resuts, else false.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public boolean execute() throws XMLDBCException;
    
    /**
     * Executes the query statement specified when this object was built with
     * the prepareStatement method (see XMLConnection interface) that returns
     * results.
     * @return an XMLResultSet object that contains the data produced by the
     * given query.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public XMLResultSet executeQuery() throws XMLDBCException;
    
}
