/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xml.xdbc;

/**
 * This interface represents a connection with a specific XML data source.
 * Within the context of an XMLConnection, XQuery or XPath statements are executed
 * and XML results are returned.<BR>
 * This interface is used by both read-only connections and storage connections,
 * which require access to a larger set of interfaces, such as interfaces representing
 * XML collections and documents. Classes providing read-only access to XML data sources
 * can avoid implementing the optional interfaces for XML storage management, by deriving
 * from the abstract class DefautReadOnlyXMLConnection.
 *
 */
public interface XMLConnection {

    /*********************************************************************/
    /*************************** Constants declaration *******************/
    /*********************************************************************/
 


    /** Indicates that transactions are not supported. */
    public static final short TRANSACTION_NONE = 0;

    /** Indicates that transactions are supported but dirty reads, non-repeatable reads and phantom reads can occur. */
    public static final short TRANSACTION_READ_UNCOMMITTED = 1;

    /** Indicates that transactions are supported and dirty reads are prevented but non-repeatable reads and phantom reads can occur. */
    public static final short TRANSACTION_READ_COMMITTED = 2;

    /** Indicates that transactions are supported, dirty reads and non-repeatable reads are prevented but phantom reads can occur. */
    public static final short TRANSACTION_REPEATABLE_READ = 4;

    /** Indicates that transactions are supported but and dirty reads, non-repeatable reads and phantom reads are prevented. */
    public static final short TRANSACTION_SERIALIZABLE = 8;

    /** Defines type for queries: XQuery in a String. */
    public static final short XQUERY_STRING_TYPE = 0;

    /** Defines type for queries: XPath in a String. */
    public static final short XPATH_STRING_TYPE = 1;

    /** Defines type for queries: mediator queries (XQuery) in a String. */
    public static final short MEDIATOR_STRING_TYPE = 2;

	/*********************************************************************/
    /*************************** Connection management *******************/
    /*********************************************************************/

    /**
     * To close the current data source connection.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void close() throws XMLDBCException;

    /**
     * To know the connection state (opened or closed).
     * @return true if the current connection is closed, else false.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public boolean isClosed() throws XMLDBCException;

    /**
     * To know the read/write mode of the current connection.
     * @return true if connection is read-only, else false.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public boolean isReadOnly() throws XMLDBCException;

    /**
     * To change the read/write mode of the connection. Some data sources can
     * make optimisations if the connection is in read-only mode.
     * @param readOnlyMode if true, the connection will pass in read-only mode, else in read/write mode.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the data source is read-only and the parameter is false.
     */
    public void setReadOnly(boolean readOnlyMode) throws XMLDBCException, XMLDBCNotSupportedException;

    /**
     * Gets the metadata regarding this connection's data source. A
     * connection's data source is able to provide information
     * describing itself. By default, metadata is loaded one time and cached.
     * Use {@link #getMetaData(boolean) getMetaData(boolean)} to enforce refresh.
     * @return an XMLDataSourceMetaData object for this connection.
     * #see #getMetaData(boolean)
     * @throws XMLDBCException if a data source access error occurs.
     */
    public XMLDataSourceMetaData getMetaData() throws XMLDBCException;

    /**
     * Gets the metadata regarding this connection's data source. A
     * connection's data source is able to provide information
     * describing itself.
     * @refresh when it is true, reload metadata.
     * @return an XMLDataSourceMetaData object for this connection.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public XMLDataSourceMetaData getMetaData(boolean refresh) throws XMLDBCException;

    /**
     * To get the current user name.
     * @return the current user name.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public String getUserName() throws XMLDBCException;

    /**
     * To get the current URL of this connection.
     * @return the current URL of this connection.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public String getURL() throws XMLDBCException;

    /*********************************************************************/
    /*************************** Create/prepare statements ***************/
    /*********************************************************************/

    /**
     * Creates a statement object for sending XQuery (query type by default)
     * statements to the data source.
     * @return a new statement.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public XMLStatement createStatement() throws XMLDBCException;

    /**
     * Creates a statement object for sending XQuery or XPath (according to the
     * specified type) statements to the data source.
     * @param queryType the type of query for this statement (see constants in this interface).
     * @return a new statement.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public XMLStatement createStatement(short queryType) throws XMLDBCException;

    /**
     * Creates a prepared statement object for sending parameterized XQuery
     * (query type by default) statements to the data source.
     * The advantage of prepared statements over statements is that the request
     * may be precompiled once and executed several times.
     * Note: this interface is optional.
     * @param query a parameterized XQuery.
     * @return a new prepared statement containing the pre-compiled statement.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the method is not supported.
     */
    public PreparedXMLStatement prepareStatement(String query)
    throws XMLDBCException, XMLDBCNotSupportedException;

    /**
     * Creates a prepared statement object for sending parameterized XQuery or
     * XPath (according to the specified type) statements to the data source.
     * The advantage of prepared statements over statements is that the request
     * may be precompiled once and executed several times.
     * Note: this interface is optional.
     * @param query a parameterized query of the specified type.
     * @param queryType the type of query for this statement (see constants in this interface).
     * @return a new Pprepared statement containing the pre-compiled statement.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the method is not supported.
     */
    public PreparedXMLStatement prepareStatement(String query, short queryType)
    throws XMLDBCException, XMLDBCNotSupportedException;

    /*********************************************************************/
    /*************************** XML collection management ***************/
    /*********************************************************************/

    /**
     * Creates a new XML Collection configuration for set features and properties before create a XML Collection.
     * @return a Configurable object which properties and features are vendor-dependant.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the data source does not support this operation (typically, when the data source is read-only)
     */
    public Configurable createCollectionConfig() throws XMLDBCException, XMLDBCNotSupportedException;

    /**
     * Creates a new XML collection. Each XML collections can be configured.
     * @param name the unique name of the new XML collection.
     * @param description the description of the new XML collection (optional).
     * @param config a XML collection configuration. If null, the XML collection will be created with default configuration.
     * @return an instance of XML collection.
     * @throws XMLDBCException if a data source access error occurs or if the XML collection name already exists.
     * @throws XMLDBCNotSupportedException if the data source does not support this operation (typically, when the data source is read-only)
     */
    public XMLCollection createCollection(String name, String description, Configurable config) throws XMLDBCException, XMLDBCNotSupportedException;

    /**
     * Deletes an XML collection by its name.
     * @param name the unique name of an existing XML collection.
     * @throws XMLDBCException if a data source access error occurs or if the XML collection does not exist.
     * @throws XMLDBCNotSupportedException if the data source does not support this operation (typically, when the data source is read-only)
     */
    public void deleteCollection(String name) throws XMLDBCException, XMLDBCNotSupportedException;

    /**
     * Renames an XML collection by its name.
     * @param oldname the unique name of an existing XML collection.
     * @param newname the new unique name of an existing XML collection.
     * @throws XMLDBCException if a data source access error errors, if the collection does not exist, or if the new name is not unique.
     * @throws XMLDBCNotSupportedException if the data source does not support this operation (typically, when the data source is read-only)
     */
    public void renameCollection(String oldname, String newname) throws XMLDBCException, XMLDBCNotSupportedException;

    /**
     * Deletes all XML collections of this connection.
     * @return the number of XML collections deleted.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the data source does not support this operation (typically, when the data source is read-only)
     */
    public int deleteAllCollections() throws XMLDBCException, XMLDBCNotSupportedException;

    /**
     * Creates an XMLcollection object featuring the XML documents collection.
     * <p>The returned object may hold resources and should be closed after
     * use.</p>
     * @param name a name of an XML documents collection.
     * @return the XML collection that has the specified name or 'null' if there is no XML collection with this name.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the data source does not support this operation (typically, when the data source is read-only)
     */
    public XMLCollection getCollection(String name) throws XMLDBCException, XMLDBCNotSupportedException;

    /*********************************************************************/
    /*************************** Transaction management ******************/
    /*********************************************************************/

    /**
     * Makes all changes made since the previous commit/rollback
     * permanent and releases any data source locks currently
     * held by the Connection. This method has no effect when
     * auto-commit mode has been enabled.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the data source does not support transactions
     */
    public void commit() throws XMLDBCException, XMLDBCNotSupportedException;

    /**
     * Sets this connection's auto-commit mode. If the parameter is true, a commit will be done after each statement execution, else
     * otherwise a commit will have to be made explicitly by calling the commit method.
     * @param autoCommitMode if true, the connection will pass in auto-commit mode, else in manual commit mode.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the data source does not support transactions
     */
    public void setAutoCommit(boolean autoCommitMode) throws XMLDBCException, XMLDBCNotSupportedException;

    /**
     * To know the current auto-commit state.
     * @return true if the current connection is in auto-commit mode, else false.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the data source does not support transactions
     */
    public boolean getAutoCommit() throws XMLDBCException, XMLDBCNotSupportedException;

    /**
     * Drops all changes made since the previous commit/rollback and releases any data source locks currently held by
     * this Connection. This method should be used only when auto-commit has been disabled.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the data source does not support transactions
     */
    public void rollback() throws XMLDBCException, XMLDBCNotSupportedException;

    /**
     * Gets this Connection's current transaction isolation level.
     * @return the current TRANSACTION_* mode value (see fields summary).
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the data source does not support transactions
     */
    public short getTransactionIsolation() throws XMLDBCException, XMLDBCNotSupportedException;

    /**
     * Attempts to change the transaction isolation level to the one given. The constants defined in the interface
     * Connection are the possible transaction isolation levels.
     * This method cannot be called while in the middle of a transaction.
     * @param level one of the TRANSACTION_* (see field summary) isolation values with the exception of
     * TRANSACTION_NONE; some data sources may not support other values.
     * @throws XMLDBCException if a data source access error occurs.
     * @throws XMLDBCNotSupportedException if the data source does not support transactions
     */
    public void setTransactionIsolation(short level) throws XMLDBCException, XMLDBCNotSupportedException;
}

