/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xml.xdbc;

/**
 * This statement interface is used for executing constant XQuery or XPath statements 
 * and obtaining the results produced (update statements are not supported yet).<BR>
 */
public interface XMLStatement {
 


    /*********************************************************************/
    /*************************** Queries execution management ************/
    /*********************************************************************/

    /**
     * Executes a query statement of the type specified when this 
     * object was built with the createStatement method (see XMLConnection interface) 
     * that may return results.
     * If the query generates results, use the <I>getResultSet()</I> or
     * <I>getDocumentSet()</I> methods to get the set of results.
     * @param query the query string.
     * @return true if there are available resuts, else false.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public boolean execute(String query) throws XMLDBCException;

    /**
     * Executes a query statement of the specified type  
     * that may return results.
     * If the query generates results, use the <I>getResultSet()</I> or
     * <I>getDocumentSet()</I> methods to get the set of results.
     * @param query the query string.
     * @param queryType the type of query for this statement (see constants in XMLConnection interface).
     * @return true if there are available resuts, else false.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public boolean execute(String query, int queryType) throws XMLDBCException;

    /**
     * Executes a query statement of the type specified when this 
     * object was built with the createStatement method (see XMLConnection interface) 
     * that returns results.
     * @param query the query string
     * @return an XMLResultSet object that contains the data produced by the given query.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public XMLResultSet executeQuery(String query) throws XMLDBCException;

    /**
     * Executes a query statement of the specified type  
     * that returns results.
     * If the query generates results, use the <I>getResultSet()</I> or
     * <I>getDocumentSet()</I> methods to get the set of results.
     * @param query the query string.
     * @param queryType the type of query for this statement (see constants in XMLConnection interface).
     * @return true if there are available resuts, else false.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public XMLResultSet executeQuery(String query, int queryType) throws XMLDBCException;

    /**
     * Executes a query statement of the type specified when this 
     * object was built with the createStatement method (see XMLConnection interface) 
     * that returns complete XML documents. It is a error to call this method with a
     * query that returns document fragments.
     * @param query a static and read-only query in the specified type.
     * @return an XMLResultSet object that contains the data produced by the given query.
     * @throws XMLDBCException if a data source access error occurs, or if the returned
     * results are not complete XML documents.
     */
    public XMLDocumentSet executeDocumentQuery(String query) throws XMLDBCException;

    /**
     * Executes a query statement of the specified type  
     * that returns complete XML documents. It is a error to call this method with a
     * query that returns document fragments.
     * @param query the query string.
     * @param queryType the type of query for this statement (see constants in XMLConnection interface).
     * @return true if there are available resuts, else false.
     * @throws XMLDBCException if a data source access error occurs, or if the returned
     * results are not complete XML documents.
     */
    public XMLDocumentSet executeDocumentQuery(String query, int queryType) throws XMLDBCException;

    /*********************************************************************/
    /*************************** Results management **********************/
    /*********************************************************************/

    /**
     * Returns the current statement results as an XMLResultSet object.
     * @return the current results as an XMLResultSet object or null if there is no result.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public XMLResultSet getResultSet() throws XMLDBCException;

    /**
     * Returns the current statement results as an XMLDocumentSet object.
     * It is a error to call this method for a statement that returned
     * document fragments.
     * @return the current results as an XMLDocumentSet object or null if there is no result.
     * @throws XMLDBCException if a data source access error occurs, or if the returned
     * results are not complete XML documents.
     */
    public XMLDocumentSet getDocumentSet() throws XMLDBCException;

    /*********************************************************************/
    /*************************** Statement management ********************/
    /*********************************************************************/

    /**
     * Returns the connection object that produced this statement.
     * @return the connection object that produced this statement.
     */
    public XMLConnection getConnection();

    /**
     * Releases the statement XMLDBC resources immediately instead of waiting for this
     * to happen when it is automatically closed.
     * @throws XMLDBCException if a data source access error occurs.
     */
    public void close() throws XMLDBCException;
    
	/*********************************************************************/
	/*************************** extension methods ***********************/
	/*********************************************************************/

	/**
	 * Sets the base URI for documents.
     * @return nothing 
     */
	public void setBaseURI(String baseURI);
	/**
	 * gets the base URI for documents.
	 * @return the base URI 
	 */
	public String getBaseURI();
    
    
}

