/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xpath;


/**
 * This class implements the model node defined in the W3C Recommendation
 * on 16 Nov, 1999 <A HREF="http://www.w3.org/TR/1999/REC-xpath-19991116">
 * XML Path Language (XPath) Version 1.0</A>
 * <p><B>Only the subset of the recommendation useful for indexing
 * metatada model is used.</B> e.g. No string-value is considered for the nodes.</p>
 */
public class XNode implements Cloneable
{
	private static final String RCSRevision = "$Revision: 1.1 $";
	private static final String RCSName = "$Name:  $";

	/** The two following attributes make up the "expanded localName" defined in the
	 * recommendation.
	 */
	private String namespace = null;
	// used also as target for Processing Instructions
	private String localName = null;

	/** the node type */
	protected byte type = NodeKind.NODE;

	/**
	 * Default constructor.
	 */
	public XNode()
	{}

	/**
	 * Constructor used for element and attributes nodes (when XNode is used as
	 * a step node test). Type is the default value: NODE.
	 * @param namespace namespace URI of the node.
	 * @param localName local name of the node
	 */
	public XNode(String namespace, String localName)
	{
		this(namespace, localName, NodeKind.NODE);
	}

	/**
	 * Constructor used for root, comment & text nodes
	 * @param the node type as a constant defined in {@link Type}.
	 */
	public XNode(byte type)
	{
		this(null, null, type);
	}

	/**
	 * Constructor used for Processing instructions nodes.
	 * @param target processing instruction target.
	 */
	public XNode(String target)
	{
		this(target, null, NodeKind.PI);
	}

	/**
	 * Constructor used for elements, attributes & namespaces nodes
	 * @param namespace namespace URI of the node.
	 * @param localName local name of the node
	 * @param the node type as a constant defined in {@link Type}.
	 */
	public XNode(String namespace, String localName, byte type)
	{
		set(namespace, localName, type);
	}

	/**
	 * Setter used for qualified elements, attributes & namespaces nodes
	 * @param namespace namespace URI of the node.
	 * @param localName local name of the node
	 * @param the node type as a constant defined in {@link Type}.
	 */
	public void set(String namespace, String localName, byte type)
	{
		if ((namespace == null) || (namespace.length() == 0))
			this.namespace = null;
		else
			this.namespace = namespace;
		this.localName = localName;
		set(type);
	}

	/**
	 * Setter for node type.
	 * @param the node type as a constant defined in {@link Type}.
	 */
	public void set(byte type)
	{
		this.type = type;
	}

	/**
	 * Parse an abbreviated string representation for attributes & elements
	 * only. Syntax is of the form : '@'? ('{' namespace '}')? localName
	 * @param exp the node abbreviated string representation.
	 */
	public void set(String exp)
	{
		int i, j = 0;

		// string step parsing
		i = 0;
		if (exp.charAt(i) == '@')
		{
			i++;
			type = NodeKind.ATTRIBUTE;
		}
		else
			type = NodeKind.ELEMENT;
		if (exp.charAt(i) == '{')
		{
			j = exp.indexOf('}');
			namespace = exp.substring(i + 1, j);
			i = j + 1;
		}
		else
			namespace = null;

		localName = exp.substring(i);
	}

	/**
	 * Accessor to the namespace uri.
	 * @return the namespace uri of the node.
	 */
	public String getNamespace()
	{
		return namespace;
	}

	/**
	 * Accessor to the local name.
	 * @return the local name of the node.
	 */
	public final String getLocalName()
	{
		return localName;
	}

	/**
	 * Returns the canonical "expanded" name of this step. The returned string
	 * matches the following formatting rules:
	 * <code>{namespace}localName</code> if the namespace is not null and
	 * <code>localName</code> if the namespace is null.
	 * @return the expanded name of the node.
	 */
	public final String getExpandedName()
	{
		if ((namespace == null) || namespace.equals(""))
			return localName;
		else
			return '{' + namespace + '}' + localName;
	}

	/**
	 * Accessor to the node type.
	 * @return the type of the node.
	 * @see Type
	 */
	public byte getType()
	{
		return type;
	}

	/**
	 * Test XNodes for equality. XNode are considered equals if their names and
	 * types are equals.
	 * @param o the XNode to compare to this object.
	 */
	public final boolean equals(Object o)
	{
		XNode node = null;
		if (o instanceof XNode)
			node = (XNode) o;
		else
			return false;
		return (
			((getNamespace() == null) && (node.getNamespace() == null))
				|| ((getNamespace() != null)
					&& getNamespace().equals(node.getNamespace())))
			&& (((getLocalName() == null) && (node.getLocalName() == null))
				|| ((getLocalName() != null)
					&& getLocalName().equals(node.getLocalName())))
			&& (type == node.getType());
	}

	/**
	 * Return the node hash code.
	 * This method is overriden because {@link XTreeNode} use XNode 
	 * as hash keys of the children map.
	 * @return the hash code for node.
	 */
	public int hashCode()
	{
		// local name hashcode is enough because it is the more likely to vary.
		if (localName != null)
			return localName.hashCode();
		else
			return toString().hashCode();
	}

	/**
	 * Return a string representation of this node using the abbreviated
	 * syntax.
	 * @return the string representation of the node.
	 */
	public String toString()
	{
		return getAbbreviatedSyntax();
	}
	
	/**
	 * Return a string representation of this node using the abbreviated
	 * syntax.
	 * @return the string representation of the node.
	 */
	public final String getAbbreviatedSyntax()
	{
		StringBuffer sb = new StringBuffer();

		switch (type)
		{
			case NodeKind.NODE :
				sb.append("node()");
				break;

			case NodeKind.COMMENT :
				sb.append("comment()");
				break;

			case NodeKind.TEXT :
				sb.append("text()");
				break;

			case NodeKind.PI :
				if (localName == null)
					sb.append("processing-instruction()");
				else
					sb.append("processing-instruction('" + localName + "')");
				break;

			case NodeKind.ATTRIBUTE :
				sb.append("@" + getExpandedName());
				break;

			case NodeKind.ELEMENT :
				sb.append(getExpandedName());
				break;
		}
		return sb.toString();
	}

	/**
	 * Whether this node matches the given step expression (used for
	 * navigation).
	 * @param step the XPath step to check for.
	 * @return true if this node matches the step.
	 */
	boolean match(StepExpr step)
	{
		return step.match(this);
	}

	/**
	 * Clone the current object.
	 * @return a new XNode object with the same contents.
	 */
	public Object clone()
	{
		try
		{
			return super.clone();
		}
		catch (CloneNotSupportedException e)
		{
			// sure that Object is cloneable !
			return null;
		}
	}

}
