/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xpath;

import java.util.*;

import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xquark.util.misc.Printing;

/**
 * Abstract XTree node.
 *
 */
public abstract class XTreeNode implements Cloneable {
	private static final String RCSRevision = "$Revision: 1.1 $";
	private static final String RCSName = "$Name:  $";
	/** */
	protected XTree tree;

	protected XNode XModelNode;
	// not designed by heritage, because XNode overrides
	// equals (to be able to be used as a key in the local children map)

	protected XTreeNode parent;

	private HashMap children;

	protected PathExpr location = null;

	/**
	 * Constructor for element and attributes nodes.
	 * Performs registration of the node to its parent node, sets the owner
	 * XTree but does not perform registration of the node to the tree.
	 * This constructor should be called by implementors.
	 * @param tree the owner tree
	 * @param parent the parent node
	 * @param namespace the namespace URI of the new node
	 * @param localName the local name of the new node
	 * @param the type of the new node
	 */
	protected XTreeNode(XTree tree, XTreeNode parent, String namespace, String localName, byte type) {
		XModelNode = new XNode(namespace, localName, type);
		set(tree, parent);
	}

	/**
	 * Accessor to the parent node.
	 * @return the parent node (normally set at construction time except for
	 * root).
	 */
	public XTreeNode getParent() {
		return parent;
	}

	/**
	 * Accessor to the owner tree.
	 * @return the owner tree.
	 */
	public XTree getTree() {
		return tree;
	}

	/**
	 * Accessor to node's children.
	 * @return a collection of XTreeNode objects. null if the node is a leaf.
	 */
	public Collection getChildren() {
		if (children == null)
			return null;
		return children.values();
	}

	/**
	 * Returns the canonical "expanded" name of this step. The returned string
	 * matches the following formatting rules:
	 * <code>{namespace}localName</code> if the namespace is not null and
	 * <code>localName</code> if the namespace is null.
	 * @return the expanded name of the node.
	 */
	public final String getExpandedName() {
		return XModelNode.getExpandedName();
	}

	/**
	 * Accessor to the local name of the node.
	 * @return the local name of the node.
	 */
	public final String getLocalName() {
		return XModelNode.getLocalName();
	}

	/**
	 * Accessor to the namespace uri of the node.
	 * @return the namespace uri of the node or null if node has no namespace.
	 */
	public String getNamespace() {
		return XModelNode.getNamespace();
	}

	/**
	 * Accessor to the XPath type of the node.
	 * @return a value defined in {@link Type}.
	 */
	public byte getType() {
		return XModelNode.getType();
	}

	/**
	 * Returns an object used as a key for the node in its parent children map.
	 * @return the data model node of the current tree node.
	 */
	public XNode getHashKey() {
		return XModelNode;
	}

	/**
	 * Whether the node is a leaf, i.e. has no child node.
	 * @return true if the node has no child.
	 */
	public boolean isLeaf() {
		return (children == null);
	}

	/**
	 * Get the depth of the node in the metadata tree.
	 * @return an int value where the tree root has 0 value.
	 */
	public int getDepth() {
		if (location == null)
			return 0;
		else
			return getLocation().getSteps().size();
	}

	/**
	 * Add a node to the children set. No order is garanteed within the
	 * children.
	 * @param child the node that will become a child node of this node.
	 */
	public final void addChild(XTreeNode child) {
		if (children == null) // to avoid to make a base class for leaf nodes
			children = new HashMap();
		children.put(child.getHashKey(), child);
		child.setParent(this);
	}

	/**
	 * Merges a given subtree with the current subtree which root is the
	 * current node.
	 * @param node the root of the subtree to merge to the current subtree.
	 */
	public final void merge(XTreeNode node) {
		if (node.children != null) {
			Iterator it = node.children.keySet().iterator();
			while (it.hasNext()) {
				XNode key = ((XTreeNode) it.next()).getHashKey();
				XTreeNode child = getChild(key);
				if (child == null)
					this.addChild(node.getChild(key));
				else
					child.merge(node.getChild(key));
			}
		}
	}

	/**
	 * Remove a given child node.
	 * @param pattern the "signature" of the children to remove.
	 */
	public final XTreeNode removeChild(XNode pattern) {
		if (children != null)
			return (XTreeNode) children.remove(getChild(pattern));
		return null;
	}

	/**
	 * Detach the node from its tree.
	 */
	public final void remove() {
		if (parent != null)
			parent.removeChild(XModelNode);
	}

	/**
	 * retrieves a particular child.
	 * @param pattern the "signature" of the children to retrieve.
	 * @return the searched node or null if not found.
	 */
	public final XTreeNode getChild(XNode pattern) {
		if (children == null)
			return null;
		return (XTreeNode) children.get(pattern);
	}

	/**
	 * Returns a <code>LocationExpression</code> corresponding to the
	 * given <code>Tree</code>.
	 * @return null if the node has no parent (ROOT or unregistered, i.e.
	 * not attached to the XTree, node)
	 */
	public final PathExpr getLocation() {
		if (location == null) {
			if (getParent() == null) // the root case or an unregistered node
				return null;
			location = new PathExprImpl();
		} else // cache
			return location;

		// construction
		for (XTreeNode node = this; node.getType() != NodeKind.NONE; node = (XTreeNode) node.getParent()) {
			switch (node.getType()) {
				case NodeKind.NAMESPACE :
					location.getSteps().add(0, new StepExprImpl(Axis.NAMESPACE, node.getHashKey()));
					break;
				case NodeKind.ATTRIBUTE :
					location.getSteps().add(0, new StepExprImpl(Axis.ATTRIBUTE, node.getHashKey()));
					break;
				default : // CHILD
					location.getSteps().add(0, new StepExprImpl(node.getHashKey()));
			}
		}
		return location;
	}

	/**
	 * Performs a navigation specified by the step expression.
	 * @param step the step that defines the elementary navigation
	 * @return the set of qualified nodes
	 */
	public final Collection navigate(StepExpr step) {
		Collection result = null;

		byte axis = step.getAxis();

		switch (axis) {
			case Axis.SELF :
				result = new ArrayList(1);
				if (step.match(XModelNode))
					result.add(this);
				break;

			case Axis.NONE :
			case Axis.CHILD :
			case Axis.ATTRIBUTE :
			case Axis.NAMESPACE :
				return getChildren(step);

			case Axis.PARENT :
				result = new ArrayList(1);
				result.add(getParent());
				break;

			case Axis.ANCESTOR_OR_SELF :
				result = getAncestors(step);
				if (step.match(XModelNode))
					result.add(this);
				break;

			case Axis.ANCESTOR :
				return getAncestors(step);

			case Axis.DESCENDANT :
				return getDescendants(step);

			case Axis.DESCENDANT_OR_SELF :
				result = getDescendants(step);
				if (step.match(XModelNode))
					result.add(this);
				break;

			case Axis.PRECEDING :
				return getPreceding(step);

			case Axis.PRECEDING_SIBLING :
				return getPrecedingSibling(step);

			case Axis.FOLLOWING :
				return getFollowing(step);

			case Axis.FOLLOWING_SIBLING :
				return getFollowingSibling(step);

//			default :
//				System.out.println("Attention  la marche: " + axis);
		}

		return result;
	}

	/**
	 * Compute the LocationExpressions matching given regular location
	 * expression.
	 * @param location a location expression
	 * @return a collection containing XTreeNode objects matching the regular
	 * expression.
	 */
	public final Collection prune(ArrayList location) {
		/*
		Set result = new HashSet();
		 
		if(location.size() == 0)
		{
		//System.out.println("adding self: "+declaration.getName());
		//result.add(this);
		        Set r = new HashSet();
		        r.add (this) ;
		        result.add (r) ;
		return result;
		}
		 
		Collection set = navigate((StepExpr)location.get(0));
		 
		//System.out.println(getDeclaration().getName()+" navigate "+location.toStringBuffer(0));
		//System.out.println(set);
		 
		LocationExpression l = (LocationExpression) location.clone();
		 
		if(l.size() > 0)
		l.remove(0);
		else
		{
		if(set != null)
		result.addAll(set);
		return result;
		}
		 
		Object[] nodes = set.toArray();
		for(int i=0; i < nodes.length; i++)
		{
		XTreeNode node = (XTreeNode)nodes[i];
		//System.out.println(node.getDeclaration().getName()+" -> "+l.toStringBuffer(0));
		result.addAll(node.prune(l));
		//result.add(node.prune(l));
		}
		return result;
		 */

		//org.xquark.xml.misc.Printing.print(location);
		//System.out.println(getNamespace()+":"+getLocalName()+" => "+location.toString(0));

		/*
		LocationExpression x = getLocation();
		String id = null;
		if(x!=null){
		id = x.toString(0);
		}
		else{
		id = "ROOT";
		}
		System.err.println(id+"  SEEK "+location.toString(0));
		 */

		Set result = new HashSet();

		if (location.size() == 0) {
			if (this.getType() != NodeKind.NONE)
				result.add(this);
			//System.err.println(" RESULT "+result);
			return result;
		}

		Collection set = navigate((StepExpr) location.get(0));

		//System.out.println(getNamespace()+":"+getLocalName()+" navigate "+location.toString(0));
		//System.out.println("SET: " +set);

		ArrayList l = (ArrayList) location.clone();

		if (l.size() > 0)
			l.remove(0);
		else {
			if (set != null)
				result.addAll(set);
			//System.out.println("RESULT2: "+result);
			return result;
		}

		Object[] nodes = set.toArray();
		for (int i = 0; i < nodes.length; i++) {
			XTreeNode node = (XTreeNode) nodes[i];
			//System.out.println(node.getNamespace()+":"+node.getLocalName()+" -> "+l.toString(0));
			Collection tmp = node.prune(l);
			result.addAll(tmp);
		}
		//System.err.println("RESULT3 "+result);
		return result;
	}

	/**
	 * Compute the LocationExpressions matching given regular location
	 * expression building a collection of collections in order to
	 * regroup same level (for using in PathResolver and next for typing))
	 * @param location a location expression
	 * @return a collection collection of collections containing XTreeNode
	 * objects matching the regular expression. First collection are results
	 * for the level of the current node.
	 */
	public final Collection pruneGroup(ArrayList location) {
		Set result = new HashSet();
		Set group = new HashSet();

		if (location.size() == 0) {
			//System.out.println("adding self: "+declaration.getName());
			/*
			group.add (this) ;
			result.add (group);
			 */
			result.add(this); // SR 4/11/02 result is a Collection of Collections : why adding the current node ?
			//        System.out.println("THIS " + this);
			return result;
		}

		Collection set = navigate((StepExpr) location.get(0));

		//System.out.println(getDeclaration().getName()+" navigate "+location.toStringBuffer(0));
		//       System.out.println("SET" + set);

		ArrayList l = (ArrayList) location.clone();

		if (l.size() > 0)
			l.remove(0);
		else {
			if (set != null) {
				group.add(set);
				result.add(group);
			}
			return result;
		}

		Object[] nodes = set.toArray();
		for (int i = 0; i < nodes.length; i++) {
			XTreeNode node = (XTreeNode) nodes[i];
			//System.out.println(node.getDeclaration().getName()+" -> "+l.toStringBuffer(0));
			if (node != null);
			{
				Collection c = node.pruneGroup(l);
				if ((c != null) && (c.size() > 0))
					result.add(node.pruneGroup(l));
			}
		}
		return result;
	}

	/**
	 * Return a string representation of this node an its descendants.
	 *
	 * @param indent the size of indentation.
	 * @return the string representation of the current subtree.
	 */
	public final String toString(int indent) {
		StringBuffer sb = new StringBuffer();
		sb.append(Printing.getIndent(indent) + this.getNamespace() + ":" + this.getLocalName() + "   " + toString());

		if (children != null) {
			Iterator it = getChildren().iterator();

			while (it.hasNext()) {
				XTreeNode child = (XTreeNode) it.next();
				sb.append(child.toString(indent + 1));
			}
		}
		return sb.toString();
	}

	/**
	 * Perform an in depth copy of the subtree which the current node is the
	 * root
	 * @return the new XTreeNode.
	 */
	public Object clone() {
		XTreeNode node = null;
		try {
			node = (XTreeNode) super.clone();
			node.XModelNode = (XNode) XModelNode.clone();
			if (children != null) {
				node.children = new HashMap();
				Iterator it = children.values().iterator();
				while (it.hasNext()) {
					node.addChild((XTreeNode) ((XTreeNode) it.next()).clone());
				}
			}
		} catch (CloneNotSupportedException e) {}
		return node;
	}

	/**
	 * Create an homothetic copy of the subtree which the current node is the
	 * root using a given tree builder.
	 * @param root the root of the new subtree.
	 * @param factory the tree builder used to produce the new subtree.
	 * @return the root of the new subtree.
	 */
	public XTreeNode duplicate(XTreeNode root, XTreeBuilder factory) {
		if (children != null) {
			Iterator it = children.values().iterator();
			XTreeNode child;
			while (it.hasNext()) {
				child = (XTreeNode) it.next();
				XTreeNode duplicate = factory.createNamedNode(root, child.getNamespace(), child.getLocalName(), child.getType()).customizeDuplicate(child);
				child.duplicate(duplicate, factory);
				// recursivity must be performed
				// after customization because children may need parent info
			}
		}
		return root;
	}

	/**
	 * Add customization to the nodes created with {@link #duplicate(XTreeNode, XTreeBuilder)}
	 * @param original the node which this node is a copy
	 * @return this node customized
	 */
	protected XTreeNode customizeDuplicate(XTreeNode original) {
		// nop by default
		return this;
	}

	/**
	 * Method call by {@link XTreeReader} to generate SAX events describing
	 * the node contents.
	 * This method is a base implementation that sends the result of toString()
	 * in a characters() SAX event.
	 * @param handler the content handler that will receive SAX events.
	 * @throws SAXException handler exception
	 */
	protected void toXML(ContentHandler handler) throws SAXException {
		String display = toString();
		handler.characters(display.toCharArray(), 0, display.length());
	}

	/**
	 * Returns the descendant accessible from this node via the given
	 * LocationExpression.
	 *
	 * This method allows only fully qualified access paths, i.e. in the
	 * LocationExpression passed in the argument, the steps should
	 * always have CHILD, ATTRIBUTE, SELF OR PARENT as axis and there should not
	 * be any wildcards in the namespaces and localnames. In another word
	 * each step should qualify only one node.
	 * @param e relative location expression from this node to the node to
	 * retrieve.
	 * @return the node specified by the location expression
	 */
	public XTreeNode getDescendant(PathExpr e) {
		if (e == null) {
			return this;
		}

		XTreeNode result = this;

		for (int i = 0; i < e.getSteps().size(); i++) {
			StepExpr step = e.getStep(i);
			byte axis = step.getAxis();
			switch (axis) {
				case Axis.ATTRIBUTE :
				case Axis.CHILD :
					result = result.getChild(step.getXNode());
					break;

				case Axis.SELF :
					break;

				case Axis.PARENT :
					result = result.getParent();
					break;

				default :
					throw new RuntimeException("Navigation axis [" + axis + "] not handled by this method.");
			}

			if (result == null) {
				break;
			}
		}
		return result;
	}

	///////////////////////////////////////////////////////////////////////////
	//                          PRIVATE UTILITIES
	///////////////////////////////////////////////////////////////////////////
	private void set(XTree tree, XTreeNode parent) {
		this.tree = tree;
		if (parent != null)
			parent.addChild(this);
	}

	private void setParent(XTreeNode parent) {
		this.parent = parent;
		location = null;
	}

	/* Override XNode and carry out problems when navigating : compare LocationExpressions instead
	   Carried out an infinite loop when calling equals() for LocationExpression
	    public boolean equals(Object o)
	    {
	        return (o instanceof XTreeNode) && getLocation().equals(((XTreeNode)o).getLocation());
	    }
	 */
	//
	//              STEP NAVIGATION
	//

	/**
	 * Returns the list of children matching the given step expression.
	 */
	private Collection getChildren(StepExpr step) {
		ArrayList set = new ArrayList();
		if (children == null)
			return set;
		Iterator it = children.values().iterator();
		while (it.hasNext()) {
			XTreeNode child = (XTreeNode) it.next();
			if (step.match(child.getHashKey()))
				set.add(child);
		}
		return set;
	}

	/**
	 * Returns the list of ancestors matching the given step expression.
	 */
	private Collection getAncestors(StepExpr step) {
		ArrayList ancestors = new ArrayList();

		XTreeNode ancestor = (XTreeNode) getParent();
		while (ancestor != null) {
			if (ancestor.getHashKey().match(step))
				ancestors.add(ancestor);
			ancestor = (XTreeNode) ancestor.getParent();
		}
		return ancestors;
	}

	/**
	 * Returns the list of descendants matching the given step expression.
	 */
	private Collection getDescendants(StepExpr step) {
		ArrayList descendants = new ArrayList();
		if (children == null)
			return descendants;

		Iterator it = children.values().iterator();
		while (it.hasNext()) {
			XTreeNode child = (XTreeNode) it.next();

			if (child.XModelNode.type != NodeKind.ELEMENT) {
				continue;
			}

			if (step.match(child.getHashKey())) {
				descendants.add(child);
			}
			descendants.addAll(child.getDescendants(step));
		}
		//System.out.println("DESC="+descendants);
		return descendants;
	}

	/**
	 * In a dynamic tree, PRECEDING, PRECEDING_SIBLING, FOLLOWING,
	 * and FOLLOWING_SIBLING are considered not well defined, since all
	 * nodes can be considered being optional.
	 */
	private Collection getPreceding(StepExpr step) {
		throw new UnsupportedOperationException();
	}

	private Collection getPrecedingSibling(StepExpr step) {
		throw new UnsupportedOperationException();
	}

	private Collection getFollowing(StepExpr step) {
		throw new UnsupportedOperationException();
	}

	private Collection getFollowingSibling(StepExpr step) {
		throw new UnsupportedOperationException();
	}

}
