/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xquery.metadata;

import java.io.Serializable;
import java.net.URL;
import java.util.*;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xquark.schema.Schema;
import org.xquark.schema.SchemaConstants;
import org.xquark.schema.SchemaManager;
import org.xquark.schema.validation.SchemaValidationContext;
import org.xquark.xml.xdbc.*;
import org.xquark.xquery.metadata.resolver.MetadataAccess;
import org.xquark.xquery.metadata.resolver.SourceMetadata;
import org.xquark.xquery.xdbc.AbstractXMLDataSourceMetaData;
import org.xquark.xquery.xdbc.PathSetXMLDocument;
import org.xquark.xquery.xdbc.StringDocument;


/**
 * Implementation of XDBC DataSourceMetadata offering other internal metadata
 * interfaces. Used by extractor and mediator.
 */
public class MetaDataImpl extends AbstractXMLDataSourceMetaData implements MetadataAccess, Serializable {
	// **********************************************************************
	// * Constant
	// **********************************************************************
	private static final String RCSRevision = "$Revision: 1.2 $";
	private static final String RCSName = "$Name:  $";
	private static final String FILE_MD_XSD = "/org/xquark/xml/xdbc/resources/XMLDBCMetaData.xsd";
	//private static final String FILE_ACC_XSD = "/org/xquark/xml/accessor/resources/accessor.xsd" ;

	// **********************************************************************
	// * Fields
	// **********************************************************************
	private HashMap metawrappers;
	private SchemaManager schemaManager = new SchemaManager();
	private SchemaManager loadingManager = null;
	private HashMap strschemas = null;

	// **********************************************************************
	// * Constructor
	// **********************************************************************

	/**
	 * Construct the class MetaData, loading default XSD
	 *
	 * @param
	 * @deprecated
	 */
	public MetaDataImpl(XMLConnection connection) {
		super(connection);
		metawrappers = new HashMap();
		strschemas = new HashMap();
		try {
			SchemaValidationContext context = new SchemaValidationContext();
			this.loadingManager = context.getSchemaManager();

			URL url = getClass().getResource(FILE_MD_XSD);
			if (url == null) {
				System.err.println("resource " + FILE_MD_XSD + " not found");
				return;
			}
			InputSource source = new InputSource(url.toString());
			loadingManager.loadSchema(context, source);

			//			url = getClass().getResource(FILE_ACC_XSD);
			//			if (url == null) {
			//				System.err.println("resource " + FILE_ACC_XSD + " not found");
			//				return;
			//			}
			//			source = new InputSource(url.toString());
			//			schema = loadingManager.loadSchema(context, source);

		} catch (SAXException saxe) {
			// must do throws MetaDataException in constructor
			//	   		throw new MetaDataException (saxe.getMessage ()) ;
			System.err.println("*** MetaDataException " + saxe.getMessage());
		}
	}

	public MetaDataImpl(XMLConnection connection, SchemaManager schemaManager) throws MetadataException {
		super(connection);
		this.schemaManager = schemaManager;
		strschemas = new HashMap();
		metawrappers = new HashMap();
	}

	// **********************************************************************
	// * Methods
	// **********************************************************************

	/**
	 *
	 *
	 * @param
	 *
	 * @return
	 */
	public void addMetaWrapper(MetaWrapper metawrapper) throws MetadataException {
		// incorrect argument
		if (metawrapper == null)
			throw new MetadataException("MetaData.addMetaWrapper : metawrapper to add is null");
		// get name of metawrapper
		String key = metawrapper.getSourceName();
		// incorrect name
		if (key == null)
			throw new MetadataException("MetaData.addMetaWrapper : metawrapper sourcename is null");
		// initialize metawrappers if necessary
		if (metawrappers == null)
			metawrappers = new HashMap();
		// verify metawrapper of same name not present
		if (metawrappers.get(key) != null) {
			System.err.println(key + " already loaded");
			return;
			// throw new MetadataException ("MetaData.addMetaWrapper : metawrapper already added") ;
		}
		// add metawrapper
		metawrappers.put(key, metawrapper);
	}

	// used exclusivily by the mediator (and mediator shouldn't use the other
	// method, because in some cases, wrapper.getSourceName() can be null
	// case of getURL of connection are null (eg. extractor with jdbc string)
	public void addMetaWrapper(MetaWrapper metawrapper, String wrappername) throws MetadataException {
		// incorrect argument
		if (metawrapper == null)
			throw new MetadataException("MetaData.addMetaWrapper : metawrapper to add is null");
		// get name of metawrapper
		String key = wrappername;
		// incorrect name
		if (key == null)
			throw new MetadataException("MetaData.addMetaWrapper : metawrapper sourcename is null");
		// initialize metawrappers if necessary
		if (metawrappers == null)
			metawrappers = new HashMap();
		// verify metawrapper of same name not present
		if (metawrappers.get(key) != null) {
			System.err.println(key + " already loaded");
			return;
			// throw new MetadataException ("MetaData.addMetaWrapper : metawrapper already added") ;
		}
		// add metawrapper
		metawrapper.setSourceName(key);
		metawrappers.put(key, metawrapper);
	}

	public MetaWrapper getMetaWrapper(String sourceName) {
		if (metawrappers == null)
			return null;
		return (MetaWrapper) metawrappers.get(sourceName);
	}

	public HashMap getMetaWrappers() {
		return metawrappers;
	}

	public void putSchema(Schema schema, String strschema) throws SAXException {
		if (schema == null)
			return;

		boolean load = schemaManager.putSchema(schema);
		strschemas.put(schema.getNamespace(), strschema);
		///*DEBUG*/System.out.println("PUTTING [" + strschema + "] :" + schema.getNamespace()) ;
	}

	public Collection getSchemas() {
		return schemaManager.getSchemas();
	}

	public Schema getSchemaSchema(String schemaname) {
		return schemaManager.getSchema(schemaname);
	}

	// returns a Collection of SourceMetadata
	public Collection getSources() {
		if (metawrappers == null)
			return null;
		return metawrappers.values();
	}

	public SourceMetadata getSourceMetadata(String sourceName) {
		if (metawrappers == null)
			return null;
		return (SourceMetadata) metawrappers.get(sourceName);
	}

	public SchemaManager getSchemaManager() {
		return schemaManager;
	}

	public SchemaManager getLoadingManager() {
		return loadingManager;
	}

	public String getStringSchema(String targetNamespace) {
		return (String) strschemas.get(targetNamespace);
	}

	// **********************************************************************
	/**
	 * Get the property with the specified name
	 * @return the value of the property
	 * @throws XMLDBCNotRecognizedException if the property is not known
	 */
	public Object getProperty(String propertyId) throws XMLDBCNotRecognizedException {
		if (propertyId.equalsIgnoreCase("xmldbc.vendor.version"))
			return new String("$Revision: 1.2 $");
		//throw new XMLDBCNotSupportedException("getProperty not yet supported");
		throw new XMLDBCNotRecognizedException("Property \"" + propertyId + "\" not defined.");
	}

	/**
	 * Retrieves the name of all XML collections in this data source.
	 * @return a list of all XML collection names as String objects.
	 * @throws XMLDBCException if a data source access error occurs.
	 * @throws XMLDBCNotSupportedException if the data source does not support this operation (typically, when the data source is read-only)
	 */
	public List getCollectionNames() throws XMLDBCException, XMLDBCNotSupportedException {
		ArrayList al = new ArrayList();
		HashMap ht = getMetaWrappers();
		for (Iterator e = ht.values().iterator(); e.hasNext();) {
			MetaWrapper mwi = (MetaWrapper) e.next();
			List ai = mwi.getCollectionsName();
			al.addAll(ai);
		}
		return al;
	}

	/**
	 * Retrieves the available schema namespaces in the data source.
	 * @return the target namespaces of all schemas available in the data source
	 * in a list of String objects.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public List getSchemaNamespaces() throws XMLDBCException {
		ArrayList al = new ArrayList();
		Collection cols = getSchemas();
		for (Iterator it = cols.iterator(); it.hasNext();) {
			Schema schema = (Schema) it.next();
			if (schema == null)
				continue;
			String namespace = schema.getNamespace();
			if (namespace != null) {
				if (namespace.equals(SchemaConstants.XMLSCHEMA_URI))
					continue;
				if (namespace.equals(SchemaConstants.XML_URI))
					continue;
				if (namespace.equals(SchemaConstants.XSI_URI))
					continue;
			}
			al.add(schema.getNamespace());
		}
		return al;
	}

	/**
	 * Retrieves the XML document describing all existing paths in all documents
	 * of the specified XML collection.
	 * @param the name of the XML collection.
	 * @return an XML document containing the path set.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public XMLDocument getPathSet(String colName) throws XMLDBCException {
		HashMap ht = getMetaWrappers();
		for (Iterator e = ht.values().iterator(); e.hasNext();) {
			MetaWrapper mwi = (MetaWrapper) e.next();
			HashMap ai = mwi.getMetaCollections();
			for (Iterator it = ai.keySet().iterator(); it.hasNext();) {
				String collectionName = (String) it.next();
				if (collectionName.equals(colName)) {
					MetaCollection mc = (MetaCollection) ai.get(collectionName);
					// beware, XMLCollection is null ! 
					PathSetXMLDocument doc = new PathSetXMLDocument(collectionName, mc.getXTree());
					return doc;
				}
			}
		}
		throw new XMLDBCException("Collection " + colName + " not found.");
	}

	/**
	 * Retrieves the XML document describing the XML schema for the given namespace.
	 * @param the target namespace of the schema.
	 * @return an XML document containing the XML schema.
	 * @throws XMLDBCException if a data source access error occurs.
	 */
	public XMLDocument getSchema(String targetNamespace) throws XMLDBCException {
		//Schema schema = getSchemaSchema (targetNamespace) ;
		// ici, si on peut faire un truc du style
		// SchemaXMLDocument xmldoc = new SchemaXMLDocument (schema) ;
		// return  xmldoc
		String strschema = getStringSchema(targetNamespace);
		// when it's null (i.e. it isn't in set of schema), it's not
		// necessary because of an error. It happens, because definition
		// namespace as "http://www.w3.org/2001/XMLSchema" are also defined,
		// but aren't considered as useful schema.
		if (strschema == null)
			return null;

		XMLDocument doc = new StringDocument(strschema, true);
		return doc;
	}

	//	public void clear() throws XMLDBCException {
	//		// clear all schema in schemamanager (but not in loading schema)
	////		Collection col = schemaManager.getSchemas();
	////		ArrayList names = new ArrayList();
	////		///  no, because, if we remove, we delete "system" schema as
	////		// http://www.w3.org/2001/XMLSchema
	////		// http://www.w3.org/XML/1998/namespace
	////		// and that's not cool !
	////		for (Iterator it = col.iterator(); it.hasNext();) {
	////			Schema schema = (Schema) it.next();
	////			names.add(schema.getNamespace());
	////		}
	////
	////		for (int i = 0; i < names.size(); i++) {
	////			schemaManager.removeSchema((String) names.get(i));
	////		}
	//
	//		metawrappers.clear();
	//		strschemas.clear();
	//	}

	// **********************************************************************
	// * Destructor
	// **********************************************************************

	/**
	 * Dereference allocated object which will be further be destroy
	 * by the garbage collector.
	 */
	protected void finalize() throws Throwable {}

//	// XPathResolver interface
//
//	public ArrayList resolveXPath(ArrayList xnodes, ArrayList steps) {
//		if (steps == null || steps.isEmpty())
//			return xnodes;
//		for (int i=0;i<steps.size();i++) {
//			xnodes = resolveXPath(xnodes,(StepExpr)steps.get(i));
//		}
//		return xnodes;
//	}
//
//	public ArrayList resolveXPath(ArrayList xnodes, StepExpr step) {
//		if (xnodes == null || xnodes.isEmpty())
//			return null;
//		ArrayList reslist = new ArrayList();
//		for (int i=0;i<xnodes.size();i++) {
//			Collection col = ((XTreeNode)xnodes.get(i)).navigate(step);
//			reslist.addAll(col);			
//		}
//		if (reslist.isEmpty())
//			reslist = null;
//		return reslist;
//	}
//
//	public QType resolveQType(byte context, QType qtype, StepExpr step) {
//		return null;
//	}
//
//	private final static String COLON = ":";
//	private final static String STAR = "*";
//	private final static String DOT = ".";
//
//	public ArrayList getCollectionRootNodes(String collectionName) {
//
//		collectionName = collectionName.trim();
//		String sourceName = STAR;
//		int index = collectionName.indexOf(COLON);
//		///*DEBUG*/ System.out.println("index: "+index);
//		if (index == -1)
//			collectionName = collectionName.substring(0, collectionName.length());
//		else {
//			sourceName = collectionName.substring(0, index);
//			collectionName = collectionName.substring(index + 1, collectionName.length());
//		}
//
//		ArrayList al = new ArrayList();
//		MetaWrapper mw = null;
//		mw = this.getMetaWrapper(sourceName);
//		if (mw != null) {
//			MetaCollection metacol = mw.getMetaCollection(collectionName);
//			if (metacol != null)
//				al.add(metacol.getXTree().getRoot());
//			else { // treat case where collection name ends with .collectionName
//				List list = mw.getCollectionsName();
//				for (Iterator it = list.iterator();it.hasNext();) {
//					String colName = (String)it.next();
//					if (STAR.equals(collectionName) || colName.endsWith(DOT+collectionName)) {
//						metacol = mw.getMetaCollection(colName);
//						al.add(metacol.getXTree().getRoot());
//						break;
//					}
//				}
//			}
//		} 
//		else {
//			HashMap ht = getMetaWrappers();
//			for (Iterator e = ht.values().iterator(); e.hasNext();) {
//				mw = (MetaWrapper) e.next();
//				MetaCollection metacol = mw.getMetaCollection(collectionName);
//				if (metacol != null)
//					al.add(metacol.getXTree().getRoot());
//				else { // treat case where collection name ends with .collectionName
//					List list = mw.getCollectionsName();
//					for (Iterator it = list.iterator();it.hasNext();) {
//						String colName = (String)it.next();
//						if (STAR.equals(collectionName) || colName.endsWith(DOT+collectionName)) {
//							metacol = mw.getMetaCollection(colName);
//							al.add(metacol.getXTree().getRoot());
//							break;
//						}
//					}
//				}
//			}
//		}
//		if (al.isEmpty())
//			al = null;
//		return al;
//	}
//
//	public XNode getDocumentRootNode(URI documentURI) {
//		return null;
//	}
//
//	public ArrayList getInputRootNodes() {
//		ArrayList al = new ArrayList();
//		HashMap ht = getMetaWrappers();
//		for (Iterator e = ht.values().iterator(); e.hasNext();) {
//			MetaWrapper mwi = (MetaWrapper) e.next();
//			HashMap metacols = mwi.getMetaCollections();
//			Collection values = metacols.values();
//			for (Iterator it = values.iterator(); it.hasNext();) {
//				al.add(((MetaCollection) it.next()).getXTree().getRoot());
//			}
//		}
//		if (al.isEmpty())
//			al = null;
//		return al;
//	}
}
