/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */


package org.xquark.xquery.metadata.resolver;

import java.util.*;

import org.xquark.xpath.Axis;
import org.xquark.xpath.TypedXTreeNode;
import org.xquark.xpath.XTree;
import org.xquark.xquery.parser.*;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.FunctionCOLLECTION;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.FunctionDOC;

// ***************************************************************************
// * Class 'PathResolver'
// ***************************************************************************

/**
 * The PathResolver is a class that allow resolving any path expression given in
 * argument. The class structure is organized as following :
 *
 * <pre>
 *  PathResolver
 * '-------------------------------------------------------------------------`
 * |    Path                                                                 |
 * |   '-----------------------------------------------------------------`   |
 * |   |                          Source                                 |   |
 * |   |                         '-----------------------------------`   |   |
 * |   |                         |                ElementPath        |   |   |
 * |   |                         |                '------------`     |   |   |
 * |   |                         |                | node       |     |   |   |
 * |   |                         | sourcename     '------------`     |   |   |
 * |   |                         |                |            |     |   |   |
 * |   |                         |                '------------`     |   |   |
 * |   |                         |                |            |     |   |   |
 * |   |                         |                '------------`     |   |   |
 * |   |   pathname              `-----------------------------------'   |   |
 * |   |                                                                 |   |
 * |   |                         '-----------------------------------`   |   |
 * |   |                         |                '------------`     |   |   |
 * |   |                         |                |            |     |   |   |
 * |   |                         |                '------------`     |   |   |
 * |   |                         |                |            |     |   |   |
 * |   |                         |                '------------`     |   |   |
 * |   |                         |                |            |     |   |   |
 * |   |                         |                '------------`     |   |   |
 * |   |                         `-----------------------------------'   |   |
 * |   `-----------------------------------------------------------------'   |
 * |                                                                         |
 * |   '-----------------------------------------------------------------`   |
 * |   |                         '-----------------------------------`   |   |
 * |   |                         |                '------------`     |   |   |
 * |   |                         |                |            |     |   |   |
 * |   |                         |                '------------`     |   |   |
 * |   |                         |                |            |     |   |   |
 * |   |                         |                '------------`     |   |   |
 * |   |                         |                |            |     |   |   |
 * |   |                         |                '------------`     |   |   |
 * |   |                         `-----------------------------------'   |   |
 *
 * [...]
 *
 *
 * '-------------------------------------------------------------------------`
 * </pre>
 *
 */
public class PathResolver {
	private static final String RCSRevision = "$Revision: 1.3 $";
	private static final String RCSName = "$Name:  $";
	// **********************************************************************
	// * Constant
	// **********************************************************************
	public final static String COLLECTION = "collection";
	public final static char OPEN_PARENTHESIS = '(';
	public final static String CLOSE_PARENTHESIS = ")";
	public final static String COMA = ",";
	public final static String COLON = ":";
	public final static String STAR = "*";

	// **********************************************************************
	// * Fields
	// **********************************************************************
	/**
	 * A reference on the metadata
	 */
	private MetadataAccess metadata = null;

	/**
	 * A vector of Path.
	 */
	private ArrayList resolvers = new ArrayList();

	// for extraction speed up
	private String path = null;
	private ArrayList restrictedCollection = new ArrayList(1);
	private ArrayList restrictedSource = new ArrayList(1);
	private String restrictedDocument = null;

	private ArrayList sourceNames = new ArrayList(1);
	private ArrayList urls = new ArrayList(1);
	private ArrayList nodes = new ArrayList(1);
	// **********************************************************************
	// * Constructor
	// **********************************************************************

	public PathResolver(MetadataAccess metadata) {
		this.metadata = metadata;
	}

	private void clear() {
		resolvers.clear();
		sourceNames = new ArrayList(1);
		urls = new ArrayList(1);
		nodes = new ArrayList(1);
	}
	
	// resolving paths
	public void resolveSteps(RootNodeFunctionCall expr) throws XQueryException {
		if (expr == null)
			return;
		clear();
		extractRestriction(expr);
		Path p = null;
		//        if ((restrictedCollection == null) || (restrictedCollection.isEmpty()))
		//            p = new Path(steps, lemaker, metadata, null, null);
		//        else
		p = new Path(expr, metadata, restrictedSource, restrictedCollection, restrictedDocument);
		p.init();
		resolvers.add(p);
	}

	public void resolveSteps(ArrayList steps) throws XQueryException {
		if (steps == null || steps.isEmpty())
			return;
		clear();
		extractRestriction((org.xquark.xquery.parser.Step) steps.get(0));
		Path p = null;
		//        if ((restrictedCollection == null) || (restrictedCollection.isEmpty()))
		//            p = new Path(steps, lemaker, metadata, null, null);
		//        else
		p = new Path(steps, metadata, restrictedSource, restrictedCollection, restrictedDocument);
		p.init();
		resolvers.add(p);
	}

	public void resolveStepsList(ArrayList stepslist) throws XQueryException {
		if (stepslist == null || stepslist.isEmpty())
			return;
		clear();
		for (int i = 0; i < stepslist.size(); i++) {
			ArrayList steps = (ArrayList) stepslist.get(i);
			if (steps == null || steps.isEmpty())
				continue;
			extractRestriction((org.xquark.xquery.parser.Step) steps.get(0));
			Path p = null;
			//            if ((restrictedCollection == null) || (restrictedCollection.isEmpty()))
			//                p = new Path(steps, lemaker, metadata, null, null);
			//            else
			p = new Path(steps, metadata, restrictedSource, restrictedCollection, restrictedDocument);
			p.init();
			resolvers.add(p);
		}
	}

	// for decomposing
	private void extractRestriction(org.xquark.xquery.parser.XQueryExpression expr) {
		restrictedCollection.clear();
		restrictedSource.clear();
		restrictedDocument = null;;
		if (expr instanceof FunctionCOLLECTION) {
			ArrayList args = ((FunctionCall) expr).getArguments();
			fillCollectionLocation(((ValueString) args.get(0)).getStringValue());
		} else if (expr instanceof FunctionDOC) {
			ArrayList args = ((FunctionCall) expr).getArguments();
			fillDocumentLocation(((ValueString) args.get(0)).getStringValue());
		}
	}
	
	private void extractRestriction(org.xquark.xquery.parser.Step step) {
		extractRestriction(step.getExpression());
	}

	//    private void fillLocation (ArrayList restrictedCollection, ArrayList restrictedSource, String location)
	private void fillCollectionLocation(String location) {
		///*DEBUG*/ System.out.println("location: "+location);
		location = location.trim();
		String source = STAR;
		String collection = null;
		int index = location.indexOf(COLON);
		///*DEBUG*/ System.out.println("index: "+index);
		if (index == -1)
			collection = location.substring(0, location.length());
		else {
			source = location.substring(0, index);
			collection = location.substring(index + 1, location.length());
		}
		///*DEBUG*/ System.out.println("source: "+source);
		///*DEBUG*/ System.out.println("collection: "+collection);
		restrictedCollection.add(collection.trim());
		restrictedSource.add(source.trim());
		///*DEBUG*/ System.out.println("restrictedCollection: "+restrictedCollection);
		///*DEBUG*/ System.out.println("restrictedSource: "+restrictedSource);
	}

	private void fillDocumentLocation(String location) {
		//restrictedCollection.add(location) ;
		//restrictedSource.add(location) ;
		restrictedDocument = location;
	}

	//	private final String EVAL_SOURCE_NAME = "EVAL";
	//	private int evalSourceIndex = 0;

	// **********************************************************************
	// * Methods
	// **********************************************************************    

//	/**
//	 * Return all path expression.
//	 *
//	 * @return a vector of all path name.
//	 */
//	public ArrayList getPathNames() {
//		if (!pathNames.isEmpty()) return pathNames;
//		for (int i = 0; i < resolvers.size(); i++) {
//			String pathname = ((Path) resolvers.get(i)).getPathName();
//			if (!pathNames.contains(pathname))
//			pathNames.add(pathname);
//		}
//		return pathNames;
//	}

	/**
	 * Return all sources names referenced.
	 *
	 * @return a vector of source name.
	 */
	public ArrayList getSourceNames() {
		if (!sourceNames.isEmpty()) return sourceNames;
		for (int i = 0; i < resolvers.size(); i++) {
			ArrayList sources = ((Path) resolvers.get(i)).getSourceNames();
			if (sources == null)
				continue;
			for (int j = 0; j < sources.size(); j++) {
				String sourcename = (String) sources.get(j);
				if (!sourceNames.contains(sourcename))
				sourceNames.add(sourcename);
			}
		}
		return sourceNames;
	}

	public ArrayList getUrls() {
		if (!urls.isEmpty()) return urls;
		for (int i = 0; i < resolvers.size(); i++) {
			String tmpstr = ((Path) resolvers.get(i)).getUrl();
			if (tmpstr == null)
				continue;
			if (!urls.contains(tmpstr))
				urls.add(tmpstr);
		}
		return urls;
	}
//	public ArrayList getArraySources() {
//		ArrayList v = new ArrayList();
//		for (int i = 0; i < resolvers.size(); i++) {
//			ArrayList sources = ((Path) resolvers.get(i)).getSourceName();
//			if (sources == null)
//				continue;
//			for (int j = 0; j < sources.size(); j++) {
//				String sourcename = (String) sources.get(j);
//				if (!v.contains(sourcename))
//					v.add(sourcename);
//			}
//		}
//		return v;
//	}

//	/**
//	 * Return all sources  names for a given path expression.
//	 *
//	 * @param pathname the path expression to search source for.
//	 *
//	 * @return a vector of source name.
//	 */
//	public ArrayList getSources(String pathname) {
//		Path pname = getPathByPathName(pathname);
//		if (pname == null)
//			return null;
//		return pname.getSourceNames();
//	}

//	// gets a path by full path name
//	private Path getPathByFullPathName(String fullpathname) {
//		for (int i = 0; i < resolvers.size(); i++) {
//			Path pi = (Path) resolvers.get(i);
//			if (fullpathname.equals(pi.getFullPathName()))
//				return pi;
//		}
//		return null;
//	}

	/**
	 * Return the Path class whith the given name.
	 *
	 * @param orig_pathname the name of the Path class to return
	 *
	 * @return a reference on the associated Path class.
	 */
	private Path getPathByPathName(String pathname) {
		for (int i = 0; i < resolvers.size(); i++) {
			Path pi = (Path) resolvers.get(i);
			if (pathname.equals(pi.getPathName()))
				return pi;
		}
		return null;
	}

	/**
	 * Return all nodes answering the set of path expression, in all sources.
	 *
	 * @return a vector of nodes.
	 */
	public ArrayList getNodes() {
		if (!nodes.isEmpty()) return nodes;
		///*DEBUG*/System.out.println("START PathResolver::getNodes()");
		///*DEBUG*/System.out.println("resolvers = " + resolvers);
		for (int i = 0; i < resolvers.size(); i++) {
			ArrayList elementnodes = ((Path) resolvers.get(i)).getNodes();
			if (elementnodes != null)
				for (int j = 0; j < elementnodes.size(); j++) {
					TypedXTreeNode nodej = (TypedXTreeNode) elementnodes.get(j);
					// Ca ca marche pas bien (Stack overflow) :  je commente en attendant, mais il va
					// y avoir des doublons...
					//if (! v.contains (nodej))
					nodes.add(nodej);
				}
		}
		///*DEBUG*/System.out.println("STOP  PathResolver::getNodes() --> " + v);
		return nodes;
	}

//	/**
//	 * Return a vector of nodes associated with the
//	 * given path expression (search in all sources).
//	 *
//	 * @param path the pathname where to find the nodes
//	 * @return a vector of nodes.
//	 */
//	public ArrayList getNodes(String pathname) {
//		Path pname = getPathByPathName(pathname);
//		if (pname == null)
//			return null;
//		return pname.getNodes();
//	}

//	/**
//	 * Return a vector of nodes associated with the given source and the
//	 * given path expression.
//	 *
//	 * @param sourcename the source where to search the nodes
//	 * @param path the pathname where to find the nodes
//	 * @return a vector of nodes.
//	 */
//	public ArrayList getNodes(String pathname, String sourcename) {
//		Path pname = getPathByPathName(pathname);
//		if (pname == null)
//			return null;
//		return pname.getNodes(sourcename);
//	}

//	/**
//	 * Return a vector of nodes associated with the given source.
//	 *
//	 * @param sourcename the source where to search the nodes
//	 * @return a vector of nodes.
//	 */
//	public ArrayList getNodesBySource(String sourcename) {
//		ArrayList v = new ArrayList();
//		for (int i = 0; i < resolvers.size(); i++) {
//			Path pathi = (Path) resolvers.get(i);
//			ArrayList elementnodes = pathi.getNodes(sourcename);
//
//			if (elementnodes != null) {
//				for (int j = 0; j < elementnodes.size(); j++) {
//					TypedXTreeNode nodej = (TypedXTreeNode) elementnodes.get(j);
//					if (!v.contains(nodej))
//						v.add(nodej);
//				}
//			}
//		}
//		return v;
//	}

//	/**
//	 *
//	 */
//	public ArrayList getNodesByCollection(String pathname, String sourcename, String colname) {
//		Path path = getPathByPathName(pathname);
//		if (path == null) {
//			System.err.println("No such path : " + pathname);
//			return null;
//		}
//		return path.getNodesByCollection(sourcename, colname);
//	}

//	/**
//	 *
//	 */
//	public ArrayList getNodesByCollection(String pathname, String colname) {
//		Path path = getPathByPathName(pathname);
//		if (path == null) {
//			System.err.println("No such path : " + pathname);
//			return null;
//		}
//		return path.getNodesByCollection(colname);
//	}

//	/**
//	 *
//	 */
//	public ArrayList getColname(String pathname, String sourcename) {
//		Path path = getPathByPathName(pathname);
//		if (path == null) {
//			System.err.println("No such path : " + pathname);
//			return null;
//		}
//		return path.getColname(sourcename);
//	}

//	/**
//	 * Display (for debugging) the state of the PathResolver.
//	 */
//	public void display() {
//		System.out.println("**** PathResolver");
//		for (int i = 0; i < resolvers.size(); i++) {
//			Path pathi = (Path) resolvers.get(i);
//			pathi.display();
//		}
//		Collection colSources = metadata.getSources();
//		Iterator it = colSources.iterator();
//		while (it.hasNext()) {
//			SourceMetadata source = (SourceMetadata) it.next();
//			System.out.println("***** SourceMetaData " + source.getSourceName());
//			Collection colcollection = source.getCollectionsMetadata();
//			Iterator it2 = colcollection.iterator();
//			while (it2.hasNext()) {
//				CollectionMetadata colmeta = (CollectionMetadata) it2.next();
//				XTree xtree = colmeta.getXTree();
//				System.out.println("****** CollectionMetadata :\n" + xtree);
//			}
//		}
//
//	}

	// **********************************************************************
	// * Destructor
	// **********************************************************************

	/**
	 * Dereference allocated object which will be further be destroy
	 * by the garbage collector.
	 */
	protected void finalize() throws Throwable {
		super.finalize();
		metadata = null;
		resolvers = null;
	}
}

/* ************************************************************************ */

/**
 * A Path class. Used to describe a "XPath" path no considering
 * the number of sources where the matching nodes are.
 *
 *
 */
class Path {
	private static final String RCSRevision = "$Revision: 1.3 $";
	private static final String RCSName = "$Name:  $";
	private String pathname = null;
	private String fullpathname = null;
	private ArrayList sources = null;
	private MetadataAccess metadata = null;
	//private LocationExpression locationexpression = null;
	private ArrayList steps = null;

	private ArrayList restrictedSource = null;
	private ArrayList restrictedCollection = null;
	private String restrictedDocument = null;
	private ArrayList nodes = null;

	/**
	 * Create a path expression.
	 *
	 * @param pathname the name of the path
	 *        (eg. <pre>//person/ * /town/ * /zip</pre>)
	 * @param metadata a reference on metadata in order to resolve the
	 *        path expression.
	 */

	public Path(ArrayList steps, MetadataAccess metadata, ArrayList restrictedSource, ArrayList restrictedCollection, String restrictedDocument) throws XQueryException {
		// temporary code
		this.steps = (ArrayList)steps.clone();
		fullpathname = printSteps(steps, 0);
		if (((org.xquark.xquery.parser.Step) steps.get(0)).getExpression() instanceof FunctionCOLLECTION) {
			pathname = printSteps(steps, 1);
			if (this.steps.size() == 1)
				this.steps = null;
			else
				this.steps.remove(0);
		}
		else {
			pathname = fullpathname;
		}
		//------------------
		//sources = new ArrayList() ;
		this.metadata = metadata;
		if (restrictedSource.isEmpty())
			restrictedSource = null;
		this.restrictedSource = restrictedSource;
		if (restrictedCollection.isEmpty())
			restrictedCollection = null;
		this.restrictedCollection = restrictedCollection;
		this.restrictedDocument = restrictedDocument;
		//this.locationexpression = lemaker.make(steps);
		//this.steps = steps;
	}

	public Path(RootNodeFunctionCall expr, MetadataAccess metadata, ArrayList restrictedSource, ArrayList restrictedCollection, String restrictedDocument) throws XQueryException {
		// temporary code
		fullpathname = expr.toString();
		pathname = "";
		//------------------
		//sources = new ArrayList() ;
		this.metadata = metadata;
		if (restrictedSource.isEmpty())
			restrictedSource = null;
		this.restrictedSource = restrictedSource;
		if (restrictedCollection.isEmpty())
			restrictedCollection = null;
		this.restrictedCollection = restrictedCollection;
		this.restrictedDocument = restrictedDocument;
		//this.locationexpression = null;
	}


	private String printSteps(ArrayList steps, int minIndex) {
		StringBuffer buf = new StringBuffer();
		//		if (expression != null)	buf.append(expression.pprint());
		for (int i = minIndex; i < steps.size(); i++) {
			org.xquark.xquery.parser.Step stepi = (org.xquark.xquery.parser.Step) steps.get(i);
			boolean separator = stepi.hasSeparator();
			int axis = stepi.getAxis();
			if (separator)
				buf.append('/');
			if (axis != -1)
				buf.append(Axis.AXISSTRINGS[axis]);
			buf.append(stepi.getExpression().toSimpleString());
		}
		return buf.toString();
	}

	/**
	 * initialize a Path structure (mandatory). Usually this should be called
	 * by the PathResolver initialization
	 */
	public void init() {
		if (restrictedDocument != null)
			return;

		Collection sourceCollection = metadata.getSources();
		// /*DEBUG*/ System.out.println("schemas: "+schemas);
		Iterator sourceIterator = sourceCollection.iterator();
		while (sourceIterator.hasNext()) {
			SourceMetadata sourceMetadata = (SourceMetadata) sourceIterator.next();
			String sourcename = sourceMetadata.getSourceName();
			// /*DEBUG*/ System.out.println("sourcename: "+sourcename);

			if (restrictedSource != null)
				if ((!restrictedSource.contains(PathResolver.STAR)) && (!restrictedSource.contains(sourcename))) {
					continue;
				}

			Collection collections = sourceMetadata.getCollectionsMetadata();
			// /*DEBUG*/ System.out.println("*DEBUG* collections: "+collections);
			Iterator collectionsIterator = collections.iterator();

			Source source = new Source(sourcename);
			boolean sourceIsValid = false;
			while (collectionsIterator.hasNext()) {
				CollectionMetadata collectionMetadata = (CollectionMetadata) collectionsIterator.next();
				String colname = collectionMetadata.getCollectionName();
				// /*DEBUG*/ System.out.println("*DEBUG* colname: "+colname);

				if (restrictedCollection != null) {
					boolean found = false;
					for (int r = 0; r < restrictedCollection.size(); r++) {
						String str = (String) restrictedCollection.get(r);
						// /*DEBUG*/ System.out.println("*DEBUG* str: "+str);
						if (!str.equals(PathResolver.STAR)) {
							if ((sourceMetadata.isCaseSensitive() && !str.equals(colname) && !colname.endsWith("." + str))) {
								continue;
							}
							if (!sourceMetadata.isCaseSensitive()) {
								if (str.equalsIgnoreCase(colname)) {
									/* found */
								} else {
									int startIndex = colname.length() - str.length() - 1;
									if (0 > startIndex) {
										continue;
									}
									String suffix = colname.substring(startIndex, colname.length());
									if (!suffix.equalsIgnoreCase("." + str)) {
										continue;
									}
								}
							}
						}

						String sname = (String) restrictedSource.get(r);
						// /*DEBUG*/ System.out.println("*DEBUG* sname: "+sname);
						if ((!sname.equals(PathResolver.STAR)) && ((!sname.equals(sourcename))))
							continue;

						found = true;
						break;
					}
					//  /*DEBUG*/ System.out.println("found: "+found);
					if (!found)
						continue;
				}

				XTree schemaTree = collectionMetadata.getXTree();
				//  /*DEBUG*/ System.out.println("*DEBUG* schemaTree: "+schemaTree);

				// /*DEBUG*/ System.out.println("*DEBUG* pathname: '"+pathname+"'");
				if (pathname.equals("")) {
					// if there are no path, use the whole root.
					Collection c = schemaTree.getRoot().getChildren();
					// /*DEBUG*/ System.out.println("c: '"+c+"'");
					if (c == null)
						continue;
					Object[] sn = c.toArray();
					TypedXTreeNode tmpNode = null;
					for (int l = 0; l < sn.length; l++) {
						tmpNode = (TypedXTreeNode) sn[l];
						// /*DEBUG*/ System.out.println("*DEBUG* firstnode: '"+firstnode+"'");
						//XMLSchemaTools.displayDeclaration (firstnode.getDeclaration ()) ;
						source.add(colname, tmpNode);
					}
					sourceIsValid = true;
				} else {
					// /*DEBUG*/ System.out.println("*DEBUG* locationexpression: "+locationexpression);
					/*
					if (locationexpression == null)
						continue;
					*/
					if (steps == null)
						continue;
					//Collection set = schemaTree.prune(locationexpression);
					Collection set = schemaTree.prune(steps);
					//  /*DEBUG*/ System.out.println("set: '"+set);
					// /*DEBUG*/ System.out.println("set.size(): '"+set.size());
					if (set.size() > 0) {
						source.add(colname, set);
						sourceIsValid = true;
					}
				}
				// /*DEBUG*/ System.out.println("source: "+source);

			}
			if (sourceIsValid == true) {
				if (sources == null)
					sources = new ArrayList();
				sources.add(source);
			}
		}
	}

//	/**
//	 * return an arraylist of class sources
//	 */
//	public ArrayList getSources() {
//		return sources;
//	}

//	/**
//	 * Display the state of the Path instance for debugging
//	 */
//	public void display() {
//		System.out.println("*** Path : [" + pathname + "]");
//		System.out.println("    restricted collection : " + restrictedCollection);
//		System.out.println("    restricted source     : " + restrictedSource);
//		System.out.println("    restricted document     : " + restrictedDocument);
//		if (sources != null)
//			for (int i = 0; i < sources.size(); i++) {
//				Source sourcei = (Source) sources.get(i);
//				sourcei.display();
//			}
//	}

	/**
	 * Return the path.
	 *
	 * @return the string respresentation of the path.
	 */
	public String getPathName() {
		return pathname;
	}

//	/**
//	 * Return the full path.
//	 *
//	 * @return the string representation of the full path.
//	 */
//	public String getFullPathName() {
//		return fullpathname;
//	}

	/**
	 * Return the names of all sources.
	 *
	 * @return the names of all sources.
	 */
	public ArrayList getSourceNames() {
		if (sources == null)
			return null;
		ArrayList v = new ArrayList(sources.size());
		for (int i = 0; i < sources.size(); i++) {
			String sourcename = ((Source) sources.get(i)).getSourceName();
			v.add(sourcename);
		}
		return v;
	}

	/**
	 * return an arraylist document names
	 */
	public String getUrl() {
		return restrictedDocument;
	}

	/**
	 * Return all nodes of all sources for the current path expression.
	 *
	 * @return all nodes of all sources for the current path expression.
	 */
	public ArrayList getNodes() {
		if (nodes != null) return nodes;
		if (sources == null)
			return null;
		nodes = new ArrayList();
		for (int i = 0; i < sources.size(); i++) {
			ArrayList elementpaths = ((Source) sources.get(i)).getNodes();
			if (elementpaths != null)
				for (int j = 0; j < elementpaths.size(); j++) {
					TypedXTreeNode nodej = (TypedXTreeNode) elementpaths.get(j);
					nodes.add(nodej);
				}
		}
		return nodes;
	}

//	/**
//	 * Return the vector of nodes asociated with the given source.
//	 *
//	 * @param sourcename the name of the source to work on.
//	 * @return a vector of nodes referenced by the source for the current
//	 *         path expression.
//	 */
//	public ArrayList getNodes(String sourcename) {
//		Source sname = getSourceByName(sourcename);
//		if (sname != null)
//			return sname.getNodes();
//		return null;
//	}

//	/**
//	 * Provided the name of one source, return a reference on the associated
//	 * source.
//	 *
//	 * @param orig_sourcename the name of the source to return.
//	 * @return a reference on the associated source. Null if no source
//	 *         match the given name.
//	 */
//	private Source getSourceByName(String orig_sourcename) {
//		if (sources == null)
//			return null;
//		for (int i = 0; i < sources.size(); i++) {
//
//			Source si = (Source) sources.get(i);
//			String sourcename = si.getSourceName();
//			if (orig_sourcename.equals(sourcename)) {
//				return si;
//			}
//		}
//		return null;
//	}

//	/**
//	 * Return all reference on schema nodes matching this source name and
//	 * this collection name.
//	 *
//	 * @param sourcename the source on which we work on.
//	 * @param colname the collection to find.
//	 */
//	public ArrayList getNodesByCollection(String sourcename, String colname) {
//		Source source = getSourceByName(sourcename);
//		if (source == null) {
//			System.err.println("No matching source " + sourcename);
//			return null;
//		}
//		return source.getNodesByCollection(colname);
//	}

//	/**
//	 * Return all reference on schema nodes matching this source name and
//	 * this collection name.
//	 *
//	 * @param colname the collection to find.
//	 */
//	public ArrayList getNodesByCollection(String colname) {
//		if (sources == null)
//			return null;
//		ArrayList v = new ArrayList();
//		for (int i = 0; i < sources.size(); i++) {
//			Source si = (Source) sources.get(i);
//			ArrayList vi = si.getNodesByCollection(colname);
//			if (vi.size() > 0) {
//				for (int j = 0; j < vi.size(); j++) {
//					v.add(vi.get(j));
//				}
//			}
//		}
//		return v;
//	}

//	/**
//	 * Return all collection's name included in the specified source.
//	 *
//	 * @param sourcename the name of the source to work on.
//	 *
//	 * @return a vector of String representing the collections' name
//	 */
//	public ArrayList getColname(String sourcename) {
//		Source source = getSourceByName(sourcename);
//		if (source == null) {
//			System.err.println("No matching source " + sourcename);
//			return null;
//		}
//		return source.getColname();
//	}
}

/* ************************************************************************ */

/**
 * This class refers <b>for a specific Expression Path</b> on of the sources
 * matching the path.
 */
class Source {
	private static final String RCSRevision = "$Revision: 1.3 $";
	private static final String RCSName = "$Name:  $";
	private String sourcename;

	/**
	 * KEY : colname
	 * VALUE : vector of paths associated with this colname
	 */
	private Hashtable elementpaths;

	/**
	 * Construct a source with an associated sourcename.
	 *
	 * @param sourcename the name of the source.
	 */
	public Source(String sourcename) {
		this.sourcename = sourcename;
		elementpaths = new Hashtable();
	}

	/**
	 * Return the hashmap colname/nodes
	 */
	public Hashtable getElementPaths() {
		return elementpaths;
	}

	/**
	 * Add a set of schema nodes to a source.
	 *
	 * @param nodes the set of schema node to associate to the source.
	 */
	public void add(String colname, Collection nodes /* , String path */
	) {
		for (Iterator iterator = nodes.iterator(); iterator.hasNext();) {
			TypedXTreeNode node = (TypedXTreeNode) iterator.next();

			ArrayList roots = (ArrayList) elementpaths.get(colname);
			if (roots == null) {
				roots = new ArrayList();
				elementpaths.put(colname, roots);
			}
			//		    roots.add (new ElementPath (node)) ;
			roots.add(node);
		}
	}

	public void add(String colname, TypedXTreeNode node /* , String path */
	) {
		ArrayList roots = null;
		if ((roots = (ArrayList) elementpaths.get(colname)) == null) {
			roots = new ArrayList();
			elementpaths.put(colname, roots);
		}
		//	    roots.add (new ElementPath (node)) ;
		roots.add(node);
	}

	/**
	 * Display the state of the Source instance for debugging
	 */
	public void display() {
		//		System.out.println ("** Source : " + sourcename) ;
		//		// System.out.println("elementpaths = " + elementpaths);
		//		for (Enumeration e = elementpaths.keys () ; e.hasMoreElements () ; )
		//		{
		//			String colname = (String) e.nextElement () ;
		//
		//			ArrayList vep = (ArrayList)  elementpaths.get (colname) ;
		//			System.out.println ("* Collection " + colname) ;
		//			for (int i = 0 ; i < vep.size () ; i ++)
		//			{
		//			    ElementPath epi = (ElementPath) vep.get (i) ;
		//			    epi.display () ;
		//			}
		//		}
	}

	/**
	 * Return the name of the source.
	 *
	 * @return the name of the source.
	 */
	public String getSourceName() {
		return sourcename;
	}

	/**
	 * Return the vector of elements associated with this source for
	 * the current expression path.
	 *
	 * @return the vector of nodes associated with this source for
	 *         the current expression path.
	 */
	public ArrayList getNodes() {
		ArrayList v = new ArrayList();
		for (Enumeration e = elementpaths.elements(); e.hasMoreElements();) {
			ArrayList vep = (ArrayList) e.nextElement();

			for (int i = 0; i < vep.size(); i++) {
				//		        ElementPath epi = (ElementPath) vep.get (i) ;
				//		        v.add (epi.getNode ()) ;
				v.add(vep.get(i));
			}
		}
		return v;
	}

	/**
	 * Return all reference on schema nodes matching a collection name.
	 *
	 * @param colname the collection to find.
	 */
	public ArrayList getNodesByCollection(String colname) {
		ArrayList v = new ArrayList();
		ArrayList vep = (ArrayList) elementpaths.get(colname);
		if (vep == null)
			return v;

		for (int i = 0; i < vep.size(); i++) {
			//		    ElementPath epi = (ElementPath) vep.get (i) ;
			//		    v.add (epi.getNode ()) ;
			v.add(vep.get(i));
		}
		return v;
	}

	/**
	 * Return all collection's name included.
	 *
	 * @return a vector of String representing the collections' name
	 */
	public ArrayList getColname() {
		ArrayList v = new ArrayList();
		for (Enumeration e = elementpaths.keys(); e.hasMoreElements();) {
			String colname = (String) e.nextElement();
			v.add(colname);
		}
		return v;
	}
}

/* ************************************************************************ */

/**
 * The atomic structure representng a corresponding schema node of one
 * particular path expression on a particular source.
 */
//class ElementPath
//{
//    // String path ;
//    TypedXTreeNode node ;
//
//    /**
//     * Construct a reference on the schema node.
//     *
//     * @param node the node to refer
//     */
//    public ElementPath (TypedXTreeNode node)
//    {
//	this.node = node ;
//    }
//
//    /**
//     * Display the state of the ElementPath instance for debugging
//     */
//    public void display ()
//    {
//		// System.out.println ("ElementPath : " + path) ;
//		System.out.println ("ElementPath : " ) ;
//		Declaration dec = ((TypedXTreeNode) node).getDeclaration () ;
//		// if (dec != null) XMLSchemaTools.displayDeclaration (dec) ;
//		// else System.out.println("Could not find declaration");
//    }
//
//    /**
//     * Return the schema node referenced.
//     *
//     * @return the referenced schema node.
//     */
//    public TypedXTreeNode getNode ()
//    {
//
//	return node ;
//    }
//}
