/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * Copyright (C) 2003 XQuark Group.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xquery.parser;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.xml.sax.SAXException;
import org.xquark.schema.*;
import org.xquark.schema.datatypes.PrimitiveType;
import org.xquark.xpath.*;
import org.xquark.xquery.metadata.StaticContext;
import org.xquark.xquery.parser.primitivefunctions.fnfunctions.*;
import org.xquark.xquery.parser.primitivefunctions.myfunctions.FunctionABS;
import org.xquark.xquery.parser.primitivefunctions.xsfunctions.*;
import org.xquark.xquery.parser.util.*;
import org.xquark.xquery.typing.*;

public class TypeVisitor extends DefaultParserVisitor {
	private static final String RCSRevision = "$Revision: 1.16 $";
	private static final String RCSName = "$Name:  $";

	private static final String XML_SCHEMA = "http://www.w3.org/2001/XMLSchema";

	public static final String SC_STRING = "string";
	public static final String SC_BOOLEAN = "boolean";
	public static final String SC_DOUBLE = "double";
	public static final String SC_DECIMAL = "decimal";
	public static final String SC_INTEGER = "integer";
	public static final String SC_FLOAT = "float";
	public static final String SC_DATE = "date";
	public static final String SC_TIME = "time";
	public static final String SC_DATETIME = "dateTime";
	public static final String SC_ANYTYPE = "anyType";
	public static final String SC_ANYSIMPLETYPE = "anySimpleType";

	private StaticContext context = null;

	// new context stuff
	private ArrayList contextXNodes = null;
	private QType contextQType = null;
	private byte contextType = QTypeDocument.ANY_ORIGIN;
	private LocatedExpression contextLocated = null;
	private int contextLocatedStepIndex = -1;

	// context stuff
	private boolean inPredicate = false;
	private boolean isAttribute = false;
	private boolean noForce = true;

	private SchemaManager typingschemamanager = null;
	private Schema typingschema = null;
	// utils
	private UnitEnvironment unitEnv = null;

	// constructor

	public TypeVisitor() throws TypeException {
		if (typingschemamanager == null) {
			this.typingschemamanager = new SchemaManager();
			this.typingschema = new Schema("http://www.xquark.org/EXECUTIONPLAN", null, typingschemamanager);
			try {
				typingschemamanager.putSchema(typingschema);
			} catch (SAXException saxe) {
				throw new TypeException("Could not add schema to manager");
			}
		}
	}

	public void setContext(StaticContext staticContext) throws TypeException {
		context = staticContext;
		//init(staticContext.getSchemaManager(),null,staticContext.getMetaData());
	}

	public SchemaManager getSchemaManager() {
		return context.getSchemaManager();
	}

	public SchemaLocator getSchemaLocator() {
		return context.getSchemaManager();
	}

	public SchemaManager getTypingSchemaManager() {
		return typingschemamanager;
	}

	// new context stuff

	private void setTypingContext(ArrayList xnodes, QType qtype, LocatedExpression contextLocated, int contextLocatedStepIndex) throws XQueryException {
		this.contextXNodes = xnodes;
		this.contextQType = qtype;
		this.contextLocated = contextLocated;
		this.contextLocatedStepIndex = contextLocatedStepIndex;
	}

	private void setContextXNodes(ArrayList xnodes) throws XQueryException {
		this.contextXNodes = xnodes;
		this.contextQType = buildQTypeFromXNodes(xnodes);
	}

	private void setContextQType(QType qtype) {
		this.contextXNodes = null;
		this.contextQType = qtype;
	}

	// visitors

	//	public void visit(AggregateFunctionCall arg) throws XQueryException

	public void visit(AttributeValuePair arg) throws XQueryException {
		// TYPE VERIFICATION
		// attribute name
		// must be a simple type xsd:string or any type derivated from it
		// attribute value
		// must be a simple type xsd:string or any type derivated from it
		// TYPE BUILDING
		// --> QType::Attribute (name,type of value)

		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:AttributeValuePair ( " + arg + " )");
		XQueryExpression attributeName = arg.getAttributeName(); // cannot be null
		XQueryExpression attributeValue = arg.getAttributeValue(); // cannot be null
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (attributeValue.getQType() == null)
				return;
		if (!noForce || attributeValue.getQType() == null)
			attributeValue.accept(this);
		// type checking
		if (attributeValue.getQType() == null)
			throw new TypeException("Expression " + attributeValue + " could not be typed."); //return;
		// CHANGE   this is to restrictive
		//		if (attributeValue.getQType().isMultiple())
		//			throw new TypeException("QType of an attribute cannot be multiple.");
		//org.xquark.xml.misc.Debug.print(this,"arg.getQType(): "+arg.getQType());
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		QType attQType = attributeValue.getQType();
		//org.xquark.xml.misc.Debug.print(this,"valueQType: "+valueQType);
		//org.xquark.xml.misc.Debug.print(this,"arg.getAttributeName(): "+arg.getAttributeName());
		//org.xquark.xml.misc.Debug.print(this,"valueQType.getSimpleType(): "+valueQType.getSimpleType());
		arg.setQType(new QTypeAttribute(arg.getAttributeName(), attQType.getSimpleType()));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:AttributeValuePair ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(BinOpANDExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// both expressions must be booleans
		// TYPE BUILDING
		// the return type is boolean

		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:BinOpANDExpression ( " + arg + " )");
		XQueryExpression expr1 = arg.getExpression1();
		XQueryExpression expr2 = arg.getExpression2();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr1.getQType() == null)
				return;
			if (expr2.getQType() == null)
				return;
		}
		if (!noForce || expr1.getQType() == null)
			expr1.accept(this);
		if (!noForce || expr2.getQType() == null)
			expr2.accept(this);
		// type verification
		if (expr1.getQType() == null)
			throw new TypeException("Expression " + expr1 + " could not be typed."); //return;
		if (expr2.getQType() == null)
			throw new TypeException("Expression " + expr2 + " could not be typed."); //return;
		/*
		 if (!expr1.getQType().isBoolean())
						throw new TypeException("Expression '" + arg + "', left operand is not a boolean.");
		if (!expr2.getQType().isBoolean())
						throw new TypeException("Expression '" + arg + "', right operand is not a boolean.");
		 */
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getBooleanType()));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:BinOpANDExpression ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(BinOpORExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// both expressions must be booleans
		// TYPE BUILDING
		// the return type is boolean

		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:BinOpORExpression ( " + arg + " )");
		XQueryExpression expr1 = arg.getExpression1();
		XQueryExpression expr2 = arg.getExpression2();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr1.getQType() == null)
				return;
			if (expr2.getQType() == null)
				return;
		}
		if (!noForce || expr1.getQType() == null)
			expr1.accept(this);
		if (!noForce || expr2.getQType() == null)
			expr2.accept(this);
		// type verification
		if (expr1.getQType() == null)
			throw new TypeException("Expression " + expr1 + " could not be typed."); //return;
		if (expr2.getQType() == null)
			throw new TypeException("Expression " + expr2 + " could not be typed."); //return;
		/*
			 if (!expr1.getQType().isBoolean())
				throw new TypeException("Expression '" + arg + "', left operand is not a boolean.");
			if (!expr2.getQType().isBoolean())
				throw new TypeException("Expression '" + arg + "', right operand is not a boolean.");
			*/
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getBooleanType()));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:BinOpORExpression ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(CastTreatExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// TYPE BUILDING
		// the type is the sequenceType type

		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:CastTreatExpression ( " + arg + " )");
		XQueryExpression expr = arg.getExpression();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		// type checking
		if (expr.getQType() == null)
			throw new TypeException("Expression " + expr + " could not be typed."); //return;
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		Type type = getType(arg.getSequenceType());
		QType qtype = getQTypeFromType(type);
		if (qtype == null)
			throw new TypeException("Could not transform schema type '" + type + "'");
		arg.setQType(qtype);
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:CastTreatExpression ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(CData arg) throws XQueryException {
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
	}

	//	public void visit(ContextDeclaration arg) throws XQueryException;

	/*
	3.7.3.1 Computed Element Constructors
	[100]    CompElemConstructor    ::=    (("element" QName "{") | ("element" "{" Expr "}" "{")) Expr? "}" 
	
	The name expression of a computed element constructor is processed as follows:
	Atomization is applied to the value of the name expression. If the result of atomization is not a single atomic value of type xs:QName, xs:string, or xdt:untypedAtomic, a type error is raised.[err:XP0004][err:XP0006]
	If the atomized value of the name expression is of type xs:QName, that value is used as the name of the constructed element.
	If the atomized value of the name expression is of type xs:string or xdt:untypedAtomic, that value is cast to the type xs:QName using the rules in [XQuery 1.0 and XPath 2.0 Functions and Operators]. 
	The resulting value is used as the name of the constructed element. A dynamic error is raised if the cast is not successful.[err:XP0021]
	
	The content expression of a computed element constructor is processed as follows:
	For each node returned by the content expression, a new deep copy of the node is constructed, including all its children, attributes, and namespace nodes (if any). 
	Each copied node has a new node identity. Copied element nodes are given the type annotation xs:anyType, and copied attribute nodes are given the type annotation xs:anySimpleType. 
	For each adjacent sequence of one or more atomic values returned by the content expression, a new text node is constructed, containing the result of casting each atomic value to a string, with a single blank character inserted between adjacent values. 
	If any of these atomic values cannot be cast into a string, a dynamic error [err:XQ0052] is raised. The resulting sequence of nodes is called the content sequence. 
	Any sequence of adjacent text nodes in the content sequence is merged into a single text node.
	If the content sequence contains a document node, a type error is raised.[err:XQ0023]
	If the content sequence contains a namespace node following a node that is not a namespace node, a type error is raised.[err:XQ0040] Namespace nodes occurring in the content sequence are attached to the constructed element node.
	If the content sequence contains an attribute node following a node that is not an attribute node or a namespace node, a type error is raised.[err:XQ0024] Attribute nodes occurring in the content sequence become attributes of the new element node. 
	If two or more of these attribute nodes have the same name, an error is raised.[err:XQ0025]
	Element, text, comment, and processing instruction nodes in the content sequence become the children of the constructed element node.
	*/
	public void visit(Element arg) throws XQueryException {
		// TYPE VERIFICATION
		// --> NONE
		// TYPE BUILDING
		// the type returned is the type constructed from the metadata elements

		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:Element ( " + arg + " )");
		ArrayList attributes = arg.getAttributes();
		ArrayList expressions = arg.getSubExpressions();
		XQueryExpression startTag = arg.getStartTag();
		// forward type checking
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (!(startTag instanceof QName) && startTag.getQType() == null)
				return;
			if (attributes != null)
				for (int i = 0; i < attributes.size(); i++)
					if (((AttributeValuePair) attributes.get(i)).getQType() == null)
						return;
			if (expressions != null)
				for (int j = 0; j < expressions.size(); j++)
					if (((XQueryExpression) expressions.get(j)).getQType() == null)
						return;
		}
		if (!(startTag instanceof QName) && (!noForce || startTag.getQType() == null)) {
			startTag.accept(this);
			if (!(startTag instanceof QName) && startTag.getQType() == null)
				throw new TypeException("Expression " + startTag + " could not be typed.");
		}
		if (attributes != null)
			for (int i = 0; i < attributes.size(); i++)
				if (!noForce || ((AttributeValuePair) attributes.get(i)).getQType() == null) {
					((AttributeValuePair) attributes.get(i)).accept(this);
					if (((AttributeValuePair) attributes.get(i)).getQType() == null)
						throw new TypeException("Expression " + (AttributeValuePair) attributes.get(i) + " could not be typed."); //return;
				}
		if (expressions != null)
			for (int j = 0; j < expressions.size(); j++) {
				XQueryExpression exprj = (XQueryExpression) expressions.get(j);
				if (!noForce || exprj.getQType() == null) {
					exprj.accept(this);
					if (exprj.getQType() == null)
						throw new TypeException("Expression " + exprj + " could not be typed."); //return;
				}
				if (exprj.getQType().isBoolean()) {
					if (exprj instanceof XQueryExpressionSequence)
						exprj = (XQueryExpression) ((XQueryExpressionSequence) exprj).getSubExpressions().get(0);
					if (!(exprj instanceof LocatedExpression) && !(exprj instanceof Variable))
						throw new TypeException("Incorrect boolean typed expression in element " + exprj); //return;
				}
			}
		// type checking
		QType tagQType = startTag.getQType();
		if (tagQType != null && !tagQType.isString()) {
			tagQType = tagQType.applyDATAFunction();
			if (!tagQType.isString())
			throw new TypeException("Expression '" + startTag + "', of element constructor has incorrect type.");
		}
		if (noForce && arg.getQType() != null)
			return;
		//        if (expressions != null)
		//            for (int j=0;j<expressions.size();j++)
		//                if (((XQueryExpression)expressions.get(j)).getQType().isCollection())
		//					if (((XQueryExpression)expressions.get(j)).getQType().getQType().getSubClass() == QType.ATTRIBUTE)
		//						throw new TypeException("Invalid XML: an element can not have many attributes with the same name.");
		// type setting
		ArrayList allElements = new ArrayList();
		ArrayListUtil.add(allElements, arg.getAttributes());
		ArrayListUtil.add(allElements, arg.getSubExpressions());
		arg.setQType(buildQTypeElementFromSubElements(arg.getStartTag(), allElements));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:Element ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FLWRExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// the where part must be a boolean
		// TYPE BUILDING
		// the type returned is the type of the return part
		// in fact it is a set of element whose type is the type of the return part

		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:FLWRExpression ( " + arg + " )");
		ArrayList variables = arg.getVariables();
		XQueryExpression whereClause = arg.getWhereClause();
		XQueryExpression returnClause = arg.getReturnClause();
		// forward type checking
		/*
		// change   27/03/2002
		if (contextNodePathList == null) {
			for (int i=0;i<variables.size();i++)
				if (((Variable)variables.get(i)).getQType() == null) return;
			if (whereClause != null && whereClause.getQType() == null) return;
			if (returnClause.getQType() == null) return;
		}
		*/
		boolean multiple = false;
		for (int i = 0; i < variables.size(); i++) {
			Variable vari = (Variable) variables.get(i);
			if (!noForce || vari.getQType() == null)
				vari.accept(this);
			if (!multiple)
				if (vari.getBindingType() == Constants.FOR_BINDINGTYPE && vari.getExpression().getQType().isMultiple())
					multiple = true;

		}
		if (whereClause != null && (!noForce || whereClause.getQType() == null))
			whereClause.accept(this);
		if (!noForce || returnClause.getQType() == null)
			returnClause.accept(this);
		// type verification
		if (returnClause.getQType() == null)
			throw new TypeException("Incorrect 'return' clause : [ " + returnClause + " ], it could not be typed.");
		if (whereClause != null && whereClause.getQType() == null)
			throw new TypeException("Incorrect 'where' clause : [ " + whereClause + " ], it could not be typed.");
		/*
		 if (whereClause != null) {
			QType qtype = whereClause.getQType();
			if ((qtype instanceof QTypeAtom) && qtype.getSimpleType() != (SimpleType) getType(SC_BOOLEAN))
				throw new TypeException("The 'where' clause is not of type 'boolean'.");
		}
		*/
		// verify sort clause typing
		if (arg.getOrderBy() != null) {
			for (int i = 0; i < arg.getOrderBy().size(); i++) {
				XQueryExpression sortiexpr = (XQueryExpression) arg.getOrderBy().get(i);
				QType qtypei = sortiexpr.getQType();
				if (qtypei == null || !qtypei.isAtomic() || qtypei.isMultiple())
					throw new TypeException("Invalid order by clause : " + sortiexpr);
			}
		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType((QType) arg.getReturnClause().getQType().changeOccurence(multiple ? QType.OCC_0_N : QType.OCC_1_1));
		//		if (multiple)
		//			arg.setQType(new QTypeCollection(arg.getReturnClause().getQType()));
		//		else
		//			arg.setQType(arg.getReturnClause().getQType());
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:FLWRExpression ( " + arg + " ) --> " + arg.getQType());

	}

	// public void visit(FunctionCall arg) throws XQueryException;

	public void visit(FunctionDefinition arg) throws XQueryException {
		// TYPE VERIFICATION
		// if there is a return sequenceType, verify that type is the same as last expression type
		// TYPE BUILDING
		// the type returned is the type of the last expression
		// or better the type returned if any

		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:FunctionDefinition ( " + arg + " )");
		// ArrayList expressions = arg.getExpressions();
		SequenceTypeExpression returnType = arg.getReturnType();
		// forward type checking --> cannot be done because of variables not binded in parameters
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		QType qtype = null;
		//org.xquark.xml.misc.Debug.print(this,"returnType = " + returnType);
		if (returnType == null) {
			SimpleType type = new SimpleType(this.typingschema, SC_ANYSIMPLETYPE, null);
			type.setPrimitive(PrimitiveType.createType(SC_ANYSIMPLETYPE));
			ArrayList list = new ArrayList(3);
			list.add(new QTypeAtom(type));
			list.add(new QTypeAttribute(new QName("*", "*", "*", this), type));
			list.add(new QTypeElement(new QName("*", "*", "*", this), getType(SC_ANYTYPE)));
			qtype = new QTypeUnion(list, QType.OCC_0_N);
		} else {
			Type type = getType(returnType);
			qtype = getQTypeFromType(type);
			if (qtype == null)
				throw new TypeException("Could not transform schema type '" + type + "'");
		}
		arg.setQType(qtype);
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:FunctionDefinition ( " + arg + " ) --> " + arg.getQType());
	}

	//	public void visit(InstanceOfExpression arg) throws XQueryException;

	public void visit(InternalFunctionCall arg) throws XQueryException {
		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:InternalFunctionCall ( " + arg + " )");
		//		SimpleType type = new SimpleType(this.typingschema, SC_ANYSIMPLETYPE, null);
		//		type.setPrimitive(PrimitiveType.createType(SC_ANYSIMPLETYPE));
		SimpleType type = context.getSchemaManager().getAnySimpleType();
		ArrayList list = new ArrayList(3);
		list.add(new QTypeAtom(type));
		list.add(new QTypeAttribute(new QName("*", "*", "*", this), type));
		list.add(new QTypeElement(new QName("*", "*", "*", this), getType(SC_ANYTYPE)));
		QType qtype = new QTypeUnion(list, QType.OCC_0_N);
		arg.setQType(qtype);
		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:InternalFunctionCall ( " + arg + " )");
	}

	public void visit(ITEExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// verify that condition is of type boolean
		// TYPE BUILDING
		// the type returned is a choice of then clause and the else clause

		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:ITEExpression ( " + arg + " )");
		XQueryExpression ifExpression = arg.getThenExpression();
		XQueryExpression thenExpression = arg.getThenExpression();
		XQueryExpression elseExpression = arg.getElseExpression();
		// forward type checking
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (ifExpression.getQType() == null)
				return;
			if (thenExpression.getQType() == null && !(thenExpression instanceof XQueryVoid))
				return;
			if (elseExpression.getQType() == null && !(elseExpression instanceof XQueryVoid))
				return;
		}
		if (!noForce || ifExpression.getQType() == null)
			ifExpression.accept(this);
		if ((!noForce || thenExpression.getQType() == null) && !(thenExpression instanceof XQueryVoid))
			thenExpression.accept(this);
		if ((!noForce || elseExpression.getQType() == null) && !(elseExpression instanceof XQueryVoid))
			elseExpression.accept(this);
		// type checking
		if (ifExpression.getQType() == null)
			throw new TypeException("Expression " + ifExpression + " could not be typed."); //return;
		if (thenExpression.getQType() == null && !(thenExpression instanceof XQueryVoid))
			throw new TypeException("Expression " + thenExpression + " could not be typed."); //return;
		if (elseExpression.getQType() == null && !(elseExpression instanceof XQueryVoid))
			throw new TypeException("Expression " + elseExpression + " could not be typed."); //return;
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		ArrayList tmpList = new ArrayList(2);
		tmpList.add(thenExpression.getQType());
		tmpList.add(elseExpression.getQType());
		arg.setQType(makeUnion(tmpList));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:ITEExpression ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(LibraryFunctionCall arg) throws XQueryException {
		// TYPE VERIFICATION
		// verify that the type of the arguments are the same as the types of the function definition
		// TYPE BUILDING
		// the type returned is the type returned by the function definition

		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:LibraryFunctionCall ( " + arg + " )");
		FunctionDefinition funcdef = arg.getFunctionDefinition(); // cannot be null
		// forward type checking
		if (funcdef.getQType() == null)
			funcdef.accept(this);
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(funcdef.getQType());
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:LibraryFunctionCall ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(ListOpAFTERExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// both expressions must be sequences of elements of the same document
		if (!arg.getExpression1().getQType().isNode() || !arg.getExpression2().getQType().isNode())
			throw new TypeException("At least one of the operands is not a node.");
		// TYPE BUILDING
		// the type returned is a sequence of elements that have the first sequence type

		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:ListOpAFTERExpression ( " + arg + " )");
		XQueryExpression expr1 = arg.getExpression1();
		XQueryExpression expr2 = arg.getExpression2();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr1.getQType() == null)
				return;
			if (expr2.getQType() == null)
				return;
		}
		if (!noForce || expr1.getQType() == null)
			expr1.accept(this);
		if (!noForce || expr2.getQType() == null)
			expr2.accept(this);
		// type verification
		if (expr1.getQType() == null)
			throw new TypeException("Expression " + expr1 + " could not be typed."); //return;
		if (expr2.getQType() == null)
			throw new TypeException("Expression " + expr2 + " could not be typed."); //return;
		if (!expr1.getQType().isNode() || !expr2.getQType().isNode())
			throw new TypeException("Expression '" + arg + "', at least one of the operands must be a node or collection of node.");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(expr1.getQType());
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:ListOpAFTERExpression ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(ListOpArithExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// both expressions must be numeric (see Function SUM for more details)
		// TYPE BUILDING
		// When both operands of an arithmetic operator are numeric, the result is straightforward.
		// When one or more operands is a node, the content of the node is extracted by an implicit call to the data function
		// and converted to a number before the operation is performed; if this conversion is not possible, an error results.
		// When one operand is a number and the other is a sequence of numbers, the result of the operator is a sequence that is created by applying the operator pairwise to the numeric operand and each individual member of the sequence operand, preserving the original order.

		//org.xquark.xml.misc.Debug.print(this, "START TypeVistor:ListOpArithExpression ( " + arg + " )");
		XQueryExpression expr1 = arg.getExpression1();
		XQueryExpression expr2 = arg.getExpression2();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr1.getQType() == null)
				return;
			if (expr2.getQType() == null)
				return;
		}
		if (!noForce || expr1.getQType() == null)
			expr1.accept(this);
		if (!noForce || expr2.getQType() == null)
			expr2.accept(this);
		// type verification
		if (expr1.getQType() == null)
			throw new TypeException("Expression " + expr1 + " could not be typed."); //return;
		if (expr2.getQType() == null)
			throw new TypeException("Expression " + expr2 + " could not be typed."); //return;
		QType qtype1 = expr1.getQType();
		QType qtype2 = expr2.getQType();
		if (qtype1 != null && qtype2 != null) {
			if (!qtype1.isAtomic() && !inPredicate)
				throw new TypeException("Expression '" + arg + "', first operand is not an atomic value.");
			if (!qtype2.isAtomic() && !inPredicate)
				throw new TypeException("Expression '" + arg + "', second operand is not an atomic value.");
			if (!qtype1.isNumeric())
				throw new TypeException("Expression '" + arg + "', first operand is not numeric.");
			if (!qtype2.isNumeric())
				throw new TypeException("Expression '" + arg + "', second operand is not numeric.");
		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_DOUBLE), (expr1.getQType().isMultiple() || expr2.getQType().isMultiple()) ? QType.OCC_0_N : QType.OCC_1_1));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:ListOpArithExpression ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(ListOpBEFOREExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// both expressions must be sequences of elements of the same document
		if (!arg.getExpression1().getQType().isNode() || !arg.getExpression2().getQType().isNode())
			throw new TypeException("At least one of the operands is not a node.");
		// TYPE BUILDING
		// the type returned is a sequence of elements that have the first sequence type

		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:ListOpBEFOREExpression ( " + arg + " )");
		XQueryExpression expr1 = arg.getExpression1();
		XQueryExpression expr2 = arg.getExpression2();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr1.getQType() == null)
				return;
			if (expr2.getQType() == null)
				return;
		}
		if (!noForce || expr1.getQType() == null)
			expr1.accept(this);
		if (!noForce || expr2.getQType() == null)
			expr2.accept(this);
		// type verification
		if (expr1.getQType() == null)
			throw new TypeException("Expression " + expr1 + " could not be typed."); //return;
		if (expr2.getQType() == null)
			throw new TypeException("Expression " + expr2 + " could not be typed."); //return;
		if (!expr1.getQType().isNode() || !expr2.getQType().isNode())
			throw new TypeException("Expression '" + arg + "', at least one of the operands must be a node or collection of node.");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(expr1.getQType());
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:ListOpBEFOREExpression ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(ListOpCompExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// both expressions must be sequences of elements of the same document
		// TYPE BUILDING
		// The =, !=, <, <=, >, and >= operators perform value comparisons.
		// If both operands are simple values of the same type, the result is straightforward.
		// If the operands are simple values of compatible types, the operand of the less-inclusive type is converted to the more-inclusive type for the purpose of comparison (for example, an integer might be converted to a float in order to be compared with a float.)
		// If one operand is a node and the other is a simple value, the content of the node is extracted by an implicit invocation of the data function before the comparison is performed.
		// If both operands are nodes, the string-values of the nodes are compared, as defined in [XPath 1.0].

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:ListOpCompExpression ( " + arg + " ) ");

		//arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));

		XQueryExpression expr1 = arg.getExpression1();
		XQueryExpression expr2 = arg.getExpression2();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr1.getQType() == null)
				return;
			if (expr2.getQType() == null)
				return;
		}
		if (!noForce || expr1.getQType() == null)
			expr1.accept(this);
		if (!noForce || expr2.getQType() == null)
			expr2.accept(this);
		// type checking
		QType qtype1 = expr1.getQType();
		QType qtype2 = expr2.getQType();
		if (qtype1 == null)
			throw new TypeException("Expression " + expr1 + " could not be typed."); //return;
		if (qtype2 == null)
			throw new TypeException("Expression " + expr2 + " could not be typed."); //return;
		if (qtype1 != null && qtype2 != null) {
			if (!qtype1.isAtom())
				qtype1 = qtype1.applyDATAFunction();
			if (!qtype2.isAtom())
				qtype2 = qtype2.applyDATAFunction();
			// if (!(qtype1.isNumeric() && qtype2.isNumeric()) && !(qtype1.canBeComparedTo(qtype2)))
			if (qtype1 == null || !(qtype1.canBeComparedTo(qtype2)))
				throw new TypeException("Could not compare types of expression --> " + arg);
		}
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:ListOpCompExpression ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(ListOpEXCEPTExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// ...
		// TYPE BUILDING

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:ListOpEXCEPTExpression ( " + arg + " ) ");
		XQueryExpression expr1 = arg.getExpression1();
		XQueryExpression expr2 = arg.getExpression2();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr1.getQType() == null)
				return;
			if (expr2.getQType() == null)
				return;
		}
		if (!noForce || expr1.getQType() == null)
			expr1.accept(this);
		if (!noForce || expr2.getQType() == null)
			expr2.accept(this);
		// type verification
		if (expr1.getQType() == null)
			throw new TypeException("Expression " + expr1 + " could not be typed."); //return;
		if (expr2.getQType() == null)
			throw new TypeException("Expression " + expr2 + " could not be typed."); //return;
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(expr1.getQType());
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:ListOpEXCEPTExpression ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(ListOpINTERSECTExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// ...
		// TYPE BUILDING

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:ListOpINTERSECTExpression ( " + arg + " ) ");
		XQueryExpression expr1 = arg.getExpression1();
		XQueryExpression expr2 = arg.getExpression2();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr1.getQType() == null)
				return;
			if (expr2.getQType() == null)
				return;
		}
		if (!noForce || expr1.getQType() == null)
			expr1.accept(this);
		if (!noForce || expr2.getQType() == null)
			expr2.accept(this);
		// type verification
		if (expr1.getQType() == null)
			throw new TypeException("Expression " + expr1 + " could not be typed."); //return;
		if (expr2.getQType() == null)
			throw new TypeException("Expression " + expr2 + " could not be typed."); //return;
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(expr1.getQType());
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:ListOpINTERSECTExpression ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(ListOpUNIONExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// ...
		// TYPE BUILDING

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:ListOpUNIONExpression ( " + arg + " ) ");
		XQueryExpression expr1 = arg.getExpression1();
		XQueryExpression expr2 = arg.getExpression2();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr1.getQType() == null)
				return;
			if (expr2.getQType() == null)
				return;
		}
		if (!noForce || expr1.getQType() == null)
			expr1.accept(this);
		if (!noForce || expr2.getQType() == null)
			expr2.accept(this);
		// type verification
		if (expr1.getQType() == null)
			throw new TypeException("Expression " + expr1 + " could not be typed."); //return;
		if (expr2.getQType() == null)
			throw new TypeException("Expression " + expr2 + " could not be typed."); //return;
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		ArrayList list = new ArrayList();
		list.add(expr1.getQType());
		list.add(expr2.getQType());
		arg.setQType(new QTypeUnion(list));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:ListOpUNIONExpression ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(LocatedExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// verify that indirect path are present in matadata( ?? )
		// TYPE BUILDING
		// type of the direct path (there should be only one!!!)

		//org.xquark.util.misc.Debug.print(this, "START TypeVisitor:LocatedExpression ( " + arg + " ) ");

		ArrayList steps = arg.getSteps();

		// checking force stuff
		if (noForce) {
			boolean alltyped = true;
			for (int i = 0; i < steps.size(); i++) {
				Step stepi = (Step) steps.get(i);
				if (stepi.getQType() == null || stepi.getAxis() == Axis.DESCENDANT || stepi.getAxis() == Axis.DESCENDANT_OR_SELF) {
					alltyped = false;
					break;
				}
			}
			// try to simplify if all steps have qtype
			if (alltyped) {
				arg.setQType(arg.getStepNum(steps.size() - 1).getQType());
				arg.setXTrees(arg.getStepNum(steps.size() - 1).getXTrees());
				return;
			}
		} else {
			arg.setQType(null);
			arg.setXTrees(null);
		}

		// look for simple paths to be typed only with their context
		boolean simplePath = arg.isRelative();

		// remember the current context 
		ArrayList tmpContextXNodes = null;
		QType tmpContextQType = null;
		LocatedExpression tmpContextLocated = null;
		int tmpContextLocatedStepIndex = -1;
		// for simple paths chek context and set current context
		if (simplePath) {
			if (arg.getStepNum(0).getAxis() == Axis.NONE)
				arg.getStepNum(0).setAxis(Axis.CHILD);
			if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
				if (arg.getInPredicate())
					return;
				else
					throw new TypeException(arg + " could not be typed because context does not exist.");
			}
			tmpContextXNodes = contextXNodes;
			tmpContextQType = contextQType;
			tmpContextLocated = contextLocated;
			tmpContextLocatedStepIndex = contextLocatedStepIndex;
			arg.setContext(true);
		}
		// if not simple path set current context anyway
		else {
			if (contextXNodes != null || contextQType != null) {
				tmpContextXNodes = contextXNodes;
				tmpContextQType = contextQType;
				tmpContextLocated = contextLocated;
				tmpContextLocatedStepIndex = contextLocatedStepIndex;
				contextXNodes = null;
				contextQType = null;
				tmpContextLocated = null;
				contextLocatedStepIndex = -1;
			}
		}

		// forward checking and typing

		boolean isMultiple = false;

		boolean done = false;
		QType qtype = null;
		Step stepi = null;

		for (int i = 0; i < steps.size(); i++) {
			// initialisation
			qtype = null;
			done = false;
			stepi = (Step) steps.get(i);
			XQueryExpression stepiExpr = stepi.getExpression();

			// test to set the contextType for the first step
			if (i == 0 && contextType == QTypeDocument.ANY_ORIGIN) {
				if (stepiExpr.getQType() instanceof QTypeDocument && ((QTypeDocument) stepiExpr.getQType()).isFromDocument())
					contextType = QTypeDocument.DOCUMENT_ORIGIN;
				else
					contextType = QTypeDocument.COLLECTION_ORIGIN;
			}

			// test for existing information
			if (noForce) {
				if (i == 0 && stepi.getQType() != null) {
					contextXNodes = stepi.getXTrees();
					contextQType = stepi.getQType();
					qtype = stepi.getQType();
					done = true;
				}
			} else {
				stepi.setQType(null);
				stepi.setXTrees(null);
			}

			// handle special case of first step
			if (!done) {
				if (i == 0 && stepi.getAxis() == Axis.NONE) {
					qtype = stepiExpr.getQType();
					if (noForce && qtype != null) {
						stepi.setXTrees(stepiExpr.getXTrees());
						done = true;
					}
				}
			}

			// handle special case of step expression being a ListOpUNIONExpression
			if (!done && stepiExpr instanceof ListOpUNIONExpression) {
				this.setTypingContext(contextXNodes, contextQType, arg, i - 1);
				stepiExpr.accept(this);
				qtype = stepiExpr.getQType();
				if (qtype != null) {
					stepi.setXTrees(stepiExpr.getXTrees());
					done = true;
				}
			}

			// special treatment to avoid bug on navigation (does not navigate on text nodes)
			boolean doNodes = true;
			boolean isTextNode = false;
			if (!done && stepiExpr instanceof NodeType) {
				if (((NodeType) stepiExpr).getKind() == NodeKind.TEXT) {
					isTextNode = true;
					doNodes = false;
				} else if (((NodeType) stepiExpr).getKind() == NodeKind.NODE)
					doNodes = false;
			}

			// get next step type
			if (!done && doNodes && contextXNodes != null) {
				ArrayList xnodes = context.resolveXPath(contextXNodes, stepi);
				if (xnodes != null) {
					qtype = this.buildQTypeFromXNodes(xnodes);
					if (qtype != null) {
						stepi.setXTrees(xnodes);
						done = true;
					}
				}
			}
			// get next step type if it could not be resolved with pathresolver
			if (!done && contextQType != null) {
				qtype = context.resolveQType(contextType, contextQType, stepi);
				if (qtype != null) {
					if (isTextNode)
						stepi.setXTrees(contextXNodes);
					else
						stepi.setXTrees(null);
					done = true;
				}

			}

			// verification of type and collection setting
			if (done) {
				// erase   01/09/2003 by extension an ttribute can have a parent
				//				if (i != steps.size() - 1 && qtype.isOnlyAttribute())
				//					throw new TypeException("Attribute step must be the last element of a located expression : " + arg);
				if (isMultiple && !qtype.isMultiple())
					qtype = qtype.changeOccurence(QType.OCC_0_N);
				// ADD   09/05/03 to make sure that self::node is not a collection
				// change !!!   29/08/2003
				/*
				if (qtype.isMultiple() && stepi.getAxis() == Axis.SELF) {
					qtype = qtype.unMultiple();
					isMultiple = false;
				}
				*/
				// DANGER 2002/09/03
				if (!isMultiple && qtype.isMultiple())
					isMultiple = true;
			} else {
				// change LARS 22/09/2003
				if (stepi.getAxis() == Axis.PARENT) {
					if (arg.startsWithVariable() || (arg.getParentExpression() != null && !arg.getParentExpression().isEmpty() && arg.getParentExpression().get(0) instanceof ListOpUNIONExpression)) {
						// add LARS 22/09/2003
						// special code to be able to type parent::node when root expression is variable
						XQueryExpression tmpExpr = arg;
						XQueryExpression firstExpr = arg.getStepNum(0).getExpression();
						ArrayList stepList = new ArrayList();
						int stepIndex = -1;
						try {
							while (true) {
								if (tmpExpr instanceof LocatedExpression) {
									LocatedExpression tmpLoc = (LocatedExpression) tmpExpr;

									if (stepIndex == -1)
										stepIndex = tmpLoc.getSteps().size() - 1;
									for (int j = stepIndex; j >= 0; j--) {
										Step clonedStep = (Step) tmpLoc.getStepNum(j).clone();
										clonedStep.setPredicates(null);
										stepList.add(0, clonedStep);
									}
									if (simplePath && (tmpLoc.getInPredicate() || (tmpLoc.getParentExpression() != null && !tmpLoc.getParentExpression().isEmpty() && tmpLoc.getParentExpression().get(0) instanceof ListOpUNIONExpression))) {
										tmpExpr = contextLocated;
										stepIndex = contextLocatedStepIndex;
										((Step) stepList.get(0)).setHasSeparator(true);
									} else {
										stepIndex = -1;
										firstExpr = tmpLoc.getStepNum(0).getExpression();
										if (firstExpr instanceof Variable) {
											tmpExpr = ((Variable) firstExpr).getExpression();
											stepList.remove(0);
										} else {
											break;
										}
									}
								} else if (tmpExpr instanceof Variable) {
									tmpExpr = ((Variable) tmpExpr).getExpression();
								} else if (tmpExpr instanceof FunctionCall) {
									stepList.add(0, tmpExpr);
									break;
								}
							}

						} catch (CloneNotSupportedException cnse) {
							stepList.clear();
						}

						if (!stepList.isEmpty() && stepList.size() != steps.size()) {
							LocatedExpression loc = new LocatedExpression(stepList, false, this, true);
							if (loc.getQType() != null)
								qtype = loc.getQType();
						}
					} else if (arg.getParentExpression() != null && !arg.getParentExpression().isEmpty() && arg.getParentExpression().get(0) instanceof ListOpUNIONExpression) {
						boolean prev = inPredicate;
						inPredicate = true;
						arg.accept(this);
						inPredicate = prev;
					}
				}
			}
			if (qtype == null)
				throw new TypeException("Could not find path : " + arg);

			// set typing context
			stepi.setQType(qtype);
			contextXNodes = stepi.getXTrees();
			contextQType = stepi.getQType();

			// handle predicates if any
			ArrayList predicates = stepi.getPredicates();
			if (predicates != null && !predicates.isEmpty()) {
				this.setTypingContext(contextXNodes, contextQType, arg, i);
				for (int j = 0; j < predicates.size(); j++) {
					XQueryExpression predicate = (XQueryExpression) predicates.get(j);
					if (!noForce || predicate.getQType() == null) {
						boolean prev = inPredicate;
						inPredicate = true;
						predicate.accept(this);
						inPredicate = prev;
					}
				}

			}
		}
		// type checking
		//		if (noForce && arg.getQType() != null) {
		//			this.setTypingContext(tmpContextXNodes, tmpContextQType);
		//			return;
		//		}
		if (qtype != null) {
			arg.setQType(qtype);
			//		} else if (noForce && arg.getQType() != null) {
			//			// do nothing

		} else {
			throw new TypeException("Could not find path : " + arg);
		}

		// reset the context type
		contextType = QTypeDocument.ANY_ORIGIN;
		// reset the context
		this.setTypingContext(tmpContextXNodes, tmpContextQType, tmpContextLocated, tmpContextLocatedStepIndex);
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:LocatedExpression ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(NodeType arg) throws XQueryException {
		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:NodeType ( " + arg + " )");

		if (contextQType == null)
			return;

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		QType qtype = Utils.getSubType(contextType, contextQType, new Step(true, Axis.CHILD, arg, null, null), context.getSchemaManager());
		arg.setQType(qtype);
		// type setting
		// remove   04/06/2003
		//arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:NodeType ( " + arg + " ) --> " + arg.getQType());
	}

	//	public void visit(Parameter arg) throws XQueryException;
	//	public void visit(PrimitiveFunctionCall arg) throws XQueryException;

	public void visit(QName arg) throws XQueryException {
		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:QName ( " + arg + " ) ");
		if (contextXNodes != null && contextQType != null && arg.getParentExpression() != null && arg.getParentExpression().get(0) instanceof SortedExpression) {
			QType qtype = null;
			//setContextNodeStepListType();
			//else setContextNodeStepListType();
			if (contextQType != null) {
				Object obj = Utils.getSubType(contextType, contextQType, new Step(false, Axis.CHILD, arg, null, null), context.getSchemaManager());
				if (obj != null) {
					if (obj instanceof Declaration)
						qtype = getQTypeFromDeclaration((Declaration) obj);
					else if (obj instanceof QType)
						qtype = (QType) obj;
				}
			}
			if (qtype == null)
				throw new TypeException("Could not type expression : " + arg);
			arg.setQType(qtype);
			//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:QName ( " + arg + " ) --> " + arg.getQType());
			return;
		}
		QType qtype = null;
		if (contextQType != null) {
			Object obj = Utils.getSubType(contextType, contextQType, new Step(false, Axis.CHILD, arg, null, null), context.getSchemaManager());
			if (obj != null) {
				if (obj instanceof Declaration)
					qtype = getQTypeFromDeclaration((Declaration) obj);
				else if (obj instanceof QType)
					qtype = (QType) obj;
			}
		}
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			return;
		if (qtype == null) {
			//			arg.setContext(true);
			//			ArrayList vectTypes = null;
			//			ArrayList steps = new ArrayList(1);
			//			steps.add(new Step(false, Axis.CHILD, arg, null, null));
			//			vectTypes = getDeclarationsFromSteps(steps, arg, true);

			qtype = buildQTypeFromXNodes(contextXNodes, new Step(false, Axis.CHILD, arg, null, null), true);
		}
		if (qtype == null)
			throw new TypeException("Could not type expression : " + arg);
		arg.setQType(qtype);
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:QName ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(QuantifiedExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// SATISFIES expression must retunr a boolean
		// TYPE BUILDING
		// type of the direct path (there should be only one!!!)

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:QuantifiedExpression ( " + arg + " )");

		//arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));

		ArrayList variables = arg.getVariables();
		XQueryExpression expr = arg.getExpression();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			for (int i = 0; i < variables.size(); i++)
				if (((Variable) variables.get(i)).getExpression().getQType() == null)
					return;
			if (expr.getQType() == null)
				return;
		}
		for (int i = 0; i < variables.size(); i++)
			if (!noForce || ((Variable) variables.get(i)).getExpression().getQType() == null) {
				((Variable) variables.get(i)).getExpression().accept(this);
				if (((Variable) variables.get(i)).getExpression().getQType() == null)
					throw new TypeException("Expression " + ((Variable) variables.get(i)).getExpression() + " could not be typed."); //return;
			}
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Expression " + expr + " could not be typed."); //return;
		// type verification
		if (!expr.getQType().isBoolean())
			throw new TypeException("SATISFIES clause must be a boolean.");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:QuantifiedExpression ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(RangeExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// both expression must be integers
		// TYPE BUILDING
		// --> NONE

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:RangeExpression ( " + arg + " )");
		XQueryExpression expr1 = arg.getExpression1();
		XQueryExpression expr2 = arg.getExpression2();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr1.getQType() == null)
				return;
			if (expr2.getQType() == null)
				return;
		}
		if (!noForce || expr1.getQType() == null)
			expr1.accept(this);
		if (!noForce || expr2.getQType() == null)
			expr2.accept(this);
		// type verification
		if (expr1.getQType() == null)
			throw new TypeException("Expression " + expr1 + " could not be typed."); //return;
		if (expr2.getQType() == null)
			throw new TypeException("Expression " + expr2 + " could not be typed."); //return;
		if (!expr1.getQType().isInteger())
			throw new TypeException("Expression '" + arg + "', first expression is not an integer.");
		if (expr2 != null && !expr2.getQType().isInteger())
			throw new TypeException("Expression '" + arg + "', second expression is not an integer.");
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:RangeExpression ( " + arg + " ) --> " + arg.getQType());
	}

	//	public void visit(RootNodeFunctionCall arg) throws XQueryException;

	public void visit(SortedExpression arg) throws XQueryException {
		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:SortedExpression ( " + arg + " )");
		// TYPE VERIFICATION
		// sort clauses must be simple xpaths expressions
		//		ArrayList sortClauses = arg.getSortClauses();
		//		for (int i = 0; i < sortClauses.size(); i++)
		//		{
		//			((XQueryExpression) sortClauses.get(i)).accept(this);
		//		}
		// TYPE BUILDING
		// build chooice type!!!

		XQueryExpression expr = arg.getExpression();
		ArrayList sortClauses = arg.getSortClauses();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr.getQType() == null)
				return;
		}
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		// set the context (but only the type is present ... not so easy!)
		// not done for now
		if (expr.getQType() == null)
			throw new TypeException("Expression " + expr + " could not be typed."); //return;

		//		if (expr.getQType() != null) {
		//			QType tmpContextType = contextQType;
		//			setContextQType(expr.getQType());
		//			for (int i = 0; i < sortClauses.size(); i++)
		//				if (!noForce || ((XQueryExpression) sortClauses.get(i)).getQType() == null)
		//					 ((XQueryExpression) sortClauses.get(i)).accept(this);
		//			setContextQType(tmpContextType);
		//		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(expr.getQType());
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:SortedExpression ( " + arg + " ) --> " + arg.getQType());
	}

	// do nothing (but implement otherwise it fails)
	public void visit(Step arg) throws XQueryException {
		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:Step ( " + arg + " )");
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:Step ( " + arg + " ) --> " + arg.getQType());
	}

	//	public void visit(TypeSwitchExpression arg) throws XQueryException;

	public void visit(UnOpMinusExpression arg) throws XQueryException {
		// TYPE VERIFICATION
		// expression must be numeric
		// TYPE BUILDING

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:UnOpMinusExpression ( " + arg + " )");
		XQueryExpression expr = arg.getExpression();
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
			if (expr.getQType() == null)
				return;
		}
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		// type verification
		if (expr.getQType() == null)
			throw new TypeException("Expression " + expr + " could not be typed."); //return;
		if (expr.getQType() != null) {
			if (!expr.getQType().isAtomic() && !inPredicate)
				throw new TypeException("Expression '" + arg + "', operand is not an atomic value.");
			if (!expr.getQType().isNumeric())
				throw new TypeException("Expression '" + arg + "', operand is not numeric.");
		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		QType qtypedata = expr.getQType().applyDATAFunction();
		if (qtypedata != null)
			arg.setQType(qtypedata);
		else
			arg.setQType(expr.getQType());
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:UnOpMinusExpression ( " + arg + " ) --> " + arg.getQType());
	}

	//	public void visit(Value arg) throws XQueryException;

	public void visit(ValueBoolean arg) throws XQueryException {
		// TYPE VERIFICATION
		// --> NONE
		// TYPE BUILDING
		// the return type is  boolean

		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:ValueBoolean ( " + arg + " )");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getBooleanType()));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:ValueBoolean ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(ValueDecimal arg) throws XQueryException {
		// TYPE VERIFICATION
		// --> NONE
		// TYPE BUILDING
		// the return type is decimal

		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:ValueInteger ( " + arg + " )");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getDecimalType()));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:ValueInteger ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(ValueDouble arg) throws XQueryException {
		// TYPE VERIFICATION
		// --> NONE
		// TYPE BUILDING
		// the return type is double

		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:ValueInteger ( " + arg + " )");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getDoubleType()));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:ValueInteger ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(ValueFloat arg) throws XQueryException {
		// TYPE VERIFICATION
		// --> NONE
		// TYPE BUILDING
		// the return type is float

		//org.xquark.xml.misc.Debug.print(this, "START TypeVistor:ValueFloat ( " + arg + " )");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getFloatType()));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:ValueFloat ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(ValueInteger arg) throws XQueryException {
		// TYPE VERIFICATION
		// --> NONE
		// TYPE BUILDING
		// the return type is integer

		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:ValueInteger ( " + arg + " )");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getIntegerType()));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:ValueInteger ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(ValueString arg) throws XQueryException {
		// TYPE VERIFICATION
		// --> NONE
		// TYPE BUILDING
		// the return type is string

		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:ValueString ( " + arg + " )");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getStringType()));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:ValueString ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(ValueParameter arg) throws XQueryException {
		// TYPE VERIFICATION
		// --> NONE
		// TYPE BUILDING
		// the return type is string

		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:ValueString ( " + arg + " )");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		if (arg.getTypeName() == null)
			return;
		arg.setQType(new QTypeAtom((SimpleType) context.getSchemaManager().getType(arg.getTypeName().getNameSpace(), arg.getTypeName().getLocalName())));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:ValueString ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(ValueText arg) throws XQueryException {
		// TYPE VERIFICATION
		// --> NONE
		// TYPE BUILDING
		// the return type is string

		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:ValueText ( " + arg + " )");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:ValueText ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(Variable arg) throws XQueryException {
		// TYPE VERIFICATION
		// --> NONE
		// TYPE BUILDING
		// the return type is the type of the associated expression, if any

		//org.xquark.util.misc.Debug.print(this, "START TypeVistor:Variable ( " + arg + " )");

		XQueryExpression expression = arg.getExpression();
		if (expression != null) {
			//org.xquark.xml.misc.Debug.print(this,"expression.getQType() = " + expression.getQType());
			//org.xquark.xml.misc.Debug.print(this,"arg.getBindingType() = " + arg.getBindingType());
			// forward checking and typing
			if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null) {
				if (expression.getQType() == null)
					return;
			}
			if (!noForce || expression.getQType() == null)
				expression.accept(this);
			// type checking
			if (expression.getQType() == null)
				throw new TypeException("Expression " + expression + " could not be typed."); //return;
			if (expression instanceof XQueryVoid)
				throw new TypeException("Variable " + arg + " could not be bound to existing path."); //return;			
			/*
			if (expression instanceof LocatedExpression && !arg.hasDocumentRoot() && (expression.getXTrees() == null || expression.getXTrees().isEmpty()))
				throw new TypeException("Variable " + arg + " could not be bound to existing path."); //return;
			*/
			if (noForce && arg.getQType() != null)
				return;
			// type setting
			if (arg.getBindingType() == Constants.LET_BINDINGTYPE)
				arg.setQType(expression.getQType());
			else if (arg.getBindingType() == Constants.FOR_BINDINGTYPE || arg.getBindingType() == Constants.QUANTIFIER_BINDINGTYPE || arg.getBindingType() == Constants.OUTERFOR_BINDINGTYPE) {
				//org.xquark.xml.misc.Debug.print(this,"FOR_BINDINGTYPE");
				//org.xquark.xml.misc.Debug.print(this,"expression.getQType().isCollection() = " + expression.getQType().isCollection());
				QType exprqtype = expression.getQType();
				if (exprqtype instanceof QTypeSequence)
					exprqtype = new QTypeUnion(exprqtype.getList());
				if (exprqtype.isMultiple()) {
					exprqtype = exprqtype.changeOccurence(QType.OCC_1_1);
				}
				arg.setQType(exprqtype);

			}
			//else if (arg.getBindingType() == Constants.FUNCTION_BINDINGTYPE)
			//	arg.setQType(new QTypeAtom((SimpleType)getType((QName)expression)));
			else
				arg.setQType(expression.getQType());
		}
		// else arg.setQType(null);
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:Variable ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(XMLComment arg) throws XQueryException {
		arg.setQType(new QTypeComment());
	}

	public void visit(XMLProcessingInstruction arg) throws XQueryException {
		arg.setQType(new QTypeProcessingInstruction());
	}

	//	public void visit(XQueryBinaryOperatorExpression arg) throws XQueryException;
	//	public void visit(XQueryBooleanOperatorExpression arg) throws XQueryException;
	public void visit(XQueryExpression arg) throws XQueryException {
		throw new TypeException("Typing of expression : " + arg + " of class " + arg.getClass().getName() + " is not supported");
	}

	public void visit(XQueryExpressionSequence arg) throws XQueryException {
		// TYPE VERIFICATION
		// --> NONE
		// TYPE BUILDING
		// the return type is the sequence of the types

		//org.xquark.xml.misc.Debug.print(this,"START TypeVistor:XQueryExpressionSequence ( " + arg + " )");
		ArrayList subExpressions = arg.getSubExpressions();
		// forward checking and typing
		for (int i = 0; i < subExpressions.size(); i++)
			if (!noForce || ((XQueryExpression) subExpressions.get(i)).getQType() == null) {
				((XQueryExpression) subExpressions.get(i)).accept(this);
				if (((XQueryExpression) subExpressions.get(i)).getQType() == null)
					throw new TypeException("Expression " + (XQueryExpression) subExpressions.get(i) + " could not be typed."); //return;
			}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		if (subExpressions.size() == 1)
			arg.setQType(((XQueryExpression) subExpressions.get(0)).getQType());
		else {
			ArrayList list = new ArrayList(subExpressions.size());
			boolean unique = true;
			for (int i = 0; i < subExpressions.size(); i++) {
				XQueryExpression expr = (XQueryExpression) subExpressions.get(i);
				QType qtype = expr.getQType();
				//org.xquark.xml.misc.Debug.print(this,"expr: " + expr);
				//org.xquark.xml.misc.Debug.print(this,"qtype: " + expr.getQType());
				if (i != 0 && unique && !((QType) list.get(i - 1)).equals(qtype))
					unique = false;
				list.add(qtype);
			}
			//org.xquark.xml.misc.Debug.print(this,"list: " + list);
			if (unique)
				arg.setQType(((QType) list.get(0)).changeOccurence(QType.OCC_0_N));
			//				arg.setQType(new QTypeCollection((QType) list.get(0)));
			else
				arg.setQType(new QTypeSequence(list));
		}
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVistor:XQueryExpressionSequence ( " + arg + " ) --> " + arg.getQType());
	}

	//	public void visit(XQueryFile arg) throws XQueryException;
	//	public void visit(XQueryListBinaryOperatorExpression arg) throws XQueryException;
	//	public void visit(XQueryListUnaryOperatorExpression arg) throws XQueryException;
	//	public void visit(XQueryUnaryOperatorExpression arg) throws XQueryException;
	//	public void visit(XQueryUnit arg) throws XQueryException;
	public void visit(XQueryVoid arg) throws XQueryException {
		arg.setQType(new QTypeSequence(null));
	}

	// primitive functions

	public void visit(FunctionABS arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric
		// TYPE BUILDING
		// the return type is a numeric

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionABS ( " + arg + " )");
		//ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call abs '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		if (!qtype.isNumeric())
			qtype = qtype.applyDATAFunction();
		if (qtype == null || !qtype.isNumeric())
			throw new TypeException("Function call abs '" + arg + "',  argument is not numeric");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(expr.getQType().getSimpleType()));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionABS ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionAVG arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric (or can be casted into a numeric)
		// each argument must be either
		//	- atom
		//	- attribute (hopefully the value can be casted into a numeric)
		//	- union (with at least a member being numeric)
		//	- collection (of one of the previous types)
		// TYPE BUILDING
		// the return type is a float

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionAVG ( " + arg + " ) --> " + arg.getQType());
		//ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call avg '" + arg + "', argument could not be typed"); // return;

		// type verification
		QType qtype = expr.getQType();
		if (!qtype.isNumeric())
			qtype = qtype.applyDATAFunction();
		if (qtype == null || !qtype.isNumeric())
			throw new TypeException("Function call avg '" + arg + "',  argument is not numeric");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_FLOAT)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionAVG ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionCEILING arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric
		// TYPE BUILDING
		// the return type is a numeric

		//org.xquark.xml.misc.Debug.print(this, "START TypeVisitor:FunctionCEILING ( " + arg + " )");
		//ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call ceiling '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		if (!qtype.isNumeric())
			qtype = qtype.applyDATAFunction();
		if (qtype == null || !qtype.isNumeric())
			throw new TypeException("Function call ceiling '" + arg + "',  argument is not numeric");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_INTEGER)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionCEILING ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionCOLLECTION arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be string
		// TYPE BUILDING
		// the return type is a node list

		//Debug.print(this, "START TypeVisitor:FunctionCOLLECTION ( " + arg + " )");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		ArrayList xnodes = context.getCollectionRootNodes(((ValueString) arg.getArgument(0)).getStringValue());
		arg.setXTrees(xnodes);
		QType qtype = null;
		if (xnodes != null) {
			ArrayList tmplist = new ArrayList();
			for (int i = 0; i < xnodes.size(); i++) {
				Collection childreni = ((XTreeNode) xnodes.get(i)).getChildren();
				if (childreni != null && !childreni.isEmpty()) {
					ArrayList childlist = new ArrayList(1);
					childlist.addAll(childreni);
					QType qtypei = buildQTypeFromXNodes(childlist, arg, true);
					if (qtypei != null)
						tmplist.add(new QTypeDocument(qtypei, QType.OCC_0_N, QTypeDocument.COLLECTION_ORIGIN));
				}
			}
			if (!tmplist.isEmpty()) {
				if (tmplist.size() == 1)
					qtype = (QType) tmplist.get(0);
				else
					qtype = new QTypeUnion(tmplist, QType.OCC_0_N);
			}
		}

		//ArrayList vectTypes = getDeclarationsFromSteps(arg);
		//QType qtype = buildTypeFromPathResolver(vectTypes, arg, true);
		// add   04/06/2003
		// ATTENTION for mediatorxql only
		//		if (qtype == null)
		//			qtype = new QTypeDocument(new QTypeElement(new QName("*", "*", "*", this), getType(SC_ANYTYPE)), QType.OCC_0_N, QTypeDocument.DOCUMENT_ORIGIN);
		// end add
		//org.xquark.xml.misc.Debug.print(this,"qtype: "+qtype);
		arg.setQType(qtype);

		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionCOLLECTION ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionCONCAT arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be collections
		// TYPE BUILDING
		// the return type is a integer

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionCONCAT ( " + arg + " )");
		ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			for (int i = 0; i < arguments.size(); i++)
				if (((XQueryExpression) arguments.get(i)).getQType() == null)
					return;
		for (int i = 0; i < arguments.size(); i++) {
			XQueryExpression expri = (XQueryExpression) arguments.get(i);
			if (!noForce || expri.getQType() == null)
				expri.accept(this);
			QType qtypei = expri.getQType();
			if (qtypei == null) //return;
				throw new TypeException("Function call concat '" + arg + "', argument " + (i + 1) + " could not be typed"); // return;
			if (!qtypei.isString())
				throw new TypeException("Function call concat '" + arg + "', argument " + (i + 1) + " is not a string.");
			if (!qtypei.isAtomic() && !inPredicate)
				throw new TypeException("Function call concat '" + arg + "', argument " + (i + 1) + " is not an atomic value.'");
		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionCONCAT ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionCONTAINS arg) throws XQueryException {
		// we treat now only signature like xf:contains(string $operand1, string $operand2) => boolean
		// but it exists signature like xf:contains(string $operand1, string $operand2, anyURI $collation) => boolean, later...
		// TYPE VERIFICATION
		// 2 arguments, and must be string
		// TYPE BUILDING
		// the return type is a boolean

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionCONTAINS ( " + arg + " )");

		//arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));

		ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			for (int i = 0; i < arguments.size(); i++)
				if (((XQueryExpression) arguments.get(i)).getQType() == null)
					return;
		for (int i = 0; i < arguments.size(); i++) {
			XQueryExpression expri = (XQueryExpression) arguments.get(i);
			if (!noForce || expri.getQType() == null)
				expri.accept(this);
			if (expri.getQType() == null) //return;
				throw new TypeException("Function call contains '" + arg + "', argument " + (i + 1) + " could not be typed"); // return;
			if (!expri.getQType().isAtomic() && !inPredicate)
				throw new TypeException("Function call contains '" + arg + "', argument " + (i + 1) + " is not an atomic value.'");
			if (!expri.getQType().isString())
				throw new TypeException("Function call contains '" + arg + "', argument " + (i + 1) + " is not a string.");
		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionCONTAINS ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionCOUNT arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be collections
		// TYPE BUILDING
		// the return type is a integer

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionCOUNT ( " + arg + " )");
		//ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call count '" + arg + "', argument could not be typed"); // return;
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_INTEGER)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionCOUNT ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionCURRENT_DATE arg) throws XQueryException {
		// but it exists signature like xf:current-date() => date
		// TYPE VERIFICATION
		// 0 arguments
		// TYPE BUILDING
		// the return type is a simpletype dateTime

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionCURRENT_DATE ( " + arg + " )");
		// ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_DATE)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionCURRENT_DATE ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionCURRENT_DATETIME arg) throws XQueryException {
		// but it exists signature like xf:current-dateTime() => date
		// TYPE VERIFICATION
		// 0 arguments
		// TYPE BUILDING
		// the return type is a simpletype dateTime

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionCURRENT_DATETIME ( " + arg + " )");
		// ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_DATETIME)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionCURRENT_DATETIME ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionCURRENT_TIME arg) throws XQueryException {
		// but it exists signature like xf:current-time() => date
		// TYPE VERIFICATION
		// 0 arguments
		// TYPE BUILDING
		// the return type is a simpletype dateTime

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionCURRENT_TIME ( " + arg + " )");
		// ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_TIME)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionCURRENT_TIME ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionDATA arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be simple type node
		// The built-in data function is implicitly invoked to extract the content of the node as a typed value.
		// If data is invoked with no argument, its implicit argument is the current (context) node.
		// For example, if the content of the current node is the integer 47, then data() has type integer and value 47.
		// If the content of its argument node cannot be expressed as a value of a simple type, the data function raises an error.
		// TYPE BUILDING
		// the return type depends on the node content

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionDATA ( " + arg + " )");
		//ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call count '" + arg + "', argument could not be typed"); // return;
		QType qtype = expr.getQType();
		//org.xquark.xml.misc.Debug.print(this,"qtype: "+qtype);
		//		if (qtype == null) throw new TypeException("Could not type Function DATA argument.");
		// DATA function must be applied successfully on qtype
		QType argqtype = qtype.applyDATAFunction();
		//org.xquark.xml.misc.Debug.print(this,"qtype: "+qtype);
		if (argqtype == null)
			throw new TypeException("Function call data '" + arg + "', could not apply data function to type '" + qtype + "'.");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(argqtype);
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionDATA ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionDATE arg) throws XQueryException {
		// but it exists signature like xf:current-dateTime() => date
		// TYPE VERIFICATION
		// 0 arguments
		// TYPE BUILDING
		// the return type is a simpleType date

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionDATE ( " + arg + " )");
		XQueryExpression argument = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (argument.getQType() == null)
				return;
		if (!(argument instanceof ValueString))
			throw new TypeException("Function call date '" + arg + "', argument is not a string literal");
		try {
			SimpleType simpleType = context.getSchemaManager().getDateType();
			String dateValue = simpleType.normalizedValue(((ValueString) argument).getValue());
			simpleType.validate(dateValue, false, null);
			((ValueString) argument).setValue(dateValue);
		} catch (SchemaException se) {
			throw new TypeException("Function call date '" + arg + "', argument has invalid date format");
		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getDateType()));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionDATE ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionDATETIME arg) throws XQueryException {
		// but it exists signature like xf:current-dateTime() => date
		// TYPE VERIFICATION
		// 0 arguments
		// TYPE BUILDING
		// the return type is a simpleType date

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionDATETIME ( " + arg + " )");
		XQueryExpression argument = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (argument.getQType() == null)
				return;
		if (!(argument instanceof ValueString))
			throw new TypeException("Function call datetime '" + arg + "', argument is not a string literal");
		try {
			SimpleType simpleType = context.getSchemaManager().getDateTimeType();
			String dateValue = simpleType.normalizedValue(((ValueString) argument).getValue());
			simpleType.validate(dateValue, false, null);
			((ValueString) argument).setValue(dateValue);
		} catch (SchemaException se) {
			throw new TypeException("Function call datetime '" + arg + "', argument has invalid datetime format");
		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getDateTimeType()));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionDATETIME ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionDECIMAL arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric
		// TYPE BUILDING
		// the return type is a numeric

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionDECIMAL ( " + arg + " )");
		//ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call decimal '" + arg + "', argument could not be typed"); // return;
		// CHANGE  
		QType qtype = expr.getQType();
		if (!qtype.isAtom()) {
			qtype = qtype.applyDATAFunction();
			if (!qtype.isAtom())
				throw new TypeException("Function call decimal '" + arg + "', argument has invalid type, not an atom");
		}
		if (!qtype.canBeCastedInto(QType.PRIMITIVE_decimal))
			throw new TypeException("Function call decimal '" + arg + "', argument has type that can not be casted");
		if (expr instanceof ValueString) {
			try {
				SimpleType simpleType = context.getSchemaManager().getDecimalType();
				String decimalValue = simpleType.normalizedValue(((ValueString) expr).getValue());
				simpleType.validate(decimalValue, false, null);
				((ValueString) expr).setValue(decimalValue);
			} catch (SchemaException se) {
				throw new TypeException("Function call decimal '" + arg + "', argument is string with invalid format");
			}
		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getDecimalType()));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionDECIMAL ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionDISTINCT_VALUES arg) throws XQueryException {
		// TYPE VERIFICATION
		// argument must be node list
		// TYPE BUILDING
		// the return type is a node list

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionDISTINCT_VALUES ( " + arg + " )");
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call distinct-values '" + arg + "', argument could not be typed"); // return;
		// type checking
		// REMOVE   16/04/2003
		if (!expr.getQType().isAtom() && !expr.getQType().isSimpleTypeNode())
			throw new TypeException("Function call distinct-values '" + arg + "', argument is not a collection of simple values");
		// type should be collection of homegeneous type (either comparable nodes or values)
		if (arg.getQType() != null) {
			//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionDISTINCT_VALUES ( " + arg + " ) --> " + arg.getQType());
			return;
		}
		// type setting
		arg.setQType(expr.getQType());
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionDISTINCT ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionDOC arg) throws XQueryException {

		// TYPE VERIFICATION
		// arguments must be string
		// TYPE BUILDING
		// the return type is a node list

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionDOCUMENT ( " + arg + " )");

		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		//arg.setQType(new QTypeDocument(null, QTypeDocument.DOCUMENT_ORIGIN));
		arg.setQType(new QTypeDocument(new QTypeElement(new QName("*", "*", "*", this), this.getType(SC_ANYTYPE)), QTypeDocument.DOCUMENT_ORIGIN));

		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionDOCUMENT ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionDOUBLE arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric
		// TYPE BUILDING
		// the return type is a numeric

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionDOUBLE ( " + arg + " )");
		//ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call double '" + arg + "', argument could not be typed"); // return;
		QType qtype = expr.getQType();
		if (!qtype.isAtom()) {
			qtype = qtype.applyDATAFunction();
			if (!qtype.isAtom())
				throw new TypeException("Function call double '" + arg + "', argument has invalid type, not an atom");
		}
		if (!qtype.canBeCastedInto(QType.PRIMITIVE_double))
			throw new TypeException("Function call double '" + arg + "', argument has type that can not be casted");
		if (expr instanceof ValueString) {
			try {
				SimpleType simpleType = context.getSchemaManager().getDoubleType();
				String value = simpleType.normalizedValue(((ValueString) expr).getValue());
				simpleType.validate(value, false, null);
				((ValueString) expr).setValue(value);
			} catch (SchemaException se) {
				throw new TypeException("Function call double '" + arg + "', argument is string with invalid format");
			}
		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getDoubleType()));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionDOUBLE ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionEMPTY arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a boolean

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionEMPTY ( " + arg + " )");

		//arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));

		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call empty '" + arg + "', argument could not be typed"); // return;
		if (!expr.getQType().isNode())
			throw new TypeException("Function call empty '" + arg + "', argument is not a node or collection of nodes");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionEMPTY ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionENDS_WITH arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a boolean

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionENDS_WITH ( " + arg + " )");

		//arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));

		ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			for (int i = 0; i < arguments.size(); i++)
				if (((XQueryExpression) arguments.get(i)).getQType() == null)
					return;
		for (int i = 0; i < arguments.size(); i++) {
			XQueryExpression expri = (XQueryExpression) arguments.get(i);
			if (!noForce || expri.getQType() == null)
				expri.accept(this);
			if (expri.getQType() == null) //return;
				throw new TypeException("Function call ends-with '" + arg + "', argument " + (i + 1) + " could not be typed"); // return;
			if (!expri.getQType().isString())
				throw new TypeException("Function call ends-with '" + arg + "', argument " + (i + 1) + " is not a string");
		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionENDS_WITH ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionEXISTS arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a boolean

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionEXISTS ( " + arg + " )");
		//ArrayList arguments = arg.getArguments();

		//arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));

		XQueryExpression expr = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call empty '" + arg + "', argument could not be typed"); // return;
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionEXISTS ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionFALSE arg) throws XQueryException {
		// TYPE VERIFICATION
		// TYPE BUILDING
		// the return type is a boolean atom

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionFALSE ( " + arg + " )");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionFALSE ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionFLOAT arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric
		// TYPE BUILDING
		// the return type is a numeric

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionFLOAT ( " + arg + " )");
		XQueryExpression expr = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call float '" + arg + "', argument could not be typed"); // return;
		QType qtype = expr.getQType();
		if (!qtype.isAtom()) {
			qtype = qtype.applyDATAFunction();
			if (!qtype.isAtom())
				throw new TypeException("Function call float '" + arg + "', argument has invalid type, not an atom");
		}
		if (!qtype.canBeCastedInto(QType.PRIMITIVE_float))
			throw new TypeException("Function call float '" + arg + "', argument has type that can not be casted");
		if (expr instanceof ValueString) {
			try {
				SimpleType simpleType = context.getSchemaManager().getFloatType();
				String value = simpleType.normalizedValue(((ValueString) expr).getValue());
				simpleType.validate(value, false, null);
				((ValueString) expr).setValue(value);
			} catch (SchemaException se) {
				throw new TypeException("Function call float '" + arg + "', argument is string with invalid format");
			}
		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getFloatType()));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionFLOAT ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionFLOOR arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric
		// TYPE BUILDING
		// the return type is a numeric

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionFLOOR ( " + arg + " )");
		//ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call floor '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		if (!qtype.isNumeric())
			qtype = qtype.applyDATAFunction();
		if (qtype == null || !qtype.isNumeric())
			throw new TypeException("Function call floor '" + arg + "',  argument is not numeric");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_INTEGER)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionFLOOR ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionGET_MONTH_FROM_DATE arg) throws XQueryException {
		// TYPE VERIFICATION
		// argument must be a date
		// TYPE BUILDING
		// the return type is an integer

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionGET_MONTH_FROM_DATE ( " + arg + " )");
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call lower-case '" + arg + "', argument could not be typed"); // return;
		// CHANGE  
		QType qtype = expr.getQType();
		if (!qtype.isAtom()) {
			qtype = qtype.applyDATAFunction();
			if (!qtype.isAtom())
				throw new TypeException("Function call get-month-from-date '" + arg + "', argument has invalid type, not an atom");
		}
		if (!qtype.canBeCastedInto(QType.PRIMITIVE_date))
			throw new TypeException("Function call get-month-from-date '" + arg + "', argument has type that can not be casted");

		//if (!(argument instanceof ValueString))
		//	throw new TypeException("Function call decimal '" + arg + "', argument is not a string literal");
		if (expr instanceof ValueString) {
			try {
				SimpleType simpleType = context.getSchemaManager().getDateType();
				String dateValue = simpleType.normalizedValue(((ValueString) expr).getValue());
				simpleType.validate(dateValue, false, null);
				((ValueString) expr).setValue(dateValue);
			} catch (SchemaException se) {
				throw new TypeException("Function call get-month-from-date '" + arg + "', argument is string with invalid format");
			}
		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getIntegerType()));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionGET_MONTH_FROM_DATE ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionGET_YEAR_FROM_DATE arg) throws XQueryException {
		// TYPE VERIFICATION
		// argument must be a date
		// TYPE BUILDING
		// the return type is an integer

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionGET_YEAR_FROM_DATE ( " + arg + " )");
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call lower-case '" + arg + "', argument could not be typed"); // return;
		// CHANGE  
		QType qtype = expr.getQType();
		if (!qtype.isAtom()) {
			qtype = qtype.applyDATAFunction();
			if (!qtype.isAtom())
				throw new TypeException("Function call get-month-from-date '" + arg + "', argument has invalid type, not an atom");
		}
		if (!qtype.canBeCastedInto(QType.PRIMITIVE_date))
			throw new TypeException("Function call get-month-from-date '" + arg + "', argument has type that can not be casted");

		//if (!(argument instanceof ValueString))
		//	throw new TypeException("Function call decimal '" + arg + "', argument is not a string literal");
		if (expr instanceof ValueString) {
			try {
				SimpleType simpleType = context.getSchemaManager().getDateType();
				String dateValue = simpleType.normalizedValue(((ValueString) expr).getValue());
				simpleType.validate(dateValue, false, null);
				((ValueString) expr).setValue(dateValue);
			} catch (SchemaException se) {
				throw new TypeException("Function call get-month-from-date '" + arg + "', argument is string with invalid format");
			}
		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getIntegerType()));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionGET_YEAR_FROM_DATE ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionINTEGER arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric
		// TYPE BUILDING
		// the return type is a numeric

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionINTEGER ( " + arg + " )");
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call integer '" + arg + "', argument could not be typed"); // return;
		QType qtype = expr.getQType();
		if (!qtype.canBeCastedInto(QType.PRIMITIVE_decimal))
			throw new TypeException("Function call integer '" + arg + "', argument has type that can not be casted");
		if (expr instanceof ValueString) {
			try {
				SimpleType simpleType = context.getSchemaManager().getIntegerType();
				String value = simpleType.normalizedValue(((ValueString) expr).getValue());
				simpleType.validate(value, false, null);
				((ValueString) expr).setValue(value);
			} catch (SchemaException se) {
				throw new TypeException("Function call float '" + arg + "', argument is string with invalid format");
			}
		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getIntegerType()));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionINTEGER ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionLOCAL_NAME arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a string

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionLOCAL_NAME ( " + arg + " )");
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call local-name'" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		if (!qtype.isNode())
			throw new TypeException("Function call local-name '" + arg + "', argument is not a node");
		if (!qtype.isAtomic() && !inPredicate)
			throw new TypeException("Function call local-name '" + arg + "', argument is not an atomic value.'");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionLOCAL_NAME ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionLOWER_CASE arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a string

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionLOWER_CASE ( " + arg + " )");
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call lower-case '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		if (!qtype.isString())
			throw new TypeException("Function call lower-case '" + arg + "', argument is not a string");
		if (!qtype.isAtomic() && !inPredicate)
			throw new TypeException("Function call lower-case '" + arg + "', argument is not an atomic value.'");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionLOWER_CASE ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionMATCHES arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a string

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionMATCH ( " + arg + " )");
		ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			for (int i = 0; i < arguments.size(); i++)
				if (((XQueryExpression) arguments.get(i)).getQType() == null)
					return;
		for (int i = 0; i < arguments.size(); i++) {
			XQueryExpression expri = (XQueryExpression) arguments.get(i);
			if (!noForce || expri.getQType() == null)
				expri.accept(this);
			if (expri.getQType() == null) //return;
				throw new TypeException("Function call match '" + arg + "', argument " + (i + 1) + " could not be typed"); // return;
			if (!expri.getQType().isString())
				throw new TypeException("Function call match '" + arg + "', argument " + (i + 1) + " is not a string");
			if (!expri.getQType().isAtomic() && !inPredicate)
				throw new TypeException("Function call match '" + arg + "', argument " + (i + 1) + " is not an atomic value.'");
		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_INTEGER), QType.OCC_0_N));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionMATCH ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionMAX arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric
		// TYPE BUILDING
		// the return type is a numeric

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionMAX ( " + arg + " )");
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call max '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		if (!qtype.isNumeric())
			qtype = qtype.applyDATAFunction();
		if (qtype == null || !qtype.isNumeric())
			throw new TypeException("Function call max '" + arg + "',  argument is not numeric");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_FLOAT)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionMAX ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionMIN arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric
		// TYPE BUILDING
		// the return type is a numeric

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionMIN ( " + arg + " )");
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call min '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		if (!qtype.isNumeric())
			qtype = qtype.applyDATAFunction();
		if (qtype == null || !qtype.isNumeric())
			throw new TypeException("Function call min '" + arg + "',  argument is not numeric");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_FLOAT)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionMIN ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionNAMESPACE_URI arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a string

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionNAMESPACE_URI ( " + arg + " )");
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call namespace-uri'" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		if (!qtype.isNode())
			throw new TypeException("Function call namespace-uri '" + arg + "', argument is not a node");
		if (!qtype.isAtomic() && !inPredicate)
			throw new TypeException("Function call local-name '" + arg + "', argument is not an atomic value.'");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionNAMESPACE_URI ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionNOT arg) throws XQueryException {
		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionNOT ( " + arg + " )");
		// TYPE VERIFICATION
		// arguments must be boolean
		// TYPE BUILDING
		// the return type is a boolean

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionNOT ( " + arg + " )");

		//arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));

		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call not '" + arg + "', argument could not be typed"); // return;
		// type verification
		if (!expr.getQType().isBoolean())
			throw new TypeException("Function call not '" + arg + "',  argument is not a boolean");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionNOT ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionNUMBER arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be an element
		// TYPE BUILDING
		// returns the content of an element expressed as a number

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionNUMBER ( " + arg + " )");
		ArrayList arguments = arg.getArguments();
		if (arguments != null) {
			XQueryExpression expr = (XQueryExpression) arguments.get(0);
			// forward checking and typing
			if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
				if (expr.getQType() == null)
					return;
			if (!noForce || expr.getQType() == null)
				expr.accept(this);
			if (expr.getQType() == null)
				throw new TypeException("Function call number '" + arg + "', argument could not be typed"); // return;
			// type verification
			if (!expr.getQType().isSimpleTypeNode())
				throw new TypeException("Function call number '" + arg + "',  argument is not a simple node or collection of simple nodes");
		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_FLOAT)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionNUMBER ( " + arg + " ) --> " + arg.getQType());
	}

	//	public void visit(FunctionPOSITION arg) throws XQueryException {
	//		// TYPE VERIFICATION
	//		// arguments must be an element
	//		// TYPE BUILDING
	//		// returns the content of an element expressed as a number
	//
	//		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionNUMBER ( " + arg + " )");
	//		ArrayList arguments = arg.getArguments();
	//		// type checking
	//		if (noForce && arg.getQType() != null)
	//			return;
	//		// type setting
	//		arg.setQType(new QTypeAtom((SimpleType) getType(SC_INTEGER)));
	//		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionNUMBER ( " + arg + " ) --> " + arg.getQType());
	//	}

	public void visit(FunctionROUND arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric
		// TYPE BUILDING
		// the return type is a numeric

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionROUND ( " + arg + " )");
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call round '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		if (!qtype.isNumeric())
			qtype = qtype.applyDATAFunction();
		if (qtype == null || !qtype.isNumeric())
			throw new TypeException("Function call round '" + arg + "',  argument is not numeric");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_INTEGER)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionROUND ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionSTARTS_WITH arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a boolean

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionSTARTS_WITH ( " + arg + " )");

		//arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));

		ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			for (int i = 0; i < arguments.size(); i++)
				if (((XQueryExpression) arguments.get(i)).getQType() == null)
					return;
		for (int i = 0; i < arguments.size(); i++) {
			XQueryExpression expri = (XQueryExpression) arguments.get(i);
			if (!noForce || expri.getQType() == null)
				expri.accept(this);
			if (expri.getQType() == null) //return;
				throw new TypeException("Function call starts-with '" + arg + "', argument " + (i + 1) + " could not be typed"); // return;
			if (!expri.getQType().isString())
				throw new TypeException("Function call starts-with '" + arg + "', argument " + (i + 1) + " is not a string");
			if (!expri.getQType().isAtomic() && !inPredicate)
				throw new TypeException("Function call starts-with '" + arg + "', argument " + (i + 1) + " is not an atomic value.'");
		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionSTARTS_WITH ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionSTRING arg) throws XQueryException {
		// TYPE VERIFICATION
		/*
		The string function converts its argument to a string according to the following rules
		A number is converted to a string as follows:
			NaN is converted to the string NaN.
			Positive zero is converted to the string 0.
			Negative zero is converted to the string 0.
			Positive infinity is converted to the string Infinity.
			Negative infinity is converted to the string -Infinity.
			If the number is an integer, the number is represented in decimal form with no decimal point and no leading zeros, preceded by a minus sign (-) if the number is negative.
			Otherwise, the number is represented in decimal form including a decimal point with at least one digit before the decimal point and at least one digit after the decimal point, preceded by a minus sign (-) if the number is negative; there must be no leading zeros before the decimal point apart possibly from the one required digit immediately before the decimal point; beyond the one required digit after the decimal point there must be as many, but only as many, more digits as are needed to uniquely distinguish the number from all other IEEE 754 numeric values.
		The boolean false value is converted to the string false. The boolean true value is converted to the string true.
		An object of a type other than the above types is converted to a string in a way that is dependent on that type.
		If the argument is omitted, it defaults to a node-set with the context node as its only member
		 */
		// each argument must be either
		//	- element
		//	- attribute
		//	- union (with at least a member being node)
		//	- collection (of one of the previous types)
		// TYPE BUILDING
		// returns the content of an element expressed as a string

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionSTRING ( " + arg + " )");
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call string '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		// type verification
		//if (!qtype.isAtom()) qtype = qtype.applyDATAFunction();
		if (qtype == null || (!qtype.isAtomic() && !inPredicate && !qtype.isSimpleTypeNode()))
			throw new TypeException("Function call string '" + arg + "', argument must be an atomic value or simple content element.");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING), qtype.isMultiple() ? QType.OCC_0_N : QType.OCC_1_1));
		//		if (qtype instanceof QTypeCollection)
		//			arg.setQType(new QTypeCollection(new QTypeAtom((SimpleType) getType(SC_STRING))));
		//		else
		//			arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionSTRING ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionSTRING_LENGTH arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a boolean

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionSTRING_LENGTH ( " + arg + " )");
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		// type verification
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null) //return;
			throw new TypeException("Function call string-length '" + arg + "', argument could not be typed"); // return;
		if (!expr.getQType().isString())
			throw new TypeException("Function call string-length '" + arg + "', argument is not a string");
		if (!expr.getQType().isAtomic() && !inPredicate)
			throw new TypeException("Function call string-length '" + arg + "', argument is not an atomic value.'");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_INTEGER)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionSTRING_LENGTH ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionSUBSTRING arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a string

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionSUBSTRING ( " + arg + " )");
		ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			for (int i = 0; i < arguments.size(); i++)
				if (((XQueryExpression) arguments.get(i)).getQType() == null)
					return;
		for (int i = 0; i < arguments.size(); i++) {
			XQueryExpression expri = (XQueryExpression) arguments.get(i);
			if (!noForce || expri.getQType() == null)
				expri.accept(this);
			if (expri.getQType() == null) //return;
				throw new TypeException("Function call substring '" + arg + "', argument " + (i + 1) + " could not be typed"); // return;
			if (i == 0 && !expri.getQType().isString())
				throw new TypeException("Function call substring '" + arg + "', argument " + (i + 1) + " is not a string");
			if (i != 0 && !expri.getQType().isNumeric())
				throw new TypeException("Function call substring '" + arg + "', argument " + (i + 1) + " is not numeric");
		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionSUBSTRING ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionSUM arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be numeric (or can be casted into a numeric)
		// each argument must be either
		//	- atom
		//	- attribute (hopefully the value can be casted into a numeric)
		//	- union (with at least a member being numeric)
		//	- collection (of one of the previous types)
		// TYPE BUILDING
		// the return type is a float

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionSUM ( " + arg + " )");
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call sum '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		// type verification
		if (!qtype.isNumeric())
			throw new TypeException("Function call sum '" + arg + "', argument is not numeric");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_FLOAT)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionSUM ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionTIME arg) throws XQueryException {
		// but it exists signature like xf:current-dateTime() => date
		// TYPE VERIFICATION
		// 0 arguments
		// TYPE BUILDING
		// the return type is a simpleType date

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionTIME ( " + arg + " )");
		XQueryExpression argument = (XQueryExpression) arg.getArguments().get(0);
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (argument.getQType() == null)
				return;
		if (!(argument instanceof ValueString))
			throw new TypeException("Function call time '" + arg + "', argument is not a string literal");
		try {
			SimpleType simpleType = context.getSchemaManager().getTimeType();
			String timeValue = simpleType.normalizedValue(((ValueString) argument).getValue());
			simpleType.validate(timeValue, false, null);
			((ValueString) argument).setValue(timeValue);
		} catch (SchemaException se) {
			throw new TypeException("Function call time '" + arg + "', argument has invalid datetime format");
		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom(context.getSchemaManager().getTimeType()));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionTIME ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionTRUE arg) throws XQueryException {
		// TYPE VERIFICATION
		// TYPE BUILDING
		// the return type is a boolean atom

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionTRUE ( " + arg + " )");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionTRUE ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionUPPER_CASE arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a string

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionUPPER_CASE ( " + arg + " )");
		ArrayList arguments = arg.getArguments();
		XQueryExpression expr = (XQueryExpression) arguments.get(0);
		// forward checking and typing
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			if (expr.getQType() == null)
				return;
		if (!noForce || expr.getQType() == null)
			expr.accept(this);
		if (expr.getQType() == null)
			throw new TypeException("Function call upper-case '" + arg + "', argument could not be typed"); // return;
		// type verification
		QType qtype = expr.getQType();
		if (!qtype.isString())
			throw new TypeException("Function call upper-case '" + arg + "', argument is not a string");
		if (!qtype.isAtomic() && !inPredicate)
			throw new TypeException("Function call upper-case '" + arg + "', argument is not an atomic value.'");
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_STRING)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionUPPER_CASE ( " + arg + " ) --> " + arg.getQType());
	}

	public void visit(FunctionDEEP_EQUALS arg) throws XQueryException {
		// TYPE VERIFICATION
		// arguments must be sequence
		// TYPE BUILDING
		// the return type is a boolean

		//org.xquark.xml.misc.Debug.print(this,"START TypeVisitor:FunctionDEEP_EQUALS ( " + arg + " )");

		//arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));

		ArrayList arguments = arg.getArguments();
		// forward checking and typing, and verification
		if ((contextXNodes == null || contextXNodes.isEmpty()) && contextQType == null)
			for (int i = 0; i < arguments.size(); i++)
				if (((XQueryExpression) arguments.get(i)).getQType() == null)
					return;
		for (int i = 0; i < arguments.size(); i++) {
			XQueryExpression expri = (XQueryExpression) arguments.get(i);
			if (!noForce || expri.getQType() == null)
				expri.accept(this);
			if (expri.getQType() == null) //return;
				throw new TypeException("Function call deep-equals '" + arg + "', argument " + (i + 1) + " could not be typed"); // return;
			if (!expri.getQType().isAtomic() && !inPredicate)
				throw new TypeException("Function call deep-equals '" + arg + "', argument " + (i + 1) + " is not an atomic value.'");
		}
		// type checking
		if (noForce && arg.getQType() != null)
			return;
		// type setting
		arg.setQType(new QTypeAtom((SimpleType) getType(SC_BOOLEAN)));
		//org.xquark.xml.misc.Debug.print(this,"STOP  TypeVisitor:FunctionDEEP_EQUALS ( " + arg + " ) --> " + arg.getQType());
	}

	//*******************************************************************************************
	// UTLITIES
	//*******************************************************************************************
	private Type getType(String typeStr) {
		return context.getSchemaManager().getType(XML_SCHEMA, typeStr);
	}

	private Type getType(SequenceTypeExpression sequenceType) throws TypeException {
		//org.xquark.xml.misc.Debug.print(this, "qname: "+qname);
		//org.xquark.xml.misc.Debug.print(this, "qname.getNameSpace(): "+qname.getNameSpace());
		//org.xquark.xml.misc.Debug.print(this, "qname.getLocalName(): "+qname.getLocalName());
		//org.xquark.xml.misc.Debug.print(this, "schemamanager.getSchemas(): "+schemamanager.getSchemas());
		Schema schema = null;
		try {
			schema = context.getSchemaManager().loadSchema(context.getSchemaManager(), sequenceType.getNameSpace());
		} catch (Exception e) {
			throw new TypeException("Could not load schema '" + sequenceType.getNameSpace() + "'");
		}
		if (schema == null)
			throw new TypeException("Could not load schema '" + sequenceType.getNameSpace() + "'");
		Type type = schema.getType(sequenceType.getLocalName());
		if (type == null)
			throw new TypeException("Could not find type '" + sequenceType + "'");
		return type;
	}

	private QType buildQTypeFromXNodes(ArrayList schemaNodes) throws XQueryException {
		return buildQTypeFromXNodes(schemaNodes, null, false);
	}

	private QType buildQTypeFromXNodes(ArrayList schemaNodes, XQueryExpression expr, boolean verify) throws XQueryException {
		//org.xquark.xml.misc.Debug.print(this,"START buildTypeFromPathResolver ( " + schemaNodes + " ) ");
		if (schemaNodes == null || schemaNodes.isEmpty())
			//		{
			//			//SimpleType type = new SimpleType(null, SC_ANYSIMPLETYPE, null);
			//			//type.setPrimitive(PrimitiveType.createType(SC_ANYSIMPLETYPE));
			//			SimpleType type = this.schemamanager.getAnySimpleType();
			//			ArrayList list = new ArrayList();
			//			list.add(new QTypeAtom(type));
			//			list.add(new QTypeAttribute(new QName("*","*", "*", this), type));
			//			list.add(new QTypeElement(new QName("*","*", "*", this), getType(SC_ANYTYPE)));
			//			QType qtype = new QTypeUnion(list);
			//			return qtype;
			//		}
			// throw new TypeException("Expression '" + expr + "' not found in metadata");
			return null;

		QType retQType = null;
		int size = schemaNodes.size();
		if (size == 1) {
			//			Declaration decl = ((TypedXTreeNode) schemaNodes.get(0)).getDeclaration();
			//			if (decl == null)
			//				throw new TypeException("Could not find declaration(s) of expression '" + expr + "'");
			//			retQType = getQTypeFromDeclaration(decl);
			//			//org.xquark.xml.misc.Debug.print(this,"STOP  buildTypeFromPathResolver ( " + schemaNodes + " ) --> " + retQType);
			TypedXTreeNode typedXTreeNode = (TypedXTreeNode) schemaNodes.get(0);
			Declaration decl = typedXTreeNode.getDeclaration();
			if (decl == null) {
				//				SchemaNode schemaNode = (SchemaNode) schemaNodes.get(0);
				if (typedXTreeNode.getLocalName() != null) {
					if (typedXTreeNode.getChildren() == null) {
						//					SimpleType type = new SimpleType(this.typingschema, SC_ANYSIMPLETYPE, null);
						//					type.setPrimitive(PrimitiveType.createType(SC_ANYSIMPLETYPE));
						SimpleType type = context.getSchemaManager().getAnySimpleType();
						//if (!isAttribute) retQType = new QTypeElement(new QName(typedXTreeNode.getNamespace(),null, typedXTreeNode.getLocalName(), null), type);
						// Olivier 2002/10/10 : add new QTypeCollection(..)
						retQType = new QTypeAttribute(new QName(typedXTreeNode.getNamespace(), null, typedXTreeNode.getLocalName(), null), type, isAttribute ? QType.OCC_1_1 : QType.OCC_0_N);
						//					if (!isAttribute)
						//						retQType = new QTypeCollection(new QTypeElement(new QName(typedXTreeNode.getNamespace(), null, typedXTreeNode.getLocalName(), null), type));
						//					else
						//						retQType = new QTypeAttribute(new QName(typedXTreeNode.getNamespace(), null, typedXTreeNode.getLocalName(), null), type);
					} else {
						//retQType = new QTypeElement(new QName(typedXTreeNode.getNamespace(),null, typedXTreeNode.getLocalName(), null), getType(SC_ANYTYPE));
						// Olivier 2002/10/10 : add new QTypeCollection(..)
						retQType = new QTypeElement(new QName(typedXTreeNode.getNamespace(), null, typedXTreeNode.getLocalName(), null), getType(SC_ANYTYPE), QType.OCC_0_N);
						//					retQType = new QTypeCollection(new QTypeElement(new QName(typedXTreeNode.getNamespace(), null, typedXTreeNode.getLocalName(), null), getType(SC_ANYTYPE)));
					}
				} else {
					// TODO find out why this case happens --> collection("ITEMS")//DESCRIPTION
					// for now do nothing
				}
			} else
				retQType = getQTypeFromDeclaration(decl);
			//org.xquark.xml.misc.Debug.print(this,"STOP  buildTypeFromPathResolver ( " + schemaNodes + " ) --> " + retQType);
			return retQType;
		}
		ArrayList list = new ArrayList();
		QType tmpQType = null;
		for (int i = 0; i < size; i++) {
			TypedXTreeNode typedXTreeNode = (TypedXTreeNode) schemaNodes.get(i);
			Declaration decl = typedXTreeNode.getDeclaration();
			if (decl == null) {
				if (typedXTreeNode.getLocalName() != null) {
					if (typedXTreeNode.getChildren() == null) {
						//					SimpleType type = new SimpleType(this.typingschema, SC_ANYSIMPLETYPE, null);
						//					type.setPrimitive(PrimitiveType.createType(SC_ANYSIMPLETYPE));
						SimpleType type = context.getSchemaManager().getAnySimpleType();
						//if (!isAttribute) tmpQType = new QTypeElement(new QName(typedXTreeNode.getNamespace(),null, typedXTreeNode.getLocalName(), null), type);
						// Olivier 2002/10/10 : add new QTypeCollection(..)
						tmpQType = new QTypeAttribute(new QName(typedXTreeNode.getNamespace(), null, typedXTreeNode.getLocalName(), null), type, isAttribute ? QType.OCC_1_1 : QType.OCC_0_N);
						//					if (!isAttribute)
						//						tmpQType = new QTypeCollection(new QTypeElement(new QName(typedXTreeNode.getNamespace(), null, typedXTreeNode.getLocalName(), null), type));
						//					else
						//						tmpQType = new QTypeAttribute(new QName(typedXTreeNode.getNamespace(), null, typedXTreeNode.getLocalName(), null), type);
					}
					//else tmpQType = new QTypeElement(new QName(typedXTreeNode.getNamespace(),null, typedXTreeNode.getLocalName(), null), getType(SC_ANYTYPE));
					// Olivier 2002/10/10 : add new QTypeCollection(..)
					else
						tmpQType = new QTypeElement(new QName(typedXTreeNode.getNamespace(), null, typedXTreeNode.getLocalName(), null), getType(SC_ANYTYPE), QType.OCC_0_N);
					//					tmpQType = new QTypeCollection(new QTypeElement(new QName(typedXTreeNode.getNamespace(), null, typedXTreeNode.getLocalName(), null), getType(SC_ANYTYPE)));
				} else {
					// TODO find out why this case happens --> collection("ITEMS")//DESCRIPTION
					// for now do nothing
				}
			} else
				tmpQType = getQTypeFromDeclaration(decl);
			//org.xquark.xml.misc.Debug.print(this,"tmpQType [" + i + "] : "+tmpQType);
			list.add(tmpQType);
		}
		if (list.isEmpty())
			if (verify)
				throw new TypeException("Could not find declaration(s) of expression '" + expr + "'");
			else
				return null;
		retQType = makeUnion(list);
		//org.xquark.xml.misc.Debug.print(this,"STOP  buildTypeFromPathResolver ( " + schemaNodes + " ) --> " + retQType);
		return retQType;
	}

	private QType getQTypeFromDeclaration(Declaration decl) throws XQueryException {
		//org.xquark.xml.misc.Debug.print(this,"START getQTypeFromDeclaration ( " + decl + " ) ");
		if (decl == null)
			return null;
		// throw new TypeException("Expression not found in metadata");
		//org.xquark.xml.misc.Debug.print(this,"decl.getClass().getName() : "+decl.getClass().getName());
		QType retQType = null;
		QName qname = new QName(decl.getNamespace(), null, decl.getName(), null);
		if (decl instanceof AttributeDeclaration)
			retQType = new QTypeAttribute(qname, (SimpleType) decl.getType());
		if (decl instanceof ElementDeclaration) {
			boolean multi = false;
			ElementDeclaration eltDecl = (ElementDeclaration) decl;
			SchemaScope scopeType = eltDecl.getScope(); // see zhang yun for always true!
			if (scopeType != null && scopeType instanceof ComplexType) {
				int size[] = { 0, 0 };
				size = ((ComplexType) scopeType).getDeclarationOccurrence(eltDecl.getNamespace(), eltDecl.getName());
				if (size != null && size[1] > 1)
					multi = true;
			}
			retQType = new QTypeElement(qname, decl.getType(), multi ? QType.OCC_0_N : QType.OCC_1_1);
			//			if (multi)
			//				retQType = new QTypeCollection(new QTypeElement(qname, decl.getType()));
			//			else
			//				retQType = new QTypeElement(qname, decl.getType());
		}
		//org.xquark.xml.misc.Debug.print(this,"STOP  getQTypeFromDeclaration ( " + decl + " ) --> " + retQType);
		return retQType;
	}

	private QType getQTypeFromParticle(Particle particle) throws XQueryException {
		//org.xquark.xml.misc.Debug.print(this,"START getQTypeFromParticle ( " + particle + " ) ");
		if (particle == null)
			return null;
		Object term = particle.getTerm();
		if (term instanceof ElementDeclaration)
			return getQTypeFromDeclaration((ElementDeclaration) term);
		return null;
	}

	private QType getQTypeFromType(Type type) throws XQueryException {
		//org.xquark.xml.misc.Debug.print(this,"START getQTypeFromType ( " + type + " ) ");
		if (type.getValueType() != null)
			return new QTypeAtom(type.getValueType());
		else // type instanceof ComplexType
			{
			ComplexType complextype = (ComplexType) type;
			//org.xquark.xml.misc.Debug.print(this,"complextype: "+complextype);
			// QType qtype = null;
			ContentModel contentmodel = complextype.getContentModel();
			//org.xquark.xml.misc.Debug.print(this,"contentmodel: "+contentmodel);
			Object model = contentmodel.getModel();
			//org.xquark.xml.misc.Debug.print(this,"model: "+model);
			//org.xquark.xml.misc.Debug.print(this,"model.getClass().getName(): "+model.getClass().getName());
			if (model instanceof Particle) {
				Particle particle = (Particle) model;
				//org.xquark.xml.misc.Debug.print(this,"particle: "+particle);
				int min = particle.getMinOccurs();
				//org.xquark.xml.misc.Debug.print(this,"min: "+min);
				int max = particle.getMaxOccurs();
				//org.xquark.xml.misc.Debug.print(this,"max: "+max);
				Object term = particle.getTerm();
				//org.xquark.xml.misc.Debug.print(this,"term: "+term);
				//org.xquark.xml.misc.Debug.print(this,"term.getClass().getName(): "+term.getClass().getName());
				if (term instanceof ElementDeclaration) {
					// by default a sequence?
					return getQTypeFromDeclaration((ElementDeclaration) term).changeOccurence(QType.OCC_0_N);
					//					return new QTypeCollection(getQTypeFromDeclaration((ElementDeclaration) term));
				}
				if (term instanceof ModelGroup) {
					ModelGroup modelgroup = (ModelGroup) term;
					//org.xquark.xml.misc.Debug.print(this,"modelgroup: "+modelgroup);
					int compositor = modelgroup.getCompositor();
					//org.xquark.xml.misc.Debug.print(this,"compositor: "+compositor);

					// CHOICE and ALL not yet do...
					switch (compositor) {
						case ModelGroup.SEQUENCE :
							//org.xquark.xml.misc.Debug.print(this,"modelgroup.size(): "+modelgroup.size());
							if (modelgroup.size() == 1) {
								if (max < 1)
									return null;
								else if (max == 1)
									return getQTypeFromParticle((Particle) modelgroup.get(0));
								else /* if (max > 1) */
									return getQTypeFromParticle((Particle) modelgroup.get(0)).changeOccurence(QType.OCC_0_N);
								//									return new QTypeCollection(getQTypeFromParticle((Particle) modelgroup.get(0)));
							} else {
								ArrayList subList = new ArrayList();
								for (int i = 0; i < modelgroup.size(); i++) {
									QType subQType = getQTypeFromParticle((Particle) modelgroup.get(i));
									if (subQType != null)
										subList.add(subQType);
								}
								if (subList.isEmpty())
									return null;
								return new QTypeSequence(subList);
							}
						case ModelGroup.CHOICE :
							//org.xquark.xml.misc.Debug.print(this,"modelgroup.size(): "+modelgroup.size());
							if (modelgroup.size() == 1) {
								if (max < 1)
									return null;
								else if (max == 1)
									return getQTypeFromParticle((Particle) modelgroup.get(0));
								else /* if (max > 1) */
									return getQTypeFromParticle((Particle) modelgroup.get(0)).changeOccurence(QType.OCC_0_N);
								//									return new QTypeCollection(getQTypeFromParticle((Particle) modelgroup.get(0)));
							} else {
								ArrayList subList = new ArrayList();
								for (int i = 0; i < modelgroup.size(); i++) {
									QType subQType = getQTypeFromParticle((Particle) modelgroup.get(i));
									if (subQType != null)
										subList.add(subQType);
								}
								if (subList.isEmpty())
									return null;
								return new QTypeUnion(subList);
							}
						case ModelGroup.ALL :
							return null;
					}
				}
			}
		}
		return null;
	}

	private QType makeUnion(ArrayList list) throws XQueryException {
		//org.xquark.xml.misc.Debug.print(this,"START makeUnion ( " + list + " ) ");
		if (list == null || list.isEmpty())
			return null;
		for (int i = 0; i < list.size() - 1; i++) {
			QType qtypei = (QType) list.get(i);
			for (int j = i + 1; j < list.size(); j++) {
				if (qtypei.equals((QType) list.get(j))) {
					list.remove(j);
					j--;
				}
			}
		}
		if (list.size() == 1)
			return (QType) list.get(0);
		return new QTypeUnion(list, QType.OCC_0_N);
	}

	//	private SimpleType getTypeFromIsAtom(QType qtype) {
	//		switch (qtype.getSubClass()) {
	//			case QType.ATOM :
	//				return qtype.getSimpleType();
	//			case QType.COLLECTION :
	//				return getTypeFromIsAtom(qtype.getQType());
	//			case QType.UNION :
	//				// in that case, we return a "string" simpletype because UNION could contain multiple qtype
	//				// (ATOM, ATTRIBUTE, ELEMENT...), and each of the atom-type (isAtom()) could be "string",
	//				// "integer"... simpletype
	//				return (SimpleType) getType(SC_STRING);
	//			default :
	//				// this case is theoretically impossible
	//				return null;
	//		}
	//	}

	private QTypeElement buildQTypeElementFromSubElements(XQueryExpression name, ArrayList allElements) throws XQueryException {
		int size = allElements.size();
		ElementTools eTools = new ElementTools(size, name);
		// case of <tag/>
		if (size == 0)
			return new QTypeElement(name, getType(SC_ANYTYPE));
		ComplexType complexType = new ComplexType(this.typingschema, null, null);
		//complexType.setRequiredAndDefaultAttributes(new ArrayList());
		ContentModel contentModel = new ContentModel();
		SequenceModelGroup sequenceModel = new SequenceModelGroup();
		QTypeElement qte = null;
		try {
			for (int i = 0; i < size; i++) {
				QType qtype = ((XQueryExpression) allElements.get(i)).getQType();
				int subclass = qtype.getSubClass();
				if (subclass == QType.ATOM) {
					// case of <tag> [simpleType] </tag>, like <a> 23 </a>
					if (size == 1)
						return new QTypeElement(name, qtype.getSimpleType());
					else
						buildQTypeAtom(eTools, qtype);
				} else if (subclass == QType.TEXT) {
					if (size == 1)
						return new QTypeElement(name, qtype.getSimpleType());
					else
						buildQTypeText(eTools, qtype);
				} else if (subclass == QType.ATTRIBUTE) {
					buildQTypeAttribute(eTools, qtype, complexType);
				} else if (subclass == QType.ELEMENT) {
					buildQTypeElement(qtype, complexType, sequenceModel, qtype.getMinOccurence() == 1, qtype.isMultiple());
				} else if (subclass == QType.UNION) {
					buildQTypeUnion(eTools, qtype, complexType, sequenceModel);
				} else if (subclass == QType.SEQUENCE) {
					qte = buildQTypeSequence(eTools, qtype, complexType, sequenceModel);
					if (qte != null)
						return qte;
				}
			}
			//org.xquark.xml.misc.Debug.print(this,"eTools: "+eTools);
			// case of <tag> [simpleType]+ </tag>, like <a> 23 toto</a>
			if (eTools.getCountAttribute() == 0 && eTools.getSize() == eTools.getCountSimpleType())
				return new QTypeElement(name, (SimpleType) getType(SC_STRING));
			// case of <tag (attr)* > [simpleType]+ </tag>, like <a b="1"> 23 toto </a>
			if ((eTools.getCountAttribute() == eTools.getSize() - eTools.getCountSimpleType()) && (eTools.getMixed())) {
				complexType.setBaseType((SimpleType) getType(SC_STRING));
				complexType.setExtension(true);
				complexType.setContentModel(new ContentModel((SimpleType) getType(SC_STRING)));
			}
			// case of <tag (attr)* > [simpleType] </tag>, like <a b="1"> 23 </a>
			else if ((eTools.getCountAttribute() == eTools.getSize() - 1) && (eTools.getMixed())) {
				complexType.setBaseType(eTools.getSimpleType());
				complexType.setExtension(true);
				complexType.setContentModel(new ContentModel(eTools.getSimpleType()));

			}
			// all others cases...
			else {
				contentModel.setModel(new Particle(sequenceModel));
				if (eTools.getMixed())
					contentModel.setContentType(SchemaConstants.MIXED);
				else
					contentModel.setContentType(SchemaConstants.ELEMENT_ONLY);
				complexType.setContentModel(contentModel);
			}
		} catch (SchemaException se) {
			complexType = (ComplexType) this.getType(SC_ANYTYPE);
		}
		return new QTypeElement(name, complexType);
	}

	private void buildQTypeAtom(ElementTools eTools, QType qtype) throws XQueryException {
		eTools.setSimpleType(qtype.getSimpleType());
		eTools.setMixed(true);
		eTools.incCountSimpleType();
		eTools.incCountSimpleList();
		// case of <tag> ([simpleType])+ </tag>, we have not yet specification
		//if (eTools.getSize() == eTools.getCountSimpleType()) throw new TypeException("The element can not be built, because all these sub-elements are simple type (NYI).");
		// case of <tag> ([simpleList])+ </tag>, we have not yet specification
		//if (eTools.getSize() == eTools.getCountSimpleList()) throw new TypeException("The element can not be built, because all these sub-elements are lists (NYI).");
	}

	private void buildQTypeText(ElementTools eTools, QType qtype) throws XQueryException {
		eTools.setSimpleType(qtype.getSimpleType());
		eTools.setMixed(true);
		eTools.incCountSimpleType();
		eTools.incCountSimpleList();
		// case of <tag> ([simpleType])+ </tag>, we have not yet specification
		//if (eTools.getSize() == eTools.getCountSimpleType()) throw new TypeException("The element can not be built, because all these sub-elements are simple type (NYI).");
		// case of <tag> ([simpleList])+ </tag>, we have not yet specification
		//if (eTools.getSize() == eTools.getCountSimpleList()) throw new TypeException("The element can not be built, because all these sub-elements are lists (NYI).");
	}

	private void buildQTypeAttribute(ElementTools eTools, QType qtype, ComplexType complexType) throws XQueryException {
		eTools.incCountAttribute();
		String name = "";
		try {
			name = ((QName) ((QTypeAttribute) qtype).getName()).getName();
		} catch (ClassCastException cce) {
			name = "mediator:TagDefineInEvaluation";
		}
		AttributeDeclaration attribute = new AttributeDeclaration(this.typingschema, name, complexType);
		attribute.setType(qtype.getSimpleType());
		try {
			complexType.register(attribute);
		} catch (SchemaException saxe) {
			throw new TypeException("Error in register(ElementDeclaration): " + saxe.getMessage());
		}
	}

	//	private QTypeElement buildQTypeCollection(ElementTools eTools, QType qtype, ComplexType complexType, SequenceModelGroup sequenceModel) throws XQueryException, SchemaException {
	//		QType ctype = qtype.getQType();
	//		if (ctype instanceof QTypeAtom) {
	//			if (eTools.getSize() == 1) {
	//				SimpleType base = ctype.getSimpleType();
	//				SimpleType list = new SimpleType(this.typingschema, "listOF" + base.getName(), null);
	//				list.setVariety(SimpleType.VARIETY_LIST);
	//				list.setItemType(base);
	//				PrimitiveType primitive = ((SimpleType) base).getPrimitive();
	//				primitive = PrimitiveType.createListType(primitive);
	//				((SimpleType) list).setPrimitive(primitive);
	//				return new QTypeElement(eTools.getName(), list);
	//			}
	//			else
	//				buildQTypeAtom(eTools, ctype);
	//		}
	//		else if (ctype instanceof QTypeAttribute) {
	//			buildQTypeAttribute(eTools, ctype, complexType);
	//		}
	//		else if (ctype instanceof QTypeElement) {
	//			buildQTypeElement(ctype, complexType, sequenceModel, ctype.isMultiple());
	//		}
	//		else if (ctype instanceof QTypeUnion) {
	//			return new QTypeElement(eTools.getName(), getType(SC_ANYTYPE));
	//		}
	//		else if (ctype instanceof QTypeSequence) {
	//			throw new TypeException("We have not specification for this case, NYI.");
	//		}
	//		else if (ctype instanceof QTypeCollection) {
	//			throw new TypeException("According QTypeCollection constructor, this case can not occur.");
	//		}
	//		return null;
	//	}

	private void buildQTypeElement(QType qtype, ComplexType complexType, ModelGroup model, boolean hasOne, boolean unbounded) throws XQueryException, SchemaException {
		String name = null;
		String uri = null;
		Schema tmpschema = this.typingschema;
		try {
			QName qname = (QName) ((QTypeElement) qtype).getName();
			uri = qname.getNameSpace();
			if (uri != null) {
				tmpschema = this.typingschemamanager.getSchema(uri);
				if (tmpschema == null) {
					tmpschema = new Schema(uri, null, typingschemamanager);
					try {
						typingschemamanager.putSchema(typingschema);
					} catch (SAXException saxe) {
						throw new TypeException("Could not add schema to manager");
					}
				}
			}
			name = qname.getName();
		} catch (ClassCastException cce) {
			name = "TagDefineInEvaluation";
		}
		ElementDeclaration element = null;
		SchemaScope scope = null;
		if (uri != null) {
			scope = tmpschema;
			element = new ElementDeclaration(tmpschema, name, scope);
			element.setQualified(true);
		} else {
			scope = complexType;
			element = new ElementDeclaration(tmpschema, name, scope);
		}
		element.setType(qtype.getType());
		// Olivier: 2002/11/21
		if (uri != null)
			complexType.register(element);
		scope.register(element);
		Particle particle = new Particle(element);
		if (!hasOne)
			particle.setMinOccurs(0);
		if (unbounded)
			particle.setMaxOccursUnbounded();
		model.add(particle);
	}

	private QTypeElement buildQTypeSequence(ElementTools eTools, QType qtype, ComplexType complexType, SequenceModelGroup sequenceModel) throws XQueryException, SchemaException {
		List list = qtype.getList();
		if (list == null)
			return null;
		int size = list.size();
		QTypeElement qte = null;
		for (int i = 0; i < size; i++) {
			QType stype = (QType) list.get(i);
			int subclass = stype.getSubClass();
			if (subclass == QType.ATOM) {
				// case of <tag> [simpleType] </tag>, like <a> 23 </a>
				if (size == 1 && eTools.getSize() == 1)
					return new QTypeElement(eTools.getName(), qtype.getSimpleType());
				else
					buildQTypeAtom(eTools, stype);
			} else if (subclass == QType.TEXT) {
				buildQTypeText(eTools, stype);
			} else if (subclass == QType.ATTRIBUTE) {
				buildQTypeAttribute(eTools, stype, complexType);
			} else if (subclass == QType.ELEMENT) {
				buildQTypeElement(stype, complexType, sequenceModel, stype.getMinOccurence() == 1, stype.isMultiple());
			} else if (subclass == QType.UNION) {
				buildQTypeUnion(eTools, stype, complexType, sequenceModel);
			} else if (subclass == QType.SEQUENCE) {
				qte = buildQTypeSequence(eTools, stype, complexType, sequenceModel);
				if (qte != null)
					return qte;
			}
			//			else if (stype instanceof QTypeCollection) {
			//				qte = buildQTypeCollection(eTools, stype, complexType, sequenceModel);
			//				if (qte != null)
			//					return qte;
			//			}

		}
		return null;
	}

	private void buildQTypeUnion(ElementTools eTools, QType qtype, ComplexType complexType, SequenceModelGroup sequenceModel) throws XQueryException, SchemaException {
		List list = qtype.getList();
		ChoiceModelGroup choiceModel = null;
		for (int i = 0; i < list.size(); i++) {
			if (i != 0)
				eTools.incSize();
			QType utype = (QType) list.get(i);
			int subclass = utype.getSubClass();
			boolean atom = false;
			if (subclass == QType.ATOM) {
				if (!atom) {
					buildQTypeAtom(eTools, utype);
					atom = true;
				}
			} else if (subclass == QType.TEXT) {
				buildQTypeText(eTools, utype);
			} else if (subclass == QType.ATTRIBUTE) {
				buildQTypeAttribute(eTools, utype, complexType);
			} else if (subclass == QType.ELEMENT) {
				if (choiceModel == null)
					choiceModel = new ChoiceModelGroup();
				buildQTypeElement(utype, complexType, choiceModel, utype.getMinOccurence() == 1, utype.isMultiple());
			} else if (subclass == QType.SEQUENCE) {
				this.buildQTypeSequence(eTools, utype, complexType, sequenceModel);
			}
			//			else if (utype instanceof QTypeCollection) {
			//				this.buildQTypeCollection(eTools, utype, complexType,new SequenceModelGroup());
			//				choice = true;
			//			}
			else
				throw new TypeException("We have not specification for this case (NYI), definition: union { (atom | element | attribute)+ }.");

		}
		if (choiceModel != null)
			sequenceModel.add(new Particle(choiceModel));
	}

	private boolean isAnyTypeKind(QType qType) {
		Type type = qType.getType();
		if (type != null && type.getNamespace() != null && type.getNamespace().equals(SchemaConstants.XMLSCHEMA_URI) && type.getName() != null && (type.getName().equals(SchemaConstants.ANY_TYPE) || type.getName().equals(SchemaConstants.SIMPLE_UR_TYPE)))
			return true;
		return false;
	}

	/*
	private boolean hasChildNodes(ArrayList xtrees) {
		if (xtrees == null || xtrees.isEmpty()) return false;
		SchemaNode schemaNode = null;
		for (int i=0;i<xtrees.size();i++) {
			schemaNode = (SchemaNode)xtrees.get(i);
			schemaNode.getChildren();
		}
		return true;
	}
	 */
}
