/*
 * This file belongs to the XQuark distribution.
 * Copyright (C) 2003 Universite de Versailles Saint-Quentin.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307.
 * You can also get it at http://www.gnu.org/licenses/lgpl.html
 * 
 * For more information on this software, see http://www.xquark.org.
 */

package org.xquark.xquery.parser;

import java.util.*;

import org.xquark.schema.SchemaManager;
import org.xquark.util.NamespaceContextStack;
import org.xquark.xquery.parser.util.ArrayListUtil;

public class XQueryUnit implements Cloneable {
	private static final String RCSRevision = "$Revision: 1.2 $";
	private static final String RCSName = "$Name:  $";

	private XQueryFile parentFile = null;
	// cannot be null or empty
	private ArrayList expressions = null;
	private NamespaceContextStack declarations = null;
	private HashMap functions = null;
	private String defaultCollation = null;

	// #############################################################################
	// VISITOR STUFF
	// #############################################################################

	public void accept(ParserVisitor visitor) throws XQueryException {
		visitor.visit(this);
	}

	// #############################################################################
	// CONSTRUCTOR STUFF
	// #############################################################################

	public XQueryUnit(NamespaceContextStack hashDeclList, HashMap hashFuncList, ArrayList expressions, String defaultCollation) throws XQueryException {
		setExpressions(expressions);
		setDeclarations(hashDeclList);
		setFunctions(hashFuncList);
		setDefaultCollation(defaultCollation);
	}

	// #############################################################################
	// ATTRIBUTE ACCESS STUFF
	// #############################################################################

	// metadata
//	public MetadataAccess getMetaData() {
//		if (parentFile == null)
//			return null;
//		return parentFile.getMetaData();
//	}

	// schemaManager
	public SchemaManager getSchemaManager() {
		if (parentFile == null)
			return null;
		return parentFile.getSchemaManager();
	}

	// parentFile
	public void setParentFile(XQueryFile xqueryFile) {
		parentFile = xqueryFile;
	}
	public XQueryFile getParentFile() {
		return parentFile;
	}

	// expressions
	public ArrayList getExpressions() {
		return expressions;
	}
	public void setExpressions(ArrayList expressions) throws XQueryException {
		if (expressions == null)
			throw new XQueryException("expressions of XQueryUnit cannot be null");
		this.expressions = expressions;
		for (int i = 0; i < this.expressions.size(); i++) {
			((XQueryExpression) this.expressions.get(i)).setParentUnit(this);
			((XQueryExpression) this.expressions.get(i)).setParentExpressions(null);
		}
	}

	public void setElementAt(XQueryExpression expr, int index) {
		expr.setParentUnit(this);
		expressions.set(index, expr);
	}

	// declarations
	public NamespaceContextStack getDeclarations() {
		return declarations;
	}
	public void setDeclarations(NamespaceContextStack declarations) {
		this.declarations = declarations;
	}

	// functions
	public HashMap getFunctions() {
		return functions;
	}
	public void setFunctions(HashMap functions) {
		this.functions = functions;
		if (functions != null) {
			Collection funcValues = functions.values();
			Iterator itValues = funcValues.iterator();
			while (itValues.hasNext())
				 ((FunctionDefinition) itValues.next()).setParentUnit(this);
		}
	}

	public FunctionDefinition getFunctionDefinition(String name) {
		if (name == null || functions == null)
			return null;
		return (FunctionDefinition) functions.get(name);
	}

	public String getNameSpace(String prefixname) {
		if (prefixname == null)
			return null;
		if (declarations == null)
			return null;
		String str = declarations.getNamespaceURI(prefixname);
		if (str == null)
			return null;
		int len = str.length();
		if (str.indexOf("\"") == 0)
			return str.substring(1, len - 2);
		else
			return str;
	}

	public void setDefaultCollation(String defaultCollation) {
		this.defaultCollation = defaultCollation;
	}

	public String getDefaultCollation() {
		return defaultCollation;
	}

	// #############################################################################
	// SOURCES STUFF
	// #############################################################################

	public ArrayList getSourceNames() {
		ArrayList sources = new ArrayList();
		for (int i = 0; i < expressions.size(); i++) {
			ArrayListUtil.add(sources, ((XQueryExpression) expressions.get(i)).getSourceNames());
		}
		return sources;
	}

	public ArrayList getUrls() {
		ArrayList urls = new ArrayList();
		for (int i = 0; i < expressions.size(); i++) {
			ArrayListUtil.add(urls, ((XQueryExpression) expressions.get(i)).getUrls());
		}
		return urls;
	}

	// #############################################################################
	// PRINT STUFF
	// #############################################################################

	public String pprint() {
		StringBuffer buf = new StringBuffer();
		//Re-add by AWR :
		if (declarations != null) {
			Collection declarationsPrefixes = declarations.getPrefixes();
			Iterator prefixes = declarationsPrefixes.iterator();
			while (prefixes.hasNext()) {
				String prefix = (String) prefixes.next();
				if ((prefix == null) || (prefix.trim().length() == 0)) {
					buf.append("default namespace ");
					buf.append("=\"");
					buf.append((String) declarations.getNamespaceURI(prefix));
					buf.append("\" ");
				} else if (prefix != "xml") {
					buf.append("namespace ");
					buf.append(prefix);
					buf.append("=\"");
					buf.append((String) declarations.getNamespaceURI(prefix));
					buf.append("\" ");
				}
			}
		}
		if (defaultCollation != null) {
			buf.append(" default collation = ");
			buf.append(defaultCollation);
		}
		if (functions != null) {
			Collection funcValues = functions.values();
			Iterator itValues = funcValues.iterator();
			while (itValues.hasNext())
				buf.append(" " + ((FunctionDefinition) itValues.next()).pprint());
		}
		for (int i = 0; i < this.expressions.size(); i++) {
			if (i != 0)
				buf.append(",");
			buf.append(" " + ((XQueryExpression) expressions.get(i)).pprint());
		}
		return buf.toString();
	}

	public String toString() {
		return pprint();
	}

	// #############################################################################
	// CLONE STUFF
	// #############################################################################
	//	public synchronized Object totalClone() throws CloneNotSupportedException {
	//		try {
	//			XQueryUnit newobj = (XQueryUnit) super.clone();
	//
	//			newobj.setExpressions((getExpressions() == null) ? null : XQueryExpressionTools.totalClone(getExpressions(), newobj));
	//			newobj.setDeclarations(getDeclarations());
	//			newobj.setFunctions((getFunctions() == null) ? null : XQueryExpressionTools.totalCloneHashMap(getFunctions(), newobj));
	//			newobj.setDefaultCollation(getDefaultCollation());
	//			return newobj;
	//		}
	//		catch (Exception e) {
	//			//System.err.println(e.getMessage()) ;
	//			throw new CloneNotSupportedException("Error in cloning XQueryUnit");
	//		}
	//	}
	//
	//	public synchronized Object clone() throws CloneNotSupportedException {
	//		try {
	//			XQueryUnit newobj = (XQueryUnit) super.clone();
	//			newobj.setExpressions((getExpressions() == null) ? null : XQueryExpressionTools.totalClone(getExpressions()));
	//			newobj.setDeclarations(getDeclarations());
	//			newobj.setFunctions((getFunctions() == null) ? null : XQueryExpressionTools.totalCloneHashMap(getFunctions()));
	//			newobj.setDefaultCollation(getDefaultCollation());
	//			return newobj;
	//		}
	//		catch (Exception e) {
	//			throw new CloneNotSupportedException("Error in cloning XQueryUnit");
	//		}
	//	}
	//
	//	public synchronized Object clone(HashMap mapvars) throws CloneNotSupportedException {
	//		try {
	//			XQueryUnit newobj = (XQueryUnit) super.clone();
	//			if (newobj == null)
	//				return null;
	//			newobj.setExpressions((getExpressions() == null) ? null : XQueryExpressionTools.totalClone(getExpressions(), mapvars));
	//			newobj.setDeclarations(getDeclarations());
	//			newobj.setFunctions((getFunctions() == null) ? null : XQueryExpressionTools.totalCloneHashMap(getFunctions(), mapvars));
	//			newobj.setDefaultCollation(getDefaultCollation());
	//			return newobj;
	//		}
	//		catch (Exception e) {
	//			//System.err.println(e.getMessage()) ;
	//			throw new CloneNotSupportedException("Error in cloning XQueryUnit");
	//		}
	//	}

}
